#include "objectrob.hpp"
#include "VimanTags.h"

double percentToReduce = 0.02;

ObjectRob::ObjectRob()
{
}

ObjectRob::~ObjectRob()
{
    for(unsigned int i = 0; i < _supportList.size(); i++)
    {
        delete _supportList.at(i);
    }
    _supportList.clear();
}

vector<string> agentNames;
vector<string> manipulableNames;
vector<string> supportNames;
vector<string> virtualNames;

void initVectors()
{
    agentNames.push_back("HUMAN");
    agentNames.push_back("ROBOT");

    manipulableNames.push_back("TAPE");
    manipulableNames.push_back("CUBE");
    manipulableNames.push_back("TOY");
    manipulableNames.push_back("BOX");
    manipulableNames.push_back("BOOK");
    manipulableNames.push_back("KIT");
    manipulableNames.push_back("VIRTUAL");
    manipulableNames.push_back("TRAY");
    manipulableNames.push_back("BOTTLE");
    manipulableNames.push_back("BRACKET");

    supportNames.push_back("TABLE");
    supportNames.push_back("STOCK");
    supportNames.push_back("SHELF");
    supportNames.push_back("KIT");
    supportNames.push_back("BOX");
    supportNames.push_back("BIN");
    supportNames.push_back("WORK_LOCATION");
    supportNames.push_back("DESK");


    virtualNames.push_back("VIRTUAL");
    virtualNames.push_back("PLACEMAT");

}
bool vectorInit = false;



ObjectRob::ObjectRob(Robot* rob)
{
    _robot = rob;
    if (!vectorInit)
    {
        vectorInit = true;
        initVectors();
    }

    string s = "";
    _isObject = false;
    _manipulable = false;
    _support = false;
    _isVirtual = false;


    for(unsigned int i = 0; i < agentNames.size(); i++)
    {
        if (_robot->getName().find(agentNames.at(i)) != std::string::npos)
        {
            s = agentNames.at(i);
            break;
        }
    }
    if (s != "")
    {
        cout << "robot " << _robot->getName() << " contains " << s << " so is not considered as object" << endl;
        _isObject = false;
        return;
    }

    for(unsigned int j = 0; j < manipulableNames.size(); j++)
    {
        if (_robot->getName().find(manipulableNames.at(j)) != std::string::npos)
        {
            s = manipulableNames.at(j);
            break;
        }
    }
    if (s != "")
    {
        cout << "robot " << _robot->getName() << " contains " << s << " so it is a manipulable object" << endl;
        _isObject = true;
        _manipulable = true;
        createRotation();
    }

    s= "";

    for(unsigned int i = 0; i < supportNames.size(); i++)
    {
        if (_robot->getName().find(supportNames.at(i)) != std::string::npos)
        {
            s = supportNames.at(i);
            break;
        }
    }
    if (s != "")
    {
        cout << "robot " << _robot->getName() << " contains " << s << " so it is a support object" << endl;
        _isObject = true;
        _support = true;
//        createSupports(_robot);
    }

    s= "";
    for(unsigned int i = 0; i < virtualNames.size(); i++)
    {
        if (_robot->getName().find(virtualNames.at(i)) != std::string::npos)
        {
            s = virtualNames.at(i);
            break;
        }
    }
    if (s != "")
    {
        cout << "robot " << _robot->getName() << " contains " << s << " so it is a virtual object" << endl;
        _isObject = true;
        _isVirtual = true;
        createVirtual();
    }

    if(_isObject){
        //add viman tags to this robot
        unsigned int tags=VimanTags::readTag(_robot,this);
        cout << "robot "<< _robot->getName() << " has " << tags << " viman tags"<<endl;
    }

}


void ObjectRob::createRotation()
{
    vector<configPt> vc;
    Robot* obj = _robot;
    configPt q_cur = p3d_get_robot_config(obj->getRobotStruct());
    if(obj->getName() == "LOTR_TAPE"
            || obj->getName() == "WALLE_TAPE"
            || obj->getName() == "BRACKET_1"
            || obj->getName() == "BRACKET_2")
    {
        configPt q1 = p3d_copy_config(obj->getRobotStruct(),q_cur);
        q1[6] = 0; q1[7] = 0; q1[8] = 0; q1[9] = 0; q1[10] = 0; q1[11] = OPEN_VALUE;
        configPt q2 = p3d_copy_config(obj->getRobotStruct(),q_cur);
        q2[6] = 0; q2[7] = 0; q2[8] = 0.2; q2[9] = 0; q2[10] = M_PI; q2[11] = OPEN_VALUE;
        configPt q3 = p3d_copy_config(obj->getRobotStruct(),q_cur);
        q3[6] = 0; q3[7] = -0.1; q3[8] = 0.06; q3[9] = -M_PI/2; q3[10] = OPEN_VALUE; q3[11] = 0;
        configPt q4 = p3d_copy_config(obj->getRobotStruct(),q_cur);
        q4[6] = 0; q4[7] = 0.1; q4[8] = 0.06; q4[9] = M_PI/2; q4[10] = OPEN_VALUE; q4[11] = 0;

        vc.push_back(q1);
        vc.push_back(q2);
        vc.push_back(q3);
        vc.push_back(q4);

    }
    else if (obj->getName().find("GREY_TAPE") != string::npos || obj->getName().find("BRACKET_3") != string::npos)
    {
        configPt q1 = p3d_copy_config(obj->getRobotStruct(),q_cur);
        q1[6] = 0; q1[7] = 0; q1[8] = 0; q1[9] = 0; q1[10] = 0; q1[11] = OPEN_VALUE;
        configPt q2 = p3d_copy_config(obj->getRobotStruct(),q_cur);
        q2[6] = 0; q2[7] = 0; q2[8] = 0.21; q2[9] = 0; q2[10] = M_PI; q2[11] = OPEN_VALUE;
        configPt q3 = p3d_copy_config(obj->getRobotStruct(),q_cur);
        q3[6] = 0; q3[7] = -0.105; q3[8] = 0.0725; q3[9] = -M_PI/2; q3[10] = OPEN_VALUE; q3[11] = 0;
        configPt q4 = p3d_copy_config(obj->getRobotStruct(),q_cur);
        q4[6] = 0; q4[7] = 0.105; q4[8] = 0.0725; q4[9] = M_PI/2; q4[10] = OPEN_VALUE; q4[11] = 0;

        vc.push_back(q1);
        vc.push_back(q2);
        vc.push_back(q3);
        vc.push_back(q4);
    }
    else if (obj->getName() == "VIRTUAL_TT" || obj->getName() == "VIRTUAL_TRAY")
    {
        configPt q1 = p3d_copy_config(obj->getRobotStruct(),q_cur);
        q1[6] = 0; q1[7] = 0; q1[8] = 0; q1[9] = 0; q1[10] = 0; q1[11] = OPEN_VALUE;

        vc.push_back(q1);
    }
    else if (obj->getName() == "TRAY" ||
             obj->getName() == "ACCESSKIT" ||
             obj->getName() == "HANDLE_BOX" ||
             obj->getName() == "SURPRISE_BOX" ||
             obj->getName() == "HANDLE_BOX2")
    {
        configPt q1 = p3d_copy_config(obj->getRobotStruct(),q_cur);
        q1[6] = 0; q1[7] = 0; q1[8] = 0; q1[9] = 0; q1[10] = 0; q1[11] = OPEN_VALUE;

        vc.push_back(q1);
    }
    else if (obj->getName() == "GREEN_CUBE"
             || obj->getName() == "RED_CUBE"
             || obj->getName() == "BLUE_CUBE"
             || obj->getName() == "YELLOW_CUBE"
             || obj->getName() == "BLACK_CUBE"
	     || obj->getName() == "GREEN_CUBE2"
	     || obj->getName() == "RED_CUBE2"
	     || obj->getName() == "BLUE_CUBE2"
	     )
    {
        configPt q1 = p3d_copy_config(obj->getRobotStruct(),q_cur);
        q1[6] = 0; q1[7] = 0; q1[8] = 0.03; q1[9] = 0; q1[10] = 0; q1[11] = OPEN_VALUE;
        vc.push_back(q1);

        configPt q2 = p3d_copy_config(obj->getRobotStruct(),q_cur);
        q2[6] = 0; q2[7] = 0; q2[8] = 0.03; q2[9] = -M_PI/2; q2[10] = OPEN_VALUE; q2[11] = 0;
        vc.push_back(q2);

        configPt q3 = p3d_copy_config(obj->getRobotStruct(),q_cur);
        q3[6] = 0; q3[7] = 0; q3[8] = 0.03; q3[9] = M_PI/2; q3[10] = OPEN_VALUE; q3[11] = 0;
        vc.push_back(q3);

        configPt q4 = p3d_copy_config(obj->getRobotStruct(),q_cur);
        q4[6] = 0; q4[7] = 0; q4[8] = 0.03; q4[9] = M_PI; q4[10] = 0; q4[11] = OPEN_VALUE;
        vc.push_back(q4);

        configPt q5 = p3d_copy_config(obj->getRobotStruct(),q_cur);
        q5[6] = 0; q5[7] = 0; q5[8] = 0.03; q5[9] = M_PI/2; q5[10] = OPEN_VALUE; q5[11] = -M_PI/2;
        vc.push_back(q5);

        configPt q6 = p3d_copy_config(obj->getRobotStruct(),q_cur);
        q6[6] = 0; q6[7] = 0; q6[8] = 0.03; q6[9] = M_PI/2; q6[10] = OPEN_VALUE; q6[11] = M_PI/2;
        vc.push_back(q6);
    }
    else if (obj->getName().find("TOP_CUBE") != string::npos)
    {
        configPt q1 = p3d_copy_config(obj->getRobotStruct(),q_cur);
        q1[6] = 0; q1[7] = 0; q1[8] = 0.05; q1[9] = 0; q1[10] = 0; q1[11] = OPEN_VALUE;
        vc.push_back(q1);
    }
    else if (obj->getName() == "GLUE_BOTTLE" || obj->getName() == "BOTTLE1")
    {
        configPt q1 = p3d_copy_config(obj->getRobotStruct(),q_cur);
        q1[6] = 0; q1[7] = 0; q1[8] = 0.0; q1[9] = 0; q1[10] = 0; q1[11] = OPEN_VALUE;
        vc.push_back(q1);
    }
    else
    {
        cout << "WARNING: no rotations found for this object " << _robot->getName() << endl;
    }
    _stableRotations = vc;
}


void ObjectRob::createSupports(Robot* r)
{
    if (_robot->getNumberOfJoints() > 2 )
    {
        return;
    }
    p3d_oriented_BB *obj_OBB = p3d_get_oriented_BB_obj(_robot->getJoint(0)->getJointStruct()->o);

    p3d_matrix3 m;
    p3d_mat3Copy(p3d_mat3IDENTITY, m);

//    p3d_point a =obj_OBB->UpBackLeft;
//    p3d_point b =obj_OBB->UpBackRight;
//    p3d_point c =obj_OBB->UpFrontLeft;
//    p3d_point d =obj_OBB->UpFrontRight;

    GeometricComputations GC = GeometricComputations();

    p3d_point a,b,c,d;
   vector<p3d_point> vs = GC.getUpperPointsInObject(_robot,0.03);
   if (vs.size() !=4)
       return;
   else
   {
       a = vs.at(0);
       b = vs.at(1);
       c = vs.at(2);
       d = vs.at(3);
   }



//    a.z = a.z - a.z*percentToReduce;
//    b.z = b.z - b.z*percentToReduce;
//    c.z = c.z - c.z*percentToReduce;
//    d.z = d.z - d.z*percentToReduce;


    RectangleForm* pr = new RectangleForm(a,b,c,d,m);

    p3d_point ct;
    ct.x= (a.x + d.x)/2;
    ct.y= (a.y + d.y)/2;
    ct.z= (a.z + d.z)/2;

    p3d_point r1;
    r1.x= (a.x + b.x)/2;
    r1.y= (a.y + b.y)/2;
    r1.z= (a.z + b.z)/2;
    p3d_point r2;
    r2.x= (a.x + c.x)/2;
    r2.y= (a.y + c.y)/2;
    r2.z= (a.z + c.z)/2;
    p3d_point r3;
    r3.x= (d.x + c.x)/2;
    r3.y= (d.y + c.y)/2;
    r3.z= (d.z + c.z)/2;
    p3d_point r4;
    r4.x= (d.x + b.x)/2;
    r4.y= (d.y + b.y)/2;
    r4.z= (d.z + b.z)/2;

    RectangleForm* a1 = new RectangleForm(a,r2,r1,ct,m);
    RectangleForm* a2 = new RectangleForm(r1,ct,b,r4,m);
    RectangleForm* a3 = new RectangleForm(r2,c,ct,r3,m);
    RectangleForm* a4 = new RectangleForm(ct,r3,r4,d,m);

    a1->isSupport=false;
    a2->isSupport=false;
    a3->isSupport=false;
    a4->isSupport=false;

    a1->object =r;
    a2->object =r;
    a3->object =r;
    a4->object =r;

    pr->supportAreas["1"] = a1;
    pr->supportAreas["2"] = a2;
    pr->supportAreas["3"] = a3;
    pr->supportAreas["4"] = a4;


    pr->object = r;

    _supportList.push_back(pr);

    p3d_destroy_oriented_BB(obj_OBB);
}

void ObjectRob::createVirtual()
{
    p3d_col_deactivate_robot(_robot->getRobotStruct());
    setToVirtual(true);
}

vector<GeometricForm*> ObjectRob::getSupports()
{
    _supportList.clear();
    createSupports(_robot);
    return _supportList;
}


//return -1 if not in stable rotation
int ObjectRob::isInStableRotation()
{
    configPt q = _robot->getCurrentPos()->getConfigStruct();
    for (unsigned int i = 0; i < _stableRotations.size(); i++)
    {
        configPt tmp = _stableRotations.at(i);
        if (fabs(q[9] - tmp[9]) < 0.01 && fabs(q[10] - tmp[10]) < 0.01)
            return 11;
        else if (fabs(q[10] - tmp[10]) < 0.01 && fabs(q[11] - tmp[11]) < 0.01)
            return 9;
        else if (fabs(q[9] - tmp[9]) < 0.01 && fabs(q[11] - tmp[11]) < 0.01)
            return 10;
    }
    return -1;
}

double ObjectRob::distToPoint(p3d_point p)
{
    //for rectangles:
    double dist = 1000;

    double tmp;
    for (unsigned int i = 0; i < _supportList.size(); i++)
    {
        GeometricForm* prec= _supportList.at(i);
        tmp = prec->getDistanceFromPoint(p);
        if (tmp < dist)
            dist = tmp;
    }
    return dist;


}

double ObjectRob::distToPointZ(p3d_point p)
{
    double dist = 1000;

    double tmp;
    for (unsigned int i = 0; i < _supportList.size(); i++)
    {
        GeometricForm* prec= _supportList.at(i);
        tmp = prec->getPlannarDistanceFromPoint(p);
        if (tmp < dist)
            dist = tmp;
    }
    return dist;
}

configPt ObjectRob::getStableRotation(int id, double rot)
{

    p3d_matrix4 K, M, res;

    if (_stableRotations.size() <= id || id < 0)
        return NULL;

    configPt q = p3d_copy_config(_robot->getRobotStruct(),_stableRotations.at(id));

    p3d_vector3 v;
    for (int i = 0; i<=2; i++)
    {
        if (q[i+9] == OPEN_VALUE)
        {
            v[i] = 1;
            q[i+9] = 0;
        }
        else
            v[i] = 0;
    }


    p3d_set_and_update_this_robot_conf(_robot->getRobotStruct(), q);

    p3d_mat4Copy(_robot->getJoint(0)->getJointStruct()->abs_pos, K);

//    p3d_mat4Print(K, "K");

    v[0] = 0; v[1] = 0; v[2] = 1;

    p3d_mat4Rot(M,v,rot);
//    p3d_mat4Print(M, "rotation matric");

    p3d_mat4Mult(M, K, res);
//    p3d_mat4Print(res, "res");


    double Tx, Ty, Tz;
    double Rx, Ry, Rz;
    p3d_mat4ExtractPosReverseOrder(res, &Tx, &Ty, &Tz,
                                   &Rx, &Ry, &Rz);

    q[6] = Tx;
    q[7] = Ty;
    q[8] = Tz;
    q[9] = Rx;
    q[10] = Ry;
    q[11] = Rz;

    p3d_set_and_update_this_robot_conf(_robot->getRobotStruct(), q);


    return q;
}

void ObjectRob::setToVirtual(bool val)
{
    if (!val)
        p3d_set_col_on(_robot->getRobotStruct());
    else
        p3d_set_col_off(_robot->getRobotStruct());
}

