#ifndef OBJECTROB_HPP
#define OBJECTROB_HPP

#define OPEN_VALUE -100

using namespace std;
#include <Eigen/Geometry>
#include "Planner-pkg.h"
#include "API/ConfigSpace/configuration.hpp"
#include "robot.hpp"
#include "GTP/GTPTools/geometricTools.hpp"
#include "Collision-pkg.h"

//class Robot;
class GeometricForm;

class ObjectRob
{
public:
    ObjectRob();
    ObjectRob(Robot* rob);

    ~ObjectRob();

    void createRotation();
    void createSupports(Robot *r);
    void createVirtual();

    //return the closest distance to a support from the object
    // for now implemented for rectangle only
    double distToPoint(p3d_point p);

    double distToPointZ(p3d_point p);

    void setManipulable (bool val) {_manipulable = val;}
    bool isManipulable() {return _manipulable;}

    void setSupport (bool val) {_support = val;}
    bool isSupport() {return _support;}

    void setContainer (bool val) {_container = val;}
    bool isContainer() {return _container;}

    void setIsObject (bool val) {_isObject = val;}
    bool isObject() {return _isObject;}

    void addStableRotation(configPt q) {_stableRotations.push_back(q);}
    vector<configPt> getStableRotations() {return _stableRotations;}
    configPt getStableRotation(int id, double rot);

    void addSupports(GeometricForm* gf) { _supportList.push_back(gf);}
    vector<GeometricForm*> getSupports();

    /// true if the object has viman tags attributed
    bool hasTags() {return _tags.size();}
    /// add a viman tag in the form of a transform
    void addTag(Eigen::Affine3d tf) {_tags.push_back(tf);}
    void clearTags(){_tags.clear();}
    vector<Eigen::Affine3d> const & getTags(){return _tags;}

    int isInStableRotation();

    bool isVirtual() {return _isVirtual;}
    void getIsVirtual(bool iv) {_isVirtual = iv;}

    void setToVirtual(bool val);

private:
    bool _manipulable;
    bool _support;
    bool _container;

    vector<configPt> _stableRotations;

    vector<GeometricForm*> _supportList;

    vector<Eigen::Affine3d> _tags; ///< list of viman tags. Applyng this tranform to a unity square in the x,y plane of the object frame gives the square representing the tag.

    bool _isObject;
    bool _isVirtual;

    Robot* _robot;

};

#endif // OBJECTROB_HPP
