#ifndef FACTS_HPP
#define FACTS_HPP

#include "P3d-pkg.h"
#include "move3d-headless.h"
#include "Planner-pkg.h"

#include "Device/robot.hpp"
#include "scene.hpp"



using namespace std;
using namespace tr1;

class WorldStateFacts;
class Fact;
//class Scene;

typedef struct struct_fact_id{
    string factType;
    string factSubType;
    bool isMainAnAgent;
} fact_id;


/**
  * @ingroup NEW_CPP_MODULE
  * @defgroup CPP_API C++ Planning API
  */

/**
 * @ingroup CPP_API
 */
class FactsManager
{
public:
    FactsManager();
    FactsManager(Scene *scen);
    ~FactsManager();

    void initAll();
    bool disableFacts(string s);
    bool enableFacts(string s);
    bool isFactEnabled(string s);

    bool computeVisibility(WorldStateFacts* WS,Robot* mainEntity, Robot* targetEntity, bool forceRecompute);
    bool computeReachability(WorldStateFacts* WS,Robot* mainEntity, Robot* targetEntity, bool forceRecompute);
    bool computeIsOn(WorldStateFacts* WS,Robot* mainEntity, Robot* targetEntity, bool forceRecompute);
    bool computeIsIn(WorldStateFacts* WS,Robot* mainEntity, Robot* targetEntity, bool forceRecompute);
    bool computeIsNextTo(WorldStateFacts* WS,Robot* mainEntity, Robot* targetEntity, bool forceRecompute);
    bool computeSize(WorldStateFacts* WS,Robot* mainEntity, Robot* targetEntity, bool forceRecompute);
    bool computeClutter(WorldStateFacts* WS,Robot* mainEntity, Robot* targetEntity, bool forceRecompute);


    bool computeRobotVisibility(WorldStateFacts* WS, Robot* mainEntity, bool forceRecompute);
    bool computeRobotReachability(WorldStateFacts* WS, Robot* mainEntity, bool forceRecompute);
    bool computeRobotIsOn(WorldStateFacts* WS, Robot* mainEntity, bool forceRecompute);
    bool computeRobotIsIn(WorldStateFacts* WS, Robot* mainEntity, bool forceRecompute);
    bool computeRobotIsNextTo(WorldStateFacts* WS, Robot* mainEntity, bool forceRecompute);
    bool computeRobotSize(WorldStateFacts* WS, Robot* mainEntity, bool forceRecompute);
    bool computeRobotClutter(WorldStateFacts* WS, Robot* mainEntity, bool forceRecompute);


    bool computeAllVisibilities(WorldStateFacts* WS, bool forceRecompute);
    bool computeAllReachabilities(WorldStateFacts* WS, bool forceRecompute);
    bool computeAllIsOn(WorldStateFacts* WS, bool forceRecompute);
    bool computeAllIsIn(WorldStateFacts* WS, bool forceRecompute);
    bool computeAllIsNextTo(WorldStateFacts* WS, bool forceRecompute);
    bool computeAllSize(WorldStateFacts* WS, bool forceRecompute);
    bool computeAllClutter(WorldStateFacts* WS, bool forceRecompute);

    bool computeFacts(WorldStateFacts* WS, bool forceRecompute);
    bool computeFacts(int WSid, bool forceRecompute);
    int computeFacts(bool forceRecompute);


    int findWorldStateFacts(WorldStateFacts* WS);
    int findOrAddWorldStateFacts(WorldStateFacts* WS);
    int findCurWorldStateFacts();

    void setFactToFalse(WorldStateFacts* WS, fact_id ft, Robot* mainEntity, Robot* targetEntity);

    bool cleanWSList();

    bool computeOneFact(WorldStateFacts* WSF, string fact, Robot *mainEntity);
    vector<Fact> computeAndGetFact(string fact, Robot *mainEntity);
    Fact computeAndGetFact(string fact, string factSubType, Robot *mainEntity, Robot* targetEntity);

    WorldStateFacts* getWSFromId(int WSid);

    vector<fact_id> getFactList() {return factList;}

private:
    static vector<fact_id> factList;
    vector<pair<int, WorldStateFacts*> > WSlist;
    int WSlistId;
    Scene* sce;

    vector<string> disabledFacts;

};



class Fact
{
public:
    Fact();
    Fact(fact_id fact_type);

    bool compareFactsDescr(Fact f);
    bool compareFactsDescr(fact_id FI, Robot* main, Robot* target);
    bool compareFactsValue(Fact f);

    fact_id getType();

    Robot* getMainEntity();
    Robot* getTargetEntity();

    void setMainEntity(Robot* r);
    void setTargetEntity(Robot* r);

    bool isItComputed();
    void setIsComputed(bool t);

    int getValueType();
    void setValueType(int v);

    bool getBoolValue();
    void setBoolValue(bool v);

    double getSpecValue();
    void setSpecValue(double v);

    void printFact();

    string getFactDescr();


private:
    fact_id fType;

    Robot* mainEntity;
    Robot* targetEntity;

    bool isComputed;

    int value_type; //0 for bool, 1 for specificValue

    bool boolValue;
    double specificValue;


};



class WorldStateFacts
{
public:
    WorldStateFacts();

    WorldStateFacts(Scene *scen);
    WorldStateFacts(vector<tr1::shared_ptr<Configuration> > confs );
    vector<Robot*> compareWS(WorldStateFacts* WS);

    void printAllFacts();
    void printAllComputedFacts();
    void printFacts(vector<Fact> fp);

    bool initFacts(vector<fact_id> factList);


    vector<Fact> getFacts();
    vector<Fact> getMeaningfulFacts();
    vector<Fact> getUpdatedFacts(WorldStateFacts* WS);
    void setFacts(vector<Fact> Facts);

    vector<Fact> getFactFromMain(string factType, string main);

    Fact getFact(fact_id Fact_Id, Robot* main, Robot* target);
    Fact getFact(string factType, string main, string target);
    void setFact(Fact fact);

    vector<shared_ptr<Configuration> > getRobotsConfs();

    shared_ptr<Configuration> getRobotConf(Robot* robot);

    shared_ptr<Configuration>  getRobotConfById(int robot_id);

    void setConfigurations(vector<shared_ptr<Configuration> > robVect) { robotsConfs = robVect;}
    void setConfiguration(shared_ptr<Configuration> conf);

private:
    vector<shared_ptr<Configuration> > robotsConfs;
    vector<Fact> facts;

    Scene* sce;
};




#endif // FACTS_HPP
