#ifndef WORLDSTATE_HPP
#define WORLDSTATE_HPP
using namespace std;
//using namespace std::tr1;
#include "P3d-pkg.h"
#include "API/Device/robot.hpp"
//#include "../virtualtask.hpp"
#include "API/scene.hpp"
#include "lightPlanner/proto/ManipulationPlanner.hpp"

#include "../../planner/plannerFunctions.hpp"
#include "../lightPlanner/proto/lightPlanner.h"
#include "../lightPlanner/proto/lightPlannerApi.h"

#include "API/facts.hpp"

#include "graspPlanning/include/gpGrasp.h"
#include "taskconstraints.hpp"
#include "tools.hpp"

#include "mightabilitiesGrids.hpp"

class TaskSolution;
class TaskConstraints;
class MightabilitiesGrid;

struct Attachement;

class WorldState
{
public:
    WorldState();
    WorldState(Scene* sce);
    WorldState(WorldState* WS);

    void load();
    void loadConf(tr1::shared_ptr<Configuration> q);
    void saveAll();
    void saveRobot(Robot* rob);
    void saveConf(tr1::shared_ptr<Configuration> q);
    bool replaceConf(tr1::shared_ptr<Configuration> q);
    void clear();

    bool isEqualTo(WorldState* WS);

    void addAttachement(Robot* rob, Robot* obj);
    void addAttachement(Robot* rob, Robot* obj, gpGrasp gr, int armId);
    void addAttachement(Attachement a);
    vector<Attachement> getAttachements() {return attachements_list;}
    void setAttachements(vector<Attachement> va) {attachements_list = va;}
    Attachement getAttachement(Robot* rob, Robot* obj);
    void clearAttachement();
    void removeAttachements(Robot* rob, Robot* obj);
    bool hasAttachement(Robot* rob);
    bool hasAttachement(Robot* rob,Robot* obj);

    void saveCurrentAttachement();
    void resetCurrentAttachements();
    void loadAttachements();
    Attachement getCurrentAttachment();

    vector<tr1::shared_ptr<Configuration> > getRobotConfs() {return _robotsConfs;}
    tr1::shared_ptr<Configuration> getRobotConf(Robot* rob);

    int getId() {return _id;}
    void setId(int id) {_id = id;}

    Scene* getScene() {return _sce;}

    vector<Fact> getFacts();
    vector<Fact> getAllFacts();
    //does not compare the values of the facts
    Fact getThisFact(Fact FI);

    TaskSolution* getTSThatEndWithThisWS() {return endingStateOf;}
    void setTSThatEndWithThisWS(TaskSolution* TS) {endingStateOf = TS;}

    TaskConstraints* getTCForNextTask() {return _TCForNextTask;}
    void setTCForNextTask(TaskConstraints* tc) {_TCForNextTask = tc;}
    void clearTCForNextTask();

    void saveToDisk(FILE *file, string text);

    void initMightabilities();
    MightabilitiesGrid* getMightabilities();

    double distanceTo(WorldState* ws);

private:
    vector<tr1::shared_ptr<Configuration> > _robotsConfs;
    vector<Attachement> attachements_list;
    //adding trajectories

    Scene* _sce;

    int _id;

    WorldStateFacts* WSF;
    bool areFactsComputed;

    TaskSolution* endingStateOf;

    TaskConstraints* _TCForNextTask;

    MightabilitiesGrid* _MG;

};

#endif // WORLDSTATE_HPP
