#include "geometricTools.hpp"

using namespace std;
using namespace tr1;

double distSample = 0.05;

extern std::vector<p3d_point> currNavTraj;

GeometricComputations::GeometricComputations()
{

}

p3d_point GeometricComputations::getRandomPointInTriangle(p3d_point A, p3d_point B, p3d_point C)
{
    p3d_point P;
    double r1 = p3d_random(0,1);
    double r2 = p3d_random(0,1);
    if(r1+r2>1) {r1 = 1-r1; r2 = 1-r2;}

//    P.x = (1-sqrt(r1)*A.x + sqrt(r1)*(1-r2)*B.x + r2*sqrt(r1)*C.x);
//    P.y = (1-sqrt(r1)*A.y + sqrt(r1)*(1-r2)*B.y + r2*sqrt(r1)*C.y);
    P.x = r1*A.x + r2*B.x + (1-r1-r2)*C.x;
    P.y = r1*A.y + r2*B.y + (1-r1-r2)*C.y;
    P.z = (B.z + C.z)/2;
    return P;
}


/**
  *a----b
  *|    |
  *c----d   .p
  *b and c should be opposite
  */
p3d_point GeometricComputations::getRandomPointInRectangle(p3d_point A, p3d_point B, p3d_point C, p3d_point D)
{
    int i = p3d_random_integer(0,1);

    if (i == 1)
        return getRandomPointInTriangle(A,B,C);
    else
        return getRandomPointInTriangle(D,B,C);
}

vector<p3d_point> GeometricComputations::getLowerPointsInObject(Robot* obj, double steepness)
{
    p3d_oriented_BB *obj_OBB = p3d_get_oriented_BB_obj(obj->getJoint(0)->getJointStruct()->o);
//    vector<p3d_point> toBeTested;
    double zMin =100;
    if(obj_OBB->BottomBackLeft.z < zMin)
        zMin = obj_OBB->BottomBackLeft.z;
    if(obj_OBB->BottomBackRight.z < zMin)
        zMin = obj_OBB->BottomBackRight.z;
    if(obj_OBB->BottomFrontLeft.z < zMin)
        zMin = obj_OBB->BottomFrontLeft.z;
    if(obj_OBB->BottomFrontRight.z < zMin)
        zMin = obj_OBB->BottomFrontRight.z;

    if(obj_OBB->UpBackLeft.z < zMin)
        zMin = obj_OBB->UpBackLeft.z;
    if(obj_OBB->UpBackRight.z < zMin)
        zMin = obj_OBB->UpBackRight.z;
    if(obj_OBB->UpFrontLeft.z < zMin)
        zMin = obj_OBB->UpFrontLeft.z;
    if(obj_OBB->UpFrontRight.z < zMin)
        zMin = obj_OBB->UpFrontRight.z;

    double minDist = steepness;
    vector<p3d_point> vp;
    if (fabs(obj_OBB->BottomBackLeft.z - zMin) < minDist)
        vp.push_back(obj_OBB->BottomBackLeft);
    if (fabs(obj_OBB->BottomBackRight.z - zMin) < minDist)
        vp.push_back(obj_OBB->BottomBackRight);
    if (fabs(obj_OBB->BottomFrontLeft.z - zMin) < minDist)
        vp.push_back(obj_OBB->BottomFrontLeft);
    if (fabs(obj_OBB->BottomFrontRight.z - zMin) < minDist)
        vp.push_back(obj_OBB->BottomFrontRight);

    if (vp.size() != 4)
    {
        if (fabs(obj_OBB->UpBackLeft.z - zMin) < minDist)
            vp.push_back(obj_OBB->UpBackLeft);
        if (fabs(obj_OBB->UpBackRight.z - zMin) < minDist)
            vp.push_back(obj_OBB->UpBackRight);
        if (fabs(obj_OBB->UpFrontLeft.z - zMin) < minDist)
            vp.push_back(obj_OBB->UpFrontLeft);
        if (fabs(obj_OBB->UpFrontRight.z - zMin) < minDist)
            vp.push_back(obj_OBB->UpFrontRight);
    }

    p3d_destroy_oriented_BB(obj_OBB);

    return vp;
}

vector<p3d_point> GeometricComputations::getUpperPointsInObject(Robot* obj, double steepness)
{
    p3d_oriented_BB *obj_OBB = p3d_get_oriented_BB_obj(obj->getJoint(0)->getJointStruct()->o);
    vector<p3d_point> toBeTested;
    double zMin =-100;
    if(obj_OBB->BottomBackLeft.z > zMin)
        zMin = obj_OBB->BottomBackLeft.z;
    if(obj_OBB->BottomBackRight.z > zMin)
        zMin = obj_OBB->BottomBackRight.z;
    if(obj_OBB->BottomFrontLeft.z > zMin)
        zMin = obj_OBB->BottomFrontLeft.z;
    if(obj_OBB->BottomFrontRight.z > zMin)
        zMin = obj_OBB->BottomFrontRight.z;

    if(obj_OBB->UpBackLeft.z > zMin)
        zMin = obj_OBB->UpBackLeft.z;
    if(obj_OBB->UpBackRight.z > zMin)
        zMin = obj_OBB->UpBackRight.z;
    if(obj_OBB->UpFrontLeft.z > zMin)
        zMin = obj_OBB->UpFrontLeft.z;
    if(obj_OBB->UpFrontRight.z > zMin)
        zMin = obj_OBB->UpFrontRight.z;

    double minDist = steepness;
    vector<p3d_point> vp;
    if (fabs(obj_OBB->BottomBackLeft.z - zMin) < minDist)
        vp.push_back(obj_OBB->BottomBackLeft);
    if (fabs(obj_OBB->BottomBackRight.z - zMin) < minDist)
        vp.push_back(obj_OBB->BottomBackRight);
    if (fabs(obj_OBB->BottomFrontLeft.z - zMin) < minDist)
        vp.push_back(obj_OBB->BottomFrontLeft);
    if (fabs(obj_OBB->BottomFrontRight.z - zMin) < minDist)
        vp.push_back(obj_OBB->BottomFrontRight);

    if (vp.size() != 4)
    {
        if (fabs(obj_OBB->UpBackLeft.z - zMin) < minDist)
            vp.push_back(obj_OBB->UpBackLeft);
        if (fabs(obj_OBB->UpBackRight.z - zMin) < minDist)
            vp.push_back(obj_OBB->UpBackRight);
        if (fabs(obj_OBB->UpFrontLeft.z - zMin) < minDist)
            vp.push_back(obj_OBB->UpFrontLeft);
        if (fabs(obj_OBB->UpFrontRight.z - zMin) < minDist)
            vp.push_back(obj_OBB->UpFrontRight);
    }

    p3d_destroy_oriented_BB(obj_OBB);

    vector<p3d_point> res;
    for (unsigned int i = 0; i < vp.size(); i++)
    {
        p3d_point a = vp.at(i);
        bool testDist = true;
        for(unsigned int j = 0; j < res.size(); j++ )
        {
            p3d_point b = res.at(j);
            if (distBetweenPoints(a,b) < steepness)
            {
                testDist = false;
                break;
            }
        }
        if (testDist)
        {
            res.push_back(a);
        }
    }

    return res;
}


p3d_point GeometricComputations::getCenterOfObj(Robot* obje)
{
    p3d_oriented_BB *obj_OBB = p3d_get_oriented_BB_obj(obje->getJoint(0)->getJointStruct()->o);
    p3d_point p;
    p.x = (obj_OBB->UpFrontLeft.x + obj_OBB->BottomBackRight.x)/2;
    p.y = (obj_OBB->UpFrontLeft.y + obj_OBB->BottomBackRight.y)/2;
    p.z = (obj_OBB->UpFrontLeft.z + obj_OBB->BottomBackRight.z)/2;
    p3d_destroy_oriented_BB(obj_OBB);

    return p;
}

double GeometricComputations::compute_triangle_area_x_y2(p3d_point a, p3d_point b, p3d_point c)
{
    return fabs((a.x*(c.y-b.y) + b.x*(a.y-c.y)+ c.x*(b.y-a.y))/2.0);
}

/**
  *a----b
  *|    |
  *c----d   .x
  *b and c should be opposite
  */
bool GeometricComputations::point_in_rect2(p3d_point a, p3d_point b, p3d_point c, p3d_point d, p3d_point x)
{
//    cout << "a ( " << a.x << " , " << a.y << " , " << a.z << " ) "
//         << "\nb ( " << b.x << " , " << b.y << " , " << b.z << " ) "
//         << "\nc ( " << c.x << " , " << c.y << " , " << c.z << " ) "
//         << "\nd ( " << d.x << " , " << d.y << " , " << d.z << " ) "
//         << "\nx ( " << x.x << " , " << x.y << " , " << x.z << " ) " << endl;




    double s1 = compute_triangle_area_x_y2(a,b,c);

    double bcx = compute_triangle_area_x_y2(b,c,x);
    double acx = compute_triangle_area_x_y2(a,c,x);
    double abx = compute_triangle_area_x_y2(a,b,x);


    if (fabs (s1 - (bcx + acx + abx)) < EPS3 && bcx != 0 && acx != 0 && abx != 0)
    {
        return true;
    }

    double s2 = compute_triangle_area_x_y2(b,c,d);

    double dcx = compute_triangle_area_x_y2(d,c,x);
    double dbx = compute_triangle_area_x_y2(d,b,x);

    if (fabs(s2 - (bcx + dcx + dbx)) < EPS3 && bcx != 0 && dcx != 0 && dbx != 0)
    {
        return true;
    }
    return false;
}

/**
  *a----b  .x
  */
bool GeometricComputations::point_in_segment2d(p3d_point a, p3d_point b, p3d_point x)
{
    return fabs((b.x - a.x) * (x.y - a.y) - (x.x - a.x) * (b.y - a.y)) < EPS3;
}


double GeometricComputations::distBetweenPoints(p3d_point a, p3d_point b)
{
    return sqrt(pow(a.x-b.x,2) + pow(a.y-b.y,2) + pow(a.z-b.z,2));
}

double GeometricComputations::distBetweenMatrix4(p3d_matrix4 a, p3d_matrix4 b)
{
    p3d_point p1, p2;
    p1.x = a[0][3];
    p1.y = a[1][3];
    p1.z = a[2][3];

    p2.x = b[0][3];
    p2.y = b[1][3];
    p2.z = b[2][3];

    return distBetweenPoints(p1,p2);
}

double GeometricComputations::distBetweenPoints2D(p3d_point a, p3d_point b)
{
    return sqrt(pow(a.x-b.x,2) + pow(a.y-b.y,2));
}


/**
  *a----b
  *| .h |
  *c----d
  *b and c should be opposite
  */
double GeometricComputations::compute_pyramid_area(p3d_point a, p3d_point b, p3d_point c, p3d_point d, p3d_point h)
{
    //calcul de l'equation du plan
    double A =    (b.y - a.y)*(c.z - a.z)
                - (b.z - a.z)*(c.y - a.y);

    double B = - ((b.x - a.x)*(c.z - a.z)
                - (b.z - a.z)*(c.x - a.x));

    double C =    (b.x - a.x)*(c.y - a.y)
                - (b.y - a.y)*(c.x - a.x);

    double D =  - (A*a.x + B *a.y + C * a.z);

    double height = fabs(A*h.x+ B*h.y + C*h.z + D) / sqrt(A*A + B*B + C*C);
    double surface = sqrt(pow(a.x-b.x,2) + pow(a.y-b.y,2) + pow(a.z-b.z,2)) * sqrt(pow(a.x-c.x,2) + pow(a.y-c.y,2) + pow(a.z-c.z,2));
    if (height == 0)
        return 0;
    return surface*height/3;
}



/**  c----d
  * /|   /|
  *a----b h
  *|/g  |/
  *e----f   .x
  *b and c should be opposite
  */
bool GeometricComputations::point_in_hexahedron(p3d_point a, p3d_point b, p3d_point c, p3d_point d,
                         p3d_point e, p3d_point f, p3d_point g, p3d_point h,
                         p3d_point x)
{

    double v =sqrt(pow(a.x-b.x,2) + pow(a.y-b.y,2) + pow(a.z-b.z,2)) * sqrt(pow(a.x-c.x,2) + pow(a.y-c.y,2) + pow(a.z-c.z,2)) *sqrt(pow(a.x-e.x,2) + pow(a.y-e.y,2) + pow(a.z-e.z,2));

    double v1 = compute_pyramid_area (a,b,c,d,x);
    double v2 = compute_pyramid_area (e,f,g,h,x);
    double v3 = compute_pyramid_area (a,b,e,f,x);
    double v4 = compute_pyramid_area (b,d,f,h,x);
    double v5 = compute_pyramid_area (d,c,h,g,x);
    double v6 = compute_pyramid_area (c,a,g,e,x);

    if (fabs(v - (v1 + v2 + v3 + v4 +v5 + v6)) < EPS5)
    {
        return true;
    }
    return false;
}

/**
  * test if four points of hexahedron y are in hexahedron x
  */
bool GeometricComputations::hexahedron_in_hexahedron(p3d_point x1,p3d_point x2, p3d_point x3, p3d_point x4,
                              p3d_point x5, p3d_point x6, p3d_point x7, p3d_point x8,
                              p3d_point y1,p3d_point y2, p3d_point y3, p3d_point y4,
                              p3d_point y5, p3d_point y6, p3d_point y7, p3d_point y8)
{

    vector<p3d_point> t;
    t.push_back(y1);
    t.push_back(y2);
    t.push_back(y3);
    t.push_back(y4);
    t.push_back(y5);
    t.push_back(y6);
    t.push_back(y7);
    t.push_back(y8);

    int cd = 0;

    for (unsigned int i = 0; i < t.size(); i ++)
    {
        if(point_in_hexahedron(x1,x2,x3,x4,x5,x6,x7,x8,t.at(i)))
            cd++;
        if (cd>=4)
            return true;
    }
    return false;
}

/**  b----d
  * /    /|
  *a-f--c h
  *|/   |/
  *e----g
  *b and c should be opposite
  */
vector<p3d_point> GeometricComputations::getRandomPointsInHexahedron(double nbPoints, p3d_point a, p3d_point b, p3d_point c, p3d_point d, p3d_point e, p3d_point f, p3d_point g, p3d_point h)
{
    vector<p3d_point> res;

    p3d_point vectX, vectY, vectZ;

    vectX.x = b.x-a.x;
    vectX.y = b.y-a.y;
    vectX.z = b.z-a.z;

    vectY.x = c.x-a.x;
    vectY.y = c.y-a.y;
    vectY.z = c.z-a.z;

    vectZ.x = e.x-a.x;
    vectZ.y = e.y-a.y;
    vectZ.z = e.z-a.z;

    for (int i = 0; i<nbPoints; i++)
    {
        double raX = p3d_random(0,distBetweenPoints(a,b));
        double raY = p3d_random(0,distBetweenPoints(a,c));
        double raZ = p3d_random(0,distBetweenPoints(a,e));
        p3d_point p;
        p.x = a.x + raX*vectX.x + raY*vectY.x + raZ*vectZ.x;
        p.y = a.y + raX*vectX.y + raY*vectY.y + raZ*vectZ.y;
        p.z = a.z + raX*vectX.z + raY*vectY.z + raZ*vectZ.z;

        res.push_back(p);
    }

    return res;
}

p3d_point GeometricComputations::getClosestBBPointFromPoint(Robot* e, p3d_point s)
{
    p3d_oriented_BB *obj_OBB = p3d_get_oriented_BB_obj(e->getJoint(0)->getJointStruct()->o);
    vector<p3d_point> vpp;
    vpp.push_back(obj_OBB->BottomBackLeft);
    vpp.push_back(obj_OBB->BottomBackRight);
    vpp.push_back(obj_OBB->BottomFrontLeft);
    vpp.push_back(obj_OBB->BottomFrontRight);
    vpp.push_back(obj_OBB->UpBackLeft);
    vpp.push_back(obj_OBB->UpBackRight);
    vpp.push_back(obj_OBB->UpFrontLeft);
    vpp.push_back(obj_OBB->UpFrontRight);
    p3d_destroy_oriented_BB(obj_OBB);

//    cout << "s = ( " << s.x << " , " << s.y << " , " << s.z << " )" << endl;
    p3d_point p = vpp.at(0);
    double dist = distBetweenPoints(s,p);
    for (unsigned int i = 0; i < vpp.size(); i++)
    {
//        cout << "vpp.at("<< i << ") = ( " << vpp.at(i).x << " , " << vpp.at(i).y << " , " << vpp.at(i).z << " )" << endl;
        double dde = distBetweenPoints(s, vpp.at(i));
//        cout << "dde = " << dde << endl;
        if (dde < dist && (vpp.at(i).x != p.x || vpp.at(i).y != p.y || vpp.at(i).z != p.z))
        {
            dist = dde;
            p = vpp.at(i);
        }
//        cout << "dist = " << dist << endl;
    }
//    cout << "p = ( " << p.x << " , " << p.y << " , " << p.z << " )" << endl;
    return p;
}

ParallelepipedForm GeometricComputations::getBoxFromObjects(vector<Robot*> vr, Robot* gripper)
{
    p3d_point a,b,c,d,e,f,g,h;
    ParallelepipedForm pr = ParallelepipedForm(a,b,c,d,e,f,g,h);
    return pr;
}

double GeometricComputations::distConfs(tr1::shared_ptr<Configuration> q1, tr1::shared_ptr<Configuration> q2)
{
    double res = 1000;
    if (q1->getRobot()->getName() != q2->getRobot()->getName())
    {
        cout << "ERROR: not same robot in conf dist" << endl;
        return res;
    }
//    if (q1->getRobot()->getHriAgent()->type != HRI_PR2)
//    {
//        cout << "ERROR: only pr2 supported for now" << endl;
//        return res;
//    }

    tr1::shared_ptr<Configuration> qcurr = q1->getRobot()->getCurrentPos();

//    for (unsigned int i = 0; i < q1->getRobot()->getNumberOfJoints(); i++)
//    {
//        cout << "id = " << i << " name = " << q1->getRobot()->getJoint(i)->getName() << endl;
//    }
    vector<int> jointToCheck;
    if (q1->getRobot()->getHriAgent()->type == HRI_PR2)
    {
    //base
    jointToCheck.push_back(0);
    //torso
    jointToCheck.push_back(1);
    //head
    jointToCheck.push_back(2);
    jointToCheck.push_back(3);
    //rarm
    jointToCheck.push_back(5);
    jointToCheck.push_back(6);
    jointToCheck.push_back(7);
    jointToCheck.push_back(8);
    jointToCheck.push_back(9);
    jointToCheck.push_back(10);
    jointToCheck.push_back(11);
    //larm
    jointToCheck.push_back(15);
    jointToCheck.push_back(16);
    jointToCheck.push_back(17);
    jointToCheck.push_back(18);
    jointToCheck.push_back(19);
    jointToCheck.push_back(20);
    jointToCheck.push_back(21);
    }
    else if (q1->getRobot()->getHriAgent()->type == HRI_KUKAROBOT)
    {
        //base
        jointToCheck.push_back(0);
        //arm
        jointToCheck.push_back(1);
        jointToCheck.push_back(2);
        jointToCheck.push_back(3);
        jointToCheck.push_back(4);
        jointToCheck.push_back(5);
        jointToCheck.push_back(6);
        jointToCheck.push_back(7);
    }



    q1->getRobot()->setAndUpdate(*q1);
    vector<p3d_point> jointP1;
    for (unsigned int i = 0; i < jointToCheck.size(); i++)
    {
        p3d_point p;
        p.x = q1->getRobot()->getJoint(jointToCheck.at(i))->getJointStruct()->abs_pos[0][3];
        p.y = q1->getRobot()->getJoint(jointToCheck.at(i))->getJointStruct()->abs_pos[1][3];
        p.z = q1->getRobot()->getJoint(jointToCheck.at(i))->getJointStruct()->abs_pos[2][3];
        jointP1.push_back(p);
    }
    q2->getRobot()->setAndUpdate(*q2);
    vector<p3d_point> jointP2;
    for (unsigned int i = 0; i < jointToCheck.size(); i++)
    {
        p3d_point p;
        p.x = q2->getRobot()->getJoint(jointToCheck.at(i))->getJointStruct()->abs_pos[0][3];
        p.y = q2->getRobot()->getJoint(jointToCheck.at(i))->getJointStruct()->abs_pos[1][3];
        p.z = q2->getRobot()->getJoint(jointToCheck.at(i))->getJointStruct()->abs_pos[2][3];
        jointP2.push_back(p);
    }

    if (jointP1.size() != jointP2.size())
        return res;

    res = 0;
//    q1->printCompare(q2.get(), true);
    for (unsigned int i = 0; i < jointP1.size(); i++)
    {
        res += distBetweenPoints(jointP1.at(i), jointP2.at(i));
//        cout << "res = " << res << endl;
    }
    q1->getRobot()->setAndUpdate(*qcurr);

    return res;
}


/**
  * sample more finely a list of points
  *
  */
vector<p3d_point> GeometricComputations::getSampled(vector<p3d_point> vp, double sampleRate)
{
    vector<p3d_point> tmpVp;

    for (unsigned int i = 0; i<vp.size()-1; i++)
    {
        tmpVp.push_back(vp.at(i));
        double ds = sampleRate;
        double totDist = distBetweenPoints2D(vp.at(i),vp.at(i+1));

        while (ds < totDist)
        {
            double d = ds/totDist;
            p3d_point p1;
            p1.x = (1-d)*vp.at(i).x + d*vp.at(i+1).x;
            p1.y = (1-d)*vp.at(i).y + d*vp.at(i+1).y;
            p1.z = vp.at(i).z;
            tmpVp.push_back(p1);
            ds += sampleRate;
        }
    }
    tmpVp.push_back(vp.at(vp.size()-1));

    return tmpVp;
}


/**
  * vHP is the list of human positions (z is the orientation) and vp is the list of points to asses (can e only one point)
  *
  */
double GeometricComputations::computeCost(vector<p3d_point> vHP, vector<p3d_point> vp, double angleWeight, double distHWeight, double distWeight)
{
    double costMaxDist = 2.3;
    double costMaxAngle = M_PI/2;

    p3d_point pI = vp.at(0);

    double cost = 0;
    vector<p3d_point> tmpVp = getSampled(vp, distSample);

    for (unsigned int i = 0; i<tmpVp.size(); i++)
    {
        p3d_point trajP = tmpVp.at(i);
        double angleCost =0;
        double distHCost =0;
        for(unsigned int j = 0; j<vHP.size(); j++)
        {
            p3d_point HP = vHP.at(j);
            double dist = distBetweenPoints2D(trajP,HP);
            double c1;
            if (dist > costMaxDist)
                c1 = 0;
            else
                c1 = -dist/costMaxDist +1;


            double angle = fabs( acos( (cos(HP.z) * trajP.x + sin(HP.z) * trajP.y)
                                       / (sqrt(trajP.x*trajP.x + trajP.y*trajP.y) * sqrt(cos(HP.z)*cos(HP.z) + sin(HP.z)*sin(HP.z))) ));
            double dp;
            if (dist < costMaxDist && angle >= costMaxAngle)
                dp = angle*c1;
            else
                dp = 0;
//            cost += (c1+dp)/2;
            angleCost += dp;
            distHCost += c1;
//            cout << " cost = " << (c1+dp)/2 << endl;
        }
        cost += (distBetweenPoints2D(trajP,pI)*distWeight + angleCost*angleWeight + distHCost*distHWeight)/(distWeight+angleWeight+distHWeight);
//        cout << "trajP (" << trajP.x << " , " << trajP.y << " , " << trajP.z << ")" << endl;
//        cout << "cost = " << cost << " i " << i << endl;
        pI = trajP;
    }

    return cost;
}

vector<p3d_point> GeometricComputations::smoothTraj(vector<p3d_point> origTraj, vector<p3d_point> vHP, Robot* r, int nbIterMax, double gainLim,
                                                    double trajDistWeight, double distToHWeight, double angleWeight)
{

    cout << "start smoothing" << endl;
    //retriving total dist of the traj
    double origDist = 0;
    for (unsigned int i = 1; i < origTraj.size(); i++)
    {
        origDist += distBetweenPoints2D(origTraj.at(i-1), origTraj.at(i));
    }
//    double tmpDist = origDist;
    double origCost = computeCost(vHP,origTraj,angleWeight,trajDistWeight,distToHWeight);
    double tmpCost = origCost;

    double gain = 0;

    vector<p3d_point> newTraj = origTraj;
    vector<p3d_point> tmpNewTraj = origTraj;

    //removing uselesspoints
    tmpNewTraj.clear();
    tmpNewTraj.push_back(newTraj.at(0));
    for (unsigned int i = 1; i < newTraj.size()-1; i++)
    {
        if (!point_in_segment2d(newTraj.at(i-1), newTraj.at(i+1), newTraj.at(i)))
        {
            tmpNewTraj.push_back(newTraj.at(i));
        }
    }

    tmpNewTraj.push_back(newTraj.at(newTraj.size()-1));
    newTraj = tmpNewTraj;

    //shortcut loop. stop when max iteraction reached
    int nbIter = -1;
    while (nbIter < nbIterMax)
    {
        nbIter++;

//        cout << "\n\naaa 1 " << nbIter << endl;
        //compute curr size
        origDist = 0;
        for (unsigned int i = 1; i < newTraj.size(); i++)
        {
            origDist += distBetweenPoints2D(newTraj.at(i-1), newTraj.at(i));
        }


        //get random dist to get corresponding points
        double d1 = p3d_random(0,origDist);
        double d2 = p3d_random(d1,origDist);

//        cout << "d1 = " << d1 << endl;
//        cout << "d2 = " << d2 << endl;
//        cout << "origDist = " << origDist << endl;

        //get corresponding points, and the previous and next point in the traj.
        int id1 = -1, id2 = -1;
        p3d_point p1,p2;
        double tmpDist = 0;
        bool foundP1 = false, foundP2=false;
        vector<p3d_point> tmpOrigTraj;

        for (unsigned int i = 0; i < newTraj.size()-1; i++)
        {
//            cout << i << " i(" << newTraj.at(i).x <<" , " << newTraj.at(i).y << " )" << endl;
            double distToStart = tmpDist;
            double newDist = distBetweenPoints2D(newTraj.at(i), newTraj.at(i+1));
            tmpDist += newDist;
//            cout << "tmpDist = " << tmpDist << endl;
            if (d1 <= tmpDist && !foundP1)
            {
                double d = (d1-distToStart)/newDist;
                p1.x = (1-d)*newTraj.at(i).x + d*newTraj.at(i+1).x;
                p1.y = (1-d)*newTraj.at(i).y + d*newTraj.at(i+1).y;
                p1.z = newTraj.at(i).z;
                id1 = i;
                foundP1 = true;
//                cout << " p1(" << p1.x <<" , " << p1.y << " )" << endl;
//                tmpOrigTraj.push_back(newTraj.at(i));
            }
            if (d2 <= tmpDist && !foundP2)
            {
                double d = (d2-distToStart)/newDist;
                p2.x = (1-d)*newTraj.at(i).x + d*newTraj.at(i+1).x;
                p2.y = (1-d)*newTraj.at(i).y + d*newTraj.at(i+1).y;
                p2.z = newTraj.at(i+1).z;
                id2 = i+1;
                foundP2 = true;
//                cout << " p2(" << p2.x <<" , " << p2.y << " )" << endl;
//                cout << i+1 << " i+1(" << newTraj.at(i+1).x <<" , " << newTraj.at(i+1).y << " )" << endl;
            }

            if (foundP1)
            {
                tmpOrigTraj.push_back(newTraj.at(i));
            }
            if (foundP2)
            {
                tmpOrigTraj.push_back(newTraj.at(i+1));
                break;
            }
        }
        if (!(foundP1 && foundP2))
            continue;

        //compute if any gain

        vector<p3d_point> tmpTraj;
        tmpTraj.push_back(newTraj.at(id1));
        tmpTraj.push_back(p1);
        tmpTraj.push_back(p2);
        tmpTraj.push_back(newTraj.at(id2));


        double oneStepOrigCost = computeCost(vHP,tmpOrigTraj,angleWeight,trajDistWeight,distToHWeight);
        double oneStepNewCost = computeCost(vHP,tmpTraj,angleWeight,trajDistWeight,distToHWeight);
        if (oneStepOrigCost < oneStepNewCost)
            continue;

//        currNavTraj = tmpTraj;

//        g3d_draw_allwin_active();
//        sleep(1);

        //testing collisions

        vector<p3d_point> collisionCheckingVect = getSampled(tmpTraj,distSample);
        shared_ptr<Configuration> q_cur (new Configuration(*r->getCurrentPos()));
        shared_ptr<Configuration> q  (new Configuration(*r->getCurrentPos()));
        bool collides = false;
        for (unsigned int i= 0; i < collisionCheckingVect.size(); i++)
        {
            (*q)[6] = collisionCheckingVect.at(i).x;
            (*q)[7] = collisionCheckingVect.at(i).y;
            (*q)[11] = collisionCheckingVect.at(i).z;
            r->setAndUpdate(*q);
//            g3d_draw_allwin_active();
//            sleep(1);
            if (r->isInCollisionWithOthersAndEnv())
            {
                cout << "aaaaaaaaaa" << endl;
                collides = true;
                break;
            }
        }
        r->setAndUpdate(*q_cur);
        if (collides)
            continue;



        //replacing in traj
        tmpNewTraj.clear();
        for(unsigned int i = 0; i < newTraj.size(); i++)
        {
            tmpNewTraj.push_back(newTraj.at(i));
            if (i == id1)
            {
                for (unsigned int j = 1; j < tmpTraj.size(); j++)
                {
                    tmpNewTraj.push_back(tmpTraj.at(j));
                }
                i = id2;
            }
        }
        newTraj = tmpNewTraj;



        //replacing gain
        tmpCost = tmpCost - (oneStepOrigCost - oneStepNewCost);
        gain = 1-tmpCost/origCost;

        //go out if the gain work
        if (gainLim != -1 && gain < gainLim)
            break;


    }

    cout << "gain = " << gain <<endl;
    cout << "origCost = " << origCost <<endl;
    cout << "newCost = " << tmpCost <<endl;





//    removing uselesspoints
    tmpNewTraj.clear();
    tmpNewTraj.push_back(newTraj.at(0));
    for (unsigned int i = 1; i < newTraj.size()-1; i++)
    {
        if (!point_in_segment2d(newTraj.at(i-1), newTraj.at(i+1), newTraj.at(i)))
        {
            tmpNewTraj.push_back(newTraj.at(i));
        }
    }

    tmpNewTraj.push_back(newTraj.at(newTraj.size()-1));
    newTraj = tmpNewTraj;

    //putting the right rotations
    tmpNewTraj.clear();
    tmpNewTraj.push_back(newTraj.at(0));
    p3d_point p1 = newTraj.at(0);
    for (unsigned int i = 1; i< newTraj.size()-1; i++)
    {
        p3d_point p2;
        p1.x = newTraj.at(i).x;
        p1.y = newTraj.at(i).y;
        p2.x = newTraj.at(i).x;
        p2.y = newTraj.at(i).y;
        p2.z = atan2((newTraj.at(i+1).y - p1.y),(newTraj.at(i+1).x - p1.x));

        tmpNewTraj.push_back(p1);
        tmpNewTraj.push_back(p2);
        p1.z = p2.z;
    }
    tmpNewTraj.push_back(newTraj.at(newTraj.size()-1));
    newTraj = tmpNewTraj;



    return newTraj;
}


//#################### rectangle ######################

RectangleForm::RectangleForm(p3d_point a, p3d_point b, p3d_point c, p3d_point d, p3d_matrix3 norm)
{
    _a = a;
    _b = b;
    _c = c;
    _d = d;
    p3d_mat3Copy(norm,_norm);
    type = "rectangle";
    isSupport = true;
}

bool RectangleForm::isPointInForm(p3d_point p)
{
    GeometricComputations GC = GeometricComputations();

    return GC.point_in_rect2(_a,_b,_c,_d,p);
}

double RectangleForm::getDistanceFromPoint(p3d_point p)
{
    double dist = 1000;

    double tmp;

    tmp = sqrt(pow(p.x - _a.x,2) + pow(p.y - _a.y,2) + pow(p.z - _a.z,2));
    if (tmp < dist)
        dist = tmp;
    tmp = sqrt(pow(p.x - _b.x,2) + pow(p.y - _b.y,2) + pow(p.z - _b.z,2));
    if (tmp < dist)
        dist = tmp;
    tmp = sqrt(pow(p.x - _c.x,2) + pow(p.y - _c.y,2) + pow(p.z - _c.z,2));
    if (tmp < dist)
        dist = tmp;
    tmp = sqrt(pow(p.x - _d.x,2) + pow(p.y - _d.y,2) + pow(p.z - _d.z,2));
    if (tmp < dist)
        dist = tmp;

    return dist;
}

double RectangleForm::getPlannarDistanceFromPoint(p3d_point p)
{

    double dist = 1000;

    double tmp;

    tmp = sqrt(pow(p.x - _a.x,2) + pow(p.y - _a.y,2));
    if (tmp < dist)
        dist = tmp;
    tmp = sqrt(pow(p.x - _b.x,2) + pow(p.y - _b.y,2));
    if (tmp < dist)
        dist = tmp;
    tmp = sqrt(pow(p.x - _c.x,2) + pow(p.y - _c.y,2));
    if (tmp < dist)
        dist = tmp;
    tmp = sqrt(pow(p.x - _d.x,2) + pow(p.y - _d.y,2));
    if (tmp < dist)
        dist = tmp;

    return dist;
}

p3d_point RectangleForm::getRandomPointInForm()
{
    GeometricComputations GC = GeometricComputations();
    return GC.getRandomPointInRectangle(_a,_b,_c,_d);
}

vector<p3d_point> RectangleForm::getFormGrid(double sampleRate)
{
    vector<p3d_point> t;
    GeometricComputations GC = GeometricComputations();
    int xlenght = GC.distBetweenPoints(_a,_b)/sampleRate;
    int ylenght = GC.distBetweenPoints(_a,_c)/sampleRate;
    double normeAB = GC.distBetweenPoints2D(_a,_b);
    double normeAC = GC.distBetweenPoints2D(_a,_c);

    p3d_point vectAB;
    vectAB.x = (_b.x - _a.x)/normeAB;
    vectAB.y = (_b.y - _a.y)/normeAB;
    p3d_point vectAC;
    vectAC.x = (_c.x - _a.x)/normeAC;
    vectAC.y = (_c.y - _a.y)/normeAC;


    for (int i = 1; i <= xlenght; i++)
    {
        for (int j = 1; j <= ylenght; j++)
        {
            p3d_point p;
            p.x = i*sampleRate*vectAB.x + j*sampleRate*vectAC.x + _a.x;
            p.y = i*sampleRate*vectAB.y + j*sampleRate*vectAC.y + _a.y;
            p.z = _a.z;
            t.push_back(p);
        }
    }

    return t;
}

//#################### Triangle ######################

TriangleForm::TriangleForm(p3d_point a, p3d_point b, p3d_point c, p3d_matrix3 norm)
{
    _a = a;
    _b = b;
    _c = c;
    p3d_mat3Copy(norm,_norm);
    type = "triangle";
    isSupport = true;
}

bool TriangleForm::isPointInForm(p3d_point p)
{
    cout << "not implemented yet" << endl;
    return false;
}

double TriangleForm::getDistanceFromPoint(p3d_point p)
{
    cout << "not implemented yet" << endl;
    return 1000;
}

double TriangleForm::getPlannarDistanceFromPoint(p3d_point p)
{
    cout << "not implemented yet" << endl;
    return 1000;
}

p3d_point TriangleForm::getRandomPointInForm()
{
    cout << "not implemented yet" << endl;
    p3d_point p;
    p.x = 0;
    p.y = 0;
    p.z = 0;
    return p;
}

vector<p3d_point> TriangleForm::getFormGrid(double sampleRate)
{
    vector<p3d_point> t;
    return t;
}


//#################### Circle ######################


CircleForm::CircleForm(p3d_point c, double r, p3d_matrix3 norm)
{
    _c = c;
    _r = r;
    p3d_mat3Copy(norm,_norm);
    type = "circle";
    isSupport = true;
}

bool CircleForm::isPointInForm(p3d_point p)
{
    cout << "not implemented yet" << endl;
    return false;
}

double CircleForm::getDistanceFromPoint(p3d_point p)
{
    cout << "not implemented yet" << endl;
    return 1000;
}

double CircleForm::getPlannarDistanceFromPoint(p3d_point p)
{
    cout << "not implemented yet" << endl;
    return 1000;
}

p3d_point CircleForm::getRandomPointInForm()
{
    cout << "not implemented yet" << endl;
    p3d_point p;
    p.x = 0;
    p.y = 0;
    p.z = 0;
    return p;
}

vector<p3d_point> CircleForm::getFormGrid(double sampleRate)
{
    vector<p3d_point> t;
    return t;
}




//#################### Parallelepiped ######################

ParallelepipedForm::ParallelepipedForm(p3d_point a, p3d_point b, p3d_point c, p3d_point d,
                                       p3d_point e, p3d_point f, p3d_point g, p3d_point h)
{
    _a = a;
    _b = b;
    _c = c;
    _d = d;
    _e = e;
    _f = f;
    _g = g;
    _h = h;
    type = "parallelepiped";
    isSupport = true;
}

bool ParallelepipedForm::isPointInForm(p3d_point p)
{
    cout << "not implemented yet" << endl;
    return false;
}

double ParallelepipedForm::getDistanceFromPoint(p3d_point p)
{
    cout << "not implemented yet" << endl;
    return 1000;
}

double ParallelepipedForm::getPlannarDistanceFromPoint(p3d_point p)
{
    cout << "not implemented yet" << endl;
    return 1000;
}

p3d_point ParallelepipedForm::getRandomPointInForm()
{
    cout << "not implemented yet" << endl;
    p3d_point p;
    p.x = 0;
    p.y = 0;
    p.z = 0;
    return p;
}

vector<p3d_point> ParallelepipedForm::getFormGrid(double sampleRate)
{
    vector<p3d_point> t;
    GeometricComputations GC = GeometricComputations();
    int xlenght = GC.distBetweenPoints(_a,_b)/sampleRate;
    int ylenght = GC.distBetweenPoints(_a,_c)/sampleRate;
    int zlenght = GC.distBetweenPoints(_a,_e)/sampleRate;
    double normeAB = GC.distBetweenPoints(_a,_b);
    double normeAC = GC.distBetweenPoints(_a,_c);
    double normeAE = GC.distBetweenPoints(_a,_e);

    p3d_point vectAB;
    vectAB.x = (_b.x - _a.x)/normeAB;
    vectAB.y = (_b.y - _a.y)/normeAB;
    vectAB.z = (_b.z - _a.z)/normeAB;
    p3d_point vectAC;
    vectAC.x = (_c.x - _a.x)/normeAC;
    vectAC.y = (_c.y - _a.y)/normeAC;
    vectAC.z = (_c.z - _a.z)/normeAC;
    p3d_point vectAE;
    vectAE.x = (_e.x - _a.x)/normeAE;
    vectAE.y = (_e.y - _a.y)/normeAE;
    vectAE.z = (_e.z - _a.z)/normeAE;

    for (int i = 1; i <= xlenght; i++)
    {
        for (int j = 1; j <= ylenght; j++)
        {
            for (int k = 1 ; k <= zlenght; k++)
            {
                p3d_point p;
                p.x = i*sampleRate*vectAB.x + j*sampleRate*vectAC.x + k*sampleRate*vectAE.x + _a.x;
                p.y = i*sampleRate*vectAB.y + j*sampleRate*vectAC.y + k*sampleRate*vectAE.y + _a.y;
                p.z = i*sampleRate*vectAB.z + j*sampleRate*vectAC.z + k*sampleRate*vectAE.z + _a.z;
                t.push_back(p);
            }
        }
    }





    return t;
}





