#include "mightabilitiesGrids.hpp"
#include "Graphic-pkg.h"

extern MightabilitiesGrid* currentMightability;

#define samplingR 0.05
#define limitAngleReach 90




vector<MightabilitiesCell*> typeCellLists::getListFromType(typeCells tc)
{
    for(unsigned int i = 0; i < lists.size(); i++)
        if (lists.at(i).first == tc)
            return lists.at(i).second;
    vector<MightabilitiesCell*> c;
    return c;
}

void typeCellLists::addCellTo(MightabilitiesCell* c, typeCells tc)
{
    int id = -1;
    for (unsigned int i = 0; i < lists.size(); i++)
    {
        if (lists.at(i).first == tc )
        {
            id = i;
            break;
        }
    }
    if (id != -1)
    {
        lists.at(id).second.push_back(c);
    }
    else
    {
        vector<MightabilitiesCell*> tmp;
        tmp.push_back(c);
        pair<typeCells, vector<MightabilitiesCell*> > p;
        p.first = tc;
        p.second = tmp;
        lists.push_back(p);
    }
}



MightabilitiesGrid::MightabilitiesGrid() : ThreeDGrid()
{
    initVariables();
}


MightabilitiesGrid::MightabilitiesGrid( Eigen::Vector3i size, std::vector<double> envSize) : ThreeDGrid(size, envSize)
{
    initVariables();
}

MightabilitiesGrid::MightabilitiesGrid( double samplingRate, std::vector<double> envSize) : ThreeDGrid(samplingRate,envSize)
{
    initVariables();
}

MightabilitiesGrid::MightabilitiesGrid( const ThreeDGrid& grid) : ThreeDGrid(grid)
{
    initVariables();
}

 void MightabilitiesGrid::initVariables()
 {
     _samplingRate = 0.05;
     _visball = NULL;
     setVisBall();
 }

void MightabilitiesGrid::init(Eigen::Vector3i size,    std::vector<double> envSize)
{
    _nbCellsX = size[0];
    _nbCellsY = size[1];
    _nbCellsZ = size[2];

    _cellSize[0] = (envSize.at(1) - envSize.at(0)) / _nbCellsX ;
    _cellSize[1] = (envSize.at(3) - envSize.at(2)) / _nbCellsY ;
    _cellSize[2] = (envSize.at(5) - envSize.at(4)) / _nbCellsZ ;

    _originCorner[0] = envSize.at(0);
    _originCorner[0] = envSize.at(2);
    _originCorner[0] = envSize.at(4);
}

void MightabilitiesGrid::init(double samplingRate ,    std::vector<double> envSize)
{
    for(unsigned int i= 0; i< envSize.size() ; i++)
    {
        cout << envSize.at(i) << " ";
    }
    cout << endl;

    if(((int)samplingRate) != 0 )
    {
        if( ( ((int) (envSize.at(1) - envSize.at(0))) % (int)samplingRate ) != 0 )
        {
            cout << "ThreeDGrid Warning : not good X disctretization " << endl;
        }

        if( ( ((int) (envSize.at(3) - envSize.at(2))) % (int)samplingRate ) != 0 )
        {
            cout << "ThreeDGrid Warning : not good Y disctretization " << endl;
        }

        if( ( ((int) (envSize.at(5) - envSize.at(4))) % (int)samplingRate ) != 0 )
        {
            cout << "ThreeDGrid Warning : not good Z disctretization " << endl;
        }
    }

    //    _cellSize.push_back( (envSize.at(1) - envSize.at(0))/pace );
    //    _cellSize.push_back( (envSize.at(3) - envSize.at(2))/pace );
    //    _cellSize.push_back( (envSize.at(5) - envSize.at(4))/pace );

    _cellSize[0] =  samplingRate ;
    _cellSize[1] =  samplingRate ;
    _cellSize[2] =  samplingRate ;

    _nbCellsX =  (envSize.at(1) - envSize.at(0)) / samplingRate ;
    _nbCellsY =  (envSize.at(3) - envSize.at(2)) / samplingRate ;
    _nbCellsZ =  (envSize.at(5) - envSize.at(4)) / samplingRate ;

    cout << " _nbCellsX = " << _nbCellsX << endl;
    cout << " _nbCellsY = " << _nbCellsY << endl;
    cout << " _nbCellsZ = " << _nbCellsZ << endl;

    _originCorner[0] = envSize.at(0);
    _originCorner[1] = envSize.at(2);
    _originCorner[2] = envSize.at(4);

    cout << "_originCorner[0] = " << _originCorner[0] <<  endl;
    cout << "_originCorner[1] = " << _originCorner[1] <<  endl;
    cout << "_originCorner[2] = " << _originCorner[2] <<  endl;
}

void MightabilitiesGrid::setVisBall()
{
    _visball = global_Project->getActiveScene()->getRobotByNameContaining("VISBALL");
    if (!_visball)
        cout << "#################################\nMightabilities not computed: no VISBALL\n##########################################" << endl;
}

void MightabilitiesGrid::placeVisball(MightabilitiesCell* c)
{
    Vector3d center = c->getCenter();
    if (!_visball)
        return;

    confPtr_t q = _visball->getCurrentPos();
    (*q)[6] = center[0];
    (*q)[7] = center[1];
    (*q)[8] = center[2];
    if (c->isItSupport())
        (*q)[8] = (*q)[8] +0.05;

    _visball->setAndUpdate(*q);
}

void MightabilitiesGrid::placeVisball(p3d_point p)
{
    if (!_visball)
        return;

    confPtr_t q = _visball->getCurrentPos();
    (*q)[6] = p.x;
    (*q)[7] = p.y;
    (*q)[8] = p.z;
    _visball->setAndUpdate(*q);
}





void MightabilitiesGrid::computeMights(vector<Robot*> agents, vector<Robot*> supports, vector<typeCells> abilities)
{

    clock_t begin = clock();
    _cellSize[0] = _samplingRate;
    _cellSize[1] = _samplingRate ;
    _cellSize[2] = _samplingRate ;

//    vector<MightabilitiesCell*> allCells;
    //TODO passing only onece in the next loop!

    bool hasSupp = false, has3d = false;

    for (unsigned int j = 0; j < abilities.size(); j++)
    {
        if (abilities.at(j) == reachDistSupp
                || abilities.at(j) == reachColSupp
                || abilities.at(j) == visDistSupp
                || abilities.at(j) == visColSupp)
        {
            hasSupp = true;
        }
        else if (abilities.at(j) == reachDist
                 || abilities.at(j) == reachCol
                 || abilities.at(j) == visDist
                 || abilities.at(j) == visCol)
        {
            has3d = true;
        }
    }

    if (hasSupp)
    {
        for (unsigned int i = 0; i < supports.size(); i++)
        {
            vector<MightabilitiesCell*> cells = getCellsFromEntity(supports.at(i));
            for (unsigned int j = 0; j< cells.size(); j++)
            {
                for (unsigned int k = 0; k<agents.size(); k++)
                {
                    cells.at(j)->addAgent(agents.at(k));
                }
            }
        }
    }
    if (has3d)
    {
        for (unsigned int i = 0; i < agents.size(); i++)
        {
            getCellsFromEntity(agents.at(i));
//                allCells = mergeVectors(), allCells);
        }
    }


    if (!_visball)
        return;

    for (unsigned int j = 0; j < abilities.size(); j++)
    {
        if (abilities.at(j) == agentC )
            continue;
        for (unsigned int l = 0; l < agents.size(); l++)
        {
            bool t1 = true, t2=true;
            if (!(agents.at(l) && agents.at(l)->getHriAgent() && agents.at(l)->getHriAgent()->manip))
                t1=false;
            if (!(agents.at(l) && agents.at(l)->getHriAgent() && agents.at(l)->getHriAgent()->perspective && agents.at(l)->getHriAgent()->perspective->camjoint))
                t2=false;

            for (unsigned int k = 0; k< _cells.size(); k++)
            {
                MightabilitiesCell* c = dynamic_cast<MightabilitiesCell*>( BaseGrid::_cells.at(k));
                if (!c->belongToTypeC(agents.at(l),agentC))
                    continue;
                placeVisball(c);
                //            Robot* r = computeVisballCollision();
                //            g3d_draw_allwin_active();
                //            sleep(3);
                if (computeVisballCollision())
                    continue;

                if (t1)
                {
                    if (abilities.at(j) == reachDist || abilities.at(j) == reachDistSupp)
                        computeReachDist(agents.at(l),c);
                    if (abilities.at(j) == reachCol || abilities.at(j) == reachColSupp)
                        computeReachCol(agents.at(l),c);
                }
                if (t2)
                {
                    if (abilities.at(j) == visDist || abilities.at(j) == visDistSupp)
                        computeVisDist(agents.at(l),c);
                    if (abilities.at(j) == visCol || abilities.at(j) == visColSupp)
                        computeVisCol(agents.at(l),c);
                }
            }
        }
    }
    if (_allWorldCells.size()>0)
        placeVisball(_allWorldCells.at(0));

    double timeT = (double)(clock() - begin)/CLOCKS_PER_SEC;
    cout << "compute mightabilities in " << timeT << " s" << endl;
}

vector<MightabilitiesCell*> MightabilitiesGrid::getCellsFromEntity(Robot* entity)
{
    vector<MightabilitiesCell*> allCells;
    if (entity->isAgent())
    {
//        manip->reach_tolerance
        double xmin = P3D_HUGE,xmax = -P3D_HUGE, ymin = P3D_HUGE,ymax = -P3D_HUGE, zmin = P3D_HUGE,zmax = -P3D_HUGE;

        if (!entity->getHriAgent()
                || !entity->getHriAgent()->manip)
        {
            cout << "ERROR: agent with no hri_agent" << endl;
            return allCells;
        }

        double reachT = entity->getHriAgent()->manip->reach_tolerance;
        vector<p3d_point> allShoulders;
        for (int i = 0; i < entity->getHriAgent()->manip->shoulder_joint_ids_no; i++)
        {
            p3d_point shoulderP;
            shoulderP.x = entity->getJoint(entity->getHriAgent()->manip->shoulder_joint_ids[i])->getJointStruct()->abs_pos[0][3];
            shoulderP.y = entity->getJoint(entity->getHriAgent()->manip->shoulder_joint_ids[i])->getJointStruct()->abs_pos[1][3];
            shoulderP.z = entity->getJoint(entity->getHriAgent()->manip->shoulder_joint_ids[i])->getJointStruct()->abs_pos[2][3];
            allShoulders.push_back(shoulderP);
            if (xmin > shoulderP.x - reachT)
                xmin = shoulderP.x - reachT;
            if (xmax < shoulderP.x + reachT)
                xmax = shoulderP.x + reachT;
            if (ymin > shoulderP.y - reachT)
                ymin = shoulderP.y - reachT;
            if (ymax < shoulderP.y + reachT)
                ymax = shoulderP.y + reachT;
            if (zmin > shoulderP.z - reachT)
                zmin = shoulderP.z - reachT;
            if (zmax < shoulderP.z + reachT)
                zmax = shoulderP.z + reachT;
        }

        p3d_point shoulderCenter;
        shoulderCenter.x = 0; shoulderCenter.y = 0; shoulderCenter.z = 0;
        for (unsigned int j = 0; j < allShoulders.size(); j++)
        {
            shoulderCenter.x += allShoulders.at(j).x;
            shoulderCenter.y += allShoulders.at(j).y;
            shoulderCenter.z += allShoulders.at(j).z;
        }
        shoulderCenter.x /= allShoulders.size(); shoulderCenter.y /= allShoulders.size(); shoulderCenter.z /= allShoulders.size();


        if (_allWorldCells.size() == 0)
        {
            vector<double> vb = global_Project->getActiveScene()->getBounds();
            p3d_point a,b,c,d,e,f,g,h;
            a.x = vb[0]; a.y = vb[2]; a.z = vb[4];
            b.x = vb[0]; b.y = vb[3]; b.z = vb[4];
            c.x = vb[1]; c.y = vb[2]; c.z = vb[4];
            d.x = vb[1]; d.y = vb[3]; d.z = vb[4];
            e.x = vb[0]; e.y = vb[2]; e.z = vb[5];
            f.x = vb[0]; f.y = vb[3]; f.z = vb[5];
            g.x = vb[1]; g.y = vb[2]; g.z = vb[5];
            h.x = vb[1]; h.y = vb[3]; h.z = vb[5];

            ParallelepipedForm PP = ParallelepipedForm(a,b,c,d,e,f,g,h);

            vector<p3d_point> allWorldPoints = PP.getFormGrid(samplingR);


            for (unsigned int i = 0; i < allWorldPoints.size(); i++)
            {
                Vector3d corner;
                corner[0] = allWorldPoints.at(i).x - samplingR/2;
                corner[1] = allWorldPoints.at(i).y - samplingR/2;
                corner[2] = allWorldPoints.at(i).z - samplingR/2;
                MightabilitiesCell* tmp = new MightabilitiesCell(i,corner,NULL);
                _allWorldCells.push_back(tmp);
            }
        }

        p3d_vector3 a,b ,cross;
    //    double a[3], b[3];
        a[0] = entity->getJoint(0)->getJointStruct()->abs_pos[0][0];
        a[1] = entity->getJoint(0)->getJointStruct()->abs_pos[1][0];
        a[2] = entity->getJoint(0)->getJointStruct()->abs_pos[2][0];







//        vector<p3d_point> allPoints;// = PP.getFormGrid(_samplingRate);
        for (unsigned int i = 0; i < _allWorldCells.size(); i++)
        {
            MightabilitiesCell* c = _allWorldCells.at(i);
            c->setGrid(this);
            Vector3d center = c->getCenter();
            if (xmin < center[0] && center[0] < xmax
                    && ymin < center[1] && center[1] < ymax
                    && zmin < center[2] && center[2] < zmax)
            {


                b[0] = center[0] - shoulderCenter.x;
                b[1] = center[1] - shoulderCenter.y;
                b[2] = center[2] - shoulderCenter.z;
                p3d_vectXprod(a,b,cross);

                double angle = asin(p3d_vectNorm(cross)/(p3d_vectNorm(a)*p3d_vectNorm(b)));
                if (p3d_vectDotProd(a,b) < 0)
                    angle = M_PI - angle;


                if (angle < limitAngleReach*M_PI/180)
                {
                    addCellToList(c,entity,agentC);
                    if (!c->belongToAtLeastOneAgent())
                        _cells.push_back(c);
                    c->addAgent(entity);
                }
            }
        }


//        int t = _cells.size();
//        _cells.resize(allPoints.size()+t); //TODO multiple supports
//        for (unsigned int k = 0; k<allPoints.size();k++)
//        {
//            Eigen::Vector3d corner;
//            corner[0] = allPoints.at(k).x - _samplingRate/2;
//            corner[1] = allPoints.at(k).y - _samplingRate/2;
//            corner[2] = allPoints.at(k).z - _samplingRate/2;
//            MightabilitiesCell* ptrCell = new MightabilitiesCell( k+t, corner , this );
////            API::ThreeDCell* ptrCell = createNewCell(k, allPoints.at(k).x, allPoints.at(k).y, allPoints.at(k).z);
//            _cells[k+t] = (API::ThreeDCell*)ptrCell;
//        }


//        std::vector<double> envSize;
//        envSize.push_back(xmin); envSize.push_back(xmax);
//        envSize.push_back(ymin); envSize.push_back(ymax);
//        envSize.push_back(zmin); envSize.push_back(zmax);
//        
//        init(_samplingRate,envSize);
//        createAllCells(); // TODO multiple agents
//        for (unsigned int j = 0; j < _cells.size(); j++)
//        {
//            allCells.push_back(dynamic_cast<MightabilitiesCell*>( BaseGrid::_cells.at(j)));
//        }
    }
    else if (entity->getObjectRob() && entity->getObjectRob()->isSupport())
    {
        vector<GeometricForm*> vG = entity->getObjectRob()->getSupports();

        vector<p3d_point> allPoints;
        for (unsigned int i = 0; i < vG.size(); i++)
        {
            vector<p3d_point> tmp = vG.at(i)->getFormGrid(_samplingRate);
            for (unsigned int j = 0; j < tmp.size(); j++)
                allPoints.push_back(tmp.at(j));
        }

        int t = _cells.size();
        _cells.resize(allPoints.size()+t); //TODO multiple supports
        for (unsigned int k = 0; k<allPoints.size();k++)
        {

            Eigen::Vector3d corner;
            corner[0] = allPoints.at(k).x - _samplingRate/2;
            corner[1] = allPoints.at(k).y - _samplingRate/2;
            corner[2] = allPoints.at(k).z - _samplingRate/2;
            MightabilitiesCell* ptrCell = new MightabilitiesCell( k+t, corner , this );
            ptrCell->setObject(entity);
            ptrCell->setIsSupport(true);
//            API::ThreeDCell* ptrCell = createNewCell(k, allPoints.at(k).x, allPoints.at(k).y, allPoints.at(k).z);
            _cells[k+t] = (API::ThreeDCell*)ptrCell;
        }
        for (unsigned int j = 0; j < _cells.size(); j++)
        {
            allCells.push_back(dynamic_cast<MightabilitiesCell*>( BaseGrid::_cells.at(j)));
        }

    }
    else
    {
        cout << "Warning: the object " << entity->getName() << " is neither an agent nor a support" << endl;
    }
    return allCells;
}

vector<MightabilitiesCell*> MightabilitiesGrid::mergeVectors(vector<MightabilitiesCell*> c1,vector<MightabilitiesCell*> c2)
{
    if (c1.size() == 0)
        return c2;
    if (c2.size() == 0)
        return c1;
    vector<MightabilitiesCell*> allCells = c1;
    for (unsigned int i = 0; i< c2.size(); i++)
    {
        bool isAlreadyIn = false;
        for (unsigned int j = 0; j < c1.size(); j++)
        {
            if (c1.at(j)->getIndex() == c2.at(i)->getIndex())
            {
                isAlreadyIn = true;
                break;
            }
        }
        if (!isAlreadyIn)
            allCells.push_back(c2.at(i));
    }

    return allCells;

}

void MightabilitiesGrid::computeReachDist(Robot* ag, MightabilitiesCell* cells)
{
    GeometricComputations GC = GeometricComputations();
    p3d_point p1;
    Vector3d center = cells->getCenter();
    p1.x = center[0];
    p1.y = center[1];
    p1.z = center[2];

    for (int i = 0; i < ag->getHriAgent()->manip->shoulder_joint_ids_no; i ++)
    {
        p3d_point shoulderP;
        shoulderP.x = ag->getJoint(ag->getHriAgent()->manip->shoulder_joint_ids[i])->getJointStruct()->abs_pos[0][3];
        shoulderP.y = ag->getJoint(ag->getHriAgent()->manip->shoulder_joint_ids[i])->getJointStruct()->abs_pos[1][3];
        shoulderP.z = ag->getJoint(ag->getHriAgent()->manip->shoulder_joint_ids[i])->getJointStruct()->abs_pos[2][3];
        double dist = GC.distBetweenPoints(p1,shoulderP);
//        cout << "dist = " << dist << endl;
        if (dist < ag->getHriAgent()->manip->reach_tolerance )
        {
            cells->addTypeC(ag,reachDist);
            addCellToList(cells, ag, reachDist);
            if (cells->isItSupport())
            {
                cells->addTypeC(ag,reachDistSupp);
                addCellToList(cells, ag, reachDistSupp);
            }

            break;
        }
    }
}

//computeReachDist needs to be called before
void MightabilitiesGrid::computeReachCol(Robot* ag, MightabilitiesCell* cells)
{
//    if (!(ag && ag->getHriAgent() && ag->getHriAgent()->manip) && cells->belongToTypeC(ag,agentC))
//        return;

    if (!cells->belongToTypeC(ag,reachDist) && !cells->belongToTypeC(ag,reachDistSupp))
        return;

//    GeometricComputations GC = GeometricComputations();
//    p3d_point p1;
//    Vector3d center = cells->getCenter();
//    p1.x = center[0];
//    p1.y = center[1];
//    p1.z = center[2];

    p3d_col_deactivate_robot_robot(ag->getRobotStruct(),_visball->getRobotStruct());
    bool reachable = true;
    for (int i = 0; i < ag->getHriAgent()->manip->shoulder_joint_ids_no; i ++)
    {
        p3d_point shoulderP;
        shoulderP.x = ag->getJoint(ag->getHriAgent()->manip->shoulder_joint_ids[i])->getJointStruct()->abs_pos[0][3];
        shoulderP.y = ag->getJoint(ag->getHriAgent()->manip->shoulder_joint_ids[i])->getJointStruct()->abs_pos[1][3];
        shoulderP.z = ag->getJoint(ag->getHriAgent()->manip->shoulder_joint_ids[i])->getJointStruct()->abs_pos[2][3];

        reachable = testLineArea(cells,shoulderP);
//        p3d_point tmp;
//        tmp.x = shoulderP.x; tmp.y = shoulderP.y; tmp.z = shoulderP.z;
//        double dist = 0.15;
//        double maxDist = GC.distBetweenPoints(p1,shoulderP);
//        p3d_point vect;
//        vect.x = (p1.x-shoulderP.x)/maxDist;
//        vect.y = (p1.y-shoulderP.y)/maxDist;
//        vect.z = (p1.z-shoulderP.z)/maxDist;



//        while (dist < maxDist)
//        {
//            placeVisball(tmp);

//            if (computeVisballCollision())
//            {
//                reachable = false;

//                break;
//            }

//            dist +=samplingR;
//            tmp.x = dist*vect.x + shoulderP.x;
//            tmp.y = dist*vect.y + shoulderP.y;
//            tmp.z = dist*vect.z + shoulderP.z;
//        }

        if (reachable)
        {
            cells->addTypeC(ag,reachCol);
            addCellToList(cells, ag, reachCol);
            if (cells->isItSupport())
            {
                cells->addTypeC(ag,reachColSupp);
                addCellToList(cells, ag, reachColSupp);
            }

            break;
        }
//        if ( < ag->getHriAgent()->manip->reach_tolerance )
//        {
//            cells->addTypeC(ag,reachDist);
//            addCellToList(cells, ag, reachDist);
//            if (cells->isItSupport())
//            {
//                cells->addTypeC(ag,reachDistSupp);
//                addCellToList(cells, ag, reachDistSupp);
//            }

//            break;
//        }
    }


//    p3d_col_activate_robot_robot(ag->getRobotStruct(),_visball->getRobotStruct());
}

void MightabilitiesGrid::computeVisDist(Robot* ag, MightabilitiesCell* cells)
{
//    if (!(ag && ag->getHriAgent() && ag->getHriAgent()->perspective && ag->getHriAgent()->perspective->camjoint) && )
//        return;

    p3d_vector3 a,b ,cross;
//    double a[3], b[3];
    a[0] = ag->getHriAgent()->perspective->camjoint->abs_pos[0][0];
    a[1] = ag->getHriAgent()->perspective->camjoint->abs_pos[1][0];
    a[2] = ag->getHriAgent()->perspective->camjoint->abs_pos[2][0];

    Vector3d center = cells->getCenter();

    b[0] = center[0] - ag->getHriAgent()->perspective->camjoint->abs_pos[0][3];
    b[1] = center[1] - ag->getHriAgent()->perspective->camjoint->abs_pos[1][3];
    b[2] = center[2] - ag->getHriAgent()->perspective->camjoint->abs_pos[2][3];


    p3d_vectXprod(a,b,cross);

    double totDist = p3d_vectNorm(b);

    if (ag->getHriAgent()->manip && totDist > ag->getHriAgent()->manip->reach_tolerance)
        return;

//    double first = (a[0]*b[0] + a[1]*b[1] + a[2]*b[2])/ (sqrt(a[0]*a[0] + a[1]*a[1] + a[2]*a[2])*sqrt(b[0]*b[0] + b[1]*b[1] + b[2]*b[2]));

    double angle = asin(p3d_vectNorm(cross)/(p3d_vectNorm(a)*totDist) );


    if (p3d_vectDotProd(a,b) < 0)
        angle = M_PI - angle;
//    cout << "angle = " << angle << endl;

    if (angle < (ag->getHriAgent()->perspective->fov/2)*M_PI/180)
    {
        cells->addTypeC(ag,visDist);
        addCellToList(cells, ag, visDist);
        if (cells->isItSupport())
        {
            cells->addTypeC(ag, visDistSupp);
            addCellToList(cells, ag, visDistSupp);
        }
    }
}

void MightabilitiesGrid::computeVisCol(Robot* ag, MightabilitiesCell* cells)
{
//    if (!(ag && ag->getHriAgent() && ag->getHriAgent()->perspective && ag->getHriAgent()->perspective->camjoint) && cells->belongToTypeC(ag,agentC))
//        return;

    if (!cells->belongToTypeC(ag,visDist) && !cells->belongToTypeC(ag,visDistSupp))
        return;

//    GeometricComputations GC = GeometricComputations();
//    p3d_point p1;
//    Vector3d center = cells->getCenter();
//    p1.x = center[0];
//    p1.y = center[1];
//    p1.z = center[2];

    p3d_col_deactivate_robot_robot(ag->getRobotStruct(),_visball->getRobotStruct());
    bool reachable = true;

    p3d_point a;
    a.x = ag->getHriAgent()->perspective->camjoint->abs_pos[0][3];
    a.y = ag->getHriAgent()->perspective->camjoint->abs_pos[1][3];
    a.z = ag->getHriAgent()->perspective->camjoint->abs_pos[2][3];

    reachable = testLineArea(cells,a);

    if (reachable)
    {
        cells->addTypeC(ag,visCol);
        addCellToList(cells, ag, visCol);
        if (cells->isItSupport())
        {
            cells->addTypeC(ag,visColSupp);
            addCellToList(cells, ag, visColSupp);
        }
    }
}

bool MightabilitiesGrid::testLineArea( MightabilitiesCell* cells, p3d_point p)
{

    GeometricComputations GC = GeometricComputations();
    p3d_point p1;
    Vector3d center = cells->getCenter();
    p1.x = center[0];
    p1.y = center[1];
    p1.z = center[2];
    if (cells->isItSupport())
        p1.z = p1.z +0.05;

//    p3d_col_deactivate_robot_robot(ag->getRobotStruct(),_visball->getRobotStruct());
    bool reachable = true;

//    cout << "testing for point = ( " << p1.x << " , " << p1.y << " , " << p1.z << " )" << endl;

    p3d_point tmp;
    tmp.x = p.x; tmp.y = p.y; tmp.z = p.z;
    double dist = 0.2;
    double maxDist = GC.distBetweenPoints(p1,p);
    p3d_point vect;
    vect.x = (p1.x-p.x)/maxDist;
    vect.y = (p1.y-p.y)/maxDist;
    vect.z = (p1.z-p.z)/maxDist;

    tmp.x = dist*vect.x + p.x;
    tmp.y = dist*vect.y + p.y;
    tmp.z = dist*vect.z + p.z;

    while (dist < maxDist)
    {
        placeVisball(tmp);

//        if (cells->getCenter()[1] > 1 && cells->getCenter()[1] < 1.15
//                && cells->getCenter()[0] < 1.544 && cells->getCenter()[0] > 1.16
//                && cells->getCenter()[2] < 1.563 && cells->getCenter()[2] > 1.250)
//        {
//            g3d_draw_allwin_active();
//        }

        if (computeVisballCollision())
        {
//            g3d_draw_allwin_active();
//            sleep(1);
            reachable = false;
            break;
        }

        dist +=0.03;
        tmp.x = dist*vect.x + p.x;
        tmp.y = dist*vect.y + p.y;
        tmp.z = dist*vect.z + p.z;
    }
    return reachable;

}

bool MightabilitiesGrid::computeVisballCollision()
{
    if (!_visball)
        return NULL;


    return _visball->isInCollisionWithOthersAndEnv();
//    for (unsigned int i = 0; i< global_Project->getActiveScene()->getNumberOfRobots(); i++)
//    {
//        if (1 == pqp_robot_robot_collision_test(_visball->getRobotStruct(),))
//            return true
//    }
//    int test = p3d_col_test_robot(_visball->getRobotStruct(),1); //1 is collision

//    if (test = 1)
//    {
//        g3d_draw_allwin_active();
//        p3d_print_col_pair();
//        sleep(3);
//        return true;
//    }
//    return false;
//    {
////        p3d_print_col_pair();

////        for (unsigned int i = 0; i < global_Project->getActiveScene()->getNumberOfRobots(); i++)
////        {
////            int realTest = p3d_col_test_robot_other(_visball->getRobotStruct(), global_Project->getActiveScene()->getRobot(i)->getRobotStruct(),1);
////            if (realTest == 1)
////            {
////                return global_Project->getActiveScene()->getRobot(i);
////            }
////        }
//    }
//    return NULL;
}

API::ThreeDCell* MightabilitiesGrid::createNewCell(unsigned int index, unsigned int x, unsigned int y, unsigned int z )
{
    //cout << " ThreeDCell " << endl;
    if (index == 0)
    {
        return new MightabilitiesCell( 0, _originCorner , this );
    }
    API::ThreeDCell* newCell = new MightabilitiesCell( index, computeCellCorner(x,y,z) , this );

//    Eigen::Vector3d corner = newCell->getCorner();
    //cout << " = (" << corner[0] <<"," << corner[1] << "," << corner[2] << ")" << endl;
    return newCell;
}


void MightabilitiesGrid::draw()
{

    cout << "drawing now!!" << endl;
    double colorvector[4];

    colorvector[0] = 1.0;       //red
    colorvector[1] = 0.5;       //green
    colorvector[2] = 0.0;       //blue
    colorvector[3] = 0.05;       //transparency


    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    //    glBlendFunc(GL_SRC_ALPHA, GL_ONE);

    glDisable(GL_LIGHTING);
    glDisable(GL_LIGHT0);

    glEnable(GL_CULL_FACE);
    glBegin(GL_QUADS);

    if (_toShow.size() == 0)
    {
        int nbCells = this->getNumberOfCells();

        for(int i=0; i<nbCells; i++)
        {
            MightabilitiesCell* cell = getCell(i);// dynamic_cast<MightabilitiesGrid*>( BaseGrid::getCell(i) );
            //        API::ThreeDCell* cell = dynamic_cast<ThreeDCell*>( BaseGrid::getCell(i) );
            glColor4dv(colorvector);
            if (cell)
                cell->draw();
            else
                cout << " cell " << i << " does not exist" << endl;
        }
    }
    else
    {
        vector<MightabilitiesCell*> cellss = getList(_toShow);
        for (unsigned int i = 0; i < cellss.size(); i++)
        {
            cellss.at(i)->draw();
        }
//        vector<MightabilitiesCell*> cellss = getList(_toShow.at(0).first,_toShow.at(0).second);
//        for (unsigned int i = 0; i < cellss.size(); i++)
//        {
//            MightabilitiesCell* cell = cellss.at(i);

//            bool toDraw = true;
//            for (unsigned int j = 0; j< _toShow.size(); j++)
//            {
//                if (!cell->belongToTypeC(_toShow.at(j).first, _toShow.at(j).second))
//                {
//                    toDraw = false;
//                    break;
//                }
//            }
//            if (toDraw)
//                cell->draw();
//        }

    }

    glEnd();

    glDisable(GL_CULL_FACE);
    glDisable(GL_BLEND);
}

MightabilitiesCell* MightabilitiesGrid::getCell(int i)
{
    return (MightabilitiesCell*)_cells[i];
}

void MightabilitiesGrid::show()
{
    currentMightability = this;
}

void MightabilitiesGrid::show(vector<pair<Robot *, typeCells> > toShow)
{
    _toShow = toShow;
    currentMightability = this;
}

vector<MightabilitiesCell*> MightabilitiesGrid::getList(Robot* ag)
{
    if (ag && ag->isAgent())
        return listMapAgentsType[ag]->getListFromType(agentC);

    vector<MightabilitiesCell*> v;
    return v;
}

vector<MightabilitiesCell*> MightabilitiesGrid::getList(Robot* ag, typeCells tc)
{
    if (ag && ag->isAgent())
    {
        if (listMapAgentsType.count(ag) == 1)
            return listMapAgentsType[ag]->getListFromType(tc);
    }
    vector<MightabilitiesCell*> v;
    return v;
}

vector<MightabilitiesCell*> MightabilitiesGrid::getList(vector< pair<Robot*,typeCells> > rpc)
{
    int id = -1;
    int minSize = P3D_HUGE;
    for (unsigned int j = 0; j < rpc.size(); j++)
    {
        if (getList(rpc.at(j).first,rpc.at(j).second).size() < minSize)
            id = j;
    }

    vector<MightabilitiesCell*> res;
    vector<MightabilitiesCell*> cellss = getList(rpc.at(id).first,rpc.at(id).second);
    for (unsigned int i = 0; i < cellss.size(); i++)
    {
        MightabilitiesCell* cell = cellss.at(i);
        bool test = true;
        for (unsigned int j = 0; j< rpc.size(); j++)
        {
            if (!cell->belongToTypeC(rpc.at(j).first, rpc.at(j).second))
            {
                test = false;
                break;
            }
        }
        if (test)
            res.push_back(cell);
    }
    return res;
}


void MightabilitiesGrid::addCellToList(MightabilitiesCell* c, Robot* ag, typeCells tc)
{
    if (listMapAgentsType.count(ag) != 1)
    {
        listMapAgentsType[ag] = new typeCellLists();

    }
    listMapAgentsType[ag]->addCellTo(c,tc);

//    if (listMapAgentsType.count(ag) == 1 && listMapAgentsType[ag].getListFromType(tc).size() > 0)
//    {
//        listMapAgentsType[ag][tc].push_back(c);
//    }
//    else if (listMapAgentsType.count(ag) == 1 && listMapAgentsType[ag].count(tc) == 0)
//    {
//        vector<MightabilitiesCell*> tmp;
//        tmp.push_back(c);
//        listMapAgentsType[ag][tc] = tmp;
//    }
//    else if (listMapAgentsType.count(ag) == 0)
//    {
//        vector<MightabilitiesCell*> tmp;
//        tmp.push_back(c);
//        map<typeCells, vector<MightabilitiesCell*> > mapTmp;
//        mapTmp[tc] = tmp;
//        listMapAgentsType[ag] = mapTmp;
//    }
//    map<Robot*,map<typeCells, vector<MightabilitiesCell*> > >::iterator it = listMapAgentsType.find(ag);
//    if (it)
}


//################### CELLS ######################

MightabilitiesCell::MightabilitiesCell() : ThreeDCell()
{
    _object= NULL;
    isSupport = false;
}

MightabilitiesCell::MightabilitiesCell(int i, MightabilitiesGrid* grid) : ThreeDCell(i,grid)
{
    _object= NULL;
    isSupport = false;
}

MightabilitiesCell::MightabilitiesCell(int i, Eigen::Vector3d corner, MightabilitiesGrid* grid) : ThreeDCell(i,corner,grid)
{
    _object= NULL;
    isSupport = false;
}

MightabilitiesCell::MightabilitiesCell(MightabilitiesCell* c)
{
    _index = c->getIndex();
    _corner = c->getCorner();
    _grid = c->getGrid();
    _object = c->getObject();
    isSupport = c->isItSupport();
}

void MightabilitiesCell::draw()
{
    double colorvector[4];
    colorvector[0] = 0.0;       //red
    colorvector[1] = 0.0;       //green
    colorvector[2] = 1.0;       //blue
    colorvector[3] = 1;       //transparency

    //        glEnable(GL_BLEND);
    g3d_set_color(Any,colorvector);

    Eigen::Vector3d center = getCenter();
    g3d_draw_solid_sphere(center[0],
                          center[1],
                          center[2],
                          0.01, 10);
}

void MightabilitiesCell::addAgent(Robot* ag)
{
    if (ag && ag->isAgent())
    {
        for(unsigned int i= 0; i < agentTypeBelongings.size(); i++)
        {
            if(agentTypeBelongings.at(i).first->getName() == ag->getName())
                return;
        }
        pair<Robot*, typeCells> p;
        p.first = ag;
        p.second = agentC;
        agentTypeBelongings.push_back(p);
    }
}

void MightabilitiesCell::removeAgent(Robot* ag)
{
    if (ag && ag->isAgent())
    {
        vector< pair<Robot*, typeCells> > tmp;
        int id = -1;
        for(unsigned int i= 0; i < agentTypeBelongings.size(); i++)
        {
            if(agentTypeBelongings.at(i).first->getName() != ag->getName())
            {
                tmp.push_back( agentTypeBelongings.at(i) );
            }
        }
        agentTypeBelongings = tmp;
    }
}

bool MightabilitiesCell::belongToAgent(Robot* ag)
{
    if (ag && ag->isAgent())
    {
        for(unsigned int i= 0; i < agentTypeBelongings.size(); i++)
        {
            if(agentTypeBelongings.at(i).first->getName() == ag->getName() &&
                    agentTypeBelongings.at(i).second == agentC)
                return true;
        }
    }
    return false;
}

bool MightabilitiesCell::belongToAtLeastOneAgent()
{
    if (agentTypeBelongings.size() >0)
        return true;
    return false;
}

void MightabilitiesCell::addTypeC(Robot* ag, typeCells tc)
{
    for(unsigned int i= 0; i < agentTypeBelongings.size(); i++)
    {
        if(agentTypeBelongings.at(i).first->getName() == ag->getName()
                && agentTypeBelongings.at(i).second == tc)
            return;
    }
    pair<Robot*, typeCells> p;
    p.first = ag;
    p.second = tc;
    agentTypeBelongings.push_back(p);
}

void MightabilitiesCell::removeTypeC(Robot* ag, typeCells tc)
{
    int id = -1;
    for(unsigned int i= 0; i < agentTypeBelongings.size(); i++)
    {
        if(agentTypeBelongings.at(i).first->getName() == ag->getName()
                && agentTypeBelongings.at(i).second == tc)
            id = i;
    }
    if (id !=-1)
        agentTypeBelongings.erase(agentTypeBelongings.begin()+id);
}

bool MightabilitiesCell::belongToTypeC(Robot* ag, typeCells tc)
{
    for(unsigned int i= 0; i < agentTypeBelongings.size(); i++)
    {
        if(agentTypeBelongings.at(i).first->getName() == ag->getName()
                && agentTypeBelongings.at(i).second == tc)
            return true;
    }
    return false;
}
