#ifndef MIGHTABILITIESGRIDS_HPP
#define MIGHTABILITIESGRIDS_HPP

#include "grid.h"
#include "API/Grids/ThreeDGrid.hpp"
#include "API/Grids/TwoDGrid.hpp"
#include "WorldState.hpp"
#include "P3d-pkg.h"

#include "API/planningAPI.hpp"

#include "move3d-headless.h"
#include "Planner-pkg.h"
#include "Collision-pkg.h"

class MightabilitiesCell;
class WorldState;

enum typeCells {
    agentC,
    reachDist,
    reachCol,
    reachDistSupp,
    reachColSupp,
    visDist,
    visCol,
    visDistSupp,
    visColSupp,
    typeCellSize
};


//vector<pair<typeCells,string> > tcvsss;

//string getTypeCName(typeCells tc);
//typeCells getTypeCType(string tcs);
//int getNbTypeC();

//class singletonAllCells
//{
//    public:
//        static singletonAllCells& getInstance()
//        {
//            static singletonAllCells    instance; // Guaranteed to be destroyed.
//                                  // Instantiated on first use.
//            return instance;
//        }

//        void computePoints();
//        vector<MightabilitiesCell*> allWorldCells;

//    private:
//        singletonAllCells() {computePoints();}                  // Constructor? (the {} brackets) are needed here.
//        // Dont forget to declare these two. You want to make sure they
//        // are unaccessable otherwise you may accidently get copies of
//        // your singleton appearing.
//        singletonAllCells(singletonAllCells const&);              // Don't Implement
//        void operator=(singletonAllCells const&); // Don't implement
//};

//class singletonAllCells
//{
//public:
//    singletonAllCells(){}
//    vector<p3d_point> allWorldPoints;
//};

class agentLinkedCells
{
public:
    agentLinkedCells(){}
    Robot* agent;
    map<typeCells, vector<MightabilitiesCell*> > cells;
};


class typeCellLists {
public:
    typeCellLists(){}
    vector<MightabilitiesCell*> getListFromType(typeCells tc);
    void addCellTo(MightabilitiesCell* c, typeCells tc);


    vector<pair<typeCells, vector<MightabilitiesCell*> > > lists;
};

class MightabilitiesGrid : public API::ThreeDGrid
{
public:
    MightabilitiesGrid();
    MightabilitiesGrid( Eigen::Vector3i size,    std::vector<double> envSize);
    MightabilitiesGrid( double samplingRate,     std::vector<double> envSize);
    MightabilitiesGrid( const ThreeDGrid& grid);

    void initVariables();

    void init(Eigen::Vector3i size,    std::vector<double> envSize);
    void init(double samplingRate ,    std::vector<double> envSize);

    void setVisBall();
    void placeVisball(MightabilitiesCell* c);
    void placeVisball(p3d_point p);

    void computeMights(vector<Robot*> agents, vector<Robot *> supports, vector<typeCells> abilities);

    vector<MightabilitiesCell*> getCellsFromEntity(Robot* entity);

    vector<MightabilitiesCell*> mergeVectors(vector<MightabilitiesCell*> c1,vector<MightabilitiesCell*> c2);

    void computeReachDist(Robot* ag, MightabilitiesCell* cells);
    void computeReachCol(Robot* ag, MightabilitiesCell* cells);
    void computeVisDist(Robot* ag, MightabilitiesCell* cells);
    void computeVisCol(Robot* ag, MightabilitiesCell* cells);
    bool computeVisballCollision();

    bool testLineArea(MightabilitiesCell* cells, p3d_point p);



//    void init(double samplingRate, vector<Robot*> ag);

    API::ThreeDCell* createNewCell(unsigned int index, unsigned int x, unsigned int y, unsigned int z );

    void draw();

    MightabilitiesCell* getCell(int i);

    WorldState* getWorldState() {return _WS;}
    void setWorldState(WorldState* ws) {_WS = ws;}
    void show();
    void show(vector< pair<Robot*,typeCells> > toShow);

    void setSampleRate(double SR) { _samplingRate = SR; }

    vector<MightabilitiesCell*> getList(Robot* ag);
    vector<MightabilitiesCell*> getList(Robot* ag, typeCells tc);
    vector<MightabilitiesCell*> getList(vector< pair<Robot*,typeCells> > rpc);

    void addCellToList(MightabilitiesCell* c, Robot* ag, typeCells tc);


private:
    WorldState* _WS;

    double _samplingRate;
    Robot* _visball;
    vector<agentLinkedCells> _ALC;

    map<Robot*, typeCellLists* > listMapAgentsType;

    vector< pair<Robot*,typeCells> > _toShow;
    vector<MightabilitiesCell*> _allWorldCells;



//    vector<p3d_point> allWorldPoints;

};


class MightabilitiesCell : public API::ThreeDCell
{
public:
    MightabilitiesCell();
    MightabilitiesCell(int i, MightabilitiesGrid* grid);
    MightabilitiesCell(int i, Eigen::Vector3d corner, MightabilitiesGrid* grid);
    MightabilitiesCell(MightabilitiesCell* c);

    void draw();

    void addAgent(Robot* ag);
    void removeAgent(Robot* ag);
    bool belongToAgent(Robot* ag);
    bool belongToAtLeastOneAgent();

    void addTypeC(Robot* ag, typeCells tc);
    void removeTypeC(Robot *ag, typeCells tc);
    bool belongToTypeC(Robot* ag, typeCells tc);

    void setObject(Robot* r){_object = r;}
    Robot* getObject(){return _object;}

    void setIsSupport(bool v) {isSupport = v;}
    bool isItSupport() {return isSupport;}

private:
    vector< pair<Robot*, typeCells> > agentTypeBelongings;

    Robot* _object;
    bool isSupport;

};


#endif // MIGHTABILITIESGRIDS_HPP
