#ifndef TASKCONSTRAINTS_HPP
#define TASKCONSTRAINTS_HPP

#include "API/facts.hpp"
#include "geometricTools.hpp"
#include "tools.hpp"
#include "WorldState.hpp"

typedef struct struct_region
{
    string nameId;
    GeometricForm* GeoForm;
}regionConst;

typedef struct struct_ressource
{
    string nameId;
    Robot* entity;
} ressourceConst;

typedef struct struct_position
{
    string nameId;
    Robot* entity;
    string objInEntity; //ex hand of PR2_ROBOT, first joint if empty
    p3d_point translations;
    p3d_point rotations;

} positionConst;

class WorldState;

class TaskConstraints
{
public:
    TaskConstraints();
    TaskConstraints(TaskConstraints* TC);

    void addConstraint(Fact f);
    void addConstraint(ressourceConst r);
    void addConstraint(regionConst r);
    void addConstraint(positionConst r);
    void addConstraint(string n, Robot* e);
    void addConstraint(string n, Robot *obj, string area, bool isSupport);
    void addConstraint(string n, GeometricForm* gf);
    void addConstraint(string n, Robot* e, string objInEntity, p3d_point t, p3d_point r );
    void addConstraint(TaskConstraints *TC);


    vector<Fact> getFactConstraints() {return factConstraints;}
    vector<ressourceConst> getRessourceConstraints() {return ressourceConstraints;}
    vector<regionConst> getRegionConstraints() {return regionConstraints;}
    vector<positionConst> getPositionConstraints() {return positionConstraints;}

    void cleanAll();

    bool areFactConstraintRespected(WorldState* WS);
    string getDescr();


private:
    vector<Fact> factConstraints;
    vector<ressourceConst> ressourceConstraints;
    vector<regionConst> regionConstraints;
    vector<positionConst> positionConstraints;

};

#endif // TASKCONSTRAINTS_HPP
