#include "tools.hpp"
#include "lightPlanner/proto/ManipulationArmData.hpp"
#include "API/project.hpp"
#include "planner/TrajectoryOptim/Classic/smoothing.hpp"

#include "../virtualtask.hpp"

//extern std::vector<Eigen::Vector2d,Eigen::aligned_allocator<Eigen::Vector2d> > path_to_draw;
extern std::vector<p3d_point> currNavTraj;

//########################### TaskSubSolution ################################


TaskSubSolution::TaskSubSolution()
{

}

TaskSubSolution::TaskSubSolution(TaskSolution* sol, string name, int id, p3d_traj* traj, Robot *r, double initRP)
{
    _subSolutionName = name;
    _id = id;
    _trajectory = traj;
    _sol = sol;
    _isLast = true;
    _isSmoothed = false;
    _robot = r;
    _initRangeParam = initRP;
    _armId = -1;

    _sol->addToSolutions(this);

    _navigation = false;

}

TaskSubSolution::~TaskSubSolution()
{
    if (_trajectory)
        p3d_del_traj(_trajectory);
}

void TaskSubSolution::addToPrevious(TaskSubSolution* prev)
{
    for(unsigned int i = 0; i < _previous.size(); i++)
    {
        if(_previous.at(i)->getId() == prev->getId())
        {
            return;
        }
    }
    _previous.push_back(prev);
}

bool TaskSubSolution::setTrajInRobotStruct()
{
    if (!_robot)
        return false;

    if (!_sol->isTrajComputed())
    {
        cout << "the trajectory is not computed for the task alternetive " << endl;
    }
    else if (_trajectory)
    {
        _robot->getRobotStruct()->tcur = _trajectory;
        currNavTraj.clear();
        return true;
    }
    else if (isNavigation())
    {
        currNavTraj = _nav_path;
        _robot->getRobotStruct()->tcur = NULL;

        //compute traj from series of points
//        path_to_draw.clear();
//        for (unsigned int i = 0; i<_nav_path.size();i++)
//        {
//            Eigen::Vector2d v;
//            v[0] = _nav_path.at(i).x; v[1] = _nav_path.at(i).y;
//            path_to_draw.push_back(v);
//        }
        API::Trajectory p3d_trajectory(_robot);
        _start_WS->load();
        shared_ptr<Configuration> q_f(_robot->getCurrentPos());
        if (_robot->getHriAgent() && _robot->getHriAgent()->type == HRI_PR2)
        {
            for (int e = 37; e<=48; e++)
                (*q_f)[e] = 0;
        }
        else if (_robot->getHriAgent() && _robot->getHriAgent()->type == HRI_KUKAROBOT)
        {
            for (int e = 21; e<=26; e++)
                (*q_f)[e] = 0;
        }
        p3d_trajectory.push_back(q_f);

        for (unsigned int i = 0; i < _nav_path.size(); i++)
        {
            shared_ptr<Configuration> q_tmp(_robot->getCurrentPos());
//            cout << " cell = " << _nav_path.at(i).x << " , " << _nav_path.at(i).y << endl;
            (*q_tmp)[6] = _nav_path.at(i).x;
            (*q_tmp)[7] = _nav_path.at(i).y;
            (*q_tmp)[11] = _nav_path.at(i).z;
            if (_robot->getHriAgent() && _robot->getHriAgent()->type == HRI_PR2)
            {
                for (int e = 37; e<=48; e++)
                    (*q_tmp)[e] = 0;
            }
            else if (_robot->getHriAgent() && _robot->getHriAgent()->type == HRI_KUKAROBOT)
            {
                for (int e = 21; e<=26; e++)
                    (*q_tmp)[e] = 0;
            }
            p3d_trajectory.push_back(q_tmp);

        }
        p3d_traj* trajPt = NULL;
        _robot->getRobotStruct()->tcur = NULL;

        p3d_trajectory.replaceP3dTraj(trajPt);
        if (!_robot->getRobotStruct()->tcur)
        {
            cout << "Fail to create traj from points" << endl;
            return false;
        }
        return true;
    }
    return false;

}

double TaskSubSolution::getDuration()
{
    return 0;
}

void TaskSubSolution::setNavigation(bool nav)
{
    _navigation = nav;
}

bool TaskSubSolution::isNavigation()
{
    return _navigation;
}

void TaskSubSolution::setNavTraj(vector<p3d_point> p)
{
    _nav_path = p;
}

vector<p3d_point> TaskSubSolution::getNavVector()
{
    return _nav_path;
}


void TaskSubSolution::setNext(vector<TaskSubSolution*> next)
{
    _next = next;
    if (next.size() > 0)
    {
        for(unsigned int i = 0; i < next.size(); i++)
        {
            next.at(i)->addToPrevious(this);
        }
    }
    _isLast = false;
}

bool TaskSubSolution::refineTraj()
{
    cout << "task to be smoothed" << endl;
    cout << getTask()->getDescr() << endl;
    if (!_robot)
        return false;
    cout << "call the smoothing for " << _subSolutionName << endl;
    if(!_isSmoothed)
    {
        API::Smoothing optimTrj(_robot,_trajectory);
        double old_cost=optimTrj.cost();
        cout<<"Previous cost: "<<old_cost<<endl;
        optimTrj.runShortCut(PlanEnv->getInt(PlanParam::smoothMaxIterations));
        optimTrj.replaceP3dTraj();
        double cost=optimTrj.cost();
        cout<<"New cost: "<<cost<<endl;
        cout<<"Smoothing gain: "<<100*(old_cost-cost)/old_cost<<" %"<<endl;




//        p3d_smoothing_function(_robot->getRobotStruct(), _trajectory, 250, 5);
    }
//    _isSmoothed = true;
    return true;
}

bool TaskSubSolution::reComputeTraj()
{
    return true;
}

p3d_traj* TaskSubSolution::getTraj()
{
//    refineTraj();
    return _trajectory;
}

string TaskSubSolution::getDescr()
{
    return "SubTask: \nName \t\t= " + _subSolutionName + "\nId \t\t= " + to_string(_id) + "\n" ;
}

void TaskSubSolution::saveToDisk(FILE *file){
    if(!file){
        printf("error: TaskSubSolution::saveToFile: file is NULL\n");
        return;
    }
    fprintf(file,"\nbeg_tss\n");
    fprintf(file,"name %s\n",getSubSolutionName().c_str());
    fprintf(file,"tssId %d\n",_id);
    fprintf(file,"armId %d\n",_armId);
    fprintf(file,"robot %s\n",_robot->getName().c_str());
    if(_start_WS)
        _start_WS->saveToDisk(file,"start");
    if(_trajectory && _trajectory->rob){
        p3d_rob *robotPt=_trajectory->rob;
        fprintf(file,"beg_traj %s\n",robotPt->name);
        p3d_localpath *lp=_trajectory->courbePt;
        while(lp){
            configPt q0=p3d_alloc_config(robotPt);
            configPt qi=lp->config_at_param(robotPt,lp,0);
            configPt qf=lp->config_at_param(robotPt,lp,lp->range_param);
            fprintf(file,"beg_localpath %s\n",p3d_get_localpath_type(lp->type_lp));
            fprintf(file,"q_init ");
            if(p3d_dist_config(robotPt,qi,q0)<EPS6)
                fprintf(file,"zero");
            else{
                for(int i=6;i<robotPt->nb_dof;i++){
                    if(fabs(qi[i])<EPS6)
                        fprintf(file,"0 ");
                    else
                        fprintf(file,"%f ",qi[i]);
                }
            }
            fprintf(file,"\nq_end ");
            if(p3d_dist_config(robotPt,qf,q0)<EPS6)
                fprintf(file,"zero");
            else{
                for(int i=6;i<robotPt->nb_dof;i++){
                    if(fabs(qf[i])<EPS6)
                        fprintf(file,"0 ");
                    else
                        fprintf(file,"%f ",qf[i]);
                }
            }
            fprintf(file,"\nend_localpath\n");
            lp=lp->next_lp;
            p3d_destroy_config(robotPt,q0);
            p3d_destroy_config(robotPt,qi);
            p3d_destroy_config(robotPt,qf);
        }
        fprintf(file,"end_traj\n");
    }
    if(_end_WS)
        _end_WS->saveToDisk(file, "end");

    fprintf(file,"end_tss\n");

}


TaskSolution::TaskSolution(TaskAllSolutions* taskAS, int id)
{
    _TAS = taskAS;
    _TAS->addAlternativeSol(this);

    alternativeId = id;
    _trajComputed = false;
    importantConfsId = 0;

    _start_WS =NULL;
    _end_WS = NULL;

}

TaskSolution::~TaskSolution()
{

}

void TaskSolution::addToSolutions(TaskSubSolution* tss)
{
    for(map<Robot*,vector<TaskSubSolution*> >::iterator it = solutionMap.begin(); it != solutionMap.end(); ++it) {
        string name1 = it->first->getName();
        string name2 = tss->getRobot()->getName();
        if(name1==name2)
        {
            it->second.push_back(tss);
            return;
        }
    }
    vector<TaskSubSolution*> vTSS;
    vTSS.push_back(tss);
    solutionMap[tss->getRobot()] = vTSS;
}

double TaskSolution::getDuration()
{
    return 0;
}

void TaskSolution::addConf(confPtr_t q)
{
    importantConfs.push_back(shared_ptr<Configuration>(new Configuration(*q)));
}

confPtr_t TaskSolution::getNextConf()
{
    confPtr_t q;
    if (importantConfs.size() > 0)
    {
        if (importantConfsId >= importantConfs.size())
            importantConfsId = 0;

        q = importantConfs.at(importantConfsId);
        importantConfsId++;
    }
    return q;
}

TaskSubSolution* TaskSolution::getFirstSubSolution()
{
    TaskSubSolution* TSS = NULL;
    vector<Robot*> v = getInvolvedRobots();
    if (v.size() == 0)
        return TSS;
    TSS = solutionMap.at(v.at(0)).at(0);
    while(TSS->getPrevious().size() > 0)
    {
        TSS = TSS->getPrevious().at(0);
    }
    return TSS;
}

TaskSubSolution* TaskSolution::getNextSubSolution(TaskSubSolution* ts)
{
    //TODO:THIS is a function that shouldn t be used
    return ts->getNext().at(0);
}

TaskSubSolution* TaskSolution::getSubSolutionWithId(int subId)
{
    vector<Robot*> v = getInvolvedRobots();
    for (unsigned int i = 0; i < v.size(); i++)
    {
        vector<TaskSubSolution*> vtss = solutionMap.at(v.at(i));
        for (unsigned int j = 0; j < vtss.size(); j++)
        {
            if (vtss.at(j)->getId() == subId)
            {
                return vtss.at(j);
            }
        }
    }
    TaskSubSolution* tss = NULL;
    return tss;
}

int TaskSolution::getNbsubSolPerRobot(Robot* rob)
{
    return solutionMap.at(rob).size();
}


vector<TaskSubSolution*> TaskSolution::getTSSByRobot(Robot* rob)
{
    return solutionMap.at(rob);
}

vector<TaskSubSolution*> TaskSolution::getAllTSS()
{
    vector<Robot*> v;
    vector<TaskSubSolution*> vrs;
    for(map<Robot*,vector<TaskSubSolution*> >::iterator it = solutionMap.begin(); it != solutionMap.end(); ++it) {

        Robot* rob = it->first;
        vector<TaskSubSolution*> tmp = solutionMap.at(rob);

        for (unsigned int i = 0; i < tmp.size(); i++)
        {
            vrs.push_back(tmp.at(i));
        }

    }
    return vrs;
}

vector<Robot*> TaskSolution::getInvolvedRobots()
{
    vector<Robot*> v;
    for(map<Robot*,vector<TaskSubSolution*> >::iterator it = solutionMap.begin(); it != solutionMap.end(); ++it) {
      v.push_back(it->first);

    }
    return v;
}

bool TaskSolution::isInPreviousDoneTSS(TaskSubSolution* tss)
{
    for (unsigned int i = 0; i < PreviousDoneTSS.size(); i++)
    {
        if(tss->getId() == PreviousDoneTSS.at(i)->getId())
        {
            return true;
        }
    }
    return false;
}
void TaskSolution::addToPreviousDoneTSS(TaskSubSolution* tss)
{
    PreviousDoneTSS.push_back(tss);
}

void TaskSolution::addTypeData(type_data td)
{
    _TData.agents.insert(td.agents.begin(), td.agents.end());
    _TData.objects.insert(td.objects.begin(), td.objects.end());
    _TData.additionalData.insert(td.additionalData.begin(), td.additionalData.end());
    _TData.points.insert(td.points.begin(), td.points.end());
}

void TaskSolution::clearTypeData()
{
    _TData.agents.clear();
    _TData.objects.clear();
    _TData.additionalData.clear();
    _TData.points.clear();
}


TaskAllSolutions::TaskAllSolutions(Task *t)
{
    _task = t;
}

TaskAllSolutions::~TaskAllSolutions()
{

}

TaskSolution* TaskAllSolutions::getAlternativeWithId(int alternativeId)
{
    for(unsigned int i = 0; i < _solutions.size(); i++)
    {
        if (_solutions.at(i)->getId() == alternativeId)
        {
            return _solutions.at(i);
        }
    }
    TaskSolution* TS = NULL;
    return TS;
}

void TaskAllSolutions::removeAlternativeWithId(int alternativeId)
{
    int id = -1;
    for(unsigned int i = 0; i < _solutions.size(); i++)
    {
        if (_solutions.at(i)->getId() == alternativeId)
        {
            id = i;
            break;
        }
    }
    delete _solutions.at(id);
    if (id != -1)
    {
        _solutions.erase(_solutions.begin()+id);
    }
}

int TaskAllSolutions::getNextId()
{
    int res = 0;
    for(unsigned int i = 0; i < _solutions.size(); i++)
    {
        if (_solutions.at(i)->getId() > res)
        {
            res = _solutions.at(i)->getId();
        }
    }
    return res + 1;
}

vector<int> TaskAllSolutions::getAlternativeIds()
{
    vector<int> vi;
    for(unsigned int i = 0; i < _solutions.size(); i++)
    {
        vi.push_back(_solutions.at(i)->getId());
    }
    return vi;
}

void TaskAllSolutions::addAlternativeSol(TaskSolution* t)
{
    _solutions.push_back(t);
}




//##################### plan management


PlanNode::PlanNode()
{
    _TS =NULL;
}

PlanNode::~PlanNode()
{

}

PlanNode::PlanNode(PlanTree* PT, TaskSolution* TS)
{
    _PT = PT;
    _TS = TS;
    _suceeded = true;
    if (!_TS)
        _suceeded = false;
    _previous =NULL;

    nbGikCalled = -1;
    nbMPCalled = -1;
    nbParameters = -1;
    nbTests = -1;
    timeComp = -1;
    nbTotPossibleTests = -1;


}

bool PlanNode::setPrevious(PlanNode* PN)
{
    //check if previous has same WS as this one
//    if (PN->getTaskSolution()->getStopWS()->isEqualTo(getTaskSolution()->getStartWS()))
//    {
        _previous = PN;
        if (PN)
            PN->addToNexts(this);
//        return PN->addToNexts(this);
//    }
//    return false;
        return true;
}

double PlanNode::getDuration()
{

    return _TS->getDuration();
}

bool PlanNode::isInNexts(PlanNode* PN)
{
    for (unsigned int i = 0; i < _nexts.size(); i ++)
    {
        PlanNode* tmp = _nexts.at(i);
        if (tmp->getId().first == PN->getId().first && tmp->getId().second == PN->getId().second)
        {
            return true;
        }
    }
    return false;
}

pair<int,int> PlanNode::getId()
{
    pair<int,int> iid;
    if (hasSucceeded() && _TS)
    {
        iid.first = _TS->getTask()->getId();
        iid.second = _TS->getId();
    }
    else
    {
        iid.first = -10;
        iid.second = -10;
    }
    return iid;
}

bool PlanNode::addToNexts(PlanNode* PN)
{
    if (PN->hasSucceeded())
        if (!PN->getTaskSolution()->getStartWS()->isEqualTo(getTaskSolution()->getStopWS()))
            return false;
    for (unsigned int i = 0; i < _nexts.size(); i++)
    {
        if (PN->getId().first == _nexts.at(i)->getId().first && PN->getId().second == _nexts.at(i)->getId().second)
        {
            cout << "Node with ID (" << PN->getId().first << " , " << PN->getId().second << ") already exist in nexts list, changing it to the new one" <<endl;
            _nexts[i] = PN;

        }
        else
            _nexts.push_back(PN);
    }
    return true;
}

template < typename T > std::string to_string4( const T& n )
{
    std::ostringstream stm ;
    stm << n ;
    return stm.str() ;
}

string PlanNode::getFormatedNumbers()
{
    string s = "Important numbers:\n";
    s += "time = " + to_string4(timeComp) +"\n";
    s += "nb GIK = " + to_string4(nbGikCalled) +"\n";
    s += "nb MP = " + to_string4(nbMPCalled) +"\n";
    s += "nb var = " + to_string4(nbParameters) +"\n";
    s += "nb tests = " + to_string4(nbTests) +"\n";
    s += "nb Tot Possible Tests = " + to_string4(nbTotPossibleTests) +"\n";
    s += "return repport = " + to_string4(returnRepport) +"\n";
    return s;
}

double PlanNode::computePriority()
{
    map<string,double>::iterator it;
    double res = 0;
    cout << "================= " << getId().first << " " << getId().second << endl;
    for(it = backtrackingData.begin(); it != backtrackingData.end(); it++)
    {
      cout << "key: "<< it->first << " value: " << it->second << endl;
      res+=it->second;
    }
    return res;
}

bool PlanNode::operator==(PlanNode* p)
{
    return (getId().first == p->getId().first && getId().second == p->getId().second);
}


OnePlan::OnePlan()
{

}

OnePlan::~OnePlan()
{

}

OnePlan::OnePlan(PlanTree* PT)
{
    _PT = PT;
    _plan.clear();
}

void OnePlan::addToPlan(PlanNode* PN)
{
    if (_plan.size() == 0)
    {
        cout << "plan size 0 " << endl;
        _plan.push_back(PN);
    }
    vector<PlanNode*> vpn;

    if (!PN->getPrevious())
    {
        cout << "ERROR, no previous for this node " << PN->getId().first << " , " << PN->getId().second << endl;
        return;
    }
    if (PN->getPrevious()->getId().first == _plan.at(_plan.size()-1)->getId().first
            && PN->getPrevious()->getId().second == _plan.at(_plan.size()-1)->getId().second )
    {
        _plan.push_back(PN);
    }
    else
    {
        PlanNode* pnTmp = PN;
        while(pnTmp)
        {
//            cout << "pnTmp id = " << pnTmp->getId().first << " , " << pnTmp->getId().second << endl;
            vpn.push_back(pnTmp);
            pnTmp = pnTmp->getPrevious();
        }
        _plan.clear();
        for(unsigned int i = vpn.size(); i > 0; i--)
        {
            _plan.push_back(vpn.at(i-1));
        }
    }


//    else
//    {
//        while (_plan.size() > 0)
//        {
//            int i = _plan.size()-1;
//            if(PN->getTaskSolution()->getStartWS()->isEqualTo(_plan.at(i)->getTaskSolution()->getStopWS()))
//            {
//                _plan.push_back(PN);
//                break;
//            }
//            else
//            {
//                _plan.pop_back();
//            }
//        }
//        if (_plan.size() == 0)
//        {
//            cout << "plan size 0 again" << endl;
//            _plan.push_back(PN);
//        }
//    }
}

void OnePlan::printPlan()
{
    cout << "printing plan:" <<endl;
    cout << "number of nodes = " << _plan.size() << endl;

    for (unsigned int i = 0 ; i < _plan.size(); i++)
    {
        cout << "node Id = (" << _plan.at(i)->getId().first << " , " <<  _plan.at(i)->getId().second << ")" << endl;
        if (_plan.at(i)->getPrevious())
            cout << "    with previous node with Id (" << _plan.at(i)->getPrevious()->getId().first << " , " <<  _plan.at(i)->getPrevious()->getId().second << ")" << endl;
        else
            cout << "    first in the tree" << endl;
    }
}

void OnePlan::removeLast()
{
    vector<PlanNode*> tmp;
    for (unsigned int i = 0; i < _plan.size()-1; i++)
    {
        tmp.push_back(_plan.at(i));
    }
    _plan = tmp;
}

TaskSolution* OnePlan::getLastSol()
{
    if(_plan.size()== 0)
        return NULL;
    return _plan.at(_plan.size()-1)->getTaskSolution();
}


