#ifndef TOOLS_HPP
#define TOOLS_HPP

using namespace std;
//using namespace std::tr1;
#include "P3d-pkg.h"
#include "API/Device/robot.hpp"
#include "WorldState.hpp"
#include "taskconstraints.hpp"
#include "API/scene.hpp"
#include "lightPlanner/proto/ManipulationPlanner.hpp"

#include "../../planner/plannerFunctions.hpp"
#include "../lightPlanner/proto/lightPlanner.h"
#include "../lightPlanner/proto/lightPlannerApi.h"

#include "API/facts.hpp"

#include "graspPlanning/include/gpGrasp.h"
#include "taskconstraints.hpp"




enum TasksId {
    noType, //dummy, in case
    pick,
    pick_multirrt,
    pickHuman,
    place,
    placeHuman,
    place_multirrt,
    placeReachable,
    pickAndPlace,
    makeAccessible,
    pickAndMakeAccessible,
    exhibit,
    pickAndShow,
    give,
    pickAndGive,
    point,
    fix,
    drop,
    gotoPlace,
    moveTo,
    stackObj,
    fakePlace,
    pickPlaceQuad,
    pickQuad,
    placeQuad,
    //this is the last one:
    nbOfTaskId
};

struct Attachement{
    Robot* agent;
    Robot* object;
    gpGrasp grasp;
    int armId;
};


typedef struct STRUCT_type_data
{
    TasksId id;
    multimap<string,string> agents;
    multimap<string,string> objects;
    multimap<string,p3d_point> points;
    multimap<string,string> additionalData;

}type_data;

typedef struct STRUCT_Support
{
    Robot* object;
    vector<p3d_point> rectangle_corner;
    //TODO: include other kind of shapes
}Support;


class Task;
class TaskSolution;
class TaskSubSolution;
class TaskConstraints;
class WorldState;


//#################### global functions #########################

template < typename T > std::string to_string( const T& n )
{
    std::ostringstream stm ;
    stm << n ;
    return stm.str() ;
}

//##################################################################

class TaskAllSolutions
{
public:
    TaskAllSolutions(Task* t);
    ~TaskAllSolutions();
    TaskSolution* getAlternativeWithId(int alternativeId);
    void removeAlternativeWithId(int alternativeId);

    Task* getTask() {return _task;}
    void addAlternativeSol(TaskSolution* t);

    int getNbAlternatives() {return _solutions.size();}
    vector<int> getAlternativeIds();

    int getNextId();

private:
    vector<TaskSolution*> _solutions;
    Task* _task;
};


class TaskSolution
{
public:
    TaskSolution(TaskAllSolutions* taskAS, int id);
    ~TaskSolution();


    TaskSubSolution* getFirstSubSolution();
    TaskSubSolution* getNextSubSolution(TaskSubSolution *ts);
    TaskSubSolution* getSubSolutionWithId(int subId);

    int getId() {return alternativeId;}
    void setId(int altId) {alternativeId = altId;}

    int getNbsubSolPerRobot(Robot* rob);
    vector<TaskSubSolution*> getTSSByRobot(Robot* rob);
    vector<TaskSubSolution*> getAllTSS();
    vector<Robot*> getInvolvedRobots();

    Task* getTask() { return _TAS->getTask();}

    bool isInPreviousDoneTSS(TaskSubSolution* tss);
    void addToPreviousDoneTSS(TaskSubSolution* tss);

    void setStartWS(WorldState* start_WS) { _start_WS = start_WS;}
    WorldState* getStartWS() {return _start_WS;}

    void setStopWS(WorldState* end_WS) { _end_WS = end_WS;}
    WorldState* getStopWS() {return _end_WS;}

    bool isTrajComputed() {return _trajComputed;}
    void setTrajComputed(bool val) {_trajComputed = val;}

    TaskConstraints* getTaskConstraints() {return _constraint;}
    void setTaskConstraints(TaskConstraints* TC) {_constraint = TC;}

    void addToSolutions(TaskSubSolution* tss);

    double getDuration();

    double getCost() {return _cost;}
    void setCost(double c) { _cost = c;}

    void addConf(confPtr_t q);
    confPtr_t getNextConf();

    type_data getTypeData() {return _TData;}
    void addTypeData(type_data td);
    void clearTypeData();


private:
    map<Robot*,vector<TaskSubSolution*> > solutionMap;
    int alternativeId;
    TaskAllSolutions* _TAS;

    vector<TaskSubSolution*> PreviousDoneTSS;

    WorldState* _start_WS;
    WorldState* _end_WS;

    bool _trajComputed;

    TaskConstraints* _constraint;

    double _cost;

    vector<confPtr_t> importantConfs;
    unsigned int importantConfsId;

    type_data _TData;

};


class TaskSubSolution
{
public:
    TaskSubSolution();
    TaskSubSolution(TaskSolution* sol, string name, int id, p3d_traj* traj, Robot* r, double initRP);

    ~TaskSubSolution();


    bool refineTraj(); //smoothing mainly
    bool reComputeTraj();

    string getSubSolutionName() {return _subSolutionName;}
    void setSubSolutionName(string name){_subSolutionName = name;}

    Robot* getRobot() {return _robot;}
    int getId() {return _id;}
    void setId(int id) {_id = id;}

    vector<TaskSubSolution*> getPrevious() {return _previous; }
    void addToPrevious(TaskSubSolution* prev);
    void setPrevious(vector<TaskSubSolution*> prev) {_previous = prev;}

    vector<TaskSubSolution*> getNext() { return _next; }
    void setNext(vector<TaskSubSolution*> next);

    p3d_traj* getTraj();// {return _trajectory;}
    void setTraj(p3d_traj* traj) {_trajectory = traj;}

    Task* getTask() {return _sol->getTask();}
    TaskSolution* getSolution() {return _sol;}

    bool isLast() {return _isLast;}

    void setStartState(WorldState* WS) { _start_WS = WS;}
    void setEndState(WorldState* WS) { _end_WS = WS;}

    WorldState* getStartState() {return _start_WS;}
    WorldState* getEndState() {return _end_WS;}

    double getInitRangeParam() {return _initRangeParam;}

    string getDescr();

    bool setTrajInRobotStruct();

    double getDuration();

    void setNavigation(bool nav);
    bool isNavigation();

    void setNavTraj(vector<p3d_point> p);
    vector<p3d_point> getNavVector();

    void saveToDisk(FILE *file);

    int getArmId() {return _armId;}
    void setArmId(int i) {_armId = i;}

private:
    string _subSolutionName;
    int _id;
    vector<TaskSubSolution*> _previous;
    vector<TaskSubSolution*> _next;

    double _initRangeParam;

    p3d_traj* _trajectory;

    TaskSolution* _sol;

    bool _isLast;

    WorldState* _start_WS;
    WorldState* _end_WS;

    bool _isSmoothed;
    Robot* _robot;

    vector<p3d_point> _nav_path;
    bool _navigation;

    //0 for right, 1 for left, 2 for both
    int _armId;
};




class PlanTree;

class PlanNode
{
public:
    PlanNode();
    ~PlanNode();

    PlanNode(PlanTree* PT, TaskSolution* TS);

    double getDuration();
    bool isInNexts(PlanNode* PN);
    bool setPrevious(PlanNode* PN);
    pair<int,int> getId();
    bool addToNexts(PlanNode* PN);

    TaskSolution* getTaskSolution() {return _TS;}
    PlanNode* getPrevious() {return _previous;}
    vector<PlanNode*> getNexts() {return _nexts;}

    double getInitTime() {return _initTime;}
    void setInitTime(double initTime) {_initTime = initTime;}

    PlanTree* getPlanTree() {return _PT;}

    bool hasSucceeded() {return _suceeded;}

    Task* getTask() {return _task;}
    void setTask(Task* t) { _task = t; }

    string getFormatedNumbers();
    double computePriority();

    Graph* getGraph() {return _explorationGraph;}
    void setGraph(Graph* g) { _explorationGraph = g;}

    bool operator==(PlanNode* p);

    int nbGikCalled;
    int nbMPCalled;
    int nbParameters;
    int nbTotPossibleTests;
    int nbTests;
    double timeComp;
    double timeMP;

    string returnRepport;


    map<string,double> backtrackingData;

private:
    TaskSolution* _TS;
    PlanNode* _previous;
    vector<PlanNode*> _nexts;

    double _initTime;
    PlanTree* _PT;

    bool _suceeded;
    Task* _task;

    Graph* _explorationGraph;

};


class OnePlan
{
public:
    OnePlan();
    ~OnePlan();

    OnePlan(PlanTree* PT);


    void addToPlan(PlanNode* PN);// {_plan.push_back(PN);}
    void removeLast();
    void clearPlan() {_plan.clear();}
    vector<PlanNode*> getPlan() {return _plan;}

    void printPlan();

    TaskSolution* getLastSol();



private:
    vector<PlanNode*> _plan;
    PlanTree* _PT;
};



#endif // TOOLS_HPP
