#include "drop.hpp"


Drop::Drop() : Place()
{
    type = "drop";
    typeId = drop;
    _armId = -1;

    noSupport = false;

    approachTraj = NULL;
    placeTraj = NULL;
    escapeTraj = NULL;
    releaseTraj = NULL;

    nbMaxLoops = 50;

    nbParameters = 5;
    nbGikCalled = 0;
    nbMPCalled = 0;
    nbTests = 0;

    maxTests = 500;

    nbTotPossibleTests = maxTests*nbMaxLoops;

    //_possibleMainAgentType.push_back("PR2_ROBOT");
    //_possibleMainAgentType.push_back("MAX_ROBOT");
    //_possibleMainAgentType.push_back("BOB_ROBOT");
    _possibleMainAgentType.push_back(HRI_PR2);

}

bool Drop::findCandidateSolutions()
{
    if (!checkAndSetArmAndObj())
        return false;

    if (!checkAttachement())
    {
        returnRepport = "no_corresponding_attachement_in_ws";
        return false;
    }

    if (!checkHRIAgent(mainAgent))
        return false;

    if (!setMainAgentHand())
        return false;

    //finding support either from set support object, or from env.
    if (!findSupports())
        return false;

    if (!findRotations())
        return false;


    _WSF = new WorldStateFacts(_WS->getRobotConfs());

    return true;
}

bool Drop::getRandomSol()
{
//    GeometricComputations GC = GeometricComputations();

    confPtr_t robotCurConf = mainAgent->getCurrentPos();
    confPtr_t objectCurConf = mainObject->getCurrentPos();

    detachObject();

    bool res = false;
    int nbTestsPlace = -1;



    int reachedReturnRepportCount = _returnRepportcount;
    while (!res && nbTestsPlace < maxTests)
    {
        nbTests++;
        nbTestsPlace++;
        _returnRepportcount = reachedReturnRepportCount;
        mainObject->setAndUpdate(*objectCurConf);
//        cout<< "test nb = " << nbTests << endl;

        if (!findCurrentTestPoint())
            continue;

        configPt q = findCurrentRotations();
        if (!q)
            continue;

        q[6] += _currentTest.x;
        q[7] += _currentTest.y;
        q[8] += _currentTest.z + 0.15;


        p3d_set_and_update_this_robot_conf(mainObject->getRobotStruct(),q);

        if (!isObjectInSurface())
            continue;

        if(isObjectInCollision(NULL))
            continue;

        if(isHandInCollision())
            continue;

        _confSol = shared_ptr<Configuration>(new Configuration(mainObject,q));

        res = true;
    }


    mainAgent->setAndUpdate(*robotCurConf);
    mainObject->setAndUpdate(*objectCurConf);
    attachObject();
    return res;

}

bool Drop::findConfigurations()
{
    removeVirtualObjects();
    // Need to detach first, otherwise computation of manipulation data is unstable and return some odd values very often
    detachObject();

//    bool tr = false;
    confPtr_t qInitA = shared_ptr<Configuration>(new Configuration(*mainAgent->getCurrentPos()));
    confPtr_t qInitO = shared_ptr<Configuration>(new Configuration(*mainObject->getCurrentPos()));

    mainObject->setAndUpdate(*_confSol);
//    g3d_draw_allwin_active();

    //HACK
    if(mainAgent->getHriAgent()->type == HRI_PR2)
    {
        _curGrasp.hand_type = GP_PR2_GRIPPER;
    }


    confPtr_t q;

    p3d_matrix4 graspMatWorldFrame;
    p3d_mat4Mult(mainObject->getJoint(0)->getJointStruct()->abs_pos, _curGrasp.frame, graspMatWorldFrame);

    activateFingertipCol(false, mainAgent, mainObject, NULL);
    bool testIk = computeInverseKinematic(mainAgent, graspMatWorldFrame  ,q , _armId );
    nbGikCalled++;
    activateFingertipCol(true, mainAgent, mainObject, NULL);


    _returnRepportcount++;
    if (!testIk)
    {
        if (_returnRepportcountMax < _returnRepportcount)
        {
            _returnRepportcountMax = _returnRepportcount;
            returnRepport = "no_IK_for_drop";
        }
        return false;
    }



    graspConfClosed = shared_ptr<Configuration>(new Configuration(*q));


    mainObject->setAndUpdate(*_confSol);
    bool foundPos = false;
    _dropConf = (shared_ptr<Configuration>)new Configuration(*_confSol);

    double dropingDist = 0.01;
    while ((*_dropConf)[8] >0)
    {
        (*_dropConf)[8] = (*_dropConf)[8] - dropingDist;

        mainObject->setAndUpdate(*_dropConf);
//        g3d_draw_allwin_active();

        if (p3d_col_test_robot(mainObject->getRobotStruct(),1) == 1)
        {
            (*_dropConf)[8] = (*_dropConf)[8] + dropingDist;
            _WSF->setConfiguration(_dropConf);
            global_Project->getActiveScene()->getFactsManager()->computeIsIn(_WSF, mainObject,_supportObject,true);
            fact_id FI3;
            FI3.factType = "IsIn";
            FI3.factSubType = "";
            FI3.isMainAnAgent = false;

            Fact f = _WSF->getFact(FI3,mainObject,_supportObject);
            if (f.getBoolValue())
            {
                foundPos = true;
                break;
            }
            else
            {
                foundPos = false;
                break;
            }
        }
    }

    _returnRepportcount++;
    if (!foundPos)
    {
        cout << "failing to find pos in obj" << endl;
        if (_returnRepportcountMax < _returnRepportcount)
        {
            _returnRepportcountMax = _returnRepportcount;
            returnRepport = "no_pos_in_obj";
        }
        return false;
    }

    mainAgent->setAndUpdate(*qInitA);
    mainObject->setAndUpdate(*qInitO);

    return true;
}

bool Drop::findTrajectories()
{

    _WS->load();
    activateFingertipCol(false, mainAgent, mainObject, NULL);
//    attachObject();

    int status = -1;

    activateFingertipCol(false, mainAgent, mainObject, NULL);
    approachTraj = compute_traj(_WS->getRobotConf(mainAgent), graspConfClosed, mainObject, status);
    activateFingertipCol(true, mainAgent, mainObject, NULL);

    _returnRepportcount++;
    if (!approachTraj)
    {
        cout << "no approach traj" << endl;
        if (_returnRepportcountMax < _returnRepportcount)
        {
            _returnRepportcountMax = _returnRepportcount;
            returnRepport = "no_approach_traj";
        }
        moveBackVirtualObjects();
        return false;
    }

    setGraph(new Graph(*API_activeGraph));

    detachObject();
    mainAgent->setAndUpdate(*graspConfClosed);

    mainObject->setAndUpdate(*_confSol);

    escapeTraj = compute_traj_using_localPath(_confSol, _dropConf, NULL, status);



    _returnRepportcount++;
    if (!escapeTraj)
    {
        cout << "no escape traj" << endl;
        if (_returnRepportcountMax < _returnRepportcount)
        {
            _returnRepportcountMax = _returnRepportcount;
            returnRepport = "no_escape_traj";
        }
        moveBackVirtualObjects();
        return false;
    }


    activateFingertipCol(true, mainAgent, mainObject, NULL);
    moveBackVirtualObjects();

    return true;
}

bool Drop::computeSolutionTrajectories(int alternativeId)
{
    TaskSolution* TS = result->getAlternativeWithId(alternativeId);
    if(!TS)
    {
        cout << "no alternative for task " << to_string(task_id) << " with alternative id = " << alternativeId << endl;
        return false;
    }

    TS->getStartWS()->load();
    TaskSubSolution* TSS0 = TS->getSubSolutionWithId(0);
    TaskSubSolution* TSS1 = TS->getSubSolutionWithId(2);

    confPtr_t q1 = TSS0->getEndState()->getRobotConf(mainAgent);
    confPtr_t q2 = TSS1->getEndState()->getRobotConf(mainObject);

    approachConf = q1;
    _dropConf = q2;

    findTrajectories();


    result->getAlternativeWithId(alternativeId)->getSubSolutionWithId(0)->setTraj(approachTraj);
    result->getAlternativeWithId(alternativeId)->getSubSolutionWithId(2)->setTraj(escapeTraj);

    TS->setTrajComputed(true);

    return true;
}

bool Drop::setToSolution(int solutionId, bool computeMP)
{
    TaskSolution* TS = new TaskSolution(result,solutionId);
    TS->setTrajComputed(computeMP);

    TS->setTaskConstraints(_currConstraints);

    TaskSubSolution* TSS1 = new TaskSubSolution(TS, "approach", 0, approachTraj, mainAgent, 0);
    TaskSubSolution* TSS2 = new TaskSubSolution(TS, "release", 1, NULL, mainAgent, 0);
    TaskSubSolution* TSS3 = new TaskSubSolution(TS, "droping", 2, escapeTraj, mainObject, 0);

    TSS1->setArmId(_armId);
    TSS2->setArmId(_armId);
    TSS3->setArmId(_armId);

    attachObject();

    vector<TaskSubSolution*> v1;
    v1.push_back(TSS2);
    TSS1->setNext(v1);
    vector<TaskSubSolution*> v2;
    v2.push_back(TSS3);
    TSS2->setNext(v2);

    TS->setStartWS(_WS);

    TSS1->setStartState(_WS);
    WorldState* WS1 = new WorldState(_WS);
    bool t = WS1->replaceConf(graspConfClosed);
    if (!t)
    {
        cout << "ERROR: error with robots names" << endl;
        return false;
    }
    _WS->load();
    attachObject();
    mainAgent->setAndUpdate(*WS1->getRobotConf(mainAgent));
    confPtr_t q_o_curr = mainObject->getCurrentPos();
    detachObject();
    WS1->replaceConf(q_o_curr);
    TSS1->setEndState(WS1);

    TSS2->setStartState(WS1);
    TSS2->setEndState(WS1);

    TSS3->setStartState(WS1);
    WorldState* WS2 = new WorldState(WS1);
    WS2->replaceConf(_dropConf);
    WS2->clearAttachement();

    TSS3->setEndState(WS2);
    TS->setStopWS(WS2);

    WS2->setTSThatEndWithThisWS(TS);
    WS2->setTCForNextTask(_currConstraints);

    return true;
}


string Drop::getTextFromValues()
{
    return "Task: " + type + ": droping an object in an other"
            + "\n    agents: "
            + "\n        mainAgent:    [AGENT ID]"
            + "\n    objects: "
            + "\n        mainObject:   [OBJECT ID]"
            + "\n        supportObject [OBJECT ID]"
            + "\n    datas:"
            + "\n        hand: [right OR left]";
}
