#include "fakeplace.hpp"



FakePlace::FakePlace() : Place()
{
    type = "fakePlace";
    typeId = fakePlace;
    _armId = -1;

    noSupport = false;

    approachTraj = NULL;
    placeTraj = NULL;
    escapeTraj = NULL;
    releaseTraj = NULL;

    nbMaxLoops = 200;

    nbParameters = 3;

    maxTests = 100;

    _supportObject = NULL;

    nbTotPossibleTests = maxTests*nbMaxLoops;


    //_possibleMainAgent.push_back("ROBOT");
    //_possibleMainAgent.push_back("HUMAN");
    _possibleMainAgentType.push_back(HRI_HERAKLES);
    _possibleMainAgentType.push_back(HRI_PR2);

    _noArm = true;
    _hands.clear();

}

bool FakePlace::initAll()
{
    result = new TaskAllSolutions(this);
    _supportObject = NULL;
    mainObject = NULL;
    mainAgent = NULL;
    _currentSurface = NULL;
    return true;
}

bool FakePlace::initialize()
{
    if (mainObject->getName().find("VIRTUAL") == string::npos)
    {
        returnRepport = "object_not_VIRTUAL";
        return false;
    }

//    p3d_col_activate_robot(mainObject->getRobotStruct());
    mainObject->getObjectRob()->setToVirtual(false);

    nbGikCalled = 0;
    nbMPCalled = 0;
    nbTests = 0;
    timeMp = 0;
    _returnRepportcountMax = 0;

    return true;
}


bool FakePlace::finalize()
{

    mainObject->getObjectRob()->setToVirtual(true);
//    p3d_col_deactivate_robot(mainObject->getRobotStruct());
    return true;
}



bool FakePlace::findCandidateSolutions()
{
    if (mainAgent)
    {
        if(!checkHRIAgent(mainAgent))
            return false;

        if(!findSupports())
            return false;
    }
    else if (_supportObject)
    {
        if (_supportObject->getObjectRob())
        {
            ObjectRob* obre = _supportObject->getObjectRob();
            if(obre->isSupport())
            {
                _supports = obre->getSupports();
            }
        }
    }
    else
    {
        cout << "ERROR: fakePlace should be called either with an agent or with a support object" << endl;
        returnRepport = "no_support_in_object_or_mainagent";
        return false;
    }


    if (!findRotations())
        return false;


    return true;
}


bool FakePlace::getRandomSol()
{
    confPtr_t objectCurConf = mainObject->getCurrentPos();

    bool res = false;
    int nbTestsPlace = -1;

    int reachedReturnRepportCount = _returnRepportcount;
    while (!res && nbTestsPlace < maxTests)
    {
        nbTests++;
        nbTestsPlace++;
        _returnRepportcount = reachedReturnRepportCount;
        mainObject->setAndUpdate(*objectCurConf);
//        cout<< "test nb = " << nbTests << endl;

        if (!findCurrentTestPoint())
            continue;

        configPt q = findObjectConf();

        _returnRepportcount++;
        if (!q)
        {
            if (_returnRepportcountMax < _returnRepportcount)
            {
                _returnRepportcountMax = _returnRepportcount;
                returnRepport = "not_conf_found";
            }
            continue;
        }

        p3d_set_and_update_this_robot_conf(mainObject->getRobotStruct(),q);

        //test reacheability


//        g3d_draw_allwin_active();
//        sleep(3);


        _returnRepportcount++;
        if (!isObjectInSurface())
        {
            if (_returnRepportcountMax < _returnRepportcount)
            {
                _returnRepportcountMax = _returnRepportcount;
                returnRepport = "not_in_surface";
            }
            continue;
        }


        mainObject->getObjectRob()->setToVirtual(false);

        p3d_set_and_update_this_robot_conf(mainObject->getRobotStruct(),q);
        p3d_col_activate_robot(mainObject->getRobotStruct());


        _returnRepportcount++;
        if (isObjectInCollision(_supportObject))
        {
            if (_returnRepportcountMax < _returnRepportcount)
            {
                _returnRepportcountMax = _returnRepportcount;
                returnRepport = "object_in_collision";
            }
            continue;
        }

        _confSol = shared_ptr<Configuration>(new Configuration(mainObject,q));

        res = true;
    }

    mainObject->setAndUpdate(*objectCurConf);
    return res;
}


bool FakePlace::findConfigurations()
{

    if (mainAgent)
    {
        confPtr_t robotCurConf = mainAgent->getCurrentPos();
        graspConfClosed = robotCurConf;
        graspConfOpen = robotCurConf;
        escapeConf = robotCurConf;
        approachConf = robotCurConf;
    }

    return true;
}

bool FakePlace::findTrajectories()
{
    return true;
}

bool FakePlace::setAgents(multimap<string,string> agents_name)
{
    if (agents_name.count("mainAgent") != 1)
    {
        mainAgent = NULL;
        cout << "no main agent set" << endl;
        return true;
    }
    mainAgent = global_Project->getActiveScene()->getRobotByName(agents_name.find("mainAgent")->second);

    if (!mainAgent)
    {
        cout << "wrong name for mainAgent, continue witout mainAgent" << endl;
        mainAgent = NULL;
    }
    else if (mainAgent->getName() != agents_name.find("mainAgent")->second)
    {
        cout << "wrong name for mainAgent, continue witout mainAgent" << endl;
        mainAgent = NULL;
    }

    return true;
}

bool FakePlace::setToSolution(int solutionId, bool computeMP)
{
    TaskSolution* TS = new TaskSolution(result,solutionId);
    TS->setTrajComputed(computeMP);

    TS->setTaskConstraints(_currConstraints);

    TaskSubSolution* TSS1 = new TaskSubSolution(TS, "approach", 0, approachTraj, mainAgent, 0);

    TS->setStartWS(_WS);

    TSS1->setStartState(_WS);


    WorldState* WS1 = new WorldState(_WS);
    bool t = WS1->replaceConf(_confSol);
    if (!t)
    {
        cout << "ERROR: error with robots names" << endl;
        return false;
    }

    WS1->setTSThatEndWithThisWS(TS);
    WS1->setTCForNextTask(_currConstraints);

    TSS1->setEndState(WS1);
    TS->setStopWS(WS1);




    return true;
}

bool FakePlace::isPossibleMainAgent(Robot* r)
{
    if (!r)
    {
        cout << "no main agent" << endl;
        return true;
    }else if (r->getHriAgent()){
        return true;
    }else{
        returnRepport = "agent_not_in_list";
        return false;
    }
    //for (unsigned int i = 0; i < _possibleMainAgent.size(); i++)
    //{
    //    if (r->getName().find(_possibleMainAgent.at(i)) != std::string::npos)
    //    {
    //        return true;
    //    }
    //}
    //returnRepport = "agent_not_in_list";
    //return false;
}

double FakePlace::computeCost(int altId)
{
    altId = 0;
    return 0;
}

string FakePlace::getDescr()
{
    string s = "Task: \nId \t\t= " + to_string(task_id) + "\nType \t\t= " + type;
    if (mainAgent)
        s += "\nMain agent \t\t= " + mainAgent->getName();
    s += "\nMain object \t= " + mainObject->getName() + "\nSupport object \t= ";
    if (_supportObject)
        s +=  _supportObject->getName();
    return s;
}

