#include "give.hpp"
#include "API/project.hpp"
#include "GTPTools/tools.hpp"
#include "GTPTools/WorldState.hpp"



#include "planner/plannerFunctions.hpp"

#ifdef LIGHT_PLANNER
#include "LightPlanner-pkg.h"
#endif

int maxTests = 100;


//####################### Give ########################
Give::Give() : CooperativeTask()
{
    type = "give";
    typeId = give;
    _armId = -1;

    nbMaxLoops = 50;

    nbParameters = 7;

    nbTotPossibleTests = nbMaxLoops;

    _hasTargedPoint = false;

    //_possibleMainAgentType.push_back("PR2_ROBOT");
    //_possibleMainAgentType.push_back("MAX_ROBOT");
    //_possibleMainAgentType.push_back("BOB_ROBOT");
    _possibleMainAgentType.push_back(HRI_PR2);
}

bool Give::initAll()
{

    result = new TaskAllSolutions(this);
    if (!checkAttachement())
    {
        returnRepport = "no_attahchment_in_ws";
        return false;
    }
    _armId = 0;
    return true;
}

bool Give::initialize()
{
    approachTraj = NULL;
    targetTraj = NULL;


    nbGikCalled = 0;
    nbMPCalled = 0;
    nbTests = 0;
    timeMp = 0;

    _currTargetHandId = -1;

    returnRepport = "ERROR";
    _returnRepportcount = 0;
    _returnRepportcountMax = 0;


    return initMotionPlaner(mainAgent);
}

bool Give::finalize()
{
    return true;
}

bool Give::findCandidateSolutions()
{
    if (!checkHRIAgent(mainAgent))
        return false;

    if (!checkHRIAgent(targetAgent))
        return false;

    if (!setMainAgentHand())
        return false;

    //check Target Hands free

    vector<int> targetHandIds;
    vector<Attachement> vA = _WS->getAttachements();
    for (unsigned int i = 0; i < vA.size(); i++)
    {
        Attachement att = vA.at(i);
        if (att.agent->getName() == targetAgent->getName())
        {
            targetHandIds.push_back(att.armId);
        }
    }
    if (targetHandIds.size() >= 2)
    {
        returnRepport ="target_agent_no_hand_free";
        return false;
    }

    GeometricComputations GC = GeometricComputations();

    _mainAgentShoulder.x = mainAgent->getJoint(mainAgent->getHriAgent()->manip->shoulder_joint_ids[getArmId()])->getJointStruct()->abs_pos[0][3];
    _mainAgentShoulder.y = mainAgent->getJoint(mainAgent->getHriAgent()->manip->shoulder_joint_ids[getArmId()])->getJointStruct()->abs_pos[1][3];
    _mainAgentShoulder.z = mainAgent->getJoint(mainAgent->getHriAgent()->manip->shoulder_joint_ids[getArmId()])->getJointStruct()->abs_pos[2][3];

    _mainAgentReach = mainAgent->getHriAgent()->manip->reach_tolerance;
    _targetAgentReach = targetAgent->getHriAgent()->manip->reach_tolerance;


    for (int i = 0; i < targetAgent->getHriAgent()->manip->shoulder_joint_ids_no; i++)
    {
        int targetarmId = i;
        for (unsigned int j = 0; j < targetHandIds.size(); j++)
        {
            if (i == targetHandIds.at(j))
                continue;
        }

        p3d_point targetShoulder;

        targetShoulder.x = targetAgent->getJoint(targetAgent->getHriAgent()->manip->shoulder_joint_ids[targetarmId])->getJointStruct()->abs_pos[0][3];
        targetShoulder.y = targetAgent->getJoint(targetAgent->getHriAgent()->manip->shoulder_joint_ids[targetarmId])->getJointStruct()->abs_pos[1][3];
        targetShoulder.z = targetAgent->getJoint(targetAgent->getHriAgent()->manip->shoulder_joint_ids[targetarmId])->getJointStruct()->abs_pos[2][3];

        double d = GC.distBetweenPoints(_mainAgentShoulder, targetShoulder);

        cout << "dist = " << d << " Main reach = " << _mainAgentReach << " target reach = " << _targetAgentReach << endl;

        if (d < _mainAgentReach + _targetAgentReach)
        {
            pair<int, p3d_point> p;
            p.first = targetarmId;
            p.second = targetShoulder;
            _targetShoulder.push_back(p);
        }
    }

    if (_targetShoulder.size() == 0)
    {
        returnRepport = "agents_too_far";
        return false;
    }
    return true;
}

bool Give::getRandomSol()
{

//    GeometricComputations GC = GeometricComputations();

    confPtr_t robotCurConf = mainAgent->getCurrentPos();
    confPtr_t objectCurConf = mainObject->getCurrentPos();

    detachObject();

    bool res = false;
    int nbTestsPlace = -1;

    int reachedReturnRepportCount = _returnRepportcount;
    while (!res && nbTestsPlace < maxTests)
    {
        nbTests++;
        nbTestsPlace++;
        _returnRepportcount = reachedReturnRepportCount;
        mainObject->setAndUpdate(*objectCurConf);
//        cout<< "test nb = " << nbTests << endl;

        if (!findCurrentTestPoint())
            continue;


        configPt q = findObjectConf();
        if (!q)
            continue;
        p3d_set_and_update_this_robot_conf(mainObject->getRobotStruct(), q);

        _confSol = shared_ptr<Configuration>(new Configuration(mainObject,q));

        if(isObjectInCollision(NULL))
            continue;

        if(isHandInCollision())
            continue;
        res = true;
    }

    mainAgent->setAndUpdate(*robotCurConf);
    mainObject->setAndUpdate(*objectCurConf);
    attachObject();
    return res;
}

bool Give::findConfigurations()
{
    removeVirtualObjects();
    detachObject();
    // Need to detach first, otherwise computation of manipulation data is unstable and return some odd values very often

    bool tr = false;
    shared_ptr<Configuration> qInitA = mainAgent->getCurrentPos();
    shared_ptr<Configuration> qInitO = mainObject->getCurrentPos();


    _returnRepportcount++;

    activateFingertipCol(false, mainAgent, mainObject, NULL);


    mainObject->setAndUpdate(*_confSol);

    MANIPULATION_TASK_MESSAGE status = ManipPlMap[mainAgent->getName()]->computeManipulationData(_armId, mainObject->getRobotStruct(), _curGrasp);
    nbGikCalled +=4;

    activateFingertipCol(true, mainAgent, mainObject, NULL);


    _returnRepportcount++;
    if (status == MANIPULATION_TASK_OK)
    {
        tr = true;
        cout << "place found" <<endl;
        graspConfClosed = shared_ptr<Configuration>(new Configuration(mainAgent, ManipPlMap[mainAgent->getName()]->getManipulationData().getGraspConfig()));
    }
    else
    {
        cout << "manipulation data fails" << endl;
        if (_returnRepportcountMax < _returnRepportcount)
        {
            _returnRepportcountMax = _returnRepportcount;
            returnRepport = "no_IK";
        }
    }

    if (!checkHumanReach())
        return false;

    mainAgent->setAndUpdate(*qInitA);
    mainObject->setAndUpdate(*qInitO);

    return tr;
}

bool Give::findTrajectories()
{
    activateFingertipCol(false, mainAgent, mainObject, NULL);
    _WS->load();
    attachObject();

    int status = -1;
    approachTraj = compute_traj(mainAgent->getCurrentPos(), graspConfClosed, mainObject, status);

    _returnRepportcount++;
    if (!approachTraj)
    {
        cout << "no approach traj" << endl;
        if (_returnRepportcountMax < _returnRepportcount)
        {
            _returnRepportcountMax = _returnRepportcount;
            returnRepport = "no_approach_traj";
        }
        activateFingertipCol(true, mainAgent, mainObject, NULL);
        moveBackVirtualObjects();
        return false;
    }

    setGraph(new Graph(*API_activeGraph));

//    mainObject->setAndUpdate(*_confSol);

    targetTraj = compute_traj_using_localPath(targetAgent->getCurrentPos(), _targetSol, mainObject, status);


    _returnRepportcount++;
    if (!targetTraj)
    {
//        cout << "no escape traj" << endl;
        if (_returnRepportcountMax < _returnRepportcount)
        {
            _returnRepportcountMax = _returnRepportcount;
            returnRepport = "no_target_traj";
        }
        activateFingertipCol(true, mainAgent, mainObject, NULL);
        moveBackVirtualObjects();
        return false;
    }

    cout << "hhhhhhhhhhhh" << endl;

    activateFingertipCol(true, mainAgent, mainObject, NULL);
    moveBackVirtualObjects();

    return true;
}

bool Give::computeSolutionTrajectories(int alternativeId)
{
    TaskSolution* TS = result->getAlternativeWithId(alternativeId);
    if(!TS)
    {
        cout << "no alternative for task " << to_string(task_id) << " with alternative id = " << alternativeId << endl;
        return false;
    }

    TS->getStartWS()->load();
    TaskSubSolution* TSS0 = TS->getSubSolutionWithId(0);
    TaskSubSolution* TSS1 = TS->getSubSolutionWithId(1);

    confPtr_t q1 = TSS0->getEndState()->getRobotConf(mainAgent);
    confPtr_t q2 = TSS1->getEndState()->getRobotConf(targetAgent);

    approachConf = q1;
    _targetSol = q2;

    findTrajectories();


    result->getAlternativeWithId(alternativeId)->getSubSolutionWithId(0)->setTraj(approachTraj);
    result->getAlternativeWithId(alternativeId)->getSubSolutionWithId(1)->setTraj(targetTraj);

    TS->setTrajComputed(true);

    return true;
}

bool Give::setToSolution(int solutionId, bool computeMP)
{
    TaskSolution* TS = new TaskSolution(result,solutionId);
    TS->setTrajComputed(computeMP);

    TS->setTaskConstraints(_currConstraints);

    TaskSubSolution* TSS1 = new TaskSubSolution(TS, "approach", 0, approachTraj, mainAgent, 0);
    TaskSubSolution* TSS2 = new TaskSubSolution(TS, "targetApproach", 1, targetTraj, targetAgent, 0);


    attachObject();

    vector<TaskSubSolution*> v1;
    v1.push_back(TSS2);
    TSS1->setNext(v1);

    TS->setStartWS(_WS);

    TSS1->setStartState(_WS);
    WorldState* WS1 = new WorldState(_WS);
    bool t = WS1->replaceConf(graspConfClosed);
    if (!t)
    {
        cout << "ERROR: error with robots names" << endl;
        return false;
    }
    _WS->load();
    attachObject();
    mainAgent->setAndUpdate(*WS1->getRobotConf(mainAgent));
    confPtr_t q_o_curr = mainObject->getCurrentPos();
    detachObject();
    WS1->replaceConf(q_o_curr);
    TSS1->setEndState(WS1);

    TSS2->setStartState(WS1);
    WorldState* WS2 = new WorldState(WS1);
    WS2->replaceConf(_targetSol);

    TSS2->setEndState(WS2);
    TS->setStopWS(WS2);

    WS2->setTSThatEndWithThisWS(TS);
    WS2->setTCForNextTask(_currConstraints);


    return true;
}

bool Give::setAgents(multimap<string,string> agents_name)
{
    return (setMainAgent(agents_name) && setTargetAgent(agents_name));
}

bool Give::setObjects(multimap<string,string> objects_name)
{
    return (setMainObject(objects_name));
}

bool Give::setPoints(multimap<string,p3d_point> points)
{
    return setTargetPoint(points);
}


bool Give::setdata(multimap<string,string> additionalData)
{
    if (additionalData.count("hand") != 1)
    {
        cout << "no arm choosen, go for right" << endl;
        setArmToUse(Right);
        return true;
    }

    string handStr = additionalData.find("hand")->second;
    if (handStr == "right")
        setArmToUse(Right);
    else if (handStr == "left")
        setArmToUse(Left);
    else
    {
        cout << "the arm selection is not rigth: choosen : " << handStr << "available : right or left." << endl;
        cout << "setting default arm to right" << endl;
        setArmToUse(Right);
    }

    return true;
}


bool Give::setTargetPoint(multimap<string,p3d_point> points)
{

    if (points.count("target") == 1)
    {
        cout << "got a target point" << endl;
        _hasTargedPoint = true;
        _targetPoint = points.find("target")->second;
        return true;
    }

//    noSupport = true;
    return true;
}

void Give::smoothSolution(int alternativeId)
{
    for (int i = 0; i < result->getAlternativeWithId(alternativeId)->getNbsubSolPerRobot(getMainAgent()); i++)
    {
        TaskSubSolution* tss = result->getAlternativeWithId(alternativeId)->getTSSByRobot(getMainAgent()).at(i);
        if (tss->getSubSolutionName() == "approach")
        {
            tss->getRobot()->getRobotStruct()->tcur = tss->getTraj();
            tss->getStartState()->load();
            attachObject();
            tss->refineTraj();
            detachObject();
            break;
        }
    }
    result->getAlternativeWithId(alternativeId)->getStartWS()->load();
}

//This function is not efficient at all, TODO having a better sampling system
int nbMaxOfTestInSphere = 100;

bool Give::findCurrentTestPoint()
{
    //finding spheres
    //finding hexadron
    GeometricComputations GC = GeometricComputations();
    int nbTestaa=0;
    bool reachablePointFound = false;
    while (nbTestaa < nbMaxOfTestInSphere && !reachablePointFound)
    {
        nbTestaa++;
        nbTests++;

        p3d_point tmp;
        tmp.x = p3d_random(_mainAgentShoulder.x - _mainAgentReach, _mainAgentShoulder.x + _mainAgentReach);
        tmp.y = p3d_random(_mainAgentShoulder.y - _mainAgentReach, _mainAgentShoulder.y + _mainAgentReach);
        tmp.z = p3d_random(_mainAgentShoulder.z - _mainAgentReach, _mainAgentShoulder.z + _mainAgentReach);

        if (GC.distBetweenPoints(tmp,_mainAgentShoulder) > _mainAgentReach)
            continue;

        int currId = -1;
        for (unsigned int i = 0; i < _targetShoulder.size(); i++)
        {
            if (GC.distBetweenPoints(tmp, _targetShoulder.at(i).second) < _targetAgentReach)
            {
                currId = i;
                break;
            }
        }
        if (currId == -1)
            continue;

        _currentTest = tmp;
        _currTargetHandId = currId;
        reachablePointFound = true;
    }

    _returnRepportcount++;
    if (!reachablePointFound)
    {
        if (_returnRepportcountMax < _returnRepportcount)
        {
            _returnRepportcountMax = _returnRepportcount;
            returnRepport = "no_point_found_in_intersection";
        }
        return false;
    }

    return true;
}

configPt Give::findObjectConf()
{
    configPt q = p3d_copy_config(mainObject->getRobotStruct(), mainObject->getCurrentPos()->getConfigStruct());
    q[6] = _currentTest.x;
    q[7] = _currentTest.y;
    q[8] = _currentTest.z;


    q[9] = p3d_random(-M_PI,M_PI);
    q[10] = p3d_random(-M_PI,M_PI);
    q[11] = p3d_random(-M_PI,M_PI);


    return q;
}



bool Give::checkHumanReach()
{
    GeometricComputations GC = GeometricComputations();

    bool res = false;

    p3d_point pr = GC.getClosestBBPointFromPoint(mainObject, _targetShoulder.at(_currTargetHandId).second);

    confPtr_t q_curr_tar = targetAgent->getCurrentPos();

    int j = _currTargetHandId;

    p3d_point agShoulder = _targetShoulder.at(_currTargetHandId).second;

    // test reach with dist
    _returnRepportcount++;
    if (GC.distBetweenPoints(pr,agShoulder) > _targetAgentReach)
    {
        if (_returnRepportcountMax < _returnRepportcount)
        {
            _returnRepportcountMax = _returnRepportcount;
            returnRepport = "not_reachable_by_target_dist";
        }
        targetAgent->setAndUpdate(*q_curr_tar);
        return false;
    }


    //test reach with IK
    HRI_GIK_TASK_TYPE task;
    if (j == 0)
        task = GIK_LATREACH;
    else if (j == 1)
        task = GIK_RATREACH;

    p3d_vector3 Tcoord;

    Tcoord[0] = pr.x;
    Tcoord[1] = pr.y;
    Tcoord[2] = pr.z;

    double distance_tolerance = 0.05;

    configPt q = p3d_get_robot_config(targetAgent->getRobotStruct());


    int IKSucceded = hri_agent_single_task_manip_move(targetAgent->getHriAgent(),
                                                      task,
                                                      &Tcoord,
                                                      distance_tolerance,
                                                      &q);
    _returnRepportcount++;
    p3d_set_and_update_this_robot_conf(targetAgent->getRobotStruct(), q);
    if (IKSucceded == 1)
    {
//        g3d_draw_allwin_active();
//        sleep(3);


        p3d_set_and_update_this_robot_conf(targetAgent->getRobotStruct(), q);
        p3d_col_deactivate_robot_robot(targetAgent->getRobotStruct(), mainObject->getRobotStruct());
        int coltest = p3d_col_test_robot(targetAgent->getRobotStruct(),0);
        p3d_col_deactivate_robot_robot(targetAgent->getRobotStruct(), mainObject->getRobotStruct());

        if (coltest != 1)
        {
            res = true;
            _targetSol = shared_ptr<Configuration>(new Configuration(targetAgent,q));
        }
        targetAgent->setAndUpdate(*q_curr_tar);

    }
    else
    {
        if (_returnRepportcountMax < _returnRepportcount)
        {
            _returnRepportcountMax = _returnRepportcount;
            returnRepport = "IK_fail_for_target";
        }
        targetAgent->setAndUpdate(*q_curr_tar);
        return false;
    }
    targetAgent->setAndUpdate(*q_curr_tar);
    return res;
}

string Give::getDescr()
{
    string s = "Task: \nId \t\t= " + to_string(task_id) + "\nType \t\t= " + type;
    if (mainAgent)
        s+= "\nMain agent \t\t= " + mainAgent->getName();
    if (targetAgent)
        s+= "\nMain agent \t\t= " + targetAgent->getName();
    if (mainObject)
        s+= "\nMain object \t= " + mainObject->getName() ;
    return s;


}

string Give::getTextFromValues()
{
    return "Task: " + type + " : Give the object to the target agent"
            + "\n    agents: "
            + "\n        mainAgent:  [AGENT ID]"
            + "\n        targetAgent:  [AGENT ID]"
            + "\n    objects: "
            + "\n        mainObject: [OBJECT ID]"
            + "\n    datas:"
            + "\n        hand: [right OR left]"
            + "\n    points:"
            + "\n        target: [X Y THETA]";
}


