#include "goto.hpp"
#include "API/project.hpp"
#include "tools.hpp"

#include "planner/plannerFunctions.hpp"

#ifdef LIGHT_PLANNER
#include "LightPlanner-pkg.h"
#endif

#include "planner/TrajectoryOptim/plannarTrajectorySmoothing.hpp"


std::vector<p3d_point> currNavTraj;

//double costMaxDist = 2;
//double costMaxAngle = M_PI/2;


vector<p3d_point> _vHP;
double _vMin,_vMax,_dMax;
p3d_point _p0;

static double distTraveledWeight, angleAndRotWeight, distToObjWeight, angleWeight, distHWeight, distWeight, ADistTraveledWeight, AAngleAndRotWeight;


//####################### GoTo ########################

GoTo::GoTo(): PrivateTask()
{
    type = "goto";
    typeId = gotoPlace;
    nbMaxLoops = 50;
    nbTotPossibleTests = nbMaxLoops;
    _idCurrentSol = -1;

    _possibleMainAgentType.push_back(HRI_PR2);
    _possibleMainAgentType.push_back(HRI_KUKAROBOT);

    _possibleMainAgentType.push_back(HRI_HERAKLES);
    mainAgent = NULL;
    _targetObject = NULL;
}

bool GoTo::initAll()
{
    result = new TaskAllSolutions(this);
    nbParameters = 3;
    _armId = 0;

    //in combineCosts
    distTraveledWeight = 1;
    angleAndRotWeight = 1;
    distToObjWeight = 1;
    //in ComputeCosts
    angleWeight = 1;
    distHWeight = 1;
    distWeight = 1;
    //in Compute A*
    ADistTraveledWeight = 1;
    AAngleAndRotWeight = 1;

    return true;
//    return true;
}

bool GoTo::initialize()
{
    extractTraj = NULL;

    nbGikCalled = 0;
    nbMPCalled = 0;
    nbTests = 0;
    timeMp = 0;

    returnRepport = "ERROR";
    _returnRepportcount = 0;
    _returnRepportcountMax = 0;

    //remove col with object in hand
    for (unsigned int i = 0; i < _WS->getAttachements().size(); i++)
    {
        Attachement att = _WS->getAttachements().at(i);
        p3d_col_deactivate_robot_robot(att.agent->getRobotStruct(),att.object->getRobotStruct());
    }


    _vHP.clear();
    for (unsigned int i = 0; i < global_Project->getActiveScene()->getNumberOfRobots(); i++)
    {
        if (global_Project->getActiveScene()->getRobot(i)->isAgent())
        {
            if (global_Project->getActiveScene()->getRobot(i)->getHriAgent()->is_human == 1)
            {
                p3d_point HP;
                confPtr_t q = global_Project->getActiveScene()->getRobot(i)->getCurrentPos();
                HP.x = (*q)[6];
                HP.y = (*q)[7];
                HP.z = (*q)[11];
//                cout << "HP ( " << HP.x << " , " << HP.y << " , " << HP.z << " )" << endl;
                _vHP.push_back(HP);
            }
        }
    }

    return  initMotionPlaner(mainAgent);
}

bool GoTo::finalize()
{
    //put back col with object in hand
    for (unsigned int i = 0; i < _WS->getAttachements().size(); i++)
    {
        Attachement att = _WS->getAttachements().at(i);
        p3d_col_activate_robot_robot(att.agent->getRobotStruct(),att.object->getRobotStruct());
    }

//    ENV.setBool(Env::drawTraj,false);
    return true;
}


bool GoTo::run(multimap<string,string> agents_name,
               multimap<string,string> objects_name,
               multimap<string,p3d_point> point3d,
               multimap<string,string> additionalData,
               WorldState *WS, bool computeMP)
{

    WS->load();
    _WS = WS;
    if (!setAgents(agents_name))
    {
        cout << "ERROR: setAgents fails for task " << type << " with id " << task_id << endl;
        return false;
    }
    if (!_targetObject)
    {
        if (!setObjects(objects_name))
        {
            if (!setPoints(point3d))
            {
                cout << "ERROR: setObjects and setPoint fails for task " << type << " with id " << task_id << endl;
                return false;
            }
        }
    }
    if (!initAll())
    {
        cout << "ERROR: init fails for task " << type << " with id " << task_id << endl;
        return false;
    }
    if (!setdata(additionalData))
    {
        cout << "ERROR: setObjects fails for task " << type << " with id " << task_id << endl;
        return false;
    }

    if (!initialize())
    {
        cout << "ERROR: initialize fails for task " << type << " with id " << task_id << endl;
        return false;
    }
    if (!findCandidateSolutions())
    {
        cout << "ERROR: findCandidateSolutions fails for task " << type << " with id " << task_id << endl;
        return false;
    }
    bool res = false;
    int count = 0;
    while (!res)
    {
        cout << "count loop = " << count << endl;
        _returnRepportcount = 0;
        if(count++ > nbMaxLoops)
            break;
        if(!getRandomSol())
           continue;
        if(!findConfigurations())
            continue;
        if (computeMP)
            if(!findTrajectories())
                continue;
        if(!setToSolution(0,computeMP))
            continue;
        if(!testConstraints(0))
            continue;

        res = true;
    }

    if (res)
        returnRepport = "OK";

    computeCost(0);
    if (!finalize())
    {
        cout << "ERROR: finalize fails for task " << type << " with id " << task_id << endl;
        return false;
    }
    return res;
}

double GoTo::computeCost(int altId)
{

    TaskSolution* TS = result->getAlternativeWithId(altId);
    if (!TS)
        return 1000;
    vector<TaskSubSolution*> vtss = TS->getTSSByRobot(mainAgent);

    vector<p3d_point> vp;
    for (unsigned int i = 0; i < vtss.size(); i++)
    {
        TaskSubSolution* TSS = vtss.at(i);
        if (TSS->getSubSolutionName().find("navigate") != string::npos)
        {
            vp = TSS->getNavVector();
        }

    }
    if (vp.size() == 0)
        return 1000;
    double cost = 0;




    GeometricComputations GC = GeometricComputations();
    TS->setCost(GC.computeCost(_vHP,vp, angleWeight,distHWeight,1));

    return cost;
}

double GoTo::combineCosts(double distTraveled, double angleAndRot, double distToObj)
{
    return (distTraveled*distTraveledWeight + angleAndRot*angleAndRotWeight + distToObj*distToObjWeight)/ (distTraveledWeight + angleAndRotWeight + distToObjWeight);
//    return angleAndRot;//(distTraveled + 3*angleAndRot + distToObj)/5;
}

bool GoTo::compareCells(GTP2DCell* i, GTP2DCell* j)
{

//    return (i->getDistCurrentRobot() < j->getDistCurrentRobot());
    switch(PlanEnv->getInt(PlanParam::GTPTypeCase))
    {
    case 0:
    {
        i->setDrawn(true);
        j->setDrawn(true);
        double iDist, iCost, iCenter, jDist, jCost, jCenter;
        GeometricComputations GC = GeometricComputations();
        //distance from init
        iDist = i->getDistCurrentRobot()/_dMax;
        jDist = j->getDistCurrentRobot()/_dMax;

        double d = (_vMax + _vMin)/2;
        //angle and distance to humans
        vector<p3d_point> vp;
        p3d_point po;
        po.x = i->getCenter()[0];
        po.y = i->getCenter()[1];
        po.z = 0;
        vp.push_back(po);
        iCost = GC.computeCost(_vHP,vp,angleWeight,distHWeight,distWeight);

        iCenter = (fabs(GC.distBetweenPoints2D(_p0, po) - d))/((_vMax-_vMin)/2);

        vp.clear();
        po.x = j->getCenter()[0];
        po.y = j->getCenter()[1];
        po.z = 0;
        vp.push_back(po);
        jCost = GC.computeCost(_vHP,vp,angleWeight,distHWeight,distWeight);
        jCenter = (GC.distBetweenPoints2D(_p0, po) - d)/((_vMax-_vMin)/2);

        i->Ecolor = GoTo::combineCosts(iDist,iCost,iCenter);
        j->Ecolor = GoTo::combineCosts(jDist,jCost,jCenter);

        return i->Ecolor<j->Ecolor;
    }
    case 1:
        return (i->getDistCurrentRobot() < j->getDistCurrentRobot());
    default:
        return (i->getDistCurrentRobot() < j->getDistCurrentRobot());
    }

}

bool GoTo::findCandidateSolutions()
{

    //building the 2D grid
    std::vector<double> m_EnvSize;
    // initCostSpace
    m_EnvSize.resize(4);
    m_EnvSize[0] = XYZ_ENV->box.x1; m_EnvSize[1] = XYZ_ENV->box.x2;
    m_EnvSize[2] = XYZ_ENV->box.y1; m_EnvSize[3] = XYZ_ENV->box.y2;
    _grid = new GTP2DGrid(0.1,m_EnvSize);

    if (!_grid->initCylinders())
    {
        returnRepport = "no_cylinders";
        return false;
    }

    _grid->addAgentAndComputeNav(mainAgent);

    if (!isTargetAPoint)
    {
        cout << "target is an object or an agent" << endl;

        if(!_targetObject)
        {
            cout << "No target object/agent!" << endl;
            returnRepport = "no_target_agent_object";
            return false;
        }

        if (_targetObject->isAgent())
        {
            cout << "goto a target agent" << endl;
            _reachableCells = _grid->getCellsArroundAgent(mainAgent,_targetObject);
            if (_reachableCells.size() == 0)
            {
                cout << "the agent object " << _targetObject->getName() << " is not reachable by the robot" << endl;
                returnRepport = "not_reachable";
                return false;
            }
        }
        else
        {
            if(!_targetObject->getObjectRob())
            {
                cout<< _targetObject->getName() << " is not an object " << endl;
                returnRepport = "object_not_object";
                return false;
            }


            if (_targetObject->getObjectRob()->isManipulable())
            {
                cout << "goto a manipulable object" << endl;
                _reachableCells = _grid->getCellsArroundObject(mainAgent,_targetObject);
                if (_reachableCells.size() == 0)
                {
                    cout << "the manipulable object " << _targetObject->getName() << " is not reachable by the robot" << endl;
                    returnRepport = "not_reachable";
                    return false;
                }
            }
            else if (_targetObject->getObjectRob()->isSupport())
            {
                cout << "goto a support object" << endl;
                _reachableCells = _grid->getCellsArroundSupport(mainAgent,_targetObject);
                if (_reachableCells.size() == 0)
                {
                    cout << "the support " << _targetObject->getName() << " is not reachable by the robot" << endl;
                    return false;
                }
            }
        }



        _grid->setCurrentRobot(mainAgent);
        _vMin = 1000;
        _vMax = 0;
        _dMax = 0;

        confPtr_t q = _targetObject->getCurrentPos();
        _p0.x = (*q)[6];
        _p0.y = (*q)[7];
        _p0.z = 0;

        GeometricComputations GC = GeometricComputations();
        for (unsigned int i = 0; i<_reachableCells.size(); i++)
        {
            p3d_point p1;
            p1.x = _reachableCells.at(i)->getCenter()[0];
            p1.y = _reachableCells.at(i)->getCenter()[1];
            p1.z = 0;

            double distw = GC.distBetweenPoints2D(_p0,p1);
            if (_vMin > distw)
                _vMin = distw;
            else if (_vMax < distw)
                _vMax = distw;

            if (_dMax < _reachableCells.at(i)->getDistCurrentRobot())
                _dMax = _reachableCells.at(i)->getDistCurrentRobot();
        }




        std::sort(_reachableCells.begin(), _reachableCells.end(), GoTo::compareCells);

        _idCurrentSol = -1;
    }

    return true;
}

bool GoTo::getRandomSol()
{
    nbTests++;
    if (isTargetAPoint)
    {
        return true;
    }
    else
    {
        if (_reachableCells.size() == 0)
        {
            cout << "the agent cannot reach the object" << endl;
            return false;
        }

//        int id = p3d_random_integer(0,_reachableCells.size()-1);

        _idCurrentSol += 1;

        if (_reachableCells.size() <=_idCurrentSol )
        {
            return false;
        }


        _targetPoint.x = _reachableCells.at(_idCurrentSol)->getCenter()[0];
        _targetPoint.y = _reachableCells.at(_idCurrentSol)->getCenter()[1];
        _targetPoint.z = atan2((*_targetObject->getCurrentPos())[7] - _targetPoint.y, (*_targetObject->getCurrentPos())[6] - _targetPoint.x );

//        cout << "not yet implemented for objects" << endl;
        return true;
    }

}

bool GoTo::findConfigurations()
{
    confPtr_t q = _WS->getRobotConf(mainAgent);
    _WS->replaceConf(q);
    _WS->load();
    mainAgent->setAndUpdate(*q);
    //extract conf
    _extractConf = setArmsToNeutralPos(mainAgent);

    mainAgent->setAndUpdate(*_extractConf);

    //juste end conf
    _targetConf = (confPtr_t)new Configuration(mainAgent,_extractConf->getConfigStruct());


    if (mainAgent->getHriAgent() && mainAgent->getHriAgent()->type == HRI_KUKAROBOT)
    {
        double distMid = 0.3;
        _targetPoint.x = _targetPoint.x - distMid*cos(_targetPoint.z);
        _targetPoint.y = _targetPoint.y - distMid*sin(_targetPoint.z);

//        (*_targetConf)[21] = mainAgent->getJoint("IkJoint")->getJointStruct()->abs_pos[0][0];
//        (*_targetConf)[22] = mainAgent->getJoint("IkJoint")->getJointStruct()->abs_pos[0][1];

//        (*_targetConf)[21] = (*_extractConf)[21] + (_targetPoint.x - (*_extractConf)[6]);
//        (*_targetConf)[22] = (*_extractConf)[22] + (_targetPoint.y - (*_extractConf)[7]);

//        p3d_matrix4 M, res;
//        p3d_mat4Copy(p3d_mat4IDENTITY,M);
//        double theta = angle_limit_PI(_targetPoint.z - (*_extractConf)[11]);
//        M[0][0] = cos(theta);
//        M[1][0] = sin(theta);
//        M[0][1] = -sin(theta);
//        M[1][1] = cos(theta);

//        M[0][3] = (_targetPoint.x - (*_extractConf)[6]);
//        M[1][3] = (_targetPoint.y - (*_extractConf)[7]);



//        p3d_mat4Print(M,"M");

//        p3d_mat4Print(mainAgent->getJoint(12)->getJointStruct()->abs_pos, "absPos");

//        p3d_matMultXform(mainAgent->getJoint(12)->getJointStruct()->abs_pos,M,res);

//        p3d_mat4Print(res,"res");

//        double Tx, Ty, Tz;
//        double Rx, Ry, Rz;
//        p3d_mat4ExtractPosReverseOrder(res, &Tx, &Ty, &Tz, &Rx, &Ry, &Rz);

//        (*_targetConf)[21] = Tx;
//        (*_targetConf)[22] = Ty;
//        //(*_targetConf)[23] = Tz;
//        (*_targetConf)[24] = Rx;
//        (*_targetConf)[25] = Ry;
//        (*_targetConf)[26] = Rz;

//        cout << "Tx " << Tx << endl;
//        cout << "Ty " << Ty << endl;
//        cout << "Tz " << Tz << endl;
//        cout << "Rx " << Rx << endl;
//        cout << "Ry " << Ry << endl;
//        cout << "Rz " << Rz << endl;


    }

    cout << "main Agent is = " << mainAgent->getName() << endl;

    (*_targetConf)[6] = _targetPoint.x;
    (*_targetConf)[7] = _targetPoint.y;
    (*_targetConf)[11] = _targetPoint.z;

    confPtr_t qw = (confPtr_t)new Configuration(*_targetConf);
    (*qw)[16] = 0.01;
//    int status = -1;
//    compute_traj(_targetConf,qw,NULL,status);

    _targetConf->setAsNotTested();

    mainAgent->setAndUpdate(*_targetConf);

    if (mainAgent->getHriAgent() && mainAgent->getHriAgent()->type == HRI_KUKAROBOT)
    {
//        (*_targetConf)[21] = mainAgent->getJoint("IkJoint")->getJointStruct()->abs_pos[0][3];
//        (*_targetConf)[22] = mainAgent->getJoint("IkJoint")->getJointStruct()->abs_pos[1][3];

//        p3d_matrix4 M;
//        p3d_mat4Copy(p3d_mat4IDENTITY,M);
//        //x
//        M[0][0] = mainAgent->getJoint("IkJoint")->getJointStruct()->abs_pos[0][1];
//        M[1][0] = mainAgent->getJoint("IkJoint")->getJointStruct()->abs_pos[1][1];
//        M[2][0] = mainAgent->getJoint("IkJoint")->getJointStruct()->abs_pos[2][1];

//        //y
//        M[0][1] = mainAgent->getJoint("IkJoint")->getJointStruct()->abs_pos[0][2];
//        M[1][1] = mainAgent->getJoint("IkJoint")->getJointStruct()->abs_pos[1][2];
//        M[2][1] = mainAgent->getJoint("IkJoint")->getJointStruct()->abs_pos[2][2];

//        //z
//        M[0][2] = mainAgent->getJoint("IkJoint")->getJointStruct()->abs_pos[0][0];
//        M[1][2] = mainAgent->getJoint("IkJoint")->getJointStruct()->abs_pos[1][0];
//        M[2][2] = mainAgent->getJoint("IkJoint")->getJointStruct()->abs_pos[2][0];

        double Tx, Ty, Tz;
        double Rx, Ry, Rz;
        p3d_mat4ExtractPosReverseOrder(mainAgent->getJoint("IkJoint")->getJointStruct()->abs_pos, &Tx, &Ty, &Tz, &Rx, &Ry, &Rz);

        (*_targetConf)[21] = Tx;
        (*_targetConf)[22] = Ty;
        (*_targetConf)[23] = Tz;
        (*_targetConf)[24] = Rx;
        (*_targetConf)[25] = Ry;
        (*_targetConf)[26] = Rz;

    }

//    g3d_draw_allwin_active();
//    sleep(1);

//    bool ret = _targetConf->isInCollision();
    bool ret = p3d_col_test_robot(mainAgent->getRobotStruct(),1);
    if (ret) //colision
    {
        p3d_print_col_pair();
        cout << "target pos in collision" << endl;
        returnRepport = "target_pos_in_col";
        return false;
    }

    mainAgent->setAndUpdate(*_targetConf);
//    g3d_draw_allwin_active();
//    sleep(1);


    mainAgent->setAndUpdate(*q);
    return true;

}

bool GoTo::findTrajectories()
{
    // traj for extract conf
    setArmToUse(Right);

    if (*_extractConf != *_WS->getRobotConf(mainAgent))
    {

        int status = -1;
        if (mainAgent->getHriAgent()->type == HRI_KUKAROBOT)
            extractTraj = compute_traj(_WS->getRobotConf(mainAgent), _extractConf, _targetObject, status);
        else
            extractTraj = compute_traj_using_localPath(_WS->getRobotConf(mainAgent), _extractConf, _targetObject, status);

        _returnRepportcount++;
        if (!extractTraj && status != PLAN_STATUS_START_EQUALS_GOAL && status != PLAN_STATUS_FOUND)
        {
            cout << "no extract traj" << endl;
            if (_returnRepportcountMax < _returnRepportcount)
            {
                _returnRepportcountMax = _returnRepportcount;
                returnRepport = "no_extract_traj";
            }
            return false;
        }

        cout << "status = " << status << endl;
        if (status == PLAN_STATUS_START_EQUALS_GOAL)
        {
            extractTraj = NULL;
        }
    }



    // list the points to the configuration
    Eigen::Vector2d pos;
    pos[0] = _targetPoint.x;
    pos[1] = _targetPoint.y;
    GTP2DCell* cell = dynamic_cast<GTP2DCell*>(_grid->getCell(pos));
    if (!cell)
    {
        cout << "ERROR, no cell found for position " << _targetPoint.x << ", " << _targetPoint.y << endl;
        return false;
    }

    angleWeight = 1;
    distHWeight = 1;
    distWeight = 1;
    //in Compute A*
    ADistTraveledWeight = 1;
    AAngleAndRotWeight = 1;

    nbMPCalled++;

    vector<GTP2DCell*> path;
    switch(PlanEnv->getInt(PlanParam::GTPTypeCase))
    {
    case 0:
        path = _grid->getHumanAwarePath(mainAgent,cell,ADistTraveledWeight,AAngleAndRotWeight,angleWeight,distHWeight,distWeight);
        break;
    case 1:
        path = _grid->getPath(mainAgent,cell);
        break;
    default:
        returnRepport = "typeCase_not_correct";
        return false;
        break;
    }


    PlannarTrajectorySmoothing PTS(mainAgent);


    std::vector<Eigen::Vector2d,Eigen::aligned_allocator<Eigen::Vector2d> > vp;

    shared_ptr<Configuration> q_cur = _WS->getRobotConf(mainAgent);

    Eigen::Vector2d vinit;
    vinit[0] = (*q_cur)[6];
    vinit[1] = (*q_cur)[7];

    vp.push_back(vinit);

    for(unsigned int i = path.size(); i > 0; i--)
    {
        Eigen::Vector2d vtmp;
        vtmp[0] = path.at(i-1)->getCenter()[0];
        vtmp[1] = path.at(i-1)->getCenter()[1];
        vp.push_back(vtmp);
    }


    Eigen::Vector2d sout;
    sout[0] = _targetPoint.x;
    sout[1] = _targetPoint.y;

    vp.push_back(sout);

//    path_to_draw = vp;
//    ENV.setBool(Env::drawTraj,true);
//    g3d_draw_allwin_active();
//    sleep(2);

//    std::vector<Eigen::Vector2d,Eigen::aligned_allocator<Eigen::Vector2d> > vp2 = PTS.smoothTrajectory(mainAgent, vp);

    std::vector<Eigen::Vector3d,Eigen::aligned_allocator<Eigen::Vector3d> > vp3 = PTS.add3DimwithoutTrajChange(vp, mainAgent, 0.05);

    _traj.clear();
    for (unsigned int j = 0; j < vp3.size(); j++)
    {
        p3d_point ptmp;
        ptmp.x= vp3.at(j)[0];
        ptmp.y= vp3.at(j)[1];
        ptmp.z= vp3.at(j)[2];
        _traj.push_back(ptmp);
    }


    p3d_point vend;
    vend.x = _targetPoint.x;
    vend.y = _targetPoint.y;
    vend.z = _targetPoint.z;
    _traj.push_back(vend);
    currNavTraj = _traj;

    return true;
}

bool GoTo::computeSolutionTrajectories(int alternativeId)
{
    TaskSolution* TS = result->getAlternativeWithId(alternativeId);
    if(!TS)
    {
        cout << "no alternative for task " << to_string(task_id) << " with alternative id = " << alternativeId << endl;
        return false;
    }
    confPtr_t q = TS->getStopWS()->getRobotConf(mainAgent);
    _targetPoint.x = (*q)[6];
    _targetPoint.y = (*q)[7];
    _targetPoint.z = (*q)[11];

    findTrajectories();

    return true;
}

bool GoTo::setToSolution(int solutionId, bool computeMP)
{

    TaskSolution* TS = new TaskSolution(result,solutionId);

    TS->setTrajComputed(computeMP);

    TS->setTaskConstraints(_currConstraints);

//    TS->nbGikCalled = nbGikCalled;
//    TS->nbMPCalled = nbMPCalled;
//    TS->nbParameters = nbParameters;
//    TS->nbTests = nbTests;
    TaskSubSolution* TSS0 = new TaskSubSolution(TS, "extract", 0, extractTraj, mainAgent, 0);
    TaskSubSolution* TSS1 = new TaskSubSolution(TS, "navigate", 1, NULL, mainAgent, 0);

    TSS0->setArmId(2);
    TSS1->setNavigation(true);
    TSS1->setNavTraj(_traj);

    vector<TaskSubSolution*> v1;
    v1.push_back(TSS1);
    TSS0->setNext(v1);

    TS->setStartWS(_WS);
    TSS0->setStartState(_WS);

    WorldState* WS1 = new WorldState(_WS);
    WS1->replaceConf(_extractConf);

    _WS->load();
    vector<Attachement> atts;
    for (unsigned int i = 0; i < _WS->getAttachements().size(); i++)
    {
        Attachement att = _WS->getAttachements().at(i);
        if (att.agent == mainAgent)
        {
            PrivateTask::attachObject(mainAgent, att.object, true ,att.armId);
            atts.push_back(att);
        }
    }
    mainAgent->setAndUpdate(*_extractConf);
    for (unsigned int i = 0; i < atts.size(); i++)
    {
        WS1->replaceConf(atts.at(i).object->getCurrentPos());
    }

    TSS0->setEndState(WS1);
    TSS1->setStartState(WS1);
    for (unsigned int i = 0; i < atts.size(); i++)
    {
        PrivateTask::attachObject(atts.at(i).agent, atts.at(i).object, false ,atts.at(i).armId);
    }

    WorldState* WS2 = new WorldState(WS1);
    WS2->replaceConf(_targetConf);

    _WS->load();

    for (unsigned int i = 0; i < atts.size(); i++)
    {
        PrivateTask::attachObject(atts.at(i).agent, atts.at(i).object, true ,atts.at(i).armId);
    }
    mainAgent->setAndUpdate(*_targetConf);
    for (unsigned int i = 0; i < atts.size(); i++)
    {
        WS2->replaceConf(atts.at(i).object->getCurrentPos());
    }

    TSS1->setEndState(WS2);

    WS2->setTCForNextTask(_currConstraints);

    TS->setStopWS(WS2);
    WS2->setTSThatEndWithThisWS(TS);

    return true;
}


bool GoTo::setAgents(multimap<string,string> agents_name)
{
    if (agents_name.count("targetAgent") == 1)
    {
        _targetObject = global_Project->getActiveScene()->getRobotByName(agents_name.find("targetAgent")->second);
        if (!_targetObject)
        {
            cout << "error in target agent name :" << agents_name.find("targetAgent")->second << endl;
            returnRepport = "error_targetagent_name";
            return false;
        }
        isTargetAPoint = false;
    }
    return setMainAgent(agents_name);
}

bool GoTo::setObjects(multimap<string,string> objects_name)
{
    if (objects_name.count("target") != 1)
    {
        cout << "one and only one target at a time" << endl;
        return false;
    }
    _targetObject = global_Project->getActiveScene()->getRobotByName(objects_name.find("target")->second);
    if (!_targetObject)
        return false;
    isTargetAPoint = false;
    return true;
}

bool GoTo::setPoints(multimap<string,p3d_point> points)
{
    if (points.count("target") != 1)
    {
        cout << "one and only one target at a time" << endl;
        return false;
    }
    _targetPoint = points.find("target")->second;
    isTargetAPoint = true;
    return true;
}

bool GoTo::setdata(multimap<string,string> additionalData)
{
    int countCost = additionalData.count("cost");
    string costStr = "";
    if (countCost > 1)
    {
        returnRepport = "too_many_costs_in_map";
        return false;
    }
    else if (countCost == 0)
    {
        cout << "using default costs: 1 10 0 2 1 0 0 1" << endl;
        costStr = "1 10 0 2 1 0 0 1";
    }
    else if (countCost == 1)
        costStr = additionalData.find("cost")->second;

    std::istringstream buf(costStr);
    std::istream_iterator<std::string> beg(buf), end;

    std::vector<std::string> tokens(beg, end);

    if (tokens.size() != 8)
    {
        returnRepport = "wrong_nb_of_costs";
        return false;
    }

    //in combineCosts
    distTraveledWeight = atof(tokens.at(0).c_str());
    angleAndRotWeight = atof(tokens.at(1).c_str());
    distToObjWeight = atof(tokens.at(2).c_str());
    //in ComputeCosts
    angleWeight = atof(tokens.at(3).c_str());
    distHWeight = atof(tokens.at(4).c_str());
    distWeight = atof(tokens.at(5).c_str());
    //in Compute A*
    ADistTraveledWeight = atof(tokens.at(6).c_str());
    AAngleAndRotWeight = atof(tokens.at(7).c_str());

//    for(unsigned int i = 0; i< tokens.size(); i++)
//        std::cout << '"' << tokens.at(i) << '"' << '\n';

//    cout <<distTraveledWeight << endl;
//    cout <<angleAndRotWeight << endl;
//    cout <<distToObjWeight << endl;
//    cout <<angleWeight << endl;
//    cout <<distHWeight << endl;
//    cout <<distWeight << endl;
//    cout <<ADistTraveledWeight << endl;
//    cout <<AAngleAndRotWeight << endl;


    return true;
}

void GoTo::smoothSolution(int alternativeId)
{

    if (mainAgent->getHriAgent()->type == HRI_KUKAROBOT)
    {
        for (int i = 0; i < result->getAlternativeWithId(alternativeId)->getNbsubSolPerRobot(getMainAgent()); i++)
        {
            TaskSubSolution* tss = result->getAlternativeWithId(alternativeId)->getTSSByRobot(getMainAgent()).at(i);
            if (tss->getSubSolutionName() == "extract")
            {
                tss->getRobot()->getRobotStruct()->tcur = tss->getTraj();
                tss->getStartState()->load();
                tss->refineTraj();
                break;
            }
        }
        result->getAlternativeWithId(alternativeId)->getStartWS()->load();
    }


    TaskSolution* TS = result->getAlternativeWithId(alternativeId);
    if (!TS)
        return;
    vector<TaskSubSolution*> vtss = TS->getTSSByRobot(mainAgent);
    TaskSubSolution* TSS = NULL;
    for (unsigned int i = 0; i < vtss.size(); i++)
    {
        TSS = vtss.at(i);
        if (TSS->getSubSolutionName().find("navigate") != string::npos)
        {
            break;
        }
    }
    if (!TSS)
        return;

    vector<p3d_point> vp3d = TSS->getNavVector();

    GeometricComputations GC = GeometricComputations();

    Robot *cyl;
//    r->getHriAgent()->type
    if (mainAgent->getHriAgent()->type == HRI_HERAKLES)
        cyl = global_Project->getActiveScene()->getRobotByName("HUMCYLINDER");
    else if (mainAgent->getHriAgent()->type == HRI_PR2 || mainAgent->getHriAgent()->type == HRI_KUKAROBOT )
        cyl = global_Project->getActiveScene()->getRobotByName("PR_2CYLINDER");

    p3d_col_deactivate_robot_robot(cyl->getRobotStruct(),mainAgent->getRobotStruct());
    _traj = GC.smoothTraj(vp3d,_vHP,cyl,500,-1,distWeight,distHWeight,angleWeight);

    currNavTraj = _traj;
//    path_to_draw.clear();
//    for (unsigned int i = 0; i<_traj.size(); i++)
//    {
//        Eigen::Vector2d v;
//        v[0] = _traj.at(i).x;
//        v[1] = _traj.at(i).y;
//        path_to_draw.push_back(v);
//    }
//    g3d_draw_allwin_active();

//    std::vector<Eigen::Vector2d,Eigen::aligned_allocator<Eigen::Vector2d> > vp;

//    for(unsigned int i =0; i < vp3d.size(); i++)
//    {
//        Eigen::Vector2d vtmp;
//        vtmp[0] = vp3d.at(i).x;
//        vtmp[1] = vp3d.at(i).y;
//        vp.push_back(vtmp);
//    }

//    PlannarTrajectorySmoothing PTS(mainAgent);
//    std::vector<Eigen::Vector2d,Eigen::aligned_allocator<Eigen::Vector2d> > vp2 = PTS.smoothTrajectory(mainAgent, vp);
//    path_to_draw = vp2;

////    g3d_draw_allwin_active();
////    sleep(1);


//    std::vector<Eigen::Vector3d,Eigen::aligned_allocator<Eigen::Vector3d> > vp3 = PTS.add3DimwithoutTrajChange(vp2, mainAgent, 0.05);

//    _traj.clear();
//    for (unsigned int j = 0; j < vp3.size(); j++)
//    {
//        p3d_point ptmp;
//        ptmp.x= vp3.at(j)[0];
//        ptmp.y= vp3.at(j)[1];
//        ptmp.z= vp3.at(j)[2];
//        _traj.push_back(ptmp);
//    }

//    p3d_point vend;
//    vend.x = _targetPoint.x;
//    vend.y = _targetPoint.y;
//    vend.z = _targetPoint.z;
//    _traj.push_back(vend);

    TSS->setNavTraj(_traj);

//    cout << "no smoothing can be applied on this task " << task_id << " alt " << alternativeId  << endl;
}

int GoTo::checkHandAttachement()
{
    int res = 0;
    bool rarm = false;
    bool larm = false;
    for (unsigned int i = 0; i < _WS->getAttachements().size(); i++)
    {
        Attachement att = _WS->getAttachements().at(i);
        if (att.agent->getName() == mainAgent->getName())
        {
            if (att.armId == 0)
                rarm = true;
            else if (att.armId == 1)
                larm = true;
        }
    }
    if (rarm && larm)
        res = 3;
    else if (rarm)
        res = 1;
    else if (larm)
        res = 2;
    return res;
}

string GoTo::getDescr()
{
    string s= "Task: \nId \t\t= " + to_string(task_id) + "\nType \t\t= " + type;
    if (mainAgent)
        s+= "\nMain agent \t\t= " + mainAgent->getName();
    if (_targetObject)
        s+= "\nMain object \t\t= " + _targetObject->getName();

    return s;
}

string GoTo::getTextFromValues()
{
    return "Task: " + type + " : navigate to target (either point or object)"
            + "\n    agents: "
            + "\n        mainAgent: [AGENT ID]"
            + "\n    objects: "
            + "\n        target: [OBJECT ID]"
            + "\n    points:"
            + "\n        target: [X Y THETA]";
}





