#include "moveto.hpp"
#include "API/project.hpp"
#include "tools.hpp"

#include "planner/plannerFunctions.hpp"

#ifdef LIGHT_PLANNER
#include "LightPlanner-pkg.h"
#endif


#include <cstdlib>
#include <iostream>
#include <fstream>
#include <vector>
#include <map>
#include <algorithm>
#include "boost/lexical_cast.hpp"
using namespace std;

//####################### parser########################

map<string, vector<double> > getConfiguration(string configurationName) {
    ifstream jsonFile;
    string str = string(getenv("HOME_MOVE3D")) + "/statFiles/PR2_configuration";
    jsonFile.open(str.c_str(), ios::in);
    map<string, vector<double> > result;
    if (jsonFile.is_open()) {
        string line;

        bool found = false;
        //look for the configuration in the file, line by line
        while (getline(jsonFile, line) && !found) {
//            cout << line << "\n";
//            unsigned int i = line.find(configurationName);
            if (line.find(configurationName) != string::npos) {
                found = true;
                break;
            }
        }
//        getline(jsonFile, line);
//        cout << "dummy line = " << line << endl;
        //if we found the configuration get the values for the joint groups
        if (found) {
            bool completed = false;
            while (getline(jsonFile, line) && !completed) {
                cout << line << "\n";
                //the format of each line is "jointGroup" : [ x y z ... ],
                if (line.find("}") != string::npos) { //configuration ends with a blank line with only a }
                    completed = true;
                } else {
                    //first we get the joint group name
                    unsigned int i = line.find("\"", 0);
                    unsigned int j = line.find("\"", i + 1);
                    if (i != string::npos && j != string::npos) {
                        string jointGroup = line.substr(i + 1, j - i - 1); //name of the joint group
                        cout << jointGroup << "\n";

                        vector<double> configuration;
                        //now we get the values
                        unsigned int i = line.find("[", j + 1);
                        while (line[i] != ']' && i < line.size()) {
                            j = i + 1;
                            i++;
                            while (line[i] != ',' && line[i] != ']' && i < line.size()) {
                                i++;
                            }
                            if (line[i] == ',' || line[i] == ']') {
                                string toConvert = line.substr(j, i - j);
//                                cout << toConvert << "\n";
                                toConvert.erase(remove_if(toConvert.begin(), toConvert.end(), ::isspace), toConvert.end());
                                double aNumber = boost::lexical_cast<double>(toConvert);
                                //  cout << aNumber << "\n";

                                configuration.push_back(aNumber);
                            }

                        }

                        result[jointGroup] = configuration;
                    }
                }
            }
        }
        return result;

        jsonFile.close();
    }
    return result;

}




//####################### Moveto ########################

MoveTo::MoveTo()
{
    type = "moveto";
    typeId = moveTo;
    _configurationName = "";
    nbMaxLoops = 3;

    nbTotPossibleTests = nbMaxLoops;

    nbParameters = 0;
    nbGikCalled = 0;
    nbMPCalled = 0;
    nbTests = 0;

    _possibleMainAgentType.push_back(HRI_PR2);
    _possibleMainAgentType.push_back(HRI_HERAKLES);
    _possibleMainAgentType.push_back(HRI_KUKAROBOT);
}

bool MoveTo::initAll()
{
    _traj = NULL;
    result = new TaskAllSolutions(this);
    _armId =0;
    _returnRepportcount = 0;

    mainAgent = NULL;
    return true;
}

bool MoveTo::initialize()
{
    _traj = NULL;
    nbGikCalled = 0;
    nbMPCalled = 0;
    nbTests = 0;
    timeMp = 0;

    returnRepport = "ERROR";
    _returnRepportcount = 0;
    _returnRepportcountMax = 0;

    _fileInfo.clear();

    //remove col with object in hand
    for (unsigned int i = 0; i < _WS->getAttachements().size(); i++)
    {
        Attachement att = _WS->getAttachements().at(i);
        p3d_col_deactivate_robot_robot(att.agent->getRobotStruct(),att.object->getRobotStruct());
    }

    armUsed = -1;

    return  initMotionPlaner(mainAgent);
}

bool MoveTo::finalize()
{
    //put back col with object in hand
    for (unsigned int i = 0; i < _WS->getAttachements().size(); i++)
    {
        Attachement att = _WS->getAttachements().at(i);
        p3d_col_activate_robot_robot(att.agent->getRobotStruct(),att.object->getRobotStruct());
    }

    ENV.setBool(Env::drawTraj,false);
    return true;
}


bool MoveTo::run(multimap<string,string> agents_name,
                 multimap<string,string> objects_name,
                 multimap<string,p3d_point> point3d,
                 multimap<string,string> additionalData,
                 WorldState *WS, bool computeMP)
{

    WS->load();
    _WS = WS;
    if (!initAll())
    {
        cout << "ERROR: init fails for task " << type << " with id " << task_id << endl;
        return false;
    }
    if (!setAgents(agents_name))
    {
        cout << "ERROR: setAgents fails for task " << type << " with id " << task_id << endl;
        return false;
    }
    if (!isPossibleMainAgent(mainAgent))
    {
        cout << "ERROR: mainAgent is not in the list of possible agent for task " << type << " with id " << task_id << endl;
        return false;
    }
    if (!setdata(additionalData))
    {
        cout << "ERROR: setData fails for task " << type << " with id " << task_id << endl;
        return false;
    }
    if (!initialize())
    {
        cout << "ERROR: initialize fails for task " << type << " with id " << task_id << endl;
        return false;
    }

    if (!findCandidateSolutions())
    {
        cout << "ERROR: findCandidateSolutions fails for task " << type << " with id " << task_id << endl;
        return false;
    }
    bool res = false;
    int count = 0;
    while (!res)
    {
        cout << "count loop = " << count << endl;
        _returnRepportcount = 0;
        if(count++ > nbMaxLoops)
            break;
        if(!getRandomSol())
           continue;
        if(!findConfigurations())
            continue;
        if (computeMP)
            if(!findTrajectories())
                continue;
        if(!setToSolution(0,computeMP))
            continue;
        if(!testConstraints(0))
            continue;

        res = true;
    }

    if (res)
        returnRepport = "OK";

    computeCost(0);
    if (!finalize())
    {
        cout << "ERROR: finalize fails for task " << type << " with id " << task_id << endl;
        return false;
    }
    return res;
}


TaskSolution* MoveTo::findAlternative(bool computeMP)
{
    int count = 0;
    if (!getResult()->getAlternativeWithId(0))
    {
        cout<< "This task has no first solution, no alternative can be computed in that case" << endl;
        return NULL;
    }
    getResult()->getAlternativeWithId(0)->getStartWS()->load(); //

    if (!initialize())
    {
        cout << "ERROR: initialize fails for task " << type << " with id " << task_id << endl;
        return false;
    }

    if (!findCandidateSolutions())
    {
        cout << "ERROR: findCandidateSolutions fails for task " << type << " with id " << task_id << endl;
        return false;
    }
    int id = -1;
    bool res = false;
    while (!res)
    {
        cout << "count loop = " << count << endl;
        _returnRepportcount = 0;
        if(count++ > nbMaxLoops)
            break;
        if(!getRandomSol())
           continue;
        if(!findConfigurations())
           continue;
        if (computeMP)
            if(!findTrajectories())
                continue;
        id = result->getNextId();

        if(!setToSolution(id,computeMP))
           continue;
        if(!testConstraints(id))
            continue;
        res = true;
    }
    if (res)
        returnRepport = "OK";

    computeCost(id);

    if (!finalize())
    {
        cout << "ERROR: finalize fails for task " << type << " with id " << task_id << endl;
        return false;
    }
    if (res && id != -1)
        return getResult()->getAlternativeWithId(id);
    return NULL;
}



bool MoveTo::findCandidateSolutions()
{
    //Read file
    if (mainAgent->getHriAgent() && mainAgent->getHriAgent()->is_human)
        return true;
    _fileInfo = getConfiguration(_configurationName);
    if (_fileInfo.size() ==0)
    {
        cout << "Notthing in the file for this configuration : " << _configurationName << endl;
        returnRepport = "no_conf_in_file";
        return false;
    }
    return true;
}

bool MoveTo::getRandomSol()
{
    return true;
}

bool MoveTo::findConfigurations()
{
    if (!mainAgent->getHriAgent())
        return false;

    if (mainAgent->getHriAgent()->type == HRI_HERAKLES)
    {
        _targetConf = mainAgent->getCurrentPos();
        for (unsigned int i = 12; i < 65; i++)
            (*_targetConf)[i] = 0;
        (*_targetConf)[18] = 1.396;
        (*_targetConf)[22] = 1.894;
        (*_targetConf)[27] = -1.396;
        (*_targetConf)[31] = -1.894;

    }
    else if (mainAgent->getHriAgent()->type == HRI_PR2)
    {
        //set conf to what was red

        shared_ptr<Configuration> q_cur =  (shared_ptr<Configuration>)(new Configuration(*_WS->getRobotConf(mainAgent)));


        if (_fileInfo.count("RARM") > 0)
        {
            vector<double> p = _fileInfo.find("RARM")->second;
            setArmToUse(Right);
            _returnRepportcount++;
            if (p.size() != 7)
            {
                if (_returnRepportcountMax < _returnRepportcount)
                {
                    _returnRepportcountMax = _returnRepportcount;
                    returnRepport = "nb_of_dof_unmatched_rarm";
                }
                return false;
            }

            for (int i = 0; i <= 6; i++)
            {
                (*q_cur)[16+i] = p.at(i);
            }

            armUsed = 0;

        }


        if (_fileInfo.count("LARM") > 0)
        {
            vector<double> p = _fileInfo.find("LARM")->second;
            setArmToUse(Left);
            _returnRepportcount++;
            if (p.size() != 7)
            {
                if (_returnRepportcountMax < _returnRepportcount)
                {
                    _returnRepportcountMax = _returnRepportcount;
                    returnRepport = "nb_of_dof_unmatched_rarm";
                }
                return false;
            }

            for (int i = 0; i <= 6; i++)
            {
                (*q_cur)[25+i] = p.at(i);
            }

            if(armUsed == 0)
                armUsed = 2;
            else
                armUsed = 1;

        }

        if (_fileInfo.count("HEAD") > 0)
        {
            vector<double> p = _fileInfo.find("HEAD")->second;
            setArmToUse(Right);
            _returnRepportcount++;
            if (p.size() != 2)
            {
                if (_returnRepportcountMax < _returnRepportcount)
                {
                    _returnRepportcountMax = _returnRepportcount;
                    returnRepport = "nb_of_dof_unmatched_rarm";
                }
                return false;
            }

            for (int i = 0; i <= 1; i++)
            {
                (*q_cur)[13+i] = p.at(i);
            }

            armUsed = 3;

        }


        _targetConf = q_cur;
    }
    else if (mainAgent->getHriAgent()->type == HRI_KUKAROBOT)
    {
        shared_ptr<Configuration> q_cur =  (shared_ptr<Configuration>)(new Configuration(*_WS->getRobotConf(mainAgent)));

        for (int i = 21; i <= 26; i++)
            (*q_cur)[i] = 0;


        if (_fileInfo.count("RARM") > 0)
        {
            vector<double> p = _fileInfo.find("RARM")->second;
            setArmToUse(Right);
            _returnRepportcount++;
            if (p.size() != 7)
            {
                if (_returnRepportcountMax < _returnRepportcount)
                {
                    _returnRepportcountMax = _returnRepportcount;
                    returnRepport = "nb_of_dof_unmatched_rarm";
                }
                return false;
            }

            for (int i = 0; i <= 6; i++)
            {
                (*q_cur)[12+i] = p.at(i);
            }
        }

        mainAgent->setAndUpdate(*q_cur);



        (*q_cur)[21] = mainAgent->getJoint("IkJoint")->getJointStruct()->abs_pos[0][3];
        (*q_cur)[22] = mainAgent->getJoint("IkJoint")->getJointStruct()->abs_pos[1][3];
        (*q_cur)[23] = mainAgent->getJoint("IkJoint")->getJointStruct()->abs_pos[2][3];

        p3d_matrix4 M;
        p3d_mat4Copy(p3d_mat4IDENTITY,M);
        //x
        M[0][0] = mainAgent->getJoint("IkJoint")->getJointStruct()->abs_pos[0][1];
        M[1][0] = mainAgent->getJoint("IkJoint")->getJointStruct()->abs_pos[1][1];
        M[2][0] = mainAgent->getJoint("IkJoint")->getJointStruct()->abs_pos[2][1];

        //y
        M[0][1] = mainAgent->getJoint("IkJoint")->getJointStruct()->abs_pos[0][2];
        M[1][1] = mainAgent->getJoint("IkJoint")->getJointStruct()->abs_pos[1][2];
        M[2][1] = mainAgent->getJoint("IkJoint")->getJointStruct()->abs_pos[2][2];

        //z
        M[0][2] = mainAgent->getJoint("IkJoint")->getJointStruct()->abs_pos[0][0];
        M[1][2] = mainAgent->getJoint("IkJoint")->getJointStruct()->abs_pos[1][0];
        M[2][2] = mainAgent->getJoint("IkJoint")->getJointStruct()->abs_pos[2][0];



        double Tx, Ty, Tz;
        double Rx, Ry, Rz;
        p3d_mat4ExtractPosReverseOrder(M, &Tx, &Ty, &Tz, &Rx, &Ry, &Rz);

        (*q_cur)[24] = Rx;
        (*q_cur)[25] = Ry;
        (*q_cur)[26] = Rz;

        _targetConf = q_cur;
    }
    else
        return false;


    return true;
}

bool MoveTo::findTrajectories()
{

    for(unsigned int i = 0; i < _WS->getAttachements().size(); i ++)
    {
        Attachement aa = _WS->getAttachements().at(i);
        PrivateTask::attachObject(aa.agent,aa.object,true);
    }
    int status = -1;




    _WS->load();
//    mainAgent->setAndUpdate(*_targetConf);
//    g3d_draw_allwin_active();
//    sleep(3);

    if (mainAgent->getHriAgent() && mainAgent->getHriAgent()->is_human)
    {
        pqp_deactivate_robot_collision(mainAgent->getRobotStruct());
        _traj = compute_traj_using_localPath(_WS->getRobotConf(mainAgent), _targetConf, NULL, status);
        pqp_activate_robot_collision(mainAgent->getRobotStruct());
    }
    else
        _traj = compute_traj(_WS->getRobotConf(mainAgent), _targetConf, NULL, status);


    _returnRepportcount++;
    if (!_traj)
    {
        cout << "no traj" << endl;
        if (_returnRepportcountMax < _returnRepportcount)
        {
            _returnRepportcountMax = _returnRepportcount;
            returnRepport = "not_traj_found";
        }
        return false;
    }

    if (mainAgent->getHriAgent() && mainAgent->getHriAgent()->is_human)
        return true;
    setGraph(new Graph(*API_activeGraph));

    return true;
}

bool MoveTo::computeSolutionTrajectories(int alternativeId)
{
    TaskSolution* TS = result->getAlternativeWithId(alternativeId);
    if(!TS)
    {
        cout << "no alternative for task " << to_string(task_id) << " with alternative id = " << alternativeId << endl;
        return false;
    }

    _WS->load();
    confPtr_t q = TS->getStopWS()->getRobotConf(mainAgent);
    _targetConf = q;
    findTrajectories();

    result->getAlternativeWithId(alternativeId)->getSubSolutionWithId(0)->setTraj(_traj);

    return true;
}

bool MoveTo::setToSolution(int solutionId, bool computeMP)
{

    TaskSolution* TS = new TaskSolution(result,solutionId);

    TS->setTrajComputed(computeMP);
    TS->setTaskConstraints(_currConstraints);
//    TS->nbGikCalled = nbGikCalled;
//    TS->nbMPCalled = nbMPCalled;
//    TS->nbParameters = nbParameters;
//    TS->nbTests = nbTests;


    TaskSubSolution* TSS1 = new TaskSubSolution(TS, "move", 0, _traj, mainAgent, 0);

    TSS1->setStartState(_WS);
    TS->setStartWS(_WS);
    WorldState* WS1 = new WorldState(_WS);

    for(unsigned int i = 0; i < _WS->getAttachements().size(); i ++)
    {
        Attachement aa = _WS->getAttachements().at(i);
        WS1->addAttachement(aa.agent,aa.object,aa.grasp,aa.armId);
    }

    bool t = WS1->replaceConf(_targetConf);
    if (!t)
    {
        cout << "ERROR: error with robots names" << endl;
        return false;
    }

//    _WS->load();
//    mainAgent->setAndUpdate(*_targetConf);
//    for(unsigned int i = 0; i < _WS->getAttachements().size(); i ++)
//    {
//        Attachement aa = _WS->getAttachements().at(i);
//        WS1->replaceConf(aa.object->getCurrentPos());
//    }

    _WS->load();
    vector<Attachement> atts;
    for (unsigned int i = 0; i < _WS->getAttachements().size(); i++)
    {
        Attachement att = _WS->getAttachements().at(i);
        if (att.agent == mainAgent)
        {
            PrivateTask::attachObject(mainAgent, att.object, true ,att.armId);
            atts.push_back(att);
        }
    }
    mainAgent->setAndUpdate(*_targetConf);
    for (unsigned int i = 0; i < atts.size(); i++)
    {
        WS1->replaceConf(atts.at(i).object->getCurrentPos());
    }


    TSS1->setEndState(WS1);

    WS1->setTCForNextTask(_currConstraints);

    TS->setStopWS(WS1);
    WS1->setTSThatEndWithThisWS(TS);

    return true;
}

bool MoveTo::setAgents(multimap<string,string> agents_name)
{
    setMainAgent(agents_name);
    if (!mainAgent || !mainAgent->getHriAgent())
    {
       returnRepport = "no_mainAgent";
       return false;
    }
    if (mainAgent->getHriAgent()->type != HRI_PR2
            && mainAgent->getHriAgent()->type != HRI_HERAKLES
            && mainAgent->getHriAgent()->type != HRI_KUKAROBOT)
    {
       cout << "this function is made to work with PR2, herakles and kuka only" << endl;
       returnRepport = "not_pr2_nor_herakles_nor_kuka_used";
       return false;
    }
    return true;

}

bool MoveTo::setObjects(multimap<string,string> objects_name)
{
    return true;
}

bool MoveTo::setPoints(multimap<string,p3d_point> points)
{
    return true;
}

bool MoveTo::setdata(multimap<string,string> additionalData)
{
    if (mainAgent->getHriAgent() && mainAgent->getHriAgent()->is_human)
        return true;
    if (additionalData.count("confName") != 1)
    {
        cout << "ERROR: no configuration name for " << type << " task." << endl;
        returnRepport = "no_confName";
        return false;
    }
    _configurationName = additionalData.find("confName")->second;
    if (_configurationName == "")
    {
        returnRepport = "configurationName_empty";
        return false;
    }
    return true;
}

void MoveTo::smoothSolution(int alternativeId)
{
    for (int i = 0; i < result->getAlternativeWithId(alternativeId)->getNbsubSolPerRobot(getMainAgent()); i++)
    {
        TaskSubSolution* tss = result->getAlternativeWithId(alternativeId)->getTSSByRobot(getMainAgent()).at(i);
        if (tss->getSubSolutionName() == "move")
        {
            tss->getRobot()->getRobotStruct()->tcur = tss->getTraj();
            tss->getStartState()->load();
            _WS->load();
            tss->refineTraj();
            break;
        }
    }
    result->getAlternativeWithId(alternativeId)->getStartWS()->load();
}


string MoveTo::getDescr()
{
    string s = "Task: \nId \t\t= " + to_string(task_id) + "\nType \t\t= " + type;
    if (mainAgent)
        s+= "\nMain agent \t\t= " + mainAgent->getName();
    return s;
}

string MoveTo::getTextFromValues()
{
    return "Task: " + type + " : move the right hand and/or the left hand to confName, defined in the file"
            + "\n    agents: "
            + "\n        mainAgent: [AGENT ID]"
            + "\n    datas: "
            + "\n        confName: [CONFIGURATION NAME]"
            + "\n            RARM: [7 double Vect]"
            + "\n            LARM: [7 double Vect]";
}
