#include "pickhuman.hpp"
#include "API/project.hpp"
#include "tools.hpp"

#include "planner/plannerFunctions.hpp"
#include <libmove3d/hri/hri.h>

#ifdef LIGHT_PLANNER
#include "LightPlanner-pkg.h"
#endif

double distToObj = 0.05;

PickHuman::PickHuman() : Pick()
{
    type = "pickHuman";
    typeId = pickHuman;
    current_grasp_id = -1;
    _armId = -1;
    nbMaxLoops = 10;

    nbGikCalled = 0;
    nbMPCalled = 0;
    nbTests = 0;
    nbParameters = 1;

    nbTotPossibleTests = nbMaxLoops;

    returnRepport = "ERROR";
    _returnRepportcount = 0;
    _returnRepportcountMax = 0;

    //_possibleMainAgentType.push_back("HERAKLES");
    _possibleMainAgentType.push_back(HRI_HERAKLES);
}

bool PickHuman::initAll()
{
    result = new TaskAllSolutions(this);

    return true;
//    return true;
}

bool PickHuman::initialize()
{
    approachTraj = NULL;
    graspTraj = NULL;
    escapeTraj = NULL;
    closeHandTraj = NULL;


    nbGikCalled = 0;
    nbMPCalled = 0;
    nbTests = 0;
    timeMp = 0;

    returnRepport = "ERROR";
    _returnRepportcount = 0;
    _returnRepportcountMax = 0;

    return true;
}


bool PickHuman::findCandidateSolutions()
{
//    setArmToUse(Right);

    if (!checkHRIAgent(mainAgent))
        return false;

    if (_noArm)
        setArmToUse(_availableArms.at(p3d_random_integer(0,_availableArms.size()-1)));

    if (!setMainAgentHand())
        return false;

    if (checkAttachement())
    {
        returnRepport = "attachement_already_exist";
        return false;
    }


    ag_reach = mainAgent->getHriAgent()->manip->reach_tolerance;


    agShoulder.x = mainAgent->getJoint(mainAgent->getHriAgent()->manip->shoulder_joint_ids[getArmId()])->getJointStruct()->abs_pos[0][3];
    agShoulder.y = mainAgent->getJoint(mainAgent->getHriAgent()->manip->shoulder_joint_ids[getArmId()])->getJointStruct()->abs_pos[1][3];
    agShoulder.z = mainAgent->getJoint(mainAgent->getHriAgent()->manip->shoulder_joint_ids[getArmId()])->getJointStruct()->abs_pos[2][3];


    if (!mainObject->getObjectRob()->isManipulable())
    {
        returnRepport = "obj_not_manipulable";
        return false;
    }
    return true;
}

bool PickHuman::getRandomSol()
{
    GeometricComputations GC = GeometricComputations();
    vector<p3d_point> vp = GC.getUpperPointsInObject(mainObject,0.03);

    if (vp.size() >= 4)
    {
        targetPoint.x = (vp.at(0).x +vp.at(3).x)/2;
        targetPoint.y = (vp.at(0).y +vp.at(3).y)/2;
        targetPoint.z = vp.at(0).z + distToObj;
    }
    else if (vp.size()>=2)
    {
        targetPoint.x = (vp.at(0).x +vp.at(1).x)/2;
        targetPoint.y = (vp.at(0).y +vp.at(1).y)/2;
        targetPoint.z = vp.at(0).z + distToObj;
    }
    else if (vp.size() >=1)
    {
        targetPoint = vp.at(0);
        targetPoint.z = vp.at(0).z + distToObj;
    }
    else
    {
        return false;
    }


    _returnRepportcount++;
    if(GC.distBetweenPoints2D(targetPoint,agShoulder) > ag_reach+0.3)
    {
        if (_returnRepportcountMax < _returnRepportcount)
        {
            _returnRepportcountMax = _returnRepportcount;
            returnRepport = "obj_too_far";
        }
        return false;
    }

//    cout << "target point = " << targetPoint.x << " " << targetPoint.y << " " << targetPoint.z << endl;

    return true;
}

bool PickHuman::findConfigurations()
{

    HRI_GIK_TASK_TYPE task;
    if (m_arm == Left)
        task = GIK_LATREACH;
    else if (m_arm == Right)
        task = GIK_RATREACH;

    p3d_vector3 Tcoord;

    Tcoord[0] = targetPoint.x;
    Tcoord[1] = targetPoint.y;
    Tcoord[2] = targetPoint.z;
    double distance_tolerance = 0.2;

    configPt q = p3d_get_robot_config(mainAgent->getRobotStruct());


    int IKSucceded = hri_agent_single_task_manip_move(mainAgent->getHriAgent(),
                                                     task,
                                                      &Tcoord,
                                                      distance_tolerance,
                                                      &q);

    IKSucceded = hri_agent_single_task_manip_move(mainAgent->getHriAgent(),
                                                         task,
                                                          &Tcoord,
                                                          distance_tolerance,
                                                          &q);
    nbGikCalled++;
    if (IKSucceded != 1)
    {
        p3d_vector3 v;
        p3d_mat4ExtractTrans(mainAgent->getRobotStruct()->joints[mainAgent->getHriAgent()->manip->shoulder_joint_ids[getArmId()]]->abs_pos,v);
        double dist = sqrt(pow(v[0] - targetPoint.x,2) + pow(v[1] - targetPoint.y,2) + pow(v[2] - targetPoint.z,2));

        _returnRepportcount++;
        if(dist > mainAgent->getHriAgent()->manip->reach_tolerance)
        {
            if (_returnRepportcountMax < _returnRepportcount)
            {
                _returnRepportcountMax = _returnRepportcount;
                returnRepport = "no_ik";
            }
            return false;
        }
    }


    p3d_set_and_update_this_robot_conf(mainAgent->getRobotStruct(),q);

    _returnRepportcount++;
    if(pqp_robot_all_no_self_collision_test(mainAgent->getRobotStruct()) == 1)
    {
        if (_returnRepportcountMax < _returnRepportcount)
        {
            _returnRepportcountMax = _returnRepportcount;
            returnRepport = "ik_in_colision";
        }
        return false;
    }

    approachConf = shared_ptr<Configuration>(new Configuration(mainAgent,q));

    return true;


}

bool PickHuman::findTrajectories()
{
    p3d_col_deactivate_robot(mainAgent->getRobotStruct());
    int status = -1;
    approachTraj = compute_traj_using_localPath(_WS->getRobotConf(mainAgent),approachConf,mainObject,status);
    nbMPCalled++;
    p3d_col_activate_robot(mainAgent->getRobotStruct());

    _returnRepportcount++;
    if(!approachTraj)
    {
        if (_returnRepportcountMax < _returnRepportcount)
        {
            _returnRepportcountMax = _returnRepportcount;
            returnRepport = "no_traj_found";
        }
        return false;
    }
    return true;

}

bool PickHuman::computeSolutionTrajectories(int alternativeId)
{
    TaskSolution* TS = result->getAlternativeWithId(alternativeId);
    cout << getDescr() << endl;
    if(!TS)
    {
        cout << "no alternative for task " << to_string(task_id) << " with alternative id = " << alternativeId << endl;
        return false;
    }



    TS->getStartWS()->load();

    confPtr_t q1 = TS->getStopWS()->getRobotConf(mainAgent);

    approachConf = q1;

    findTrajectories();

    result->getAlternativeWithId(alternativeId)->getSubSolutionWithId(0)->setTraj(approachTraj);

    TS->setTrajComputed(true);

    return true;
}

bool PickHuman::setToSolution(int solutionId, bool computeMP)
{
    TaskSolution* TS = new TaskSolution(result,solutionId);
    TS->setTrajComputed(computeMP);

    TS->setTaskConstraints(_currConstraints);

    TaskSubSolution* TSS1 = new TaskSubSolution(TS, "approach", 0, approachTraj, mainAgent, 0);
    detachObject();


    TSS1->setStartState(_WS);
    TS->setStartWS(_WS);
    WorldState* WS1 = new WorldState(_WS);
    bool t = WS1->replaceConf(approachConf);
    if (!t)
    {
        cout << "ERROR: error with robots names" << endl;
        return false;
    }

    TSS1->setEndState(WS1);

    gpGrasp gr = gpGrasp();
    WS1->addAttachement(mainAgent, mainObject, gr, getArmId());

    TS->setStopWS(WS1);
    WS1->setTSThatEndWithThisWS(TS);
    WS1->setTCForNextTask(_currConstraints);

    return true;
}

void PickHuman::smoothSolution(int alternativeId)
{
    alternativeId = 0;
}
