#ifndef PICKPLACEQUAD_HPP
#define PICKPLACEQUAD_HPP

using namespace std;
#include "GTP/virtualtask.hpp"
#include "Planner-pkg.h"

class PickPlaceQuad : public PrivateTask{

public:

    PickPlaceQuad();

    void print_msg(const char *fct,const char *msg,int verbose_level=1);

    bool initAll();
    bool initialize();
    bool setAgents(multimap<string,string> agents_name);
    bool setObjects(multimap<string,string> objects_name);

    bool findPickConf();
    bool findPickApproachConf();
    bool findPickTraj();
    bool findPickExtractConf();
    bool findPickExtractTraj();
    bool findPickTargetConf();
    bool findPickTargetPlacedArmConf();

    bool findPlaceConf();
    bool findPlaceApproachConf();
    bool findPlaceTraj();
    bool findPlaceExtractConf();
    bool findPlaceExtractTraj();
    bool findPlaceTargetConf();

    bool findConfigurations();

    bool computeSolutionTrajectories(int alternativeId);

    bool findPickApproachTraj();
    bool findPickArmTraj();
    bool findPickGotoTraj();

    bool findPlaceArmTraj();
    bool findPlaceGotoTraj();

    bool findTrajectories();

    bool setToSolution(int solutionId, bool computeMP);

    string getDescr();
    string getTextFromValues();

    bool run(multimap<string,string> agents_name,
             multimap<string,string> objects_name,
             multimap<string,p3d_point> point3d,
             multimap<string,string> additionalData,
             WorldState* WS, bool computeMP);

    double computeCost(int altId);
    bool findCandidateSolutions();
    bool getRandomSol();
    bool setPoints(multimap<string,p3d_point> points);
    bool setdata(multimap<string,string> additionalData);
    void smoothSolution(int alternativeId);

protected:

    confPtr_t _initConf;

    confPtr_t _pick_targetGhostConf;
    confPtr_t _pick_targetConf;
    confPtr_t _pick_targetPlacedArmConf;
    confPtr_t _pick_approachConf;
    confPtr_t _pickConf;
    confPtr_t _pick_extractConf;
    confPtr_t _pick_extractObjectConf;

    confPtr_t _place_targetConf;
    confPtr_t _place_targetObjectConf;
    confPtr_t _place_approachConf;
    confPtr_t _place_approachObjectConf;
    confPtr_t _placeConf;
    confPtr_t _placeObjectConf;
    confPtr_t _place_extractConf;

    p3d_traj *_pick_goto_traj;
    p3d_traj *_pick_arm_traj;
    p3d_traj *_pick_approach_traj;
    p3d_traj *_pick_traj;
    p3d_traj *_pick_extract_traj;

    p3d_traj *_place_goto_traj;
    p3d_traj *_place_arm_traj;
    p3d_traj *_place_traj;
    p3d_traj *_place_extract_traj;

    Robot* _mainObject;
    Robot* _supportObject;
    Robot* _ghost;

    lm_quadspline_str *_param;

    p3d_vector3 _pick_graspDir;
    p3d_vector3 _place_graspDir;

    bool use_ghost;
    bool is_pick;
    bool is_place;

    int _idCurrentSol;

    char _className[50];

};

#endif // PICKPLACEQUAD_HPP
