#include "place.hpp"
#include "API/project.hpp"
#include "tools.hpp"



#include "planner/plannerFunctions.hpp"

#ifdef LIGHT_PLANNER
#include "LightPlanner-pkg.h"
#endif

#define APPROACHDISTPLACE 0.1
#define ESCAPEDISTPLACE 0.1

//####################### Place ########################
Place::Place() : ManipulationTask()
{
    type = "place";
    typeId = place;

    noSupport = false;


    nbMaxLoops = 100;
    maxTests = 200;

    nbTotPossibleTests = maxTests*nbMaxLoops;

    nbParameters = 5;

    _hasTargedPoint = false;
    _possibleMainAgentType.push_back(HRI_PR2);
    _possibleMainAgentType.push_back(HRI_ROMEO);
    _possibleMainAgentType.push_back(HRI_KUKAROBOT);


    _noArm = true;
    _hands.clear();

    _taskGraph = NULL;

    _supportObject = NULL;

}


bool Place::initAll()
{

    result = new TaskAllSolutions(this);
    _armId = 0;

    _supportObject = NULL;
    mainObject = NULL;
    mainAgent = NULL;
    _currentSurface = NULL;
    return true;
}

bool Place::initialize()
{
    approachTraj = NULL;
    placeTraj = NULL;
    escapeTraj = NULL;
    releaseTraj = NULL;




    nbGikCalled = 0;
    nbMPCalled = 0;
    nbTests = 0;
    timeMp = 0;

    returnRepport = "not OK";
    _returnRepportcount = 0;
    _returnRepportcountMax = 0;

    //moving all virtual object to 0 and storing confs

    return initMotionPlaner(mainAgent);
}

bool Place::finalize()
{
    activateFingertipCol(true, mainAgent, mainObject, NULL);
    return true;
}


bool Place::findCandidateSolutions()
{
    if (!checkAndSetArmAndObj())
        return false;
    //check if seted hand is the same as the one with the object

    if (!checkAttachement())
    {
        returnRepport = "no_corresponding_attachement_in_ws";
        return false;
    }

    if (!checkHRIAgent(mainAgent))
        return false;

    if (!setMainAgentHand())
        return false;

    switch(PlanEnv->getInt(PlanParam::GTPTypeCase))
    {
    case 0:
        //finding support either from set support object, or from env.
        if (!findSupports())
            return false;
        break;
    case 1:
        if (!initMightabilities())
            return false;
        break;
    default:
        returnRepport = "typeCase_not_correct";
        return false;
        break;
    }


    if (!findRotations())
        return false;

    return true;
}

bool Place::getRandomSol()
{
    confPtr_t robotCurConf = shared_ptr<Configuration>(new Configuration(*mainAgent->getCurrentPos()));
    confPtr_t objectCurConf = shared_ptr<Configuration>(new Configuration(*mainObject->getCurrentPos()));
    bool res = false;
    detachObject();

    int nbTestsPlace = -1;

    int reachedReturnRepportCount = _returnRepportcount;

//    switch(PlanEnv->getInt(PlanParam::GTPTypeCase))
//    {
//    case 0:

        while (!res && nbTestsPlace < maxTests)
        {
            nbTests++;
            nbTestsPlace++;
            _returnRepportcount = reachedReturnRepportCount;
            mainObject->setAndUpdate(*objectCurConf);
    //        cout<< "test nb = " << nbTests << endl;

            switch(PlanEnv->getInt(PlanParam::GTPTypeCase))
            {
            case 0:
                if (!findCurrentTestPoint())
                    continue;
                break;
            case 1:
                if (!getNextMightPoint())
                    return false;
                break;
            default:
                returnRepport = "typeCase_not_correct";
                return false;
                break;
            }



//            configPt q = findCurrentRotations();
//            if (!q)
//                continue;

            configPt q = findObjectConf();
            if (!q)
                continue;

            p3d_set_and_update_this_robot_conf(mainObject->getRobotStruct(),q);

//            g3d_draw_allwin_active();
//            sleep(3);

            if (!isObjectInSurface())
                continue;

            if(isObjectInCollision(_supportObject))
                continue;

            if(isHandInCollision())
                continue;

            if (!testFactConstraints())
                continue;

            _confSol = shared_ptr<Configuration>(new Configuration(mainObject,q));

            res = true;
        }
//        break;
//    case 1:
//        cout << "walalalalalala" << endl;
//        break;
//    default:
//        returnRepport = "typeCase_not_correct";
//        return false;
//        break;
//    }

    mainAgent->setAndUpdate(*robotCurConf);
    mainObject->setAndUpdate(*objectCurConf);
//    attachObject();
    return res;
}

bool Place::findConfigurations()
{
    removeVirtualObjects();
    // Need to detach first, otherwise computation of manipulation data is unstable and return some odd values very often
    detachObject();

//    bool tr = false;
    confPtr_t qInitA = shared_ptr<Configuration>(new Configuration(*mainAgent->getCurrentPos()));
    confPtr_t qInitO = shared_ptr<Configuration>(new Configuration(*mainObject->getCurrentPos()));

    mainObject->setAndUpdate(*_confSol);
//    g3d_draw_allwin_active();


    activateFingertipCol(false, mainAgent, mainObject, _supportObject);

    //HACK
    if(mainAgent->getHriAgent()->type == HRI_PR2)
    {
        _curGrasp.hand_type = GP_PR2_GRIPPER;
    }


    p3d_vector3 v1, v2;
    _curGrasp.direction(v1);

    v2[0] = 0;
    v2[1] = 0;
    v2[2] = 1;



//    activateFingertipCol(false, mainAgent, mainObject, _supportObject);
    bool test = computeManipulationConfs(escapeConf,graspConfClosed,graspConfOpen,approachConf,mainObject->getJoint(0)->getJointStruct()->abs_pos,
                                         _curGrasp, v1, APPROACHDISTPLACE, v2, ESCAPEDISTPLACE);
    activateFingertipCol(true, mainAgent, mainObject, _supportObject);



    detachObject();
    mainAgent->setAndUpdate(*qInitA);
    mainObject->setAndUpdate(*qInitO);
    attachObject();

    return test;
}

bool Place::findTrajectories()
{
    detachObject();
    _WS->load();
    activateFingertipCol(false, mainAgent, mainObject, _supportObject);

    attachObject();

    int status = -1;
    approachTraj = compute_traj(_WS->getRobotConf(mainAgent), approachConf, mainObject, status);

    _returnRepportcount++;
    if (!approachTraj)
    {
        cout << "no approach traj" << endl;
        if (_returnRepportcountMax < _returnRepportcount)
        {
            _returnRepportcountMax = _returnRepportcount;
            returnRepport = "no_approach_traj";
        }
        moveBackVirtualObjects();
        activateFingertipCol(true, mainAgent, mainObject, _supportObject);
        return false;
    }

    setGraph(new Graph(*API_activeGraph));
    p3d_col_deactivate_robot_robot(mainObject->getRobotStruct(), _supportObject->getRobotStruct());

    placeTraj = compute_traj_using_localPath(approachConf, graspConfOpen, mainObject, status);


    _returnRepportcount++;
    if (!placeTraj)
    {
        cout << "no place traj" << endl;
        if (_returnRepportcountMax < _returnRepportcount)
        {
            _returnRepportcountMax = _returnRepportcount;
            returnRepport = "no_place_traj";
        }
        moveBackVirtualObjects();
        activateFingertipCol(true, mainAgent, mainObject, _supportObject);
        return false;
    }

    detachObject();



    mainObject->setAndUpdate(*_confSol);
    escapeTraj = compute_traj_using_localPath(graspConfOpen, escapeConf, mainObject, status);
    p3d_col_activate_robot_robot(mainObject->getRobotStruct(), _supportObject->getRobotStruct());
    
    _returnRepportcount++;
    if (!escapeTraj)
    {
        cout << "no escape traj" << endl;
        if (_returnRepportcountMax < _returnRepportcount)
        {
            _returnRepportcountMax = _returnRepportcount;
            returnRepport = "no_escape_traj";
        }
        moveBackVirtualObjects();
        activateFingertipCol(true, mainAgent, mainObject, _supportObject);
        return false;
    }

    releaseTraj = NULL;

//    gpActivate_object_fingertips_collisions(mainAgent->getRobotStruct(), mainObject->getRobotStruct()->joints[1]->o, armHandProp , _armId);

    activateFingertipCol(true, mainAgent, mainObject, _supportObject);
    moveBackVirtualObjects();

    return true;
}

bool Place::computeSolutionTrajectories(int alternativeId)
{
    TaskSolution* TS = result->getAlternativeWithId(alternativeId);
    if(!TS)
    {
        cout << "no alternative for task " << to_string(task_id) << " with alternative id = " << alternativeId << endl;
        return false;
    }

    TS->getStartWS()->load();
    TaskSubSolution* TSS0 = TS->getSubSolutionWithId(0);
    TaskSubSolution* TSS1 = TS->getSubSolutionWithId(1);
    TaskSubSolution* TSS3 = TS->getSubSolutionWithId(3);

    confPtr_t q1 = TSS0->getEndState()->getRobotConf(mainAgent);
    confPtr_t q2 = TSS1->getEndState()->getRobotConf(mainAgent);
    confPtr_t q3 = TSS3->getEndState()->getRobotConf(mainAgent);

    approachConf = q1;
    graspConfClosed = q2;
    escapeConf = q3;

    findTrajectories();


    result->getAlternativeWithId(alternativeId)->getSubSolutionWithId(0)->setTraj(approachTraj);
    result->getAlternativeWithId(alternativeId)->getSubSolutionWithId(1)->setTraj(placeTraj);
    result->getAlternativeWithId(alternativeId)->getSubSolutionWithId(3)->setTraj(escapeTraj);

    TS->setTrajComputed(true);

    return true;
}

bool Place::setToSolution(int solutionId, bool computeMP)
{
    TaskSolution* TS = new TaskSolution(result,solutionId);
    TS->setTrajComputed(computeMP);

    TS->setTaskConstraints(_currConstraints);

    TaskSubSolution* TSS1 = new TaskSubSolution(TS, "approach", 0, approachTraj, mainAgent, 0);
    TaskSubSolution* TSS2 = new TaskSubSolution(TS, "engage", 1, placeTraj, mainAgent, 0);
    TaskSubSolution* TSS3 = new TaskSubSolution(TS, "release", 2, releaseTraj, mainAgent, 0);
    TaskSubSolution* TSS4 = new TaskSubSolution(TS, "escape", 3, escapeTraj, mainAgent, 0);

    TSS1->setArmId(_armId);
    TSS2->setArmId(_armId);
    TSS3->setArmId(_armId);
    TSS4->setArmId(_armId);


    attachObject();

    vector<TaskSubSolution*> v1;
    v1.push_back(TSS2);
    TSS1->setNext(v1);
    vector<TaskSubSolution*> v2;
    v2.push_back(TSS3);
    TSS2->setNext(v2);
    vector<TaskSubSolution*> v3;
    v3.push_back(TSS4);
    TSS3->setNext(v3);

    TS->setStartWS(_WS);

    TSS1->setStartState(_WS);
    WorldState* WS1 = new WorldState(_WS);
    bool t = WS1->replaceConf(approachConf);
    if (!t)
    {
        cout << "ERROR: error with robots names" << endl;
        return false;
    }
    _WS->load();
    attachObject();
    mainAgent->setAndUpdate(*WS1->getRobotConf(mainAgent));
    confPtr_t q_o_curr = mainObject->getCurrentPos();
    detachObject();
    WS1->replaceConf(q_o_curr);
    TSS1->setEndState(WS1);

    TSS2->setStartState(WS1);
    WorldState* WS2 = new WorldState(WS1);
    t = WS2->replaceConf(graspConfClosed);
    if (!t)
    {
        cout << "ERROR: error with robots names" << endl;
        return false;
    }
    WS2->replaceConf(_confSol);
    TSS2->setEndState(WS2);
    WS2->clearAttachement();
    TSS3->setStartState(WS2);

    WorldState* WS3 = new WorldState(WS2);
    t = WS3->replaceConf(graspConfOpen);
    if (!t)
    {
        cout << "ERROR: error with robots names" << endl;
        return false;
    }
    TSS3->setEndState(WS3);

    TSS4->setStartState(WS3);
    WorldState* WS4 = new WorldState(WS3);
    WS4->replaceConf(_confSol);
    WS4->clearAttachement();
    t = WS4->replaceConf(escapeConf);
    if (!t)
    {
        cout << "ERROR: error with robots names" << endl;
        return false;
    }
    TSS4->setEndState(WS4);
    TS->setStopWS(WS4);

    WS4->setTSThatEndWithThisWS(TS);
    WS4->setTCForNextTask(_currConstraints);

    TS->addConf(_WS->getRobotConf(mainAgent));
    TS->addConf(approachConf);
    TS->addConf(graspConfClosed);
    TS->addConf(escapeConf);

    return true;
}

bool Place::setAgents(multimap<string,string> agents_name)
{
    return setMainAgent(agents_name);
}

bool Place::setMainObject(multimap<string,string> objects_name)
{
    if (objects_name.count("mainObject") != 1)
    {
        cout << "no main object for " << type << " task." << endl;
        return true;
    }
    mainObject = global_Project->getActiveScene()->getRobotByName(objects_name.find("mainObject")->second);
    if (!mainObject)
    {
        returnRepport = "no_main_object_name";
        return false;
    }
    return true;
}



bool Place::setObjects(multimap<string,string> objects_name)
{
    return (setMainObject(objects_name) && setSupportObject(objects_name));
}

bool Place::setPoints(multimap<string,p3d_point> points)
{
    return setTargetPoint(points);
}

bool Place::setdata(multimap<string,string> additionalData)
{
    if (additionalData.count("hand") != 1)
    {
        _noArm = true;
        return true;
    }

    string handStr = additionalData.find("hand")->second;
    if (handStr == "right")
        setArmToUse(Right);
    else if (handStr == "left")
        setArmToUse(Left);
    else
    {
        cout << "the arm selection is not rigth: choosen : " << handStr << "available : right or left." << endl;
        _noArm = true;
        return false;
    }

    _noArm = false;
    return true;
}

bool Place::setSupportObject(multimap<string,string> objects_name)
{
    if (objects_name.count("supportObject") == 1)
    {
        cout << "got a support name" << endl;
        Robot* r = global_Project->getActiveScene()->getRobotByName(objects_name.find("supportObject")->second);
        bool test = r->getName() == objects_name.find("supportObject")->second;

        if (test)
        {
            _supportObject = r;
            cout << "the object with name Id " << objects_name.find("supportObject")->second << " is now the support" <<endl;
            noSupport = false;

        }
        else
        {
            cout << "Support object with name " << objects_name.find("supportObject")->second <<
                    " does not exist, proceed without support" << endl;
            noSupport = true;
        }
        return true;
    }


//    cout << "NO SUPPORT" << endl;
    noSupport = true;
    return true;
}

bool Place::setTargetPoint(multimap<string,p3d_point> points)
{

    if (points.count("target") == 1)
    {
        cout << "got a target point" << endl;
        _hasTargedPoint = true;
        _targetPoint = points.find("target")->second;
        return true;
    }

//    noSupport = true;
    return true;
}

void Place::smoothSolution(int alternativeId)
{
    for (int i = 0; i < result->getAlternativeWithId(alternativeId)->getNbsubSolPerRobot(getMainAgent()); i++)
    {
        TaskSubSolution* tss = result->getAlternativeWithId(alternativeId)->getTSSByRobot(getMainAgent()).at(i);
        if (tss->getSubSolutionName() == "approach")
        {
            tss->getRobot()->getRobotStruct()->tcur = tss->getTraj();
            tss->getStartState()->load();
            attachObject();
            activateFingertipCol(false, mainAgent, mainObject, NULL);
            tss->refineTraj();
            activateFingertipCol(true, mainAgent, mainObject, NULL);
            detachObject();
            break;
        }
    }
    result->getAlternativeWithId(alternativeId)->getStartWS()->load();
}

bool Place::findSupports()
{
    if (noSupport || !_supportObject)
    {
        //find a support
        //find all support in environment

        double ag_reach = mainAgent->getHriAgent()->manip->reach_tolerance;

        p3d_point agShoulder;

        agShoulder.x = mainAgent->getJoint(mainAgent->getHriAgent()->manip->shoulder_joint_ids[_armId])->getJointStruct()->abs_pos[0][3];
        agShoulder.y = mainAgent->getJoint(mainAgent->getHriAgent()->manip->shoulder_joint_ids[_armId])->getJointStruct()->abs_pos[1][3];
        agShoulder.z = mainAgent->getJoint(mainAgent->getHriAgent()->manip->shoulder_joint_ids[_armId])->getJointStruct()->abs_pos[2][3];


        for(unsigned int i = 0; i < global_Project->getActiveScene()->getNumberOfRobots(); i++)
        {
            Robot* r = global_Project->getActiveScene()->getRobot(i);
            if (r->getObjectRob())
            {
                ObjectRob* obre = r->getObjectRob();
                if(obre->isSupport())
                {
                    //select support that are close enouph
                    vector<GeometricForm*> sss =obre->getSupports();
                    if (obre->distToPoint(agShoulder) < ag_reach)
                    {
                        for (unsigned int j = 0; j < sss.size(); j++)
                        {
                            _supports.push_back(sss.at(j));
                        }
                    }
                }
            }
        }

        noSupport = false;
    }
    else
    {
        if (_supportObject->getObjectRob())
        {
            ObjectRob* obre = _supportObject->getObjectRob();
            if(obre->isSupport())
            {
                _supports = obre->getSupports();
            }
        }
    }


    if (_supports.size() == 0)
    {
        cout << "ERROR: there is a supportObject but no support on it" << endl;
        returnRepport = "no_support_in_object";
        return false;
    }
    return true;
}

bool Place::findRotations()
{

    if (mainObject->getObjectRob() && mainObject->getObjectRob()->isManipulable())
        _rotations = mainObject->getObjectRob()->getStableRotations();
//    _rotations = getConfigPlacement(mainObject);

    if(_rotations.size() == 0)
    {
        cout << "no rotations defined for the object : " << mainObject->getName() << endl;
        returnRepport = "no_rotations";
        return false;
    }
    return true;
}

vector<regionConst> Place::getForceInConstraints()
{
    vector<regionConst> vrg = _currConstraints->getRegionConstraints();
    vector<regionConst> vfrg;
    for (unsigned int i = 0; i < vrg.size(); i++)
    {
        if (vrg.at(i).nameId == "force_place")
        {
            vfrg.push_back(vrg.at(i));
        }
    }
    return vfrg;
}

vector<regionConst> Place::getForbidConstraints()
{
    vector<regionConst> vrg = _currConstraints->getRegionConstraints();
    vector<regionConst> vfrg;
    for (unsigned int i = 0; i < vrg.size(); i++)
    {
        if (vrg.at(i).nameId == "lock")
        {
            vfrg.push_back(vrg.at(i));
        }
    }
    return vfrg;
}

bool Place::findCurrentTestPoint()
{
    GeometricComputations GC = GeometricComputations();
    //if there is a target point
    if (_hasTargedPoint && !_supportObject)
    {
        returnRepport = "no_support_object_while_needed";
        return false;
    }
    else if (_hasTargedPoint && _supportObject)
    {
        _currentTest.x = _targetPoint.x;
        _currentTest.y = _targetPoint.y;
        _currentSurface = _supports.at(p3d_random_integer(0,_supports.size()-1));
        _currentTest.z = _currentSurface->getRandomPointInForm().z;
        return true;
    }


    vector<regionConst> vfic = getForceInConstraints();

    if (vfic.size() > 0) // if force in constraint
    {
        _currentSurface = vfic.at(0).GeoForm;
        _currentTest = _currentSurface->getRandomPointInForm();
        _returnRepportcount++;
        for (unsigned int i = 1; i < vfic.size(); i++)
        {
            if(!vfic.at(i).GeoForm->isPointInForm(_currentTest))
            {
                if (_returnRepportcountMax < _returnRepportcount)
                {
                    _returnRepportcountMax = _returnRepportcount;
                    returnRepport = "no_overlapping_force_in";
                }
                return false;
            }
        }
        return true;
    }
    else // normal
    {
        int r1 = p3d_random_integer(0,_supports.size()-1);
        _currentSurface = _supports.at(r1);
        _supportObject = _currentSurface->object;
        _currentTest  = _currentSurface->getRandomPointInForm();

        vector<regionConst> vfibc = getForbidConstraints(); //for forbid constraint
        _returnRepportcount++;
        for (unsigned int i = 1; i < vfibc.size(); i++)
        {
            if (vfibc.at(i).GeoForm->isPointInForm(_currentTest))
            {
                if (_returnRepportcountMax < _returnRepportcount)
                {
                    _returnRepportcountMax = _returnRepportcount;
                    returnRepport = "no_outside_of_forbid";
                }
                return false;
            }
        }

        // Test if placement point is within the reach sphere of agent (if exist)
//        if (mainAgent)
//        {
//            p3d_point shoulderPos;
//            shoulderPos.x = mainAgent->getJoint(mainAgent->getHriAgent()->manip->shoulder_joint_ids[_armId])->getJointStruct()->abs_pos[0][3];
//            shoulderPos.y = mainAgent->getJoint(mainAgent->getHriAgent()->manip->shoulder_joint_ids[_armId])->getJointStruct()->abs_pos[1][3];
//            shoulderPos.z = mainAgent->getJoint(mainAgent->getHriAgent()->manip->shoulder_joint_ids[_armId])->getJointStruct()->abs_pos[2][3];

//            confPtr_t qwe = mainObject->getCurrentPos();
//            (*qwe)[6] = _currentTest.x;
//            (*qwe)[7] = _currentTest.y;
//            (*qwe)[8] = _currentTest.z;
//            (*qwe)[9] = 0;
//            (*qwe)[10] = 0;
//            (*qwe)[11] = 0;
//            mainObject->setAndUpdate(*qwe);


//            _returnRepportcount++;
//            if (GC.distBetweenPoints(_currentTest, shoulderPos) > mainAgent->getHriAgent()->manip->reach_tolerance)
//            {
////                cout << "distance1 = " << GC.distBetweenPoints(_currentTest, shoulderPos) << endl;
////                cout << "distance2 = " << mainAgent->getHriAgent()->manip->reach_tolerance << endl;
//                if (_returnRepportcountMax < _returnRepportcount)
//                {
//                    _returnRepportcountMax = _returnRepportcount;
//                    returnRepport = "distance_is_supperior";
//                }
//                return false;
//            }
//            return true;
//        }
//        else
            return true;
    }
}

configPt Place::findCurrentRotations()
{
    if (_rotations.size()<= 0)
        return NULL;
    int rotationNb = p3d_random_integer(0,_rotations.size()-1);

    configPt q = p3d_copy_config(mainObject->getRobotStruct(),_rotations.at(rotationNb));
    p3d_set_and_update_this_robot_conf(mainObject->getRobotStruct(),q);

    p3d_matrix4 rotMat;
    p3d_mat4Copy(mainObject->getJoint(0)->getJointStruct()->abs_pos, rotMat);

    double rot;
    if (_hasTargedPoint)
        rot = _targetPoint.z;
    else
        rot = p3d_random(-M_PI,M_PI);

    if (q[9] == OPEN_VALUE)
        q[9] = rot;
    else if (q[10] == OPEN_VALUE)
        q[10] = rot;
    else if (q[11] == OPEN_VALUE)
        q[11] = rot;
    return q;
}

configPt Place::findObjectConf()
{
    int rotationNb = p3d_random_integer(0,_rotations.size()-1);
    double rot;
    if (_hasTargedPoint)
        rot = _targetPoint.z;
    else
        rot = p3d_random(-M_PI,M_PI);

    configPt a = mainObject->getObjectRob()->getStableRotation(rotationNb,rot);


    if (!a)
        return NULL;

    a[6] += _currentTest.x;
    a[7] += _currentTest.y;
    a[8] += _currentTest.z;


    p3d_set_and_update_this_robot_conf(mainObject->getRobotStruct(),a);
    return a;
}

bool Place::isObjectInSurface()
{
    GeometricComputations GC = GeometricComputations();
//    g3d_draw_allwin_active();
//    sleep(3);
    vector<p3d_point> vp = GC.getLowerPointsInObject(mainObject, 0.03);

    bool isOnSupport = true;
    for (unsigned int i = 0; i < vp.size(); i++)
    {
//        cout << "test for : (" << vp.at(i).x << ", " <<  vp.at(i).y << ", " << vp.at(i).z << ")" << endl;
        if(!_currentSurface->isPointInForm(vp.at(i)))
        {
            isOnSupport = false;
            break;
        }
    }

    _returnRepportcount++;
    if (!isOnSupport)
    {
        if (_returnRepportcountMax < _returnRepportcount)
        {
            _returnRepportcountMax = _returnRepportcount;
            returnRepport = "not_on_support";
        }
        //TODO adding a function to detect that support is too small
        return false;
    }
    return true;
}



Robot* Place::getMainObjectFromAtt()
{
    if (_WS->getAttachements().size() == 0)
    {
        returnRepport = "no_attachement_in_ws";
        return false;
    }
    vector<Attachement> vatt = _WS->getAttachements();
    for (unsigned int i = 0; i < vatt.size(); i++)
    {
        Attachement att = vatt.at(i);
        if (att.agent && att.object)
        {
            if (att.agent->getName() == mainAgent->getName() && att.armId == getArmId())
            {
                return att.object;
            }
        }
    }
    return NULL;
}

int Place::getArmIdFromAtt()
{
    if (_WS->getAttachements().size() == 0)
    {
        returnRepport = "no_attachement_in_ws";
        return false;
    }
    vector<Attachement> vatt = _WS->getAttachements();
    for (unsigned int i = 0; i < vatt.size(); i++)
    {
        Attachement att = vatt.at(i);
        if (att.agent && att.object)
        {
            if (att.agent->getName() == mainAgent->getName() && att.object->getName() == mainObject->getName())
            {
                return att.armId;
            }
        }
    }
    return -1;
}

bool Place::checkArmAtt()
{
    if (_WS->getAttachements().size() == 0)
    {
        return false;
    }
    Attachement att = _WS->getAttachement(mainAgent,mainObject);
    if(att.object == NULL || att.agent == NULL)
    {
        return false;
    }
    else if (att.armId != getArmId())
    {
        return false;
    }
    return true;
}

bool Place::checkAndSetArmAndObj()
{
    if (!mainObject && _noArm)
    {
        returnRepport ="no_obj_and_no_arm";
        return false;
    }
    else if (!mainObject && !_noArm)//if no obj setobjfrom attachment
    {
        mainObject = getMainObjectFromAtt();
        if (!mainObject)
        {
            returnRepport = "no_attachement_for_arm";
            return false;
        }
    }
    else if (mainObject && _noArm)
    {
        int aI = getArmIdFromAtt();
        if (aI == -1)
        {
            returnRepport = "no_attachement_for_object";
            return false;
        }
        setArmToUse(getArmLabelFromId(aI));
    }
    else if (mainObject && !_noArm)
    {
        if (!checkArmAtt())
        {
            returnRepport = "wrong_arm_choice";
            return false;
        }
    }
    else
    {
        returnRepport = "unexpected_error";
        return false;
    }
    return true;
}

bool Place::initMightabilities()
{
    if (!_supportObject)
    {
        returnRepport ="no_support_set_for_mightabilities";
        return false;
    }

    _currentSurface = _supportObject->getObjectRob()->getSupports().at(0);

    _WS->initMightabilities();
    MightabilitiesGrid* MG = _WS->getMightabilities();

    MG->setSampleRate(0.05);

    vector<Robot*> ags,supp;
    ags.push_back(mainAgent);
    supp.push_back(_supportObject);

    vector<typeCells> TC;
    TC.push_back(reachDistSupp);
    TC.push_back(reachColSupp);
    MG->computeMights(ags,supp,TC);

    _cellId = 0;

    vector<pair<Robot*,typeCells> > rpc;
    rpc.push_back(make_pair(mainAgent,reachColSupp));
    _cells = MG->getList(rpc);

    cout << "number of cells = " << _cells.size() << endl;
    if (_cells.size() == 0)
    {
        returnRepport = "no_cell_in_list";
        return false;
    }
    return true;

}

bool Place::getNextMightPoint()
{
    MightabilitiesCell* c = _cells.at(_cellId);
    _currentTest.x = c->getCenter()[0];
    _currentTest.y = c->getCenter()[1];
    _currentTest.z = c->getCenter()[2];
    _cellId++;
    if (_cellId >= _cells.size())
        _cellId = 0;

    return true;
}



string Place::getDescr()
{
    string s = "Task: \nId \t\t= " + to_string(task_id) + "\nType \t\t= " + type;
    if (mainAgent)
        s+= "\nMain agent \t\t= " + mainAgent->getName();
    if (mainObject)
        s+= "\nMain object \t= " + mainObject->getName() ;
    if (_supportObject)
        s +=  "\nSupport object \t= " + _supportObject->getName();
    return s;
}

string Place::getTextFromValues()
{
    return "Task: " + type + " : place the object, depending on input, on support, or at a point ..."
            + "\n    agents: "
            + "\n        mainAgent:  [AGENT ID]"
            + "\n    objects: "
            + "\n        mainObject: [OBJECT ID]"
            + "\n        supportObject [OBJECT ID] -optional-"
            + "\n    datas:"
            + "\n        hand: [right OR left]"
            + "\n    points:"
            + "\n        target: [X Y THETA]"
            + "\n    constraints:"
            + "\n        lock: [geometric form]"
            + "\n        force_place: [geometric form]"
            + "\n        noUse: [Entity]";
}

bool Place::testFactConstraints()
{
    WorldState *wsTmp = new WorldState(_WS);
    wsTmp->replaceConf(mainObject->getCurrentPos());
    bool testConst = _currConstraints->areFactConstraintRespected(wsTmp);
    delete wsTmp;

    _returnRepportcount++;
    if (!testConst)
    {
        if (_returnRepportcountMax < _returnRepportcount)
        {
            _returnRepportcountMax = _returnRepportcount;
            returnRepport = "fact_not_respected";
        }
        return false;
    }
    return true;
}

bool Place::testConstraints(int alternativeId)
{

    TaskSolution* TS = result->getAlternativeWithId(alternativeId);
    if (!TS)
    {
        cout << "test of constraint fail, no solution with id " << alternativeId << endl;
        return false;
    }

    //test fact
    if (!TS->getTaskConstraints()->areFactConstraintRespected(TS->getStopWS()))
    {
        result->removeAlternativeWithId(alternativeId);
        return false;
    }


    //test ressources
    vector<ressourceConst> vrs = TS->getTaskConstraints()->getRessourceConstraints();
    for (unsigned int k = 0; k < vrs.size(); k++)
    {
        ressourceConst rcs = vrs.at(k);
        if (rcs.nameId == "noUse")
        {
            if (rcs.entity->getName() == mainObject->getName() ||
                    rcs.entity->getName() == mainAgent->getName() ||
                    rcs.entity->getName() == _supportObject->getName())
            {
                result->removeAlternativeWithId(alternativeId);
                return false;
            }
        }
    }

    //test regions
    GeometricComputations GC = GeometricComputations();
    vector<regionConst> vrg = TS->getTaskConstraints()->getRegionConstraints();
    confPtr_t q_curr = mainObject->getCurrentPos();
    mainObject->setAndUpdate(*_confSol);
    for (unsigned int i = 0; i < vrg.size(); i++)
    {
        regionConst rc = vrg.at(i);
        if (rc.nameId == "lock")
        {
            vector<p3d_point> vp = GC.getLowerPointsInObject(mainObject, 0.03);
            p3d_point p;
            p.x = (vp[0].x + vp[3].x) /2;
            p.y = (vp[0].y + vp[3].y) /2;
            p.z = (vp[0].z + vp[3].z) /2;
            if(rc.GeoForm->isPointInForm(p))
            {
                result->removeAlternativeWithId(alternativeId);
                return false;
            }
        }
        else if (rc.nameId == "force_place")
        {
            vector<p3d_point> vp = GC.getLowerPointsInObject(mainObject, 0.03);
            p3d_point p;
            p.x = (vp[0].x + vp[3].x) /2;
            p.y = (vp[0].y + vp[3].y) /2;
            p.z = (vp[0].z + vp[3].z) /2;
            if(!rc.GeoForm->isPointInForm(p))
            {
                result->removeAlternativeWithId(alternativeId);
                return false;
            }
        }
    }
    mainObject->setAndUpdate(*q_curr);

    return true;
}

double Place::getSSSizeLeft()
{
    double t = 0;
    switch(PlanEnv->getInt(PlanParam::GTPTypeCase))
    {
    case 0:
        t = 36*_rotations.size();
        if (_currentSurface)
            t =t*_currentSurface->getFormGrid(0.05).size();
        break;
//    case 1:
//        break;
//    default:
//        break;
    }
    cout << "size place = " << t << endl;
    return t;
}

