#ifndef PLACE_HPP
#define PLACE_HPP

using namespace std;
#include "GTP/virtualtask.hpp"
#include "Planner-pkg.h"


#define MAX_NBR_TEST 1000

class Place : public ManipulationTask
{
public:
    Place();
    bool initAll();
    virtual bool initialize();
    bool finalize();
    bool findCandidateSolutions();
    bool getRandomSol();
    bool findConfigurations();
    bool findTrajectories();
    bool computeSolutionTrajectories(int alternativeId);
    bool setToSolution(int solutionId, bool computeMP);
    bool testConstraints(int alternativeId);
//    virtual vector<TaskSubSolution*> setNextTraj(TaskSubSolution* doneTSS);
    virtual bool setAgents(multimap<string,string> agents_name);
    bool setObjects(multimap<string,string> objects_name);
    bool setPoints(multimap<string,p3d_point> points);
    bool setdata(multimap<string,string> additionalData);
    bool setSupportObject(multimap<string,string> objects_name);
    bool setMainObject(multimap<string,string> objects_name);
    void smoothSolution(int alternativeId);
    string getDescr();
    string getTextFromValues();

    bool setTargetPoint(multimap<string,p3d_point> points);

    bool findSupports();
    bool findRotations();
    vector<regionConst> getForceInConstraints();
    vector<regionConst> getForbidConstraints();
    bool findCurrentTestPoint();
    configPt findCurrentRotations();
    configPt findObjectConf();
    bool isObjectInSurface();
    Robot* getMainObjectFromAtt();
    int getArmIdFromAtt();
    bool checkArmAtt();
    bool checkAndSetArmAndObj();
    bool testFactConstraints();

    bool initMightabilities();
    bool getNextMightPoint();

    virtual double getSSSizeLeft();


protected:
    int maxTests;
    Robot* _supportObject;

    p3d_traj* approachTraj;
    p3d_traj* placeTraj;
    p3d_traj* escapeTraj;
    p3d_traj* releaseTraj;

    vector<GeometricForm*> _supports;
    confPtr_t _confSol;

    bool noSupport; //false of a support is given


    bool _hasTargedPoint;
    p3d_point _targetPoint;

    p3d_point _currentTest;
    GeometricForm* _currentSurface;

    vector<configPt> _rotations;

    bool _noArm;

    vector<MightabilitiesCell*> _cells;
    int _cellId;
};



#endif // PLACE_HPP
