#include "placereachable.hpp"


PlaceReachable::PlaceReachable() : Place()
{
    type = "placeReachable";
    typeId = placeReachable;
    _armId = -1;

    noSupport = false;

    approachTraj = NULL;
    placeTraj = NULL;
    escapeTraj = NULL;
    releaseTraj = NULL;

    nbMaxLoops = 20;

    nbParameters = 5;
    nbGikCalled = 0;
    nbMPCalled = 0;
    nbTests = 0;

    maxTests = 500;

    nbTotPossibleTests = maxTests*nbMaxLoops;

    //_possibleMainAgent.push_back("PR2_ROBOT");
    //_possibleMainAgent.push_back("MAX_ROBOT");
    //_possibleMainAgent.push_back("BOB_ROBOT");
    _possibleMainAgentType.push_back(HRI_PR2);

}

bool PlaceReachable::findCandidateSolutions()
{
    if (!checkAndSetArmAndObj())
        return false;
    //check if seted hand is the same as the one with the object

    if (!checkAttachement())
    {
        returnRepport = "no_corresponding_attachement_in_ws";
        return false;
    }

    if (!checkHRIAgent(mainAgent))
        return false;

    if (!setMainAgentHand())
        return false;

    //finding support either from set support object, or from env.
    if (!findSupports())
        return false;

    if (!findRotations())
        return false;

    //computing human reach
    int nbArms = _targetAgent->getHriAgent()->manip->shoulder_joint_ids_no;
    double agReach = _targetAgent->getHriAgent()->manip->reach_tolerance;
    //TODO: there is a problem if multiple support in an object

    vector<GeometricForm*> newVs;
    for (unsigned int i = 0; i < _supports.size(); i++)
    {
        GeometricForm* supp = _supports.at(i);

        for(int j = 0; j < nbArms; j++)
        {
            p3d_point agShoulder;

            agShoulder.x = _targetAgent->getJoint(_targetAgent->getHriAgent()->manip->shoulder_joint_ids[j])->getJointStruct()->abs_pos[0][3];
            agShoulder.y = _targetAgent->getJoint(_targetAgent->getHriAgent()->manip->shoulder_joint_ids[j])->getJointStruct()->abs_pos[1][3];
            agShoulder.z = _targetAgent->getJoint(_targetAgent->getHriAgent()->manip->shoulder_joint_ids[j])->getJointStruct()->abs_pos[2][3];

            //this is the real distance, with taking z into acount
            //double disttoP = supp.object->getObjectRob()->distToPoint(agShoulder);
            //this doesn t take z into account
            double disttoP = supp->object->getObjectRob()->distToPointZ(agShoulder);
            cout << "objects tested = " << supp->object->getName() << " distance = " << disttoP << endl;
            if ( disttoP < agReach)
            {
                newVs.push_back(supp);
                cout << "object reachable = " << supp->object->getName() << endl;
            }
        }
    }

    _supports = newVs;


    if (_supports.size() == 0)
    {
        cout << "no reachable object for target Agent " << type << " is not feasible" << endl;
        return false;
    }

    _WSF = new WorldStateFacts(_WS->getRobotConfs());

    return true;
}

bool PlaceReachable::getRandomSol()
{

    confPtr_t robotCurConf = mainAgent->getCurrentPos();
    confPtr_t objectCurConf = mainObject->getCurrentPos();

    detachObject();

    bool res = false;
    int nbTestsPlace = -1;



    int reachedReturnRepportCount = _returnRepportcount;
    while (!res && nbTestsPlace < maxTests)
    {
        nbTests++;
        nbTestsPlace++;
        _returnRepportcount = reachedReturnRepportCount;
        mainObject->setAndUpdate(*objectCurConf);
//        cout<< "test nb = " << nbTests << endl;

        if (!findCurrentTestPoint())
            continue;

//        configPt q = findCurrentRotations();
//        if (!q)
//            continue;

        configPt q = findObjectConf();
        if (!q)
            continue;

        p3d_set_and_update_this_robot_conf(mainObject->getRobotStruct(),q);


        if (!isObjectInSurface())
            continue;

        if(isObjectInCollision(_supportObject))
            continue;

        if(isHandInCollision())
            continue;

         _confSol = shared_ptr<Configuration>(new Configuration(mainObject,q));

         p3d_col_deactivate_robot_robot(mainObject->getRobotStruct(), _targetAgent->getRobotStruct());
        if(!checkHumanReach())
        {
//            cout << "noooooooo" << endl;
            continue;
        }
        p3d_col_activate_robot_robot(mainObject->getRobotStruct(), _targetAgent->getRobotStruct());


        if (!testFactConstraints())
            continue;


//        _confSol = shared_ptr<Configuration>(new Configuration(mainObject,q));

        res = true;
    }


    mainAgent->setAndUpdate(*robotCurConf);
    mainObject->setAndUpdate(*objectCurConf);
    attachObject();
    return res;

}


bool PlaceReachable::setTargetAgent(multimap<string,string> agents_name)
{
    if (agents_name.count("targetAgent") != 1)
    {
        cout << "ERROR: no target agent for " << type << " task." << endl;
        return false;
    }
    _targetAgent = global_Project->getActiveScene()->getRobotByName(agents_name.find("targetAgent")->second);
    if (!_targetAgent)
        return false;
    return true;
}

bool PlaceReachable::setAgents(multimap<string,string> agents_name)
{
    return setMainAgent(agents_name)&& setTargetAgent(agents_name);
}


bool PlaceReachable::checkHumanReach()
{

    _WSF->setConfiguration(_confSol);
    global_Project->getActiveScene()->getFactsManager()->computeReachability(_WSF,_targetAgent,mainObject,true);

    fact_id FI3;
    FI3.factType = "IsReachable";
    FI3.factSubType = "GIK";
    FI3.isMainAnAgent = true;

    Fact F = _WSF->getFact(FI3, _targetAgent, mainObject);
//    F.printFact();
    return F.getBoolValue();
//    GeometricComputations GC = GeometricComputations();

//    p3d_point pr = GC.getCenterOfObj(mainObject);
//    pr.z = GC.getLowerPointsInObject(mainObject,0.01).at(0).z;

//    confPtr_t q_curr_tar = _targetAgent->getCurrentPos();

//    if (!checkHRIAgent(_targetAgent))
//        return false;

//    int nbArms = _targetAgent->getHriAgent()->manip->shoulder_joint_ids_no;
//    double agReach = _targetAgent->getHriAgent()->manip->reach_tolerance;


//    bool res = false;
//    int reachedReturnRepportCount = _returnRepportcount;
//    for(int j = 0; j < nbArms; j++)
//    {
//        _returnRepportcount = reachedReturnRepportCount;
//        p3d_point agShoulder;

//        agShoulder.x = _targetAgent->getJoint(_targetAgent->getHriAgent()->manip->shoulder_joint_ids[j])->getJointStruct()->abs_pos[0][3];
//        agShoulder.y = _targetAgent->getJoint(_targetAgent->getHriAgent()->manip->shoulder_joint_ids[j])->getJointStruct()->abs_pos[1][3];
//        agShoulder.z = _targetAgent->getJoint(_targetAgent->getHriAgent()->manip->shoulder_joint_ids[j])->getJointStruct()->abs_pos[2][3];

//        // test reach with dist
//        _returnRepportcount++;
//        if (GC.distBetweenPoints2D(pr,agShoulder) > agReach)
//        {
//            res = false;
//            if (_returnRepportcountMax < _returnRepportcount)
//            {
//                _returnRepportcountMax = _returnRepportcount;
//                returnRepport = "not_reachable_by_target_dist";
//            }
//            continue;
//        }
//        //test reach with IK
//        HRI_GIK_TASK_TYPE task;
//        if (j == 0)
//            task = GIK_LATREACH;
//        else if (j == 1)
//            task = GIK_RATREACH;

//        p3d_vector3 Tcoord;

////        p3d_point clos = GC.getClosestBBPointFromPoint(mainObject,agShoulder);
//        p3d_point clos = GC.getCenterOfObj(mainObject);
//        clos.z = GC.getUpperPointsInObject(mainObject,0.03).at(0).z +0.05;

//        Tcoord[0] = clos.x;
//        Tcoord[1] = clos.y;
//        Tcoord[2] = clos.z;
//        double distance_tolerance = 0.05;

//        configPt q = p3d_get_robot_config(_targetAgent->getRobotStruct());


//        int IKSucceded = hri_agent_single_task_manip_move(_targetAgent->getHriAgent(),
//                                                          task,
//                                                          &Tcoord,
//                                                          distance_tolerance,
//                                                          &q);
//        _returnRepportcount++;
//        p3d_set_and_update_this_robot_conf(_targetAgent->getRobotStruct(), q);
//        if (IKSucceded == 1)
//        {
////            g3d_draw_allwin_active();
////            sleep(3);


//            p3d_set_and_update_this_robot_conf(_targetAgent->getRobotStruct(), q);
//            p3d_col_deactivate_robot_robot(_targetAgent->getRobotStruct(), mainObject->getRobotStruct());
//            int coltest = p3d_col_test_robot(_targetAgent->getRobotStruct(),0);
//            p3d_col_deactivate_robot_robot(_targetAgent->getRobotStruct(), mainObject->getRobotStruct());
//            if (coltest != 1)
//            {
//                res = true;
//                _targetAgent->setAndUpdate(*q_curr_tar);
//                break;
//            }
//            else
//                p3d_print_col_pair();
////            cout << "aaaa "<< j << endl;
//            _targetAgent->setAndUpdate(*q_curr_tar);
////            g3d_draw_allwin_active();
////            sleep(3);
////            break;
//        }
//        else
//        {
//            if (_returnRepportcountMax < _returnRepportcount)
//            {
//                _returnRepportcountMax = _returnRepportcount;
//                returnRepport = "IK_fail_for_target";
//            }
//        }
//        _targetAgent->setAndUpdate(*q_curr_tar);
//    }
//    return res;
}


string PlaceReachable::getTextFromValues()
{
    return "Task: " + type
            + "\n    agents: "
            + "\n        mainAgent:    [AGENT ID]"
            + "\n        targetAgent:  [AGENT ID]"
            + "\n    objects: "
            + "\n        mainObject:   [OBJECT ID]"
            + "\n        supportObject [OBJECT ID] -optional-"
            + "\n    datas:"
            + "\n        hand: [right OR left]";
}
