#ifndef TASKMANAGER_HPP
#define TASKMANAGER_HPP

using namespace std;
//using namespace tr1;

#include "GTPTools/tools.hpp"
#include "GTPTools/WorldState.hpp"
#include "GTPTools/gtpplan.hpp"
#include "GTPTools/geometricTools.hpp"
#include "virtualtask.hpp"
#include "Tasks/task-pkg.hpp"
#include "GTPTools/taskconstraints.hpp"




class TaskManager
{
public:
    TaskManager();
    ~TaskManager();
    void init();

    PlanNode *computeTask(TasksId task,
                    multimap<string,string> agents_name,
                    multimap<string,string> objects_name,
                    multimap<string, p3d_point> point3d,
                    multimap<string, string> additionalData,
                    WorldState *WC, bool computeMP, TaskConstraints *taskConst);

    double computeTaskHeuristic(TasksId taskId,
                    multimap<string,string> agents_name,
                    multimap<string,string> objects_name,
                    multimap<string, p3d_point> point3d,
                    multimap<string, string> additionalData,
                    WorldState *WC, bool computeMP, TaskConstraints *taskConst);

    PlanNode *findTaskAlternative(int taskId, bool computeMP);

    Task* createTask(TasksId task);
    Task* createHumanTask(TasksId taskId);

    bool computeTrajInTasks(int taskId, int alternativeId);

    Task* getTask(int id);

    bool setArm(armLabel a);

    void clearTasks();

    int addWorldState(WorldState* WC);
    WorldState* getWorldStateById(int id);
    void clearWorldStates();

    TasksId getTaskIdFromString(string value);
    TasksId getTaskIdFromString(char* value);


    bool forceArmChoice;

    PlanTree* getPlanTree() {return _PT;}
    OnePlan* getOnePlan() {return _OP;}

    void printAllTasks(bool verbose);
    void printOnePlan(bool verbose);
    string printAllTextNeeded();

    vector<Task*> getTaskList() { return task_list;}
    void addTask(Task* t) {task_list.push_back(t);}

    void printTd(type_data td);

    void setCurrentId(int t){current_id = t;}

private:
    armLabel arm_l;

    int current_id;
    vector<Task*> task_list;
    vector<WorldState*> worldState_list;

    int current_worldState_id;

    PlanTree* _PT;
    OnePlan* _OP;

};

#endif // TASKMANAGER_HPP
