#include "taskmanagerinterface.h"
#include <boost/algorithm/string.hpp>
#include <boost/lexical_cast.hpp>

//############################# taskManagerInterface #####################

taskManagerInterface* TMI=0;


taskManagerInterface::taskManagerInterface()
{
    current_computing_id = -1;
    _PS = Iddle;
    current_sub_solution = NULL;
    current_executing_traj = NULL;
    setSolutionId(-1);
    _WS = NULL;
    _initWS = NULL;
    _computeMP = true;
    _constraints = new TaskConstraints();
    _isComputePlanSuccess = false;
    _currentHeuristic = P3D_HUGE;
}

void taskManagerInterface::initTaskManager()
{
    m_TM = new TaskManager();
    m_TM->init();
}

void taskManagerInterface::setTypeStr(string str)
{
    _typeStr = str;
}

void taskManagerInterface::setType(TasksId type)
{
    _type = type;
}

void taskManagerInterface::setAgent(string id, string name)
{
    if (_agents.find(id) == _agents.end() || _agents.find(id)->second != name) {
        _agents.erase(id);
        _agents.insert(std::pair<string,string>(id,name));
    }
}

void taskManagerInterface::setObject(string id, string name)
{
    if (_objects.find(id) == _objects.end() || _objects.find(id)->second != name) {
        _objects.erase(id);
        _objects.insert(std::pair<string,string>(id,name));
    }
}

void taskManagerInterface::addPoint(string id, p3d_point p)
{
    if (_points.find(id) == _points.end() || _points.find(id)->second.x != p.x) {
        _points.erase(id);
        _points.insert(std::pair<string,p3d_point>(id,p));
    }

//    _points.insert(std::pair<string,p3d_point>(id,p));
}

void taskManagerInterface::addAdditinalData(string id, string value)
{
    _additionalData.insert(std::pair<string,string>(id,value));
}


void taskManagerInterface::setWorldState(WorldState* WS)
{
    _WS = WS;
}

void taskManagerInterface::callGTP()
{
    _lastInputs.additionalData = _additionalData;
    _lastInputs.agents = _agents;
    _lastInputs.constraints = _constraints;
    _lastInputs.isAlreadyComputed = true;
    _lastInputs.MPComputation = _computeMP;
    _lastInputs.nbAlt = 0;
    _lastInputs.objects = _objects;
    _lastInputs.points = _points;
    _lastInputs.TaskId = -10;
    _lastInputs.type = _type;
    _lastInputs.ws = _WS;

    PlanNode* PN = m_TM->computeTask(_type, _agents, _objects, _points, _additionalData, _WS, _computeMP,_constraints);
    setCurrentSol(PN->getTaskSolution());
    setCurrentPN(PN);
}

void taskManagerInterface::callAlternative(int taskId)
{
    PlanNode* PN = m_TM->findTaskAlternative(taskId,_computeMP);

    //todo: fill in _lastInputs.

    setCurrentSol(PN->getTaskSolution());
    setCurrentPN(PN);
}

void taskManagerInterface::desactiveConstraints(Robot* ag)
{
    p3d_cntrt* ct;
    // over all constraints
    for(int i=0; i<ag->getRobotStruct()->cntrt_manager->ncntrts; i++)
    {
        string name = ag->getRobotStruct()->cntrt_manager->cntrts[i]->namecntrt;

        if( name == "p3d_fix_jnts_relpos" ){
            continue;
        }
        // get constraint from the cntrts manager
        ct = ag->getRobotStruct()->cntrt_manager->cntrts[i];
        p3d_desactivateCntrt( ag->getRobotStruct(), ct );
    }


}


double taskManagerInterface::getTaskHeuristic()
{
    _currentHeuristic=  m_TM->computeTaskHeuristic(_type, _agents, _objects, _points, _additionalData, _WS, _computeMP,_constraints);
    return _currentHeuristic;
}


Task* taskManagerInterface::getLastSolutionFound()
{
    if (current_computing_id  == -1)
    {
        cout << "no last solution!!" << endl;
        Task* t = NULL;
        return t;
    }
    int i = current_computing_id;
    current_computing_id = -1;
    return m_TM->getTask(i);

}

void taskManagerInterface::resetAll()
{
    _agents.clear();
    _objects.clear();
    _additionalData.clear();
    _points.clear();
    _constraints->cleanAll();

    current_computing_id = -1;
    _PS = Iddle;
    current_sub_solution =NULL;
    current_executing_traj = NULL;

    m_TM->clearTasks();
    setSolutionId(-1);

    getTaskManager()->clearWorldStates();

}

void taskManagerInterface::clearAllInputs()
{
    _agents.clear();
    _objects.clear();
    _additionalData.clear();
    _points.clear();
    _constraints->cleanAll();

    current_computing_id = -1;
    _PS = Iddle;
    current_sub_solution = NULL;
    current_executing_traj = NULL;


}

string taskManagerInterface::getCurrentInputsDescr()
{
    string s;
    s = _typeStr + " " + to_string(_type);

    for ( std::multimap< string, string, std::less< string > >::const_iterator iter =_agents.begin(); iter != _agents.end(); ++iter )
        s = s + "\n" + iter->first + " : " + iter->second;

    for ( std::multimap< string, string, std::less< string > >::const_iterator iter =_objects.begin(); iter != _objects.end(); ++iter )
        s+= "\n" + iter->first + " : " + iter->second;

    for ( std::multimap< string, p3d_point >::const_iterator iter =_points.begin(); iter != _points.end(); ++iter )
    {
        p3d_point p = iter->second;
        s+= "\n" + iter->first + " : ( " + to_string(p.x) + ", " + to_string(p.y) + ", " + to_string(p.z) + ")";
    }

    for ( std::multimap< string, string, std::less< string > >::const_iterator iter =_additionalData.begin(); iter != _additionalData.end(); ++iter )
        s+= "\n" + iter->first + " : " + iter->second;

    s+="\n" + getConstraints()->getDescr();
//    cout << s << endl;
    return s;
}

void taskManagerInterface::removeConstraintFromPrevTS(int taskId, int altId)
{
    m_TM->getPlanTree()->getNodeFromId(taskId,altId)->getTaskSolution()->getStopWS()->clearTCForNextTask();
}


void taskManagerInterface::addCurrentTaskToList()
{
    task_in TI;
    TI.typeStr = _typeStr;
    TI.type = _type;
    TI.agents = _agents;
    TI.objects = _objects;
    TI.additionalData = _additionalData;
    TI.points = _points;
    TI.MPComputation = true;
    TI.constraints = new TaskConstraints(_constraints);
    TI.isAlreadyComputed = false;
    TI.nbAlt = 0;

    _tasksList.push_back(TI);
}

void taskManagerInterface::loadCurrentTaskFrom(task_in TI)
{
    _typeStr = TI.typeStr;
    _type = TI.type;
    _agents = TI.agents;
    _objects = TI.objects;
    _additionalData = TI.additionalData;
    _points = TI.points;
    _constraints = TI.constraints;
}

void taskManagerInterface::clearList()
{
    for (unsigned int i = 0; i < _tasksList.size(); i++)
    {
        delete _tasksList.at(i).constraints;
    }
    _tasksList.clear();
}

void taskManagerInterface::removeLastTaskFromList()
{
    _tasksList.erase(_tasksList.end());
}

void taskManagerInterface::resetAllFromScratch()
{
    resetAll();
    clearList();
    delete m_TM;
    m_TM = new TaskManager();
    m_TM->init();
}



string taskManagerInterface::getOneTaskDescr(task_in t)
{
    string s = "";
    s+= "\n\nstype:\n";
    s+= "\t" + t.typeStr + " \n";

    s+= "agents:\n";
    for ( std::multimap< string, string, std::less< string > >::const_iterator iter =t.agents.begin();
          iter != t.agents.end(); ++iter )
        s+=  "\t" + iter->first + " : " + iter->second;


    s+= "\nobjects:\n";
    for ( std::multimap< string, string, std::less< string > >::const_iterator iter =t.objects.begin();
          iter != t.objects.end(); ++iter )
        s+=  "\t" + iter->first + " : " + iter->second;

    s+= "\ndatas:\n";
    for ( std::multimap< string, string, std::less< string > >::const_iterator iter =t.additionalData.begin();
          iter != t.additionalData.end(); ++iter )
        s+=  "\t" + iter->first + " : " + iter->second;

    s+= "\npoints:\n";
    for ( std::multimap< string, p3d_point >::const_iterator iter =t.points.begin(); iter != t.points.end(); ++iter )
        s+=  "\t" + iter->first + " : (" + to_string(iter->second.x) + " , " + to_string(iter->second.y)
                + " , " + to_string(iter->second.z) + " )";

    s+= "\nconstraints:\n";
    s+= "\tnot yet printed";
    return s;
}

string taskManagerInterface::getListDescr()
{
    string s;
    for (unsigned int i = 0; i < _tasksList.size(); i++)
    {
        s+=getOneTaskDescr(_tasksList.at(i));
    }
    return s;
}


TaskConstraints* taskManagerInterface::getGoalConstraints()
{
    return _goalConstraints;
}

bool taskManagerInterface::testGoalConstraints()
{
    return _goalConstraints->areFactConstraintRespected(getTaskManager()->getOnePlan()->getLastSol()->getStopWS());
}


bool taskManagerInterface::computePlan(bool computeMP, int init_id)
{
    int maxNbAlt = 7;
    double timeLimit = 10*maxNbAlt*_tasksList.size();
    clock_t begin = clock();
    _computeMP = computeMP;

    WorldState* initWS = _WS;

    unsigned int cId = init_id;
    while (cId < _tasksList.size() && timeLimit > (double)(clock() - begin)/CLOCKS_PER_SEC )
    {
        cout << "id count = " << cId << endl;
        if (!_tasksList.at(cId).isAlreadyComputed)
        {
            if (cId == 0)
                _WS = initWS;
            else
            {
                if (!getTaskManager()->getOnePlan()->getLastSol())
                {
                    cout << "ERROR: the plan is not filled correctly" << endl;
                    return false;
                }
                _WS = getTaskManager()->getOnePlan()->getLastSol()->getStopWS();
            }
            loadCurrentTaskFrom(_tasksList.at(cId));
            callGTP();
            _tasksList.at(cId).nbAlt = 0;
        }
        else
        {
            _computeMP = computeMP;
            callAlternative(_tasksList.at(cId).TaskId);
        }
        if (current_solution)
        {
            _tasksList.at(cId).TaskId = current_solution->getTask()->getId();
            _tasksList.at(cId).nbAlt = _tasksList.at(cId).nbAlt + 1;

            int task_id = current_solution->getTask()->getId();
            int alternativeId = current_solution->getId();
            getTaskManager()->getOnePlan()->addToPlan(getTaskManager()->getPlanTree()->getNodeFromId(task_id,alternativeId));

            current_solution->getStopWS()->load();
            g3d_draw_allwin_active();

            _tasksList.at(cId).isAlreadyComputed = true;
            cId += 1;

        }
        else
        {
            if (cId == 0)
                break;
            cId -=1;
            while (_tasksList.at(cId).nbAlt > maxNbAlt && cId != 0)
            {
                _tasksList.at(cId).nbAlt = 0;
                _tasksList.at(cId).isAlreadyComputed = false;
                cId -= 1;
            }
            if (cId == 0)
                break;

        }

        getTaskManager()->getOnePlan()->printPlan();
//        sleep(2);
    }

    if (cId >= _tasksList.size())
    {
        cout << "#####################"<< endl;
        cout << "plan computation success " << (double)(clock() - begin)/CLOCKS_PER_SEC << " s" << endl;
        cout << "#####################"<< endl;

        return true;
    }

    cout << "#####################"<< endl;
    cout << "fail to compute plan, time reached = " << (double)(clock() - begin)/CLOCKS_PER_SEC << " s and max time = "
         << timeLimit << " s" << endl;
    return false;
}


bool taskManagerInterface::computeJustMP()
{
    vector<PlanNode*> vpn = getTaskManager()->getOnePlan()->getPlan();

    bool test = true;
    int endId = -1;
    for(unsigned int i = 0; i <vpn.size(); i++)
    {
        TaskSolution* TS = vpn.at(i)->getTaskSolution();
        test = getTaskManager()->computeTrajInTasks(TS->getTask()->getId(), TS->getId());
        if (!test)
        {
            endId = i;
            break;
        }
    }

    if (!test)
    {
        return computePlan(true,endId);
    }
    return true;
}

bool taskManagerInterface::getWSFromRangeParam(int taskId, int altId, WorldState* ws, double rangeParam)
{
    Task* t = getTaskManager()->getTask(taskId);
    if (!t)
    {
        cout << "no task with id = " << taskId << endl;
        return false;
    }
    TaskSolution* TS = t->getResult()->getAlternativeWithId(altId);
    if (!TS)
    {
        cout << "no task alternative with id = " << altId << " in task with id " << taskId << endl;
        return false;
    }
    if (!ws)
    {
        cout << "ws is empty" << endl;
        return false;
    }


    if (ws->getRobotConfs().size() == 0)
    {
        cout << "ws is not set" << endl;
        return false;
    }

    TS->getStartWS()->load();
    double curr_range_param = 0;
    vector<TaskSubSolution*> vTSS = TS->getAllTSS();
    configPt tmp = NULL;
    TaskSubSolution* curr_tss = NULL;
    for (unsigned int i = 0; i < vTSS.size(); i++)
    {
        TaskSubSolution* tss = vTSS.at(i);
        if (!tss)
            continue;

        p3d_traj* tr;
        if (tss->isNavigation() && !tss->getTraj())
        {
            API::Trajectory p3d_trajectory(tss->getRobot());
            tss->getStartState()->load();
            shared_ptr<Configuration> q_f(tss->getRobot()->getCurrentPos());
            p3d_trajectory.push_back(q_f);
            vector<p3d_point> _nav_path = tss->getNavVector();
            for (unsigned int i = 0; i < _nav_path.size(); i++)
            {
                shared_ptr<Configuration> q_tmp(tss->getRobot()->getCurrentPos());
                cout << " cell = " << _nav_path.at(i).x << " , " << _nav_path.at(i).y << endl;
                (*q_tmp)[6] = _nav_path.at(i).x;
                (*q_tmp)[7] = _nav_path.at(i).y;
                (*q_tmp)[11] = _nav_path.at(i).z;
                p3d_trajectory.push_back(q_tmp);

            }
            p3d_traj* trajPt = NULL;
            tss->getRobot()->getRobotStruct()->tcur = NULL;

            p3d_trajectory.replaceP3dTraj(trajPt);
            if (!tss->getRobot()->getRobotStruct()->tcur)
            {
                cout << "Fail to create traj from points" << endl;
                return false;
            }
            tss->setTraj(tss->getRobot()->getRobotStruct()->tcur);
        }

        tr = tss->getTraj();
        if (!tr)
        {
            if (!TS->isTrajComputed())
                cout << "this trajectory is not computed (by choice) only the end state will be shown" << endl;
            continue;
        }
//        cout << "rangeParam = " << rangeParam  << " tr->range_param = " << tr->range_param << " curr_range_param = " << curr_range_param << endl;
        if (rangeParam > tr->range_param + curr_range_param) //case where range_param bigger then actual lenght
        {
            if (tss->getNext().size() ==0) //case where this is the last tss
            {
                cout << "traj ended" << endl;
                return false;
            }
            curr_range_param += tr->range_param;
            continue; //case where there is more tss
        }
        else //case where range param is within this traj
        {
            tmp = p3d_config_at_distance_along_traj(tr, rangeParam - curr_range_param);
            curr_tss = tss;
            break;
        }
    }
    if (!tmp ||  !curr_tss)
    {
        return false;
    }
    vector< pair<p3d_matrix4,Attachement> > atts;
    if (curr_tss->getStartState()->getAttachements().size() > 0) //getting objects attachements
    {
        curr_tss->getStartState()->load();
//        cout << curr_tss->getSubSolutionName()<< " nb of attachement = " << curr_tss->getStartState()->getAttachements().size() << endl;

        for (unsigned int i = 0; i < curr_tss->getStartState()->getAttachements().size(); i++)
        {
            Attachement att = curr_tss->getStartState()->getAttachements().at(i);
            p3d_matrix4 matRot, objPos, handPos, invHandPos;

            p3d_mat4Copy(att.object->getJoint(0)->getJointStruct()->abs_pos, objPos);
            int f = att.agent->getHriAgent()->manip->shoulder_joint_ids[att.armId]+7;

            p3d_mat4Copy(att.agent->getRobotStruct()->joints[f]->abs_pos, handPos);
            p3d_matInvertXform(handPos,invHandPos);
            p3d_mat4Mult(invHandPos, objPos, matRot);

            pair<p3d_matrix4, Attachement> p;
            p3d_mat4Copy( matRot, p.first);
//            p3d_mat4Print(p.first, "p.first");
            p.second = att;

            atts.push_back(p);

        }
        ws->load();
    }


    confPtr_t qa = (tr1::shared_ptr<Configuration>)(new Configuration(curr_tss->getRobot(), tmp));
    ws->replaceConf(qa);
    ws->load();
    for (unsigned int i = 0; i < atts.size(); i++)
    {
        p3d_matrix4  objPos, handPos;
        int f = atts.at(i).second.agent->getHriAgent()->manip->shoulder_joint_ids[atts.at(i).second.armId]+7;

        p3d_mat4Copy(atts.at(i).second.agent->getRobotStruct()->joints[f]->abs_pos, handPos);
        p3d_mat4Mult(handPos, atts.at(i).first, objPos);

        double Tx, Ty, Tz;
        double Rx, Ry, Rz;

//        p3d_mat4Print(handPos, "handPos");
//        p3d_mat4Print(atts.at(i).first, "atts.at(i).first");
//        p3d_mat4Print(objPos, "objpose");
        p3d_mat4ExtractPosReverseOrder(objPos, &Tx, &Ty, &Tz,
                                       &Rx, &Ry, &Rz);
        confPtr_t q = atts.at(i).second.object->getCurrentPos();
        (*q)[6] = Tx;
        (*q)[7] = Ty;
        (*q)[8] = Tz;
        (*q)[9] = Rx;
        (*q)[10] = Ry;
        (*q)[11] = Rz;
        ws->replaceConf(q);
    }
    return true;

}


vector<pair<pair<int,int>, double> > taskManagerInterface::getBacktrackPriorities()
{

    //size of the remaining solution space in the task
    vector<PlanNode*> tree = m_TM->getPlanTree()->getAllNodes();
    for (unsigned int i = 0; i < m_TM->getTaskList().size(); i++)
    {
        Task* t =m_TM->getTaskList().at(i);
        double sSSize = t->getSSSizeLeft();
        for (int j = 0; j< t->getResult()->getNbAlternatives(); j++)
        {
            m_TM->getPlanTree()->getNodeFormTS(t->getResult()->getAlternativeWithId(j))->backtrackingData["size"] = sSSize;
        }
    }
    TaskSolution* TS = _lastInputs.ws->getTSThatEndWithThisWS();

    PlanNode* PTmp = m_TM->getPlanTree()->getNodeFormTS(TS);

    double pr = 0;
    bool firstGoto = false;
    bool firstPick = false;
    bool firstMove = false;
    string currMoveObj = _lastInputs.objects.find("supportObject")->second;
    double prMove =0;
    while (PTmp)
    {
        //manipulation and navigation antagonism
        pr = 0;
        if ((_lastInputs.type == pick || _lastInputs.type == place || _lastInputs.type == drop || _lastInputs.type == stackObj) && !firstGoto)
        {
            if(PTmp->getTask()->getTypeId() == gotoPlace)
            {
                pr = 1;
                firstGoto = true;
            }
        }
        PTmp->backtrackingData["goto"] = pr;


        //when placing an object on another manipulated one.
        bool test = false;
        if (_lastInputs.type == place || _lastInputs.type == stackObj)
        {
            if (_lastInputs.objects.count("supportObject") &&
                    (PTmp->getTask()->getTypeId() == place || PTmp->getTask()->getTypeId() == stackObj) &&
                    PTmp->getTaskSolution()->getTypeData().objects.find("mainObject")->second
                    == currMoveObj)
            {
                prMove++;
                currMoveObj = PTmp->getTaskSolution()->getTypeData().objects.find("supportObject")->second;
                test = true;
            }
        }

        double t = prMove;
        if(!test)
            t = 0;
        PTmp->backtrackingData["moveObj"] = t;


        //if object in hand, how did i get it
        pr=0;
        if ((_lastInputs.type == place || _lastInputs.type == drop || _lastInputs.type == stackObj) && !firstPick)
        {
            if(PTmp->getTask()->getTypeId() == pick)
            {
                pr = 1;
                firstPick = true;
            }
        }
        PTmp->backtrackingData["pick"] = pr;


        //where was the object i wanted to pick?
        pr=0;
        if (_lastInputs.type == pick && !firstMove)
        {
            if((PTmp->getTask()->getTypeId() == place
                || PTmp->getTask()->getTypeId() == drop
                || PTmp->getTask()->getTypeId() == stackObj)
                    && PTmp->getTaskSolution()->getTypeData().objects.find("mainObject")->second
                    == _lastInputs.objects.find("mainObject")->second )
            {
                pr = 1;
                firstMove =true;
            }
        }
        PTmp->backtrackingData["move"] = pr;


        //shared workspace
        pr=0;
        if (_lastInputs.type == place)
        {
            if(PTmp->getTask()->getTypeId() == place
                    && PTmp->getTaskSolution()->getTypeData().objects.find("supportObject")->second
                    == _lastInputs.objects.find("supportObject")->second )
            {
                pr = 1;
            }
        }
        PTmp->backtrackingData["shared"] = pr;

        PTmp = PTmp->getPrevious();
    }

    //remetre sur 1
    map<string,double> maxValues;
    for (unsigned int i = 0; i< tree.size(); i++)
    {
        PlanNode* PN = tree.at(i);

        map<string,double>::iterator it;
        for(it = PN->backtrackingData.begin(); it != PN->backtrackingData.end(); it++)
        {
            if (maxValues.count(it->first) != 1 || maxValues[it->first] < it->second)
                maxValues[it->first] = it->second;
        }
    }

    for (unsigned int i = 0; i< tree.size(); i++)
    {
        PlanNode* PN = tree.at(i);
        map<string,double>::iterator it;
        for(it = PN->backtrackingData.begin(); it != PN->backtrackingData.end(); it++)
        {
            if (maxValues[it->first] != 0)
                PN->backtrackingData[it->first] = it->second/maxValues[it->first];

        }
    }

    //mettre en forme
    vector<pair<pair<int,int>, double> > iid;
    for (unsigned int i = 0; i< tree.size(); i++)
    {
        PlanNode* PN = tree.at(i);
        pair<int,int> er = PN->getId();
        pair<pair<int,int>,double> v;
        v.first = er;
        v.second = PN->computePriority();
        iid.push_back(v);
    }

    return iid;

}

void taskManagerInterface::savePlanToDisk(){
    char filename[200];
    sprintf(filename,"%s/gtp_plan",getenv("HOME"));
    FILE *file=fopen(filename,"w");
    if(!file){
        printf("error: taskManagerInterface::savePlanToDisk: cannot open file %s\n",filename);
        return;
    }
    fprintf(file,"# GTP plan generated by move3d::taskManagerInterface::savePlanToDisk()\n\n");
    printf("Saving current plan to file %s ...\n",filename);
    if(!this->getTaskManager()|| !this->getTaskManager()->getPlanTree())
        return;
//    vector<PlanNode*> VPN=this->getTaskManager()->getOnePlan()->getPlan();
    vector<PlanNode*> VPN = this->getTaskManager()->getPlanTree()->getAllNodes();
//    if(!VPN.empty()){
//        PlanNode *PN=VPN.at(0);
//        if(PN
//        && PN->getTaskSolution()
//        && PN->getTaskSolution()->getFirstSubSolution()
//        && PN->getTaskSolution()->getFirstSubSolution()->getStartState())
//            PN->getTaskSolution()->getFirstSubSolution()->getStartState()->saveToDisk(file);
//    }

    for(unsigned int i=0;i<VPN.size();i++){
        PlanNode *PN=VPN.at(i);
        if(!PN || !PN->getTask()) continue;
        fprintf(file,"\n\nbeg_task");
        fprintf(file,"\ntype %s",PN->getTask()->getType().c_str());
        fprintf(file,"\ntask_id %d",PN->getId().first);
        fprintf(file,"\nalternative_id %d",PN->getId().second);
        if (PN->getPrevious())
        {
            fprintf(file,"\ntask_prec_id %d",PN->getPrevious()->getId().first);
            fprintf(file,"\nalternative_prec_id %d\n",PN->getPrevious()->getId().second);
        }
        else
        {
            fprintf(file,"\ntask_prec_id %d",-1);
            fprintf(file,"\nalternative_prec_id %d\n",-1);
        }
        if(!PN->getTaskSolution()) continue;
        TaskSubSolution* TSS=PN->getTaskSolution()->getFirstSubSolution();
        while(TSS){
            TSS->saveToDisk(file);
            vector<TaskSubSolution*> next=TSS->getNext();
            if(!next.empty())
                TSS=next.at(0);
            else
                TSS=NULL;
        }
        fprintf(file,"end_task\n");
    }
    fclose(file);
    printf("Done\n");

    return;

}

void taskManagerInterface::loadPlanFromDisk()
{

    //TODO: the next paragraph is a hack, the planning does not work without it.
    Robot* dummy = global_Project->getActiveScene()->getActiveRobot();
    configPt qw=p3d_copy_config(dummy->getRobotStruct(),dummy->getCurrentPos()->getConfigStruct());
    if (dummy->getHriAgent() && dummy->getHriAgent()->type == HRI_PR2)
        qw[22] = qw[22] +0.1;
    p3d_plan(dummy->getRobotStruct(), dummy->getCurrentPos()->getConfigStruct(), qw, PLAN_RIGHT_ARM, false);


    char filename[200];
    sprintf(filename,"%s/gtp_plan",getenv("HOME"));
    FILE *file=fopen(filename,"r");
    if(!file){
        printf("error: taskManagerInterface::loadPlanFromDisk: cannot open file %s\n",filename);
        return;
    }

    string text= "";
    char buffer [100];
    while ( ! feof (file) )
    {
      if ( fgets (buffer , 100 , file) == NULL ) break;
      string ad = string(buffer);
      text.append(ad);
    }

    fclose(file);

    vector<string> vt;
    boost::split(vt, text, boost::is_any_of("\n"));

    bool inTask = false;
    vector<string> oneTask;
    for (unsigned int i = 0; i< vt.size(); i++)
    {
        boost::algorithm::trim(vt.at(i));
        if (vt.at(i) == "beg_task")
        {
            inTask = true;
        }
        else if (vt.at(i) == "end_task")
        {
            inTask = false;
            readTask(oneTask);
            oneTask.clear();
        }

        if (inTask)
            oneTask.push_back(vt.at(i));
    }



    printf("Done\n");
    return;
}

void taskManagerInterface::readTask(vector<string> vt)
{
    string type = "";
    int taskId,altId,precTaskId,precAltId;
    bool inTSS = false;
    vector<string> oneTSS;

    Task* task;
    TaskAllSolutions* result;
    TaskSolution* TS;

    vector<TaskSubSolution*> TSSvector;

    for (unsigned int i = 0; i< vt.size(); i++)
    {
        string l = vt.at(i);
        if (l.find("type") != std::string::npos && !inTSS)
        {
            type = l;
            type.erase(0,5);
        }
        else if (l.find("task_id") != std::string::npos && !inTSS)
        {
            string tmp = l;
            taskId = boost::lexical_cast<int>(tmp.erase(0,8));
            TasksId taId = m_TM->getTaskIdFromString(type);
            task = m_TM->createTask(taId);
            result = new TaskAllSolutions(task);
        }
        else if (l.find("alternative_id") != std::string::npos && !inTSS)
        {
            string tmp = l;
            altId = boost::lexical_cast<int>(tmp.erase(0,15));
            TS = new TaskSolution(result,altId);
            TS->setTrajComputed(true);
        }

        else if (l.find("task_prec_id") != std::string::npos && !inTSS)
        {
            string tmp = l;
            precTaskId = boost::lexical_cast<int>(tmp.erase(0,13));
        }

        else if (l.find("alternative_prec_id") != std::string::npos && !inTSS)
        {
            string tmp = l;
            precAltId = boost::lexical_cast<int>(tmp.erase(0,20));
        }
        else if (l == "beg_tss" && !inTSS)
        {
            inTSS = true;
        }
        else if (l == "end_tss" && inTSS)
        {
            if (type != "")
            {
//                TaskSubSolution* TSS = readTSS(TS, oneTSS);
                TSSvector.push_back(readTSS(TS, oneTSS));
                inTSS = false;
                oneTSS.clear();
            }
        }
        else if (inTSS)
        {
            oneTSS.push_back(l);
        }
    }

    for (unsigned int i = 0; i< TSSvector.size()-1;i++)
    {
        TaskSubSolution* tss = TSSvector.at(i);
        if (tss)
        {
            vector<TaskSubSolution*> tmptss;
            tmptss.push_back(TSSvector.at(i+1));
            tss->setNext(tmptss);
        }
    }
    if(TSSvector.size() >0)
    {
        TS->setStartWS(TSSvector.at(0)->getStartState());
        TSSvector.at(TSSvector.size()-1)->getEndState()->setTSThatEndWithThisWS(TS);
        TS->setStopWS(TSSvector.at(TSSvector.size()-1)->getEndState());

    }

    //creating the task:

    if (!task)
        return;

    task->setPlanTree(m_TM->getPlanTree());

    task->setResult(result);
    task->setId(taskId);
    m_TM->setCurrentId(taskId+1);
    m_TM->addTask(task);

    PlanNode* PN = new PlanNode(m_TM->getPlanTree(),TS);
    PN->setTask(task);

    TaskSolution* prevTS = NULL;
    if (precTaskId < 0 || precAltId < 0)
        prevTS = NULL;
    else
        prevTS = m_TM->getPlanTree()->getNodeFromId(precTaskId,precAltId)->getTaskSolution();

    bool test = m_TM->getPlanTree()->addPlanNode(PN, prevTS);

    if (!test)
        cout << "faillure due to node already existing in the tree" << endl;

    TS->getStopWS()->load();
}

TaskSubSolution* taskManagerInterface::readTSS(TaskSolution* TS, vector<string> vt)
{
    bool inWS = false;
    string name = "";
    int tssId,armId;
    WorldState* WSstart, *WSend;
    vector<string> oneWS;

    Robot* r;

    bool inTraj = false;
    vector<string> oneTraj;
    p3d_traj* t = NULL;

    for (unsigned int i = 0; i< vt.size(); i++)
    {
        string l = vt.at(i);
//        cout << l << endl;
        vector<string> vmtmp;
        boost::split(vmtmp, l, boost::is_any_of(" "));
        if (vmtmp.at(0) == "name" && !inWS && !inTraj)
        {
            name = vmtmp.at(1);
        }
        else if (vmtmp.at(0) == "tssId" && !inWS && !inTraj)
        {
            tssId = boost::lexical_cast<int>(vmtmp.at(1));
        }
        else if (vmtmp.at(0) == "armId"  && !inWS && !inTraj)
        {
            armId = boost::lexical_cast<int>(vmtmp.at(1));
        }
        else if (vmtmp.at(0) == "robot"  && !inWS && !inTraj)
        {
            r = global_Project->getActiveScene()->getRobotByName(vmtmp.at(1));
        }
        else if (l == "beg_world_state" && !inWS && !inTraj)
        {
            inWS = true;
        }
        else if (l == "end_world_state" && inWS && !inTraj)
        {
            inWS = false;
            pair<string,WorldState*> ww = readWS(oneWS);
            if (ww.first == "start")
            {
                WSstart = ww.second;
            }
            else if (ww.first == "end")
            {
                WSend = ww.second;
            }
            else
            {
                cout << "ERROR: not handled type of WS" << endl;
            }

            oneWS.clear();

        }
        else if (inWS && !inTraj)
        {
            oneWS.push_back(l);
        }
        else if(vmtmp.at(0) == "beg_traj" && !inTraj && !inWS)
        {
            inTraj = true;
        }
        else if(vmtmp.at(0) == "end_traj" && inTraj && !inWS)
        {
            t = readTraj(oneTraj, r);
            oneTraj.clear();
            inTraj = false;
        }
        else if (inTraj && !inWS)
        {
            oneTraj.push_back(l);
        }
        else
        {
            cout << "not good" << endl;
        }
    }


    TaskSubSolution* TSS = new TaskSubSolution(TS,name,tssId,t,r,0);
    TSS->setArmId(armId);
    TSS->setStartState(WSstart);
    TSS->setEndState(WSend);


    return TSS;
}

pair<string, WorldState *> taskManagerInterface::readWS(vector<string> vt)
{
    WorldState* WS = new WorldState(global_Project->getActiveScene());
    WS->saveAll();

    pair<string, WorldState *> ww;
    vector<Attachement> atts;
    for (unsigned int i = 0; i< vt.size(); i++)
    {
        string l = vt.at(i);
        vector<string> vmtmp;
        boost::split(vmtmp, l, boost::is_any_of(" "));

        if (vmtmp.at(0) ==  "type")
        {
            ww.first = vmtmp.at(1);
        }
        else if (vmtmp.at(0) ==  "config")
        {
            confPtr_t q = readConf(vmtmp,NULL);
            WS->replaceConf(q);
        }
        else if (vmtmp.at(0) ==  "attachement")
        {
            Attachement a;
            a.agent = global_Project->getActiveScene()->getRobotByName(vmtmp.at(1));
            a.object = global_Project->getActiveScene()->getRobotByName(vmtmp.at(2));
//            a.grasp = NULL;
            a.armId = boost::lexical_cast<int>(vmtmp.at(3));
            atts.push_back(a);
        }
    }

    WS->setAttachements(atts);
    ww.second = WS;


    WS->load();
    m_TM->addWorldState(WS);
    return ww;

}

confPtr_t taskManagerInterface::readConf(vector<string> ligne, Robot* r)
{
    vector<string> vmt = ligne;
    vmt.erase(vmt.begin());
    if (!r)
    {
        r = global_Project->getActiveScene()->getRobotByName(vmt.at(0));
        vmt.erase(vmt.begin());
    }
    confPtr_t q = r->getCurrentPos();

    unsigned int nbDof= 0;
    unsigned int njnt = r->getNumberOfJoints();
    for (unsigned int i = 0; i < njnt; i++)
        nbDof += r->getJoint(i)->getNumberOfDof();

//    cout << "nbDof for " << r->getName() << " is " << nbDof << " and vmt size is = " << vmt.size() << endl;
//    for (unsigned int i=0;i<vmt.size();i++)
//        cout << vmt.at(i) << " | ";
//    cout<<endl<<"-------------------"<< endl;

    if (vmt.size() == 1 && vmt.at(0) == "zero")
    {
        for (unsigned int i=0;i<vmt.size();i++)
            (*q)[i+6] = 0;
    }
    else if (vmt.size() == nbDof)
    {
        for (unsigned int i=0;i<vmt.size();i++)
        {
            double newP = boost::lexical_cast<double>(vmt.at(i));
            (*q)[i+6] = newP;
        }
    }
    else
    {
        cout << "ERROR, reading a conf" << endl;
    }


    return q;

}

p3d_traj* taskManagerInterface::readTraj(vector<string> vt, Robot* r)
{

    bool inLP = false;
    vector<string> oneLP;
    vector<p3d_localpath*> vLP;
    string localPathType;
    for (unsigned int i = 0; i< vt.size(); i++)
    {
        string l = vt.at(i);
        vector<string> vmtmp;
        boost::split(vmtmp, l, boost::is_any_of(" "));

        if (vmtmp.at(0) ==  "beg_localpath" && !inLP)
        {
            localPathType = vmtmp.at(1);
            inLP = true;
        }
        else if (l ==  "end_localpath" && inLP)
        {
            vLP.push_back(readLocalpath(oneLP,localPathType,r));
            inLP = false;
            oneLP.clear();
        }
        else if (inLP)
        {
            oneLP.push_back(l);
        }
    }


    if (vLP.size()<= 0)
        return NULL;

    double rp = 0;
    for (unsigned int i = 0; i<vLP.size()-1;i++)
    {
        p3d_localpath* lp = vLP.at(i);
        rp += lp->range_param;
        lp->next_lp = vLP.at(i+1);
    }
    rp+=vLP.at(vLP.size()-1)->range_param;

    p3d_traj *traj = p3d_create_empty_trajectory(r->getRobotStruct());
    traj->courbePt = vLP.at(0);
    traj->nlp = vLP.size();
    traj->range_param = rp;


    return traj;
}

p3d_localpath *taskManagerInterface::readLocalpath(vector<string> vt, string type, Robot* r)
{
    confPtr_t q1,q2;
    for (unsigned int i = 0; i< vt.size(); i++)
    {
        string l = vt.at(i);
        vector<string> vmtmp;
        boost::split(vmtmp, l, boost::is_any_of(" "));

        if (vmtmp.at(0) ==  "q_init")
        {
            q1 = readConf(vmtmp,r);
        }
        else if (vmtmp.at(0) ==  "q_end")
        {
            q2 = readConf(vmtmp,r);
        }
    }

//    cout << "reading localpath" << endl;

    p3d_set_this_robot_steering_method(r->getRobotStruct(), type.c_str());

    p3d_localpath *lp = p3d_local_planner(r->getRobotStruct(), q1->getConfigStruct(), q2->getConfigStruct());

    return lp;
}




bool taskManagerInterface::addAttachementsFromTask(int taskId, int alternativeId)
{
    PlanNode* PN = m_TM->getPlanTree()->getNodeFromId(taskId,alternativeId);
    if (!PN)
        return false;
    _attachements = PN->getTaskSolution()->getStopWS()->getAttachements();
    return true;
}

bool taskManagerInterface::addAttachementFromObject(Robot* robot, Robot *object, int armId)
{

    if (robot->getHriAgent()->type != HRI_PR2)
    {
        cout << "only PR2 is implemented for now" << endl;
        return false;
    }

    std::list<gpGrasp> gl;
    gpHand_type handType = GP_PR2_GRIPPER;;

    int te = gpGet_grasp_list(object->getName(), handType, gl);

    if (te != GP_OK || gl.size() == 0)
    {
        cout << "no grasps for this object" <<endl;
        return false;
    }
    Attachement a;
    a.agent = robot;
    a.object = object;
    a.armId = armId;
    a.grasp = *gl.begin();

    _attachements.push_back(a);
    return true;

}

void taskManagerInterface::clearAttachements()
{
    _attachements.clear();
}
