
#include "../API/HPP_Wrap/Romeo_Wrap.hpp"
#include "virtualtask.hpp"
#include "P3d-pkg.h"

//#include <string>
//#include <sstream>

#include "API/project.hpp"


#include "planner/plannerFunctions.hpp"
#include "plannerFunctions.hpp"

#ifdef LIGHT_PLANNER
#include "LightPlanner-pkg.h"
#endif

#define MAXIKLOOP 200

#ifdef HPP
RomeoWrap *_romeoWrap;
#endif



#include <iostream>
#include <map>

Task::Task()
{
    _taskGraph = NULL;
    _planTree = NULL;
    timeMp = 0;
    result = NULL;
    returnRepport = "ERROR";
}

int Task::getId()
{
    return task_id;
}

void Task::setId(int id)
{
    task_id = id;
}



bool Task::run(multimap<string,string> agents_name,
               multimap<string,string> objects_name,
               multimap<string,p3d_point> point3d,
               multimap<string,string> additionalData,
               WorldState* WS,bool computeMP)
{
    WS->load();
    _WS = WS;
    initAll();
    if (!setAgents(agents_name))
        return false;
    if (!setObjects(objects_name))
        return false;
    if (!setPoints(point3d))
        return false;
    if (!setdata(additionalData))
        return false;
    if (!findConfigurations())
        return false;
    if (computeMP)
        if(!findTrajectories())
            return false;

    return true;
}

double Task::heuristic(multimap<string,string> agents_name,
                       multimap<string,string> objects_name,
                       multimap<string,p3d_point> point3d,
                       multimap<string,string> additionalData,
                       WorldState* WS, bool computeMP)
{
    WS->load();
    _WS = WS;
    if (!initAll())
    {
        cout << "ERROR: init fails for task " << type << " with id " << task_id << endl;
        return false;
    }
    if (!setAgents(agents_name))
    {
        cout << "ERROR: setAgents fails for task " << type << " with id " << task_id << endl;
        return false;
    }
    if (!setObjects(objects_name))
    {
        cout << "ERROR: setObjects fails for task " << type << " with id " << task_id << endl;
        return false;
    }
    if (!setPoints(point3d))
    {
        cout << "ERROR: setPoints fails for task " << type << " with id " << task_id << endl;
        return false;
    }
    if (!setdata(additionalData))
    {
        cout << "ERROR: setData fails for task " << type << " with id " << task_id << endl;
        return false;
    }
    if (!initialize())
    {
        cout << "ERROR: initialize fails for task " << type << " with id " << task_id << endl;
        return false;
    }


    if (!findCandidateSolutions())
    {
        cout << "ERROR: findCandidateSolutions fails for task " << type << " with id " << task_id << endl;
        return false;
    }

    double res =  computeHeuristicCost();

    if (!finalize())
    {
        cout << "ERROR: initialize fails for task " << type << " with id " << task_id << endl;
        return false;
    }

    return res;
}

double Task::computeHeuristicCost()
{
    return P3D_HUGE;
}

p3d_traj* Task::compute_traj(tr1::shared_ptr<Configuration> q_init, tr1::shared_ptr<Configuration> q_goal, int &status)
{
    if (q_init->getRobot()->getName() != q_goal->getRobot()->getName())
        return false;

    //this should be implemented in the near future
    return false;
}

p3d_traj* Task::compute_traj_using_localPath(tr1::shared_ptr<Configuration> q_init, tr1::shared_ptr<Configuration> q_goal, int& status)
{
    return false;
}

bool Task::computeManipulationConfs(confPtr_t& , confPtr_t& , confPtr_t& , confPtr_t& , p3d_matrix4 , gpGrasp,
                                      p3d_vector3 , double , p3d_vector3 , double )
{
    cout << "wrong fct" << endl;
    return true;
}

bool Task::computeInverseKinematic(Robot* ag, p3d_matrix4 target_pos, confPtr_t &solution, int armId)
{
    cout << "wrong fct" << endl;
    return true;
}

TaskSubSolution* Task::getSubSolAt(int alternativeId, int id)
{
    TaskSolution* ts = result->getAlternativeWithId(alternativeId);
    TaskSubSolution* tSubSol = NULL;
    if (!ts)
        return tSubSol;
    return ts->getSubSolutionWithId(id);
}

int Task::getSubSolNumberOfTrajs(int alternativeId)
{
    TaskSolution* TS = result->getAlternativeWithId(alternativeId);
    if (!TS)
        return 0;

    int nbRes = 0;
    vector<Robot*> vr = TS->getInvolvedRobots();
    for (unsigned int i = 0; i<vr.size(); i++)
    {
        nbRes += TS->getNbsubSolPerRobot(vr.at(i));
    }
    return nbRes;
}

double Task::getSSSizeLeft()
{
    return 0;
}

PrivateTask::PrivateTask() : Task()
{
    mainAgent = NULL;
}

bool PrivateTask::initAll()
{
    return true;
}

bool PrivateTask::initialize()
{
    return true;
}

bool PrivateTask::finalize()
{
    return true;
}

bool PrivateTask::setMainAgent(multimap<string,string> agents_name)
{
    if (agents_name.count("mainAgent") != 1)
    {
        cout << "ERROR: no main agent for " << type << " task." << endl;
        returnRepport = "no_main_agent";
        return false;
    }
    mainAgent = global_Project->getActiveScene()->getRobotByName(agents_name.find("mainAgent")->second);
    if (!mainAgent)
    {
        returnRepport = "no_main_agent_wrong_name";
        return false;
    }
    if (mainAgent->getName() == "ROMEO")
    {
#ifdef HPP
        if (!_romeoWrap)
            _romeoWrap = new RomeoWrap(mainAgent);
        if (!_romeoWrap)
#endif
            return false;
    }
    return true;
}

bool PrivateTask::isPossibleMainAgent(Robot* r)
{
    bool ok = false;
    if(r->getHriAgent()){
        if(_possibleMainAgentType.end() != std::find(_possibleMainAgentType.begin(), _possibleMainAgentType.end(), r->getHriAgent()->type)){
            ok=true;
        }else{
            returnRepport = "agent_not_in_list";
        }
    }else{
        returnRepport = "agent_has_no_hri_struct";
    }
    return ok;
}

void PrivateTask::openGripper(confPtr_t &q, bool open, int armId)
{
    if (q->getRobot()->getHriAgent() && q->getRobot()->getHriAgent()->type == HRI_PR2)
    {
        double v = 0.02;
        if (open)
            v = 0.5;

        if (armId == 0)
        {
            (*q)[23] = v;
            (*q)[24] = v;
        }
        else
        {
            (*q)[32] = v;
            (*q)[33] = v;
        }
    }
    else if (q->getRobot()->getHriAgent() && q->getRobot()->getHriAgent()->type == HRI_KUKAROBOT)
    {
        double v = 0.02;
        if (open)
            v = 0.5;

        (*q)[19] = v;
        (*q)[20] = v;

    }
}

bool PrivateTask::testConstraints(int alternativeId)
{
    TaskSolution* TS = result->getAlternativeWithId(alternativeId);
    _returnRepportcount++;
    if (!TS)
    {
        cout << "test of constraint fail, no solution with id " << alternativeId << endl;
        if (_returnRepportcountMax < _returnRepportcount)
        {
            _returnRepportcountMax = _returnRepportcount;
            returnRepport = "alternative_creation_fail";
        }
        return false;
    }
    _returnRepportcount++;
    if (TS->getTaskConstraints()->areFactConstraintRespected(TS->getStopWS()))
        return true;
    else
    {
        if (_returnRepportcountMax < _returnRepportcount)
        {
            _returnRepportcountMax = _returnRepportcount;
            returnRepport = "constraints_not_respected";
        }
    }

    result->removeAlternativeWithId(alternativeId);
    return false;
}

bool PrivateTask::run(multimap<string,string> agents_name,
                           multimap<string,string> objects_name,
                           multimap<string,p3d_point> point3d,
                           multimap<string,string> additionalData,
                           WorldState *WS, bool computeMP)
{

    WS->load();
    _WS = WS;
    if (!initAll())
    {
        cout << "ERROR: init fails for task " << type << " with id " << task_id << endl;
        return false;
    }
    if (!setAgents(agents_name))
    {
        cout << "ERROR: setAgents fails for task " << type << " with id " << task_id << endl;
        return false;
    }
    if (!isPossibleMainAgent(mainAgent))
    {
        cout << "ERROR: mainAgent is not in the list of possible agent for task " << type << " with id " << task_id << endl;
        return false;
    }
    if (!setObjects(objects_name))
    {
        cout << "ERROR: setObjects fails for task " << type << " with id " << task_id << endl;
        return false;
    }
    if (!setPoints(point3d))
    {
        cout << "ERROR: setPoints fails for task " << type << " with id " << task_id << endl;
        return false;
    }
    if (!setdata(additionalData))
    {
        cout << "ERROR: setData fails for task " << type << " with id " << task_id << endl;
        return false;
    }
    if (!initialize())
    {
        cout << "ERROR: initialize fails for task " << type << " with id " << task_id << endl;
        return false;
    }


    if (!findCandidateSolutions())
    {
        cout << "ERROR: findCandidateSolutions fails for task " << type << " with id " << task_id << endl;
        return false;
    }
    bool res = false;
    int count = 0;
    while (!res)
    {
        cout << "count loop = " << count << endl;
        _returnRepportcount = 0;
        if(count++ > nbMaxLoops)
            break;
        if(!getRandomSol())
           continue;
        if(!findConfigurations())
            continue;
        if (computeMP)
            if(!findTrajectories())
                continue;
        if(!setToSolution(0,computeMP))
            continue;
        if(!testConstraints(0))
            continue;

        res = true;
    }

    if (res)
        returnRepport = "OK";

    computeCost(0);
    if (!finalize())
    {
        cout << "ERROR: finalize fails for task " << type << " with id " << task_id << endl;
        return false;
    }
    return res;
}


TaskSolution* PrivateTask::findAlternative(bool computeMP)
{
    int count = 0;
    if (!getResult()->getAlternativeWithId(0))
    {
        cout<< "This task has no first solution, no alternative can be computed in that case" << endl;
        return NULL;
    }
    getResult()->getAlternativeWithId(0)->getStartWS()->load(); //
    int id = -1;
    bool res = false;


    if (!initialize())
    {
        cout << "ERROR: initialize fails for task " << type << " with id " << task_id << endl;
        return false;
    }

    while (!res)
    {
        cout << "count loop = " << count << endl;
        _returnRepportcount = 0;
        if(count++ > nbMaxLoops)
            break;
        if(!getRandomSol())
           continue;
        if(!findConfigurations())
           continue;
        if (computeMP)
            if(!findTrajectories())
                continue;
        id = result->getNextId();

        if(!setToSolution(id,computeMP))
           continue;
        if(!testConstraints(id))
            continue;
        res = true;
    }

    if (res)
        returnRepport = "OK";

    computeCost(id);

    if (!finalize())
    {
        cout << "ERROR: finalize fails for task " << type << " with id " << task_id << endl;
        return false;
    }

    if (res && id != -1)
        return getResult()->getAlternativeWithId(id);
    return NULL;
}


armLabel PrivateTask::getArmLabelFromId(int armId)
{
    if (armId == 1)
        return Left;
    else
        return Right;
}

bool PrivateTask::setArmToUse(armLabel arm_t)
{
    m_arm = arm_t;
    if (!mainAgent)
        return false;
    //define, from the arm string and the type of main robot the hand type:
    // need to be changed!!!!!!!!!!!!!! should be stocked in agent.h

    bool t = false;
    if (mainAgent->getHriAgent()->type == HRI_PR2 ||
           mainAgent->getHriAgent()->type == HRI_KUKAROBOT )
    {

        t =true;
        handType = GP_PR2_GRIPPER;
        if (ManipPlMap.count(mainAgent->getName()) > 0)
        {
            if (arm_t == Left)
                fixAllJointsWithoutArm(ManipPlMap[mainAgent->getName()]->robot(),1);
            else
                fixAllJointsWithoutArm(ManipPlMap[mainAgent->getName()]->robot(),0);
        }
    }
    else if (mainAgent->getName().find("HUMAN") !=std::string::npos)
    {
        t =true;
        if (arm_t == Left)
            handType = GP_SAHAND_LEFT;
        else
            handType = GP_SAHAND_RIGHT; //by default right hand
    }
    else if (mainAgent->getName().find("ROMEO") !=std::string::npos)
    {
        t = true;
        if (arm_t == Left)
            handType = GP_ROMEO_LEFT_HAND;
        else
            handType = GP_ROMEO_RIGHT_HAND; //by default right hand
    }

    if (t)
    {
        for (int i = 0; i < mainAgent->getHriAgent()->manip->external_hand_no; i++)
        {
            if (!mainAgent->getHriAgent()->manip->external_hand)
                return false;
            if (!mainAgent->getHriAgent()->manip->external_hand[i])
                return false;
            if (mainAgent->getHriAgent()->manip->external_hand[i]->hand_l == m_arm)
            {
                _armId = mainAgent->getHriAgent()->manip->external_hand[i]->hand_id;
                break;
            }
        }
    }
//    cout << "armid = " << _armId << endl;
    return t;
}

int PrivateTask::getArmId()
{
    return _armId;
//    if (m_arm == Left)
//        return 1;
//    else if (m_arm == Right)
//        return 0;
//    else
//    {
//        cout << "arm not defined as it should be, by default choosing id 0, string defined = " << m_arm << endl;
//        return 0;
//    }
}

bool PrivateTask::initMotionPlaner(Robot* rob)
{
//    cout << "init" << endl;
//    if (!rob)
//        return false;
//    if(!rob->isAgent())
//        return false;
//    if (ManipPlMap[rob->getName()])
//        return true;

//    ManipPlMap[rob->getName()] = new ManipulationPlanner( rob->getRobotStruct() );
//    if (!ManipPlMap[rob->getName()])
//        return false;

////    ManipulationPlanner* mp = new ManipulationPlanner( mainAgent->getRobotStruct() );


//    ManipPlMap[rob->getName()]->setPlanningMethod( p3d_planner_function );
//    ManipPlMap[rob->getName()]->setSmoothingMethod( p3d_smoothing_function_dummy );
//    ManipPlMap[rob->getName()]->setCleanningRoadmaps( false );
//    //ManipPlMap[rob->getName()]->setArmCartesian( getArmId(), true );
//    setPlanningTimeLimit(60);


//    if (_armId > 1 || _armId < 0)
//    {
//        cout << "wrong _armId param, " << _armId << " going with right arm" <<endl;
//        _armId = 0;
//    }
//    fixAllJointsWithoutArm(ManipPlMap[rob->getName()]->robot(),_armId);

//    p3d_multiLocalPath_disable_all_groupToPlan(rob->getRobotStruct(), false);
//    p3d_multiLocalPath_set_groupToPlan(rob->getRobotStruct(), ManipPlMap[rob->getName()]->getUpBodyMLP(), 1, false);

//    return true;
    if (rob->getHriAgent() && rob->getHriAgent()->type == HRI_PR2)
    {
        if (_armId > 1 || _armId < 0)
        {
            cout << "wrong _armId param, " << _armId << " going with right arm" <<endl;
            _armId = 0;
        }

        p3d_cntrt* ct;
        // over all constraints
        for(int i=0; i<rob->getRobotStruct()->cntrt_manager->ncntrts; i++)
        {
            string name = rob->getRobotStruct()->cntrt_manager->cntrts[i]->namecntrt;

            if( name == "p3d_fix_jnts_relpos" ){
                continue;
            }
            // get constraint from the cntrts manager
            ct = rob->getRobotStruct()->cntrt_manager->cntrts[i];
            p3d_desactivateCntrt( rob->getRobotStruct(), ct );
        }

        p3d_multiLocalPath_disable_all_groupToPlan( rob->getRobotStruct() , false );
        p3d_multiLocalPath_set_groupToPlan( rob->getRobotStruct(), 2, 1, false);


        //Fix every joint except the active arm(s)
        for (int i = 0; i < rob->getRobotStruct()->njoints + 1; i++) {
            p3d_jnt * joint = rob->getRobotStruct()->joints[i];
            if (joint->type != P3D_BASE && joint->type != P3D_FIXED) {
                fixJoint(rob->getRobotStruct(), joint, joint->jnt_mat);
            }
        }


        ArmManipulationData& mData = (*rob->getRobotStruct()->armManipulationData)[_armId];
        ct = mData.getCcCntrt();
        for(int i = 0; i < ct->npasjnts; i++){
            p3d_jnt * joint = ct->pasjnts[i];
            if(joint->type != P3D_FIXED){
                unFixJoint(rob->getRobotStruct(), joint);
            }
        }
        if((!strcmp(ct->namecntrt, "p3d_kuka_arm_ik")) ||
                (!strcmp(ct->namecntrt, "p3d_lwr_arm_ik")) ||
                (!strcmp(ct->namecntrt, "p3d_pr2_arm_ik"))){
            p3d_jnt * joint = rob->getRobotStruct()->joints[ct->argu_i[0]];
            unFixJoint(rob->getRobotStruct(), joint);
        }




        desactivateTwoJointsFixCntrt(rob->getRobotStruct(),mData.getManipulationJnt(), mData.getCcCntrt()->pasjnts[ mData.getCcCntrt()->npasjnts-1 ]);

        activateCcCntrts(rob->getRobotStruct(), _armId, false);
        ManipulationUtils::unfixManipulationJoints(rob->getRobotStruct(), _armId);
        shootTheObjectArroundTheBase(rob->getRobotStruct(), rob->getRobotStruct()->baseJnt, mData.getManipulationJnt(), 2.0);
        mData.getManipulationJnt()->dist  = 0.10;
    }


    return true;

}

p3d_traj* PrivateTask::compute_traj(tr1::shared_ptr<Configuration> q_init, tr1::shared_ptr<Configuration> q_goal, Robot* object, int& status)
{

    clock_t begin = clock();
//    p3d_plan();


    nbMPCalled++;
    if (q_init->getRobot()->getName() != q_goal->getRobot()->getName())
        return NULL;

    if (!q_init->getRobot()->getHriAgent())
    {
        returnRepport = "not_an_agent";
        status = 5;
        return NULL;
    }

    int part = PLAN_FULL;
    switch (q_init->getRobot()->getHriAgent()->type)
    {
        case HRI_ROMEO:
        {
#ifdef HPP
            double *timer;
            p3d_traj *outputTraj;

            bool result = false;
            if (PlanEnv->getBool(PlanParam::useGraph))
            {

            }
            else
            {
                result = _romeoWrap->findTrajectory(q_init, q_goal, outputTraj, timer);
            }
            std::cout << "Path planning : " << (result ? "SUCCESS" : "FAIL") << ", in " << timer << "s" << std::endl;
            if (!result)
            {
                returnRepport = "no_path_found";
                return NULL;
            }
            timeMp += (double)(clock() - begin)/CLOCKS_PER_SEC;
            return outputTraj;
#else
            return NULL;
#endif
        }
        case HRI_PR2:
        {
            for (int i = 0; i < q_init->getRobot()->getRobotStruct()->nb_dof; i++)
            {
    //            cout << " for " << i << " diff = " << fabs(q_init->at(i) - q_goal->at(i)) << endl;
                if (fabs(q_init->at(i) - q_goal->at(i)) > EPS3)
                {
                    if (5<i && i<12)
                    {
                        part = PLAN_BASE;
                        i = 12;
                    }

                    else if (14<i && i<23)
                    {
                        if (part == PLAN_BASE)
                        {
                            part=PLAN_FULL;
                            break;
                        }
                        part=PLAN_RIGHT_ARM;
                        i=23;
                    }

                    else if (24<i && i<32)
                    {
                        if (part == PLAN_BASE || part == PLAN_RIGHT_ARM)
                        {
                            part=PLAN_FULL;
                            break;
                        }
                        part=PLAN_LEFT_ARM;
                        i=32;
                    }
                }
            }

            mainAgent->setAndUpdate(*q_init);

            double Tx, Ty, Tz;
            double Rx, Ry, Rz;
            p3d_mat4ExtractPosReverseOrder(p3d_get_robot_jnt_by_name(q_init->getRobot()->getRobotStruct(), "rIkJoint")->abs_pos, &Tx, &Ty, &Tz, &Rx, &Ry, &Rz);

            q_init->getConfigStruct()[37] = Tx;
            q_init->getConfigStruct()[38] = Ty;
            q_init->getConfigStruct()[39] = Tz;
            q_init->getConfigStruct()[40] = Rx;
            q_init->getConfigStruct()[41] = Ry;
            q_init->getConfigStruct()[42] = Rz;

            p3d_mat4ExtractPosReverseOrder(p3d_get_robot_jnt_by_name(q_init->getRobot()->getRobotStruct(), "lIkJoint")->abs_pos, &Tx, &Ty, &Tz, &Rx, &Ry, &Rz);

            q_init->getConfigStruct()[43] = Tx;
            q_init->getConfigStruct()[44] = Ty;
            q_init->getConfigStruct()[45] = Tz;
            q_init->getConfigStruct()[46] = Rx;
            q_init->getConfigStruct()[47] = Ry;
            q_init->getConfigStruct()[48] = Rz;



            int ress = -1;
            p3d_traj* t = NULL;
            bool testOn = true;
            if (PlanEnv->getBool(PlanParam::useGraph))
            {
                testOn = false;
                PlanNode* PN = _planTree->getClosestNode(_WS,currentDataType);
                if (!PN)
                {
                    testOn = true;
                }
                else
                {

                }
            }
            if (testOn)
            {
                confPtr_t q_cur = q_init->getRobot()->getCurrentPos();
                ress = p3d_plan(q_init->getRobot()->getRobotStruct(),q_init->getConfigStruct(), q_goal->getConfigStruct(), part, ENV.getBool(Env::useCartesian));
                mainAgent->setAndUpdate(*q_cur);

                if (ress == PLAN_STATUS_FOUND)
                    t = p3d_create_traj_by_copy(q_init->getRobot()->getRobotStruct()->tcur);
            }

            status = ress;
            timeMp += (double)(clock() - begin)/CLOCKS_PER_SEC;
            return t;
        }
    case HRI_KUKAROBOT:
    {
        for (unsigned int i = 0; i< q_init->getRobot()->getJoint(0)->getNumberOfDof();i++)
        {
            q_init->getRobot()->getJoint(0)->getJointStruct()->dof_data[i].is_active_for_planner = 0;
        }
        double Tx, Ty, Tz;
        double Rx, Ry, Rz;
        p3d_mat4ExtractPosReverseOrder(p3d_get_robot_jnt_by_name(q_init->getRobot()->getRobotStruct(), "IkJoint")->abs_pos, &Tx, &Ty, &Tz, &Rx, &Ry, &Rz);

        q_init->getConfigStruct()[21] = Tx;
        q_init->getConfigStruct()[22] = Ty;
        q_init->getConfigStruct()[23] = Tz;
        q_init->getConfigStruct()[24] = Rx;
        q_init->getConfigStruct()[25] = Ry;
        q_init->getConfigStruct()[26] = Rz;


        int ress = -1;
        p3d_traj* t = NULL;

        confPtr_t q_cur = q_init->getRobot()->getCurrentPos();

//        q_init->print(true);



        ress = p3d_plan(q_init->getRobot()->getRobotStruct(),q_init->getConfigStruct(), q_goal->getConfigStruct(), part, ENV.getBool(Env::useCartesian));
        mainAgent->setAndUpdate(*q_cur);

        if (ress == PLAN_STATUS_FOUND)
            t = p3d_create_traj_by_copy(q_init->getRobot()->getRobotStruct()->tcur);

        status = ress;
        timeMp += (double)(clock() - begin)/CLOCKS_PER_SEC;
        return t;
    }
        default:
            return NULL;
    }
}

p3d_traj* PrivateTask::compute_traj_using_localPath(confPtr_t q_init, confPtr_t q_goal, Robot* object, int& status)
{
    q_init->getRobot()->setAndUpdate(*q_init);

    if (q_init->getRobot()->getHriAgent() && q_init->getRobot()->getHriAgent()->type == HRI_PR2 )
    {


        for(int i=0; i<q_init->getRobot()->getRobotStruct()->cntrt_manager->ncntrts; i++)
        {
            string name = q_init->getRobot()->getRobotStruct()->cntrt_manager->cntrts[i]->namecntrt;

            if( name == "p3d_fix_jnts_relpos" ){
                continue;
            }

            cout << "deactivate : " << name << endl;

            // get constraint from the cntrts manager
            p3d_cntrt* ct = q_init->getRobot()->getRobotStruct()->cntrt_manager->cntrts[i];
            p3d_desactivateCntrt( q_init->getRobot()->getRobotStruct(), ct );
        }


//        cout << "arm to use = " << _armId << endl;
        ArmManipulationData& mData = (*q_init->getRobot()->getRobotStruct()->armManipulationData)[_armId];
        p3d_cntrt* ct = mData.getCcCntrt();
        for(int i = 0; i < ct->npasjnts; i++){
            p3d_jnt * joint = ct->pasjnts[i];
            if(joint->type != P3D_FIXED){
                unFixJoint(q_init->getRobot()->getRobotStruct(), joint);
            }
        }



        if((!strcmp(ct->namecntrt, "p3d_kuka_arm_ik")) ||
            (!strcmp(ct->namecntrt, "p3d_lwr_arm_ik")) ||
            (!strcmp(ct->namecntrt, "p3d_pr2_arm_ik"))){
            p3d_jnt * joint = q_init->getRobot()->getRobotStruct()->joints[ct->argu_i[0]];
            unFixJoint(q_init->getRobot()->getRobotStruct(), joint);
        }
//        mData.setCartesian( true );
        mData.setCartesian( false );


        ArmManipulationData& mDataOther = (*q_init->getRobot()->getRobotStruct()->armManipulationData)[(_armId+1)%2];
        mDataOther.setCartesian( false );

        for (int i=0; i < int((*q_init->getRobot()->getRobotStruct()->armManipulationData).size()); i++)
        {
            ArmManipulationData& armData  = (*q_init->getRobot()->getRobotStruct()->armManipulationData)[i];

            desactivateTwoJointsFixCntrt(q_init->getRobot()->getRobotStruct(),armData.getManipulationJnt(), armData.getCcCntrt()->pasjnts[ armData.getCcCntrt()->npasjnts-1 ]);


            if (armData.getCartesian())
            {
                cout << "Arm Data " << i <<  " is set cartesian" << endl;
                activateCcCntrts(q_init->getRobot()->getRobotStruct(), i, false);
                ManipulationUtils::unfixManipulationJoints(q_init->getRobot()->getRobotStruct(), i);
                shootTheObjectArroundTheBase(q_init->getRobot()->getRobotStruct(), q_init->getRobot()->getRobotStruct()->baseJnt, armData.getManipulationJnt(), 2.0);
                armData.getManipulationJnt()->dist  = 0.10;

            } else {
                deactivateCcCntrts(q_init->getRobot()->getRobotStruct(), i);
                p3d_set_and_update_this_robot_conf(q_init->getRobot()->getRobotStruct(), q_init->getConfigStruct());
                setAndActivateTwoJointsFixCntrt(q_init->getRobot()->getRobotStruct(),armData.getManipulationJnt(), armData.getCcCntrt()->pasjnts[ armData.getCcCntrt()->npasjnts-1 ]);
                ManipulationUtils::fixManipulationJoints(q_init->getRobot()->getRobotStruct(), i, q_init->getConfigStruct(), NULL);
                shootTheObjectInTheWorld(q_init->getRobot()->getRobotStruct(), armData.getManipulationJnt());
            }
        }



        double Tx, Ty, Tz;
        double Rx, Ry, Rz;
        p3d_mat4ExtractPosReverseOrder(p3d_get_robot_jnt_by_name(q_init->getRobot()->getRobotStruct(), "rIkJoint")->abs_pos, &Tx, &Ty, &Tz, &Rx, &Ry, &Rz);

        q_init->getConfigStruct()[37] = Tx;
        q_init->getConfigStruct()[38] = Ty;
        q_init->getConfigStruct()[39] = Tz;
        q_init->getConfigStruct()[40] = Rx;
        q_init->getConfigStruct()[41] = Ry;
        q_init->getConfigStruct()[42] = Rz;

        p3d_mat4ExtractPosReverseOrder(p3d_get_robot_jnt_by_name(q_init->getRobot()->getRobotStruct(), "lIkJoint")->abs_pos, &Tx, &Ty, &Tz, &Rx, &Ry, &Rz);

        q_init->getConfigStruct()[43] = Tx;
        q_init->getConfigStruct()[44] = Ty;
        q_init->getConfigStruct()[45] = Tz;
        q_init->getConfigStruct()[46] = Rx;
        q_init->getConfigStruct()[47] = Ry;
        q_init->getConfigStruct()[48] = Rz;



    }
    else if (q_init->getRobot()->getHriAgent() && q_init->getRobot()->getHriAgent()->type == HRI_KUKAROBOT )
    {
        p3d_cntrt* ct = mainAgent->getRobotStruct()->cntrt_manager->cntrts[0];
        p3d_desactivateCntrt(mainAgent->getRobotStruct(), ct);


        double Tx, Ty, Tz;
        double Rx, Ry, Rz;
        p3d_mat4ExtractPosReverseOrder(p3d_get_robot_jnt_by_name(q_init->getRobot()->getRobotStruct(), "IkJoint")->abs_pos, &Tx, &Ty, &Tz, &Rx, &Ry, &Rz);

        q_init->getConfigStruct()[21] = Tx;
        q_init->getConfigStruct()[22] = Ty;
        q_init->getConfigStruct()[23] = Tz;
        q_init->getConfigStruct()[24] = Rx;
        q_init->getConfigStruct()[25] = Ry;
        q_init->getConfigStruct()[26] = Rz;
    }





    confPtr_t q_curr = q_init->getRobot()->getCurrentPos();
    if (q_init->getRobot()->getName() != q_goal->getRobot()->getName())
        return NULL;

    if (q_init->equal(*q_goal))
    {
        status = PLAN_STATUS_START_EQUALS_GOAL;
        return NULL;
    }


    p3d_localpath *lp = p3d_local_planner(q_init->getRobot()->getRobotStruct(), q_init->getConfigStruct(), q_goal->getConfigStruct());
    if (!lp)
        return NULL;

    p3d_traj *traj = p3d_create_empty_trajectory(q_init->getRobot()->getRobotStruct());
    traj->courbePt = lp;
    traj->nlp = 1;
    traj->range_param = lp->range_param;
    double d = 0;
    if (object)
        activateFingertipCol(false,q_init->getRobot(),object,NULL);
//        (q_init->getRobot()->getRobotStruct(), object->getRobotStruct());

    while (d < traj->range_param)
    {
        d += traj->range_param / 100;
        configPt q =p3d_config_at_param_along_traj(traj,d);
        p3d_set_and_update_this_robot_conf(q_init->getRobot()->getRobotStruct(),q);


        bool testCol = p3d_col_test_robot(q_init->getRobot()->getRobotStruct(),0) == 1;
        if (q_init->getRobot()->getHriAgent() &&q_init->getRobot()->getHriAgent()->is_human == 1)
            testCol = pqp_robot_all_no_self_collision_test(q_init->getRobot()->getRobotStruct());

        if (testCol) //col
        {
            p3d_print_col_pair();
//            g3d_draw_allwin_active();
//            sleep(3);
            traj = NULL;
            status = PLAN_STATUS_NOT_FOUND;
            break;
        }
    }
    if (object)
        activateFingertipCol(true,q_init->getRobot(),object,NULL);
//        p3d_col_activate_robot_robot(q_init->getRobot()->getRobotStruct(), object->getRobotStruct());

    q_init->getRobot()->setAndUpdate(*q_curr);

    if (q_init->getRobot()->getHriAgent() && q_init->getRobot()->getHriAgent()->type == HRI_PR2)
    {
        ArmManipulationData& mData = (*q_init->getRobot()->getRobotStruct()->armManipulationData)[0];
        mData.setCartesian( false );
        desactivateTwoJointsFixCntrt(q_init->getRobot()->getRobotStruct(),mData.getManipulationJnt(), mData.getCcCntrt()->pasjnts[ mData.getCcCntrt()->npasjnts-1 ]);
        deactivateCcCntrts(q_init->getRobot()->getRobotStruct(), 0);
        p3d_set_and_update_this_robot_conf(q_init->getRobot()->getRobotStruct(), q_init->getConfigStruct());
        setAndActivateTwoJointsFixCntrt(q_init->getRobot()->getRobotStruct(),mData.getManipulationJnt(), mData.getCcCntrt()->pasjnts[ mData.getCcCntrt()->npasjnts-1 ]);
        ManipulationUtils::fixManipulationJoints(q_init->getRobot()->getRobotStruct(), 0, q_init->getConfigStruct(), NULL);
        shootTheObjectInTheWorld(q_init->getRobot()->getRobotStruct(), mData.getManipulationJnt());


        ArmManipulationData&  mData1 = (*q_init->getRobot()->getRobotStruct()->armManipulationData)[1];
        mData1.setCartesian( false );
        desactivateTwoJointsFixCntrt(q_init->getRobot()->getRobotStruct(),mData1.getManipulationJnt(), mData1.getCcCntrt()->pasjnts[ mData1.getCcCntrt()->npasjnts-1 ]);
        deactivateCcCntrts(q_init->getRobot()->getRobotStruct(), 1);
        p3d_set_and_update_this_robot_conf(q_init->getRobot()->getRobotStruct(), q_init->getConfigStruct());
        setAndActivateTwoJointsFixCntrt(q_init->getRobot()->getRobotStruct(),mData1.getManipulationJnt(), mData1.getCcCntrt()->pasjnts[ mData1.getCcCntrt()->npasjnts-1 ]);
        ManipulationUtils::fixManipulationJoints(q_init->getRobot()->getRobotStruct(), 1, q_init->getConfigStruct(), NULL);
        shootTheObjectInTheWorld(q_init->getRobot()->getRobotStruct(), mData1.getManipulationJnt());


    }

    if (traj)
        status = PLAN_STATUS_FOUND;
    return traj;
}

double PrivateTask::computeCost(int altId)
{
    //get the taskSolution
    //get subtrajs
    //get beg and end confs
    //compute dist
    //add dists
    TaskSolution* TS = result->getAlternativeWithId(altId);
    if (!TS)
        return 1000;
    vector<TaskSubSolution*> vtss = TS->getTSSByRobot(mainAgent);
    double cost = 0;
    GeometricComputations GC = GeometricComputations();
    for (unsigned int i = 0; i < vtss.size(); i++)
    {
        TaskSubSolution* TSS = vtss.at(i);
        tr1::shared_ptr<Configuration> q1 = TSS->getStartState()->getRobotConf(mainAgent);
        tr1::shared_ptr<Configuration> q2 = TSS->getEndState()->getRobotConf(mainAgent);

        cost += GC.distConfs(q1,q2);
//        cost += p3d_dist_config(mainAgent->getRobotStruct(), q1->getConfigStruct(), q2->getConfigStruct(), true);
//        q1->

    }
    TS->setCost(cost);

    return cost;
}

bool PrivateTask::attachObject(Robot* rob, Robot* obj, bool attach)
{
    return attachObject(rob,obj,attach,getArmId());
}

bool PrivateTask::attachObject(Robot* rob, Robot* obj, bool attach, int  armId)
{
    if (!rob || ! obj)
        return false;
    int jointId = -1;
    if (attach)
    {
        if (!rob->isAgent())
            return false;
        if (!rob->getHriAgent())
            return false;
        if (!rob->getHriAgent()->manip)
            return false;
        if (!rob->getHriAgent()->manip->external_hand[armId])
            return false;
        jointId = rob->getHriAgent()->manip->external_hand[armId]->joint_id_in_rob;

        if (!rob->isObjectAttached(obj,jointId))
        {
            rob->attachObj(obj,jointId);
        }
    }
    else
        rob->dettachObj(obj);

    return rob->whereObjectAttached(obj);
    //    int r = hri_agent_is_grasping_obj_rob_pt(robot, attach , c , armId);

    //    return !r;
}


bool PrivateTask::computeManipulationConfs(confPtr_t& approachC, confPtr_t& graspC, confPtr_t& graspOC, confPtr_t& escapeC, p3d_matrix4 objectPosMat,
                                           gpGrasp grasp, p3d_vector3 approachDir, double approachDist, p3d_vector3 escapeDir, double escapeDist)
{
    shared_ptr<Configuration> qInit = _WS->getRobotConf(mainAgent);

    confPtr_t q;

    p3d_matrix4 graspMatWorldFrame, approachMat, escapeMat, /*rotMat,*/ reelTarg;
    p3d_mat4Mult(objectPosMat, grasp.frame, graspMatWorldFrame);

    p3d_mat4Copy(graspMatWorldFrame, reelTarg);

    p3d_vector3 v_new;

    p3d_matrix3 matRot;
    p3d_mat4ExtractRotMatrix(objectPosMat, matRot);
    p3d_xform3(matRot, approachDir,  v_new);

    p3d_mat4Copy(reelTarg, approachMat);
    approachMat[0][3] = reelTarg[0][3] - v_new[0]*approachDist;
    approachMat[1][3] = reelTarg[1][3] - v_new[1]*approachDist;
    approachMat[2][3] = reelTarg[2][3] - v_new[2]*approachDist;

//    cout << "x " << v_new[0] << " y " << v_new[1] << " z " << v_new[2] << endl;
//    cout << "approach dist = " << approachDist << endl ;

    p3d_mat4Copy(reelTarg, escapeMat);
    escapeMat[0][3] = reelTarg[0][3] + escapeDir[0]*escapeDist;
    escapeMat[1][3] = reelTarg[1][3] + escapeDir[1]*escapeDist;
    escapeMat[2][3] = reelTarg[2][3] + escapeDir[2]*escapeDist;


    bool testIk = computeInverseKinematic(mainAgent, reelTarg, q, _armId );
    nbGikCalled++;

    _returnRepportcount++;
    if (!testIk)
    {
        if (_returnRepportcountMax < _returnRepportcount)
        {
            _returnRepportcountMax = _returnRepportcount;
            returnRepport = "no_IK_for_grasp";
        }
        return false;
    }

    openGripper(q,true,_armId);

    graspC = graspOC = shared_ptr<Configuration>(new Configuration(*q));

    mainAgent->setAndUpdate(*graspC);



    mainAgent->setAndUpdate(*graspC);
    testIk = computeInverseKinematic(mainAgent, approachMat  ,q , _armId );
    nbGikCalled++;

    _returnRepportcount++;
    if (!testIk)
    {
        if (_returnRepportcountMax < _returnRepportcount)
        {
            _returnRepportcountMax = _returnRepportcount;
            returnRepport = "no_IK_for_hand_free";
        }
        return false;
    }

    openGripper(q,true,_armId);
    approachC = shared_ptr<Configuration>(new Configuration(*q));

    //    mainAgent->setAndUpdate(*approachConf);

    mainAgent->setAndUpdate(*graspC);
    testIk = computeInverseKinematic(mainAgent, escapeMat  ,q , _armId );
    nbGikCalled++;

    _returnRepportcount++;
    if (!testIk)
    {
        if (_returnRepportcountMax < _returnRepportcount)
        {
            _returnRepportcountMax = _returnRepportcount;
            returnRepport = "no_IK_for_obj_in_hand";
        }
        return false;
    }

    openGripper(q,true,_armId);
    escapeC = shared_ptr<Configuration>(new Configuration(*q));

    mainAgent->setAndUpdate(*approachC);
//    if (true)
//    {
//        //BIIIG HAAAAACKKKK // for virtual object in robot hand
//        (*qInit)[37] = (*approachC)[37];
//        (*qInit)[38] = (*approachC)[38];
////        (*qInit)[39] = (*approachC)[39];
//        (*qInit)[43] = (*approachC)[43];
//        (*qInit)[44] = (*approachC)[44];
////        (*qInit)[45] = (*approachC)[45];
//    }
    _WS->replaceConf(qInit);

    mainAgent->setAndUpdate(*qInit);
    return true;
}


bool PrivateTask::computeInverseKinematic(Robot* ag, p3d_matrix4 target_pos, confPtr_t &solution, int armId)
{
    if (!ag->getHriAgent())
    {
        returnRepport = "not_an_agent";
        return false;
    }

    p3d_matrix4 rotMat, tmp;
    rotMat[0][0] = 0;
    rotMat[0][1] = 1;
    rotMat[0][2] = 0;
    rotMat[0][3] = 0;

    rotMat[1][0] = 0;
    rotMat[1][1] = 0;
    rotMat[1][2] = 1;
    rotMat[1][3] = 0;

    rotMat[2][0] = 1;
    rotMat[2][1] = 0;
    rotMat[2][2] = 0;
    rotMat[2][3] = 0;

    rotMat[3][0] = 0;
    rotMat[3][1] = 0;
    rotMat[3][2] = 0;
    rotMat[3][3] = 1;

    p3d_mat4Mult(target_pos, rotMat, tmp);


//    p3d_mat4Print(target_pos,"target_pos");

    switch(ag->getHriAgent()->type)
    {
        bool found, i_can;
        int nbIKloop;
        case HRI_ROMEO:
        {
            p3d_mat4Copy(tmp, target_pos);
#ifdef HPP
            bool randomize = true;
            if (!getResult()->getAlternativeWithId(0))
                randomize = false;
            found = i_can = false;
            nbIKloop = 0;

            confPtr_t outputConf;
            confPtr_t inputConf = shared_ptr<Configuration>(new Configuration(*ag->getCurrentPos()));
            i_can = _romeoWrap->generateConfig(inputConf, outputConf, target_pos, armId, randomize);
            if (i_can)
            {
                // std::cout << "RV -- Stability of generated config : "
                //           << (_romeoWrap->isPositionStable(outputConf) ? "Stable" : "Not stable")
                //           << std::endl;
                if (p3d_col_test_robot(ag->getRobotStruct(),0) == 0)
                    found = true;
                else
                    i_can = false;
            }

            if (!i_can)
            {
                while (nbIKloop < MAXIKLOOP)
                {
                    nbIKloop++;
                    i_can = _romeoWrap->generateConfig(inputConf, outputConf, target_pos, armId, true);
                    if (i_can)
                    {
                        // std::cout << "RV -- Stability of generated config : "
                        //           << (_romeoWrap->isPositionStable(outputConf) ? "Stable" : "Not stable")
                        //           << std::endl;
                        if (p3d_col_test_robot(ag->getRobotStruct(),0) == 0)
                        {
                            found = true;
                            break;
                        }
                    }
                }
            }

            if (found)
            {
    //            q = ag->getCurrentPos()->getConfigStruct();
                solution = shared_ptr<Configuration>(new Configuration(*outputConf));
                return true;
            }
#endif
            break;
        }
        case HRI_PR2:
        {
            p3d_mat4Copy(tmp, target_pos);
            found = i_can = false;
            double Tx, Ty, Tz;
            double Rx, Ry, Rz;
            nbIKloop = 0;
            p3d_mat4ExtractPosReverseOrder(target_pos, &Tx, &Ty, &Tz,
                                           &Rx, &Ry, &Rz);

            configPt q = p3d_copy_config(ag->getRobotStruct(), _WS->getRobotConf(ag)->getConfigStruct());
            int ffjntIndex = (*ag->getRobotStruct()->armManipulationData)[armId].getManipulationJnt()->index_dof;

            q[ffjntIndex + 0] = Tx;
            q[ffjntIndex + 1] = Ty;
            q[ffjntIndex + 2] = Tz;
            q[ffjntIndex + 3] = Rx;
            q[ffjntIndex + 4] = Ry;
            q[ffjntIndex + 5] = Rz;

    //        print_config(ag->getRobotStruct(), q);

            int *ikSol = NULL;
            p3d_set_and_update_this_robot_conf( ag->getRobotStruct(), q );

            p3d_cntrt* ct;
            // over all constraints
            for (int i=0; i<ag->getRobotStruct()->cntrt_manager->ncntrts; i++)
            {
                string name = ag->getRobotStruct()->cntrt_manager->cntrts[i]->namecntrt;

                if ( name == "p3d_fix_jnts_relpos" )
                    continue;
                // get constraint from the cntrts manager
                ct = ag->getRobotStruct()->cntrt_manager->cntrts[i];
                p3d_desactivateCntrt( ag->getRobotStruct(), ct );
            }

            p3d_multiLocalPath_disable_all_groupToPlan( ag->getRobotStruct() , false );
            p3d_multiLocalPath_set_groupToPlan( ag->getRobotStruct(), 2, 1, false);


            //Fix every joint except the active arm(s)
            for (int i = 0; i < ag->getRobotStruct()->njoints + 1; i++)
            {
                p3d_jnt * joint = ag->getRobotStruct()->joints[i];
                if (joint->type != P3D_BASE && joint->type != P3D_FIXED)
                    fixJoint(ag->getRobotStruct(), joint, joint->jnt_mat);
            }


            ArmManipulationData& mData = (*ag->getRobotStruct()->armManipulationData)[armId];
            ct = mData.getCcCntrt();
            for (int i = 0; i < ct->npasjnts; i++)
            {
                p3d_jnt * joint = ct->pasjnts[i];
                if (joint->type != P3D_FIXED)
                    unFixJoint(ag->getRobotStruct(), joint);
            }
            if ((!strcmp(ct->namecntrt, "p3d_kuka_arm_ik")) ||
                (!strcmp(ct->namecntrt, "p3d_lwr_arm_ik"))  ||
                (!strcmp(ct->namecntrt, "p3d_pr2_arm_ik")))
            {
                p3d_jnt * joint = ag->getRobotStruct()->joints[ct->argu_i[0]];
                unFixJoint(ag->getRobotStruct(), joint);
            }

//            mData.setCartesian(true);

            desactivateTwoJointsFixCntrt(ag->getRobotStruct(),mData.getManipulationJnt(), mData.getCcCntrt()->pasjnts[ mData.getCcCntrt()->npasjnts-1 ]);

            activateCcCntrts(ag->getRobotStruct(), armId, false);
            ManipulationUtils::unfixManipulationJoints(ag->getRobotStruct(), armId);
            shootTheObjectArroundTheBase(ag->getRobotStruct(), ag->getRobotStruct()->baseJnt, mData.getManipulationJnt(), 2.0);
            mData.getManipulationJnt()->dist  = 0.10;

            p3d_set_and_update_this_robot_conf(ag->getRobotStruct(), q);

//            print_config(ag->getRobotStruct(), q);
            i_can = p3d_update_this_robot_pos_multisol(ag->getRobotStruct(), NULL, 0, ikSol);
//            g3d_draw_allwin_active();
//            sleep(3);

            int begId = -1;
            if (armId == 0)
                begId = 5;
            else
                begId = 15;


            if (p3d_col_test_robot(ag->getRobotStruct(),0) == 0
                    && !p3d_isOutOfBounds(ag->getRobotStruct(), ag->getCurrentPos()->getConfigStruct()))
                found =true;
            else
                i_can = false;


            // in case of faillure, trying different initial pose before resolving the IK
            if (!i_can)
            {


                while (nbIKloop < MAXIKLOOP)
                {
                    nbIKloop++;


                    for(int i=begId; i<begId+7; i++)
                    {
                        double vmin,vmax;
                        ag->getJoint(i)->getDofBounds(0,vmin,vmax);
                        q[ag->getJoint(i)->getIndexOfFirstDof()] = p3d_random(vmin,vmax);
                    }

                    q[ffjntIndex + 0] = Tx;
                    q[ffjntIndex + 1] = Ty;
                    q[ffjntIndex + 2] = Tz;
                    q[ffjntIndex + 3] = Rx;
                    q[ffjntIndex + 4] = Ry;
                    q[ffjntIndex + 5] = Rz;

                    p3d_set_and_update_this_robot_conf( ag->getRobotStruct(), q );
                    i_can = p3d_update_this_robot_pos_multisol(ag->getRobotStruct(), NULL, 0, ikSol);
//                    print_config(ag->getRobotStruct(), q);
//                    g3d_draw_allwin_active();

                    if (i_can && (p3d_col_test_robot(ag->getRobotStruct(),0) == 0)
                            && !p3d_isOutOfBounds(ag->getRobotStruct(), ag->getCurrentPos()->getConfigStruct()))
                    {
                        found = true;
//                        cout << "not direct " << endl;
//                        g3d_draw_allwin_active();
//                        sleep(3);
                        break;
                    }
//                   else
//                   {
//                       p3d_print_col_pair();
////                       if (p3d_col_test_robot(ag->getRobotStruct(),0) != 0)
////                       {
////                           g3d_draw_allwin_active();
////                           sleep(1);
////                       }
//                   }

                }
            }

            if (i_can && found)
            {
    //            q = ag->getCurrentPos()->getConfigStruct();

                solution = shared_ptr<Configuration>(new Configuration(*ag->getCurrentPos()));
    //            solution->print(true);
                return true;
            }
            q =NULL;
            break;
        }
        case HRI_KUKAROBOT:
        {

//            p3d_matrix4 rotMat, tmp;
            rotMat[0][0] = 0;
            rotMat[0][1] = 0;
            rotMat[0][2] = -1;
            rotMat[0][3] = 0;

            rotMat[1][0] = 0;
            rotMat[1][1] = 1;
            rotMat[1][2] = 0;
            rotMat[1][3] = 0;

            rotMat[2][0] = 1;
            rotMat[2][1] = 0;
            rotMat[2][2] = 0;
            rotMat[2][3] = 0;

            rotMat[3][0] = 0;
            rotMat[3][1] = 0;
            rotMat[3][2] = 0;
            rotMat[3][3] = 1;

            p3d_mat4Mult(target_pos, rotMat, tmp);
            p3d_mat4Copy(tmp, target_pos);

            rotMat[0][0] = 0;
            rotMat[0][1] = 1;
            rotMat[0][2] = 0;
            rotMat[0][3] = 0;

            rotMat[1][0] = -1;
            rotMat[1][1] = 0;
            rotMat[1][2] = 0;
            rotMat[1][3] = 0;

            rotMat[2][0] = 0;
            rotMat[2][1] = 0;
            rotMat[2][2] = 1;
            rotMat[2][3] = 0;

            rotMat[3][0] = 0;
            rotMat[3][1] = 0;
            rotMat[3][2] = 0;
            rotMat[3][3] = 1;

            p3d_mat4Mult(target_pos, rotMat, tmp);
            p3d_mat4Copy(tmp, target_pos);




            found = i_can = false;
            double Tx, Ty, Tz;
            double Rx, Ry, Rz;
            nbIKloop = 0;
            p3d_mat4ExtractPosReverseOrder(target_pos, &Tx, &Ty, &Tz,
                                           &Rx, &Ry, &Rz);

            configPt q = p3d_copy_config(ag->getRobotStruct(), _WS->getRobotConf(ag)->getConfigStruct());
            int ffjntIndex = (*ag->getRobotStruct()->armManipulationData)[armId].getManipulationJnt()->index_dof;


            q[ffjntIndex + 0] = Tx;
            q[ffjntIndex + 1] = Ty;
            q[ffjntIndex + 2] = Tz;
            q[ffjntIndex + 3] = Rx;
            q[ffjntIndex + 4] = Ry;
            q[ffjntIndex + 5] = Rz;

//            print_config(ag->getRobotStruct(), q);

            int *ikSol = NULL;
            p3d_set_and_update_this_robot_conf( ag->getRobotStruct(), q );

            p3d_cntrt* ct;
            // over all constraints
//            for (int i=0; i<ag->getRobotStruct()->cntrt_manager->ncntrts; i++)
//            {
//                string name = ag->getRobotStruct()->cntrt_manager->cntrts[i]->namecntrt;

//                if ( name == "p3d_fix_jnts_relpos" )
//                    continue;
//                // get constraint from the cntrts manager
//                ct = ag->getRobotStruct()->cntrt_manager->cntrts[i];
//                p3d_desactivateCntrt( ag->getRobotStruct(), ct );
//            }

//            p3d_multiLocalPath_disable_all_groupToPlan( ag->getRobotStruct() , false );
//            p3d_multiLocalPath_set_groupToPlan( ag->getRobotStruct(), 2, 1, false);


            //Fix every joint except the active arm(s)
//            for (int i = 0; i < ag->getRobotStruct()->njoints + 1; i++)
//            {
//                p3d_jnt * joint = ag->getRobotStruct()->joints[i];
//                if (joint->type != P3D_BASE && joint->type != P3D_FIXED)
//                    fixJoint(ag->getRobotStruct(), joint, joint->jnt_mat);
//            }


            ArmManipulationData& mData = (*ag->getRobotStruct()->armManipulationData)[armId];
            ct = mData.getCcCntrt();
//            for (int i = 0; i < ct->npasjnts; i++)
//            {
//                p3d_jnt * joint = ct->pasjnts[i];
//                if (joint->type != P3D_FIXED)
//                    unFixJoint(ag->getRobotStruct(), joint);
//            }
            if ((!strcmp(ct->namecntrt, "p3d_kuka_arm_ik")) ||
                (!strcmp(ct->namecntrt, "p3d_lwr_arm_ik"))  ||
                (!strcmp(ct->namecntrt, "p3d_pr2_arm_ik")))
            {
                p3d_jnt * joint = ag->getRobotStruct()->joints[ct->argu_i[0]];
                unFixJoint(ag->getRobotStruct(), joint);
            }

//            mData.setCartesian(true);

            desactivateTwoJointsFixCntrt(ag->getRobotStruct(),mData.getManipulationJnt(), mData.getCcCntrt()->pasjnts[ mData.getCcCntrt()->npasjnts-1 ]);


            activateCcCntrts(ag->getRobotStruct(), armId, false);
            ManipulationUtils::unfixManipulationJoints(ag->getRobotStruct(), armId);
            shootTheObjectArroundTheBase(ag->getRobotStruct(), ag->getRobotStruct()->baseJnt, mData.getManipulationJnt(), 2.0);
            mData.getManipulationJnt()->dist  = 0.10;

            p3d_set_and_update_this_robot_conf(ag->getRobotStruct(), q);

//            print_config(ag->getRobotStruct(), q);
            i_can = p3d_update_this_robot_pos_multisol(ag->getRobotStruct(), NULL, 0, ikSol);
//            if (i_can)
//            {
//                cout << "aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa" << endl;
//                g3d_draw_allwin_active();
//                sleep(1);
//            }


            int begId = 1;



            if (p3d_col_test_robot(ag->getRobotStruct(),0) == 0
                    && !p3d_isOutOfBounds(ag->getRobotStruct(), ag->getCurrentPos()->getConfigStruct()))
                found =true;
            else
                i_can = false;


            // in case of faillure, trying different initial pose before resolving the IK
            if (!i_can)
            {


                while (nbIKloop < MAXIKLOOP)
                {
                    nbIKloop++;


                    for(int i=begId; i<begId+7; i++)
                    {
                        double vmin,vmax;
                        ag->getJoint(i)->getDofBounds(0,vmin,vmax);
                        q[ag->getJoint(i)->getIndexOfFirstDof()] = p3d_random(vmin,vmax);
                    }

                    q[ffjntIndex + 0] = Tx;
                    q[ffjntIndex + 1] = Ty;
                    q[ffjntIndex + 2] = Tz;
                    q[ffjntIndex + 3] = Rx;
                    q[ffjntIndex + 4] = Ry;
                    q[ffjntIndex + 5] = Rz;

                    p3d_set_and_update_this_robot_conf( ag->getRobotStruct(), q );
                    i_can = p3d_update_this_robot_pos_multisol(ag->getRobotStruct(), NULL, 0, ikSol);
//                    print_config(ag->getRobotStruct(), q);
//                    g3d_draw_allwin_active();
//                    sleep(2);

                    if (i_can && (p3d_col_test_robot(ag->getRobotStruct(),0) == 0)
                            && !p3d_isOutOfBounds(ag->getRobotStruct(), ag->getCurrentPos()->getConfigStruct()))
                    {
                        found = true;
//                        cout << "not direct " << endl;
//                        g3d_draw_allwin_active();
//                        sleep(3);
                        break;
                    }
//                   else
//                   {
//                       p3d_print_col_pair();
////                       if (p3d_col_test_robot(ag->getRobotStruct(),0) != 0)
////                       {
////                           g3d_draw_allwin_active();
////                           sleep(1);
////                       }
//                   }

                }
            }

            if (i_can && found)
            {
    //            q = ag->getCurrentPos()->getConfigStruct();

                solution = shared_ptr<Configuration>(new Configuration(*ag->getCurrentPos()));
    //            solution->print(true);
                return true;
            }
            q =NULL;
            break;
        }
        default:
            cout << "inverse kinematic not implemented for robot : " << mainAgent->getName() << " yet" << endl;
            break;
    }

    return false;
}


confPtr_t PrivateTask::setArmsToNeutralPos(Robot* ag)
{
    confPtr_t q = (confPtr_t)(new Configuration(*ag->getCurrentPos()));
    if (!ag || !ag->getHriAgent())
        return q;
    if (ag->getHriAgent()->type == HRI_PR2)
    {

        for (int i = 37; i <= 48; i++)
            (*q)[i] = 0;

        ag->setAndUpdate(*q);

        (*q)[16] = 0;
        (*q)[17] = 1.39;
        (*q)[18] = 0;
        (*q)[19] = -2.32;
        (*q)[20] = 3.14;
        (*q)[21] = -0.78;
        (*q)[22] = 0;

        (*q)[25] = 0;
        (*q)[26] = 1.39;
        (*q)[27] = 0;
        (*q)[28] = -2.32;
        (*q)[29] = 3.14;
        (*q)[30] = -0.78;
        (*q)[31] = 0;


        double Tx, Ty, Tz;
        double Rx, Ry, Rz;
        p3d_mat4ExtractPosReverseOrder(p3d_get_robot_jnt_by_name(q->getRobot()->getRobotStruct(), "rIkJoint")->abs_pos, &Tx, &Ty, &Tz, &Rx, &Ry, &Rz);

        q->getConfigStruct()[37] = Tx;
        q->getConfigStruct()[38] = Ty;
        q->getConfigStruct()[39] = Tz;
        q->getConfigStruct()[40] = Rx;
        q->getConfigStruct()[41] = Ry;
        q->getConfigStruct()[42] = Rz;

        p3d_mat4ExtractPosReverseOrder(p3d_get_robot_jnt_by_name(q->getRobot()->getRobotStruct(), "lIkJoint")->abs_pos, &Tx, &Ty, &Tz, &Rx, &Ry, &Rz);

        q->getConfigStruct()[43] = Tx;
        q->getConfigStruct()[44] = Ty;
        q->getConfigStruct()[45] = Tz;
        q->getConfigStruct()[46] = Rx;
        q->getConfigStruct()[47] = Ry;
        q->getConfigStruct()[48] = Rz;


    }
    else if (ag->getHriAgent()->type == HRI_HERAKLES)
    {
        for (unsigned int i = 12; i < 65; i++)
            (*q)[i] = 0;
        (*q)[18] = 1.396;
        (*q)[22] = 1.894;
        (*q)[27] = -1.396;
        (*q)[31] = -1.894;
    }
    else if (ag->getHriAgent()->type == HRI_KUKAROBOT)
    {

        for (int i = 21; i <= 26; i++)
            (*q)[i] = 0;

        ag->setAndUpdate(*q);

        (*q)[12] = 0;
        (*q)[13] = -M_PI/2;
        (*q)[14] = M_PI/2;
        (*q)[15] = M_PI/2;
        (*q)[16] = 0;
        (*q)[17] = 0;
        (*q)[18] = 0;

        ag->setAndUpdate(*q);


//        (*q)[21] = ag->getJoint("IkJoint")->getJointStruct()->abs_pos[0][3];
//        (*q)[22] = ag->getJoint("IkJoint")->getJointStruct()->abs_pos[1][3];
//        (*q)[23] = ag->getJoint("IkJoint")->getJointStruct()->abs_pos[2][3];

        double Tx, Ty, Tz;
        double Rx, Ry, Rz;
        p3d_mat4ExtractPosReverseOrder(ag->getJoint("IkJoint")->getJointStruct()->abs_pos, &Tx, &Ty, &Tz, &Rx, &Ry, &Rz);

        (*q)[21] = Tx;
        (*q)[22] = Ty;
        (*q)[23] = Tz;
        (*q)[24] = Rx;
        (*q)[25] = Ry;
        (*q)[26] = Rz;


//        p3d_matrix4 M;
//        p3d_mat4Copy(p3d_mat4IDENTITY,M);
//        //x
//        M[0][0] = ag->getJoint("IkJoint")->getJointStruct()->abs_pos[0][1];
//        M[1][0] = ag->getJoint("IkJoint")->getJointStruct()->abs_pos[1][1];
//        M[2][0] = ag->getJoint("IkJoint")->getJointStruct()->abs_pos[2][1];

//        //y
//        M[0][1] = ag->getJoint("IkJoint")->getJointStruct()->abs_pos[0][2];
//        M[1][1] = ag->getJoint("IkJoint")->getJointStruct()->abs_pos[1][2];
//        M[2][1] = ag->getJoint("IkJoint")->getJointStruct()->abs_pos[2][2];

//        //z
//        M[0][2] = ag->getJoint("IkJoint")->getJointStruct()->abs_pos[0][0];
//        M[1][2] = ag->getJoint("IkJoint")->getJointStruct()->abs_pos[1][0];
//        M[2][2] = ag->getJoint("IkJoint")->getJointStruct()->abs_pos[2][0];



//        double Tx, Ty, Tz;
//        double Rx, Ry, Rz;
//        p3d_mat4ExtractPosReverseOrder(M, &Tx, &Ty, &Tz, &Rx, &Ry, &Rz);

//        (*q)[24] = Rx;
//        (*q)[25] = Ry;
//        (*q)[26] = Rz;

//        ag->setAndUpdate(*q);

//        g3d_draw_allwin_active();
//        sleep(2);


//        double Tx, Ty, Tz;
//        double Rx, Ry, Rz;
//        p3d_mat4ExtractPosReverseOrder(p3d_get_robot_jnt_by_name(q->getRobot()->getRobotStruct(), "IkJoint")->abs_pos, &Tx, &Ty, &Tz, &Rx, &Ry, &Rz);

//        (*q)[21] = Tx;
//        (*q)[22] = Ty;
//        (*q)[23] = Tz;
//        (*q)[24] = Rx;
//        (*q)[25] = Ry;
//        (*q)[26] = Rz;
    }

    return q;
}

void PrivateTask::activateFingertipCol(bool val, Robot* robi , Robot* object, Robot* supportObj)
{
    if (val)
    {
        if (supportObj)
            gpActivate_object_fingertips_collisions( robi->getRobotStruct(), supportObj->getRobotStruct()->joints[1]->o, _armId);
        gpActivate_object_fingertips_collisions( robi->getRobotStruct(), object->getRobotStruct()->joints[1]->o, _armId);
        gpActivate_object_fingertips_collisions( robi->getRobotStruct(), object->getJoint(0)->getJointStruct()->o, _armId);
    }
    else
    {
        if (supportObj)
            gpDeactivate_object_fingertips_collisions( robi->getRobotStruct(), supportObj->getRobotStruct()->joints[1]->o, _armId);
        gpDeactivate_object_fingertips_collisions( robi->getRobotStruct(), object->getRobotStruct()->joints[1]->o, _armId);
        gpDeactivate_object_fingertips_collisions( robi->getRobotStruct(), object->getJoint(0)->getJointStruct()->o, _armId);
    }
}



//############################################# MANIPULATIONTASK ######################################################

ManipulationTask::ManipulationTask() : PrivateTask()
{
    currentTransfo = 0;
    mainObject = NULL;
}

bool ManipulationTask::setMainObject(multimap<string,string> objects_name)
{
    if (objects_name.count("mainObject") != 1)
    {
        cout << "ERROR: no main object for " << type << " task." << endl;
        returnRepport = "no_main_object";
        return false;
    }
    mainObject = global_Project->getActiveScene()->getRobotByName(objects_name.find("mainObject")->second);
    if (!mainObject)
    {
        returnRepport = "no_main_object_name";
        return false;
    }
    return true;
}

bool asd = true;

bool ManipulationTask::run(multimap<string,string> agents_name,
                           multimap<string,string> objects_name,
                           multimap<string,p3d_point> point3d,
                           multimap<string,string> additionalData,
                           WorldState *WS, bool computeMP)
{

    WS->load();

    _WS = WS;
    if (!initAll())
    {
        cout << "ERROR: init fails for task " << type << " with id " << task_id << endl;
        return false;
    }
    if (!setAgents(agents_name))
    {
        cout << "ERROR: setAgents fails for task " << type << " with id " << task_id << endl;
        return false;
    }
    if (!isPossibleMainAgent(mainAgent))
    {
        cout << "ERROR: mainAgent is not in the list of possible agent for task " << type << " with id " << task_id << endl;
        return false;
    }
    if (!setObjects(objects_name))
    {
        cout << "ERROR: setObjects fails for task " << type << " with id " << task_id << endl;
        return false;
    }
    if (!setPoints(point3d))
    {
        cout << "ERROR: setPoints fails for task " << type << " with id " << task_id << endl;
        return false;
    }
    if (!setdata(additionalData))
    {
        cout << "ERROR: setData fails for task " << type << " with id " << task_id << endl;
        return false;
    }
    if (!initialize())
    {
        cout << "ERROR: initialize fails for task " << type << " with id " << task_id << endl;
        return false;
    }

    if (!findCandidateSolutions())
    {
        cout << "ERROR: findCandidateSolutions fails for task " << type << " with id " << task_id << endl;
        return false;
    }
    bool res = false;
    int count = 0;



    while (!res)
    {
        cout << "count loop = " << count << endl;
        _returnRepportcount = 0;
        if(count++ > nbMaxLoops)
            break;
        if(!getRandomSol())
           continue;
        if(!findConfigurations())
            continue;
        if (computeMP)
            if(!findTrajectories())
                continue;
        if(!setToSolution(0,computeMP))
            continue;
        if(!testConstraints(0))
            continue;
        res = true;
    }
    if (res)
        returnRepport = "OK";

    computeCost(0);

    if (!finalize())
    {
        cout << "ERROR: finalize fails for task " << type << " with id " << task_id << endl;
        return false;
    }
    return res;
}

TaskSolution* ManipulationTask::findAlternative(bool computeMP)
{
    int count = 0;
    if (!getResult()->getAlternativeWithId(0))
    {
        cout<< "This task has no first solution, no alternative can be computed in that case" << endl;
        return NULL;
    }
    getResult()->getAlternativeWithId(0)->getStartWS()->load(); //

    if (!initialize())
    {
        cout << "ERROR: initialize fails for task " << type << " with id " << task_id << endl;
        return false;
    }

    int id = -1;
    bool res = false;
    while (!res)
    {
        cout << "count loop = " << count << endl;
        _returnRepportcount = 0;
        if(count++ > nbMaxLoops)
            break;
        if(!getRandomSol())
           continue;
        if(!findConfigurations())
           continue;
        if (computeMP)
            if(!findTrajectories())
                continue;
        id = result->getNextId();

        if(!setToSolution(id,computeMP))
           continue;
        if(!testConstraints(id))
            continue;
        res = true;
    }
    if (res)
        returnRepport = "OK";

    computeCost(id);

    if (!finalize())
    {
        cout << "ERROR: finalize fails for task " << type << " with id " << task_id << endl;
        return false;
    }
    if (res && id != -1)
        return getResult()->getAlternativeWithId(id);
    return NULL;
}


void ManipulationTask::showMainObjectRandomTransform()
{
    double Tx, Ty, Tz, Rx, Ry, Rz;
    p3d_matrix4 res;
    
    confPtr_t qO = mainObject->getCurrentPos();
    p3d_mat4Mult(mainObject->getRobotStruct()->joints[1]->abs_pos, mainObject->getRobotStruct()->objectTransform[currentTransfo].value, res);
    p3d_mat4ExtractPosReverseOrder(res, &Tx, &Ty, &Tz, &Rx, &Ry, &Rz);
    (*qO)[6] = Tx; (*qO)[7] = Ty; (*qO)[8] = Tz; (*qO)[9] = Rx; (*qO)[10] = Ry; (*qO)[11] = Rz;
    mainObject->setAndUpdate(*qO);

    if (currentTransfo < (mainObject->getRobotStruct()->objectTransform.size() - 1))
        currentTransfo ++;
    else
        currentTransfo = 0;
}

bool ManipulationTask::attachObject()
{
    return PrivateTask::attachObject(mainAgent, mainObject, true);
}

bool ManipulationTask::detachObject()
{
    return PrivateTask::attachObject(mainAgent, mainObject, false);
}



bool ManipulationTask::checkHRIAgent(Robot* r)
{
    if (!r->getHriAgent())
    {
        cout << "hri agent doesn't exist for "<< r->getName()<<" , unable to compute "<< type << endl;
        returnRepport = "no_hri_agent";
        return false;
    }
    if (!mainAgent->getHriAgent()->manip)
    {
        cout << "hri agent doesn't have manip data, unable to compute task" << endl;
        returnRepport = "no_manip_data_in_agent";
        return false;
    }
    return true;
}

bool ManipulationTask::setMainAgentHand()
{
    if (mainAgent->getHriAgent()->manip->external_hand_no <= 0)
    {
        cout << "hri agent doesn't have external hands" << endl;
        returnRepport = "no_external_hand_in_struct";
        return false;
    }

    for (int i = 0; i < mainAgent->getHriAgent()->manip->external_hand_no; i++)
    {
        if (!mainAgent->getHriAgent()->manip->external_hand)
            return false;
        if (!mainAgent->getHriAgent()->manip->external_hand[i])
            return false;
        if (mainAgent->getHriAgent()->manip->external_hand[i]->hand_l == m_arm)
        {
            _armId = mainAgent->getHriAgent()->manip->external_hand[i]->hand_id;
            _hands[_armId] = global_Project->getActiveScene()->getRobotByName(mainAgent->getHriAgent()->manip->external_hand[i]->rob_hand->name);
            break;
        }
    }
    if (!_hands[_armId])
    {
        cout << "ERROR, the hand of the robot was not found" << endl;
        returnRepport = "no_external_hand_in_world";
        return false;
    }
    return true;
}


bool ManipulationTask::isObjectInCollision(Robot* supportObject)
{
    if (supportObject)
        p3d_col_deactivate_robot_robot(mainObject->getRobotStruct(), supportObject->getRobotStruct());

    _returnRepportcount++;
    bool isColliding = false;
//    if(p3d_col_test_robot(mainObject->getRobotStruct(),0) == 1)//there is collision
    if (mainObject->getCurrentPos()->isInCollision())
    {
        if (_returnRepportcountMax < _returnRepportcount)
        {
            _returnRepportcountMax = _returnRepportcount;
            returnRepport = "object_in_collision";
        }
            p3d_print_col_pair();
        isColliding = true;
    }
    if (supportObject)
        p3d_col_activate_robot_robot(mainObject->getRobotStruct(), supportObject->getRobotStruct());
    if (isColliding)
        return true;
    else
        return false;
}


bool ManipulationTask::isHandInCollision()
{

    p3d_matrix4 rot;
    p3d_mat4Mult(mainObject->getRobotStruct()->joints[1]->abs_pos,_curGrasp.frame, rot);

    double Tx, Ty, Tz;
    double Rx, Ry, Rz;
    p3d_mat4ExtractPosReverseOrder(rot, &Tx, &Ty, &Tz,
                                   &Rx, &Ry, &Rz);


    confPtr_t q_hand = _hands[_armId]->getCurrentPos();


    double Itx = (*q_hand)[6], Ity = (*q_hand)[7],  Itz = (*q_hand)[8], Irx = (*q_hand)[9], Iry = (*q_hand)[10], Irz = (*q_hand)[11];

    (*q_hand)[6] = Tx; (*q_hand)[7] = Ty; (*q_hand)[8] = Tz;
    (*q_hand)[9] = Rx; (*q_hand)[10] = Ry; (*q_hand)[11] = Rz; (*q_hand)[12] = 0.15;
    _hands[_armId]->setAndUpdate(*q_hand);


    p3d_col_deactivate_robot_robot(mainObject->getRobotStruct(), _hands[_armId]->getRobotStruct());

    bool colTest = p3d_col_test_robot(_hands[_armId]->getRobotStruct(),0) == 1;
//    if (colTest)
//    {
//        cout << "hand col" << endl;
//        p3d_print_col_pair();

//        g3d_draw_allwin_active();
//        sleep(3);
//    }

    p3d_col_activate_robot_robot(mainObject->getRobotStruct(), _hands[_armId]->getRobotStruct());


    (*q_hand)[6] = Itx; (*q_hand)[7] = Ity; (*q_hand)[8] = Itz;
    (*q_hand)[9] = Irx; (*q_hand)[10] = Iry; (*q_hand)[11] = Irz;
    _hands[_armId]->setAndUpdate(*q_hand);


    _returnRepportcount++;
    if(colTest) // there is collision
    {
        if (_returnRepportcountMax < _returnRepportcount)
        {
            _returnRepportcountMax = _returnRepportcount;
            returnRepport = "hand_in_collision";
        }
        p3d_col_activate_robot_robot(mainObject->getRobotStruct(), _hands[_armId]->getRobotStruct());
        return true;
    }
    return false;
}


void ManipulationTask::removeVirtualObjects()
{
    virtualObjs.clear();
    for (unsigned int v = 0; v < global_Project->getActiveScene()->getNumberOfRobots(); v++)
    {
        Robot* re = global_Project->getActiveScene()->getRobot(v);
        if (re->getName().find("VIRTUAL") != string::npos)
        {
            pair<Robot*,configPt> rcp;
            rcp.first = re;
            rcp.second = re->getCurrentPos()->getConfigStruct();
            virtualObjs.push_back(rcp);

            configPt qtmp = p3d_copy_config(re->getRobotStruct(), re->getCurrentPos()->getConfigStruct());
            qtmp[6] = 0;
            qtmp[7] = 0;
            qtmp[8] = 0;
            p3d_set_and_update_this_robot_conf(re->getRobotStruct(),qtmp);
        }
    }
}

void ManipulationTask::moveBackVirtualObjects()
{
    for (unsigned int v = 0; v < virtualObjs.size(); v++)
    {
        p3d_set_and_update_this_robot_conf(virtualObjs.at(v).first->getRobotStruct(),virtualObjs.at(v).second);
    }
}


bool ManipulationTask::checkAttachement()
{
    if (_WS->getAttachements().size() == 0)
    {
        return false;
    }
    Attachement att = _WS->getAttachement(mainAgent,mainObject);
    if(att.object == NULL || att.agent == NULL)
    {
        return false;
    }
    _curGrasp = att.grasp;
    return true;
}


CooperativeTask::CooperativeTask()
{

}

bool CooperativeTask::setTargetAgent(multimap<string,string> agents_name)
{
    if (agents_name.count("targetAgent") != 1)
    {
        cout << "ERROR: no target agent for " << type << " task." << endl;
        returnRepport = "no_target_agent";
        return false;
    }
    targetAgent = global_Project->getActiveScene()->getRobotByName(agents_name.find("targetAgent")->second);
    if (!targetAgent)
        return false;
    return true;
}

//bool CooperativeTask::run(multimap<string,string> agents_name,
//                           multimap<string,string> objects_name,
//                           multimap<string,p3d_point> point3d,
//                           multimap<string,string> additionalData,
//                           WorldState *WS, bool computeMP)
//{

//    WS->load();
//    _WS = WS;
//    if (!setAgents(agents_name))
//    {
//        cout << "ERROR: setAgents fails for task " << type << " with id " << task_id << endl;
//        return false;
//    }
//    if (!isPossibleMainAgent(mainAgent))
//    {
//        cout << "ERROR: mainAgent is not in the list of possible agent for task " << type << " with id " << task_id << endl;
//        return false;
//    }
//    if (!setObjects(objects_name))
//    {
//        cout << "ERROR: setObjects fails for task " << type << " with id " << task_id << endl;
//        return false;
//    }
//    if (!setPoints(point3d))
//    {
//        cout << "ERROR: setPoint fails for task " << type << " with id " << task_id << endl;
//        return false;
//    }
//    if (!setdata(additionalData))
//    {
//        cout << "ERROR: setData fails for task " << type << " with id " << task_id << endl;
//        return false;
//    }
//    if (!initAll())
//    {
//        cout << "ERROR: init fails for task " << type << " with id " << task_id << endl;
//        return false;
//    }

//    if (!initialize())
//    {
//        cout << "ERROR: initialize fails for task " << type << " with id " << task_id << endl;
//        return false;
//    }

//    if (!findCandidateSolutions())
//    {
//        cout << "ERROR: findCandidateSolutions fails for task " << type << " with id " << task_id << endl;
//        return false;
//    }
//    bool res = false;
//    int count = 0;
//    while (!res)
//    {
//        cout << "count loop = " << count << endl;
//        _returnRepportcount = 0;
//        if(count++ > nbMaxLoops)
//            break;
//        if(!getRandomSol())
//           continue;
//        if(!findConfigurations())
//            continue;
//        if (computeMP)
//            if(!findTrajectories())
//                continue;
//        res = true;
//    }

//    if (res)
//        returnRepport = "OK";

//    computeCost(0);

//    if (!finalize())
//    {
//        cout << "ERROR: finalize fails for task " << type << " with id " << task_id << endl;
//        return false;
//    }
//    return res;
//}


//####################### Default ########################

DefaultTask::DefaultTask()
{
    type = "default";
}

bool DefaultTask::initAll()
{

    cout << "a problem in task name?" << endl;
    nbParameters = 0;
    nbGikCalled = 0;
    nbMPCalled = 0;
    nbTests = 0;
    return true;
}

bool DefaultTask::initialize()
{
    return true;
}

bool DefaultTask::finalize()
{
    return true;
}

bool DefaultTask::findConfigurations()
{
    cout<< "Default dummy task return false!!" << endl;
    return false;
}

bool DefaultTask::findCandidateSolutions()
{
    cout<< "Default dummy task return false!!" << endl;
    return false;
}

bool DefaultTask::findTrajectories()
{
    cout << "No trajectory to find" << endl;
    return true;
}

double DefaultTask::computeCost(int altId)
{
    return altId;
}

bool DefaultTask::computeSolutionTrajectories(int alternativeId)
{
    cout << "No trajectory to find" << endl;
    return true;
}

bool DefaultTask::setAgents(multimap<string,string> agents_name)
{
    cout << "No agents to set" << endl;
    return true;
}

bool DefaultTask::setObjects(multimap<string,string> objects_name)
{
    cout << "No objects to set" << endl;
    return true;
}

bool DefaultTask::setPoints(multimap<string,p3d_point> points)
{
    cout << "No points to set" << endl;
    return true;
}

bool DefaultTask::setdata(multimap<string,string> additionalData)
{
    cout << "No data to set" << endl;
    return true;
}

string DefaultTask::getDescr()
{
    cout << "a problem in task name?" << endl;
    return "Task: \nId \t\t= " + to_string(task_id) + "\nType \t\t= " + type;
}


bool DefaultTask::setToSolution(int solutionId, bool computeMP)
{

    solutionId = 0;
    computeMP = 0;
    return true;
}

bool DefaultTask::testConstraints(int alternativeId)
{
    return false;
}

TaskSolution* DefaultTask::findAlternative(bool computeMP)
{
    return NULL;
}

string DefaultTask::getTextFromValues()
{
    return "";
}

void DefaultTask::showMainObjectRandomTransform()
{

}

void DefaultTask::smoothSolution(int alternativeId)
{
    std::cout << "not implemented" << std::endl;
}


