#ifndef TASK_HPP
#define TASK_HPP

#include "GTPTools/grid.h"
#include "GTPTools/tools.hpp"
#include "GTPTools/WorldState.hpp"
#include "GTPTools/geometricTools.hpp"
#include "GTPTools/taskconstraints.hpp"
#include "lightPlanner/proto/ManipulationPlanner.hpp"



using namespace std;
#include "move3d-headless.h"
#include "Planner-pkg.h"
#include "Collision-pkg.h"
#include "../planner/planEnvironment.hpp"

#include "../API/Device/robot.hpp"
#include "GTPTools/gtpplan.hpp"


#include <iostream>
#include <map>

class WorldState;
class TaskSubSolution;

template < typename T > std::string to_string( const T& n );
class TaskAllSolutions;
class TaskSolution;
class TaskConstraints;
class PlanTree;

class Task
{
public:
    Task();
    virtual bool initAll() = 0;
    virtual bool initialize() = 0;
    virtual bool finalize() = 0;
    virtual bool findCandidateSolutions() = 0;
    virtual bool findConfigurations() =0;
    virtual bool findTrajectories() =0;
    virtual bool setToSolution(int solutionId, bool computeMP) = 0;
    virtual bool testConstraints(int alternativeId) = 0;
    //validatePoint();
    virtual bool setAgents(multimap<string,string> agents_name) = 0;
    virtual bool setObjects(multimap<string,string> objects_name) = 0;
    virtual bool setPoints(multimap<string,p3d_point> points) = 0;
    virtual bool setdata(multimap<string,string> additionalData) = 0;
    virtual bool run(multimap<string,string> agents_name,
                     multimap<string,string> objects_name,
                     multimap<string,p3d_point> point3d,
                     multimap<string,string> additionalData,
                     WorldState* WS, bool computeMP);
    virtual double heuristic(multimap<string,string> agents_name,
                             multimap<string,string> objects_name,
                             multimap<string,p3d_point> point3d,
                             multimap<string,string> additionalData,
                             WorldState* WS, bool computeMP);
    virtual TaskSolution* findAlternative(bool computeMP) = 0;
    virtual double computeCost(int altId) = 0;
    virtual double computeHeuristicCost();
    virtual bool computeSolutionTrajectories(int alternativeId) = 0;
    p3d_traj* compute_traj(tr1::shared_ptr<Configuration> q_init, tr1::shared_ptr<Configuration> q_goal, int& status);
    virtual bool computeManipulationConfs(confPtr_t& approachC, confPtr_t& graspC, confPtr_t& graspOC, confPtr_t& escapeC, p3d_matrix4 objectPosMat,
                                          gpGrasp grasp, p3d_vector3 approachDir, double approachDist , p3d_vector3 escapeDir, double escapeDist);
    virtual bool computeInverseKinematic(Robot* ag, p3d_matrix4 target_pos, confPtr_t &solution, int armId);
    p3d_traj* compute_traj_using_localPath(tr1::shared_ptr<Configuration> q_init, tr1::shared_ptr<Configuration> q_goal, int& status);
//    virtual vector<TaskSubSolution*> setNextTraj(TaskSubSolution* doneTSS) =0;
    virtual void smoothSolution(int alternativeId) = 0;
    int getId();
    void setId(int id);
    string getType() { return type; }
    TasksId getTypeId() { return typeId; }
    virtual string getDescr() = 0;
    virtual string getTextFromValues() = 0;
    TaskSubSolution* getSubSolAt(int alternativeId, int id);
    int getSubSolNumberOfTrajs(int alternativeId);

    virtual double getSSSizeLeft();

    TaskAllSolutions* getResult(){return result; }
    void setResult(TaskAllSolutions* r){result = r; }

    void setConstraints(TaskConstraints* c) {_currConstraints = c;}
    TaskConstraints* getCurrConstraint(){return _currConstraints;}

    void setGraph(Graph* g) {_taskGraph = g;}
    Graph* getGraph() {return _taskGraph;}

    void setPlanTree(PlanTree* pt) {_planTree = pt;}
    PlanTree* getPlanTree() {return _planTree;}

    void setCurrentDataType(type_data pt) {currentDataType = pt;}
    type_data getCurrentDataType() {return currentDataType;}

    int nbGikCalled;
    int nbMPCalled;
    int nbParameters;
    int nbTests;
    double timeComp;
    double timeMp;
    int nbTotPossibleTests;
    string returnRepport;

protected:
    int task_id;
    string type;
    TasksId typeId;
    TaskAllSolutions* result;
//    std::vector<TaskSubSolution*> result;
    WorldState* _WS;

    int nbMaxLoops;
    TaskConstraints* _currConstraints;

    int _returnRepportcount;
    int _returnRepportcountMax;

    Graph* _taskGraph;
    PlanTree* _planTree;
    type_data currentDataType;

};

class PrivateTask : public Task
{
public:
    PrivateTask();
    virtual bool initAll();
    virtual bool initialize();
    virtual bool finalize();
    virtual bool findCandidateSolutions() = 0;
    virtual bool getRandomSol() = 0;
    virtual bool testConstraints(int alternativeId);
//    virtual vector<TaskSubSolution*> setNextTraj(TaskSubSolution* doneTSS) =0;

    virtual bool run(multimap<string,string> agents_name,
             multimap<string,string> objects_name,
             multimap<string,p3d_point> point3d,
             multimap<string,string> additionalData,
             WorldState* WS, bool computeMP);
    virtual double computeCost(int altId);
    virtual TaskSolution* findAlternative(bool computeMP);
    armLabel getArmLabelFromId(int armId);
    bool setArmToUse(armLabel arm_t);
    bool initMotionPlaner(Robot *rob);
    p3d_traj* compute_traj(tr1::shared_ptr<Configuration> q_init, tr1::shared_ptr<Configuration> q_goal, Robot* object, int& status);
    p3d_traj* compute_traj_using_localPath(tr1::shared_ptr<Configuration> q_init, tr1::shared_ptr<Configuration> q_goal, Robot* object, int& status);

    virtual bool attachObject(Robot* rob, Robot* obj, bool attach);
    virtual bool attachObject(Robot *rob, Robot *obj, bool attach, int armId); //true for attaching, false for detaching

    bool computeInverseKinematic(Robot* ag, p3d_matrix4 target_pos, confPtr_t  &solution, int armId);
    bool computeManipulationConfs(confPtr_t &approachC, confPtr_t &graspC, confPtr_t &graspOC, confPtr_t &escapeC, p3d_matrix4 objectPosMat,
                                  gpGrasp grasp, p3d_vector3 approachDir, double approachDist, p3d_vector3 escapeDir, double escapeDist);

    //this is a specific function used in case we need ids.
    int getArmId();

    bool setMainAgent(multimap<string,string> agents_name);

    Robot* getMainAgent(){return mainAgent;}

    virtual bool isPossibleMainAgent(Robot* r);

    void openGripper(confPtr_t &q, bool open, int armId);

    confPtr_t setArmsToNeutralPos(Robot* ag);

    void activateFingertipCol(bool val, Robot* robi, Robot *object, Robot *supportObj);



protected:
    std::vector<GTPCell> candidatePoints;
    //trajs
    //configurations
    //WorldState
    Robot* mainAgent;
    armLabel m_arm;
    gpHand_type handType;
    int _armId;

    map<string, ManipulationPlanner*> ManipPlMap;

    vector<HRI_AGENT_TYPE> _possibleMainAgentType;

};

class ManipulationTask : public PrivateTask
{
public:
    ManipulationTask();
    //virtual bool init();  //should be called after setting main agent
    virtual bool findCandidateSolutions() = 0;
    virtual bool getRandomSol() = 0;
//    virtual vector<TaskSubSolution*> setNextTraj(TaskSubSolution* doneTSS) =0;

    bool run(multimap<string,string> agents_name,
             multimap<string,string> objects_name,
             multimap<string,p3d_point> point3d,
             multimap<string,string> additionalData,
             WorldState* WS, bool computeMP);

    TaskSolution* findAlternative(bool computeMP);

    std::vector <confPtr_t> compute_manipulation_data(int armId, Robot *object, gpGrasp grasp);

    bool setMainObject(multimap<string,string> objects_name);

    Robot* getMainObject(){return mainObject;}

    virtual void showMainObjectRandomTransform();


    bool attachObject();
    bool detachObject();
    bool checkHRIAgent(Robot* r);
    bool setMainAgentHand();
    bool isObjectInCollision(Robot *supportObject);
    bool isHandInCollision();

    void removeVirtualObjects();
    void moveBackVirtualObjects();



    //return true if attachement exist between mainagent and mainobject
    bool checkAttachement();

protected:

    Robot* mainObject;

    tr1::shared_ptr<Configuration> graspConfClosed;
    tr1::shared_ptr<Configuration> graspConfOpen;
    tr1::shared_ptr<Configuration> approachConf;
    tr1::shared_ptr<Configuration> escapeConf;

    int currentTransfo;
    map<int, Robot*> _hands;
    gpGrasp _curGrasp;

    vector<pair<Robot*,configPt> > virtualObjs;


};

class CooperativeTask : public ManipulationTask
{
public:
    CooperativeTask();
    virtual bool findCandidateSolutions() = 0;
    virtual bool getRandomSol() = 0;
//    virtual vector<TaskSubSolution*> setNextTraj(TaskSubSolution* doneTSS) = 0;

//    bool run(multimap<string,string> agents_name,
//             multimap<string,string> objects_name,
//             multimap<string,p3d_point> point3d,
//             multimap<string,string> additionalData,
//             WorldState* WS, bool computeMP);


    bool setTargetAgent(multimap<string,string> agents_name);

    Robot *getTargetAgent() { return targetAgent; }

protected:
    Robot* targetAgent;
};


class DefaultTask : public Task
{
public:
    DefaultTask();
    virtual bool initAll();
    virtual bool initialize();
    virtual bool finalize();
    virtual bool findCandidateSolutions();
    virtual bool findConfigurations();
    virtual bool findTrajectories();
    virtual double computeCost(int altId);
    virtual bool computeSolutionTrajectories(int alternativeId);
    virtual bool setToSolution(int solutionId, bool computeMP);
    virtual bool testConstraints(int alternativeId);
    virtual bool setAgents(multimap<string,string> agents_name);
    virtual bool setObjects(multimap<string,string> objects_name);
    virtual bool setPoints(multimap<string,p3d_point> points);
    virtual bool setdata(multimap<string,string> additionalData);
    virtual string getDescr();
    virtual TaskSolution* findAlternative(bool computeMP);
    virtual string getTextFromValues();
    virtual void showMainObjectRandomTransform();
    virtual void smoothSolution(int alternativeId);
//    virtual vector<TaskSubSolution*> setNextTraj(TaskSubSolution* doneTSS)
//    {
//         vector<TaskSubSolution*> ts;
//         return ts;
//    }
};




#endif // TASK_HPP
