#include "HRICS_MultiHandOver.h"

#include "P3d-pkg.h"
#include "Collision-pkg.h"
#include "planEnvironment.hpp"
#include "utils/multiHandOver/graphAlgorithms.h"
#include "GraspPlanning-pkg.h"
#include "LightPlanner-pkg.h"
#include "utils/MultiHandOverUtils.hpp"
#include "utils/Chronometer.hpp"
#include "MultiPathPlanner.h"
#include <boost/random/normal_distribution.hpp>
#include <boost/random/uniform_int.hpp>
#include <boost/random/mersenne_twister.hpp>
#include <boost/random/variate_generator.hpp>
#include "planner/plannerFunctions.hpp"
#include <sstream>
#include "VimanTags.h"
#include "taskmanagerinterface.h"

#include "Logging/Logger.h"


using namespace std;
namespace HRICS {
using namespace MHO;

INIT_MOVE3D_STATIC_LOGGER(MultiHandOver,"planners.hrics.mho");

class AlgoLoggerInterfaceFull : public MHO::AlgoLoggerInterface{
public:
    //virtual void textlog
    AlgoLoggerInterfaceFull(MultiHandOver *mho,MultiAgentGrid *grid,bool draw) :
        MHO::AlgoLoggerInterface(),
        _mho(mho),_grid(grid),_draw(draw)
    {
    }
    virtual void drawcells(MultiAgentCell *c1, MultiAgentCell *c2, bool marker){
        if(_draw){
#ifdef DRAW_OPTIMIZE_FRONTIER
            _mho->resetAgentsConf();
            _grid->setDrawOptimizeFrontier(c1,c2,marker);
#endif
            this->draw();
        }
    }
    virtual void draw(){
        if(_draw){
            g3d_draw_allwin_active();
        }
    }
    virtual void resetdisp(){
        _grid->unsetDrawGridForOptim();
        _grid->unsetDrawOptimizeFrontier();
    }

private:
    MultiHandOver *_mho;
    MultiAgentGrid* _grid;
    bool _draw;
};

MultiHandOver::MultiHandOver(double cell_size) : HumanAwareMotionPlanner() /*, m_PathExist(false) , m_HumanPathExist(false) , initTime(10)*/
{
    _showText = PlanEnv->getBool(PlanParam::env_showText);
    M3D_INFO("MultiHandOver::MultiHandOver()");
    _search_algo=LWA_SEARCH;
    _a_star_heuristic = DIST_BEST_AGENT;
    this->_random_first_target_agents=false;

    _draw_object_frontiers_checks=false;
    _draw_2d_path=false;
    _draw_all_2d_pathes=false;
    _draw_agents_holding_conf_during_search=false;
#ifdef DRAW_OPTIMIZE_FRONTIER
    _draw_optimize_frontier=false;
#endif
    _draw_lwa=false;
    _display_lwa_in_grid_obj=0;
    _mho_display = 0;

    _overheat_temperature=100;
    _time_limit=-1;
    _max_iterations=1000;

    _epsilon=2;

    _max_frontier_check_status = FrontierCells::FCS_HANDOVER;

    _first_run=true;
    _compute_traj_only=false;
    _play_solution_task_traj=false;
    _play_solution_task=false;
    _solution_to_compute_traj=0;
    _step_to_compute_traj=0;
    _compute_for_all_robots=false;

    _use_post_process=false;
    _use_rrt_for_frontier_check=true;
    _use_dummy_smooth_for_frontier_check=true;
    _check_frontier_before=true;

    _best_solution=0;
    _tool_set=0;

    _h_h_ho_time=_r_h_ho_time=_r_r_ho_time=5;


    if(TMI){
        delete TMI;
    }
    TMI = new taskManagerInterface();
    TMI->initTaskManager();

    setCostsFactors(1,1,1,1);

    global_Project->getActiveScene()->createAgents();

    global_Project->getActiveScene()->createObjects();

    this->findAgents();

    _grid=0;

    this->setGridCellSize(cell_size);

    this->createGrid();

    M3D_DEBUG("MultiHandOver::MultiHandOver() returns");
//    initAll();
//    m_OTPsuceed =false;
}

MultiHandOver::~MultiHandOver(){
    //for(unsigned int i=0;i<_agents.size();++i){
        //delete _agents[i];
    //}
    this->clearSolutions();
    delete _grid;
    if(_tool_set){
        delete _tool_set->handoverObjectChecker;
        delete _tool_set->solutionTools;
        delete _tool_set;
    }
}

void MultiHandOver::testForReinit(){
    bool conf_chged=false;
    for(unsigned int i=0;i<this->_mho_agents.size();i++){
        if(this->_agents_init_confs[i] != *this->getAgents()[i]->getCurrentPos()){
            conf_chged=true;
            break;
        }
    }
    if(conf_chged){
        M3D_DEBUG("confs have changed, reinit init/target confs");
        this->fetchAgentsConfsInitGoTo();
        this->computeGrid();
    }
}

void MultiHandOver::createGrid(){
    _grid=0; //TODO remove this, handle same grid over multiples runs
    this->initGrid();
    this->setDrawAgentAccessibilityGrid(false);
    this->setDrawFrontiers(false);
    API_activeGrid = this->getPlanGrid();
}

void MultiHandOver::dumpVar(){
    cout << "============ MHO variable ===========" <<endl;

    cout << "fused grid = "  << PlanEnv->getBool(PlanParam::env_fusedGridRand) << endl;
    cout << "fused grid and rotation = "  << PlanEnv->getBool(PlanParam::env_fusedGridAndRotRand) << endl;
    cout << "normal rand = "  << PlanEnv->getBool(PlanParam::env_normalRand) << endl;
    cout << "oriented slice = "  << PlanEnv->getBool(PlanParam::env_useOrientedSlice) << endl;
    cout << "slices = "  << PlanEnv->getBool(PlanParam::env_useSlice) << endl;
    cout << "all grid = "  << PlanEnv->getBool(PlanParam::env_useAllGrid) << endl << endl;

#ifdef MULTILOCALPATH
    cout << " plotSoftMotionCurve = " << ENV.getBool(Env::plotSoftMotionCurve) << endl;
    cout << " writeSoftMotionFiles = " << ENV.getBool(Env::writeSoftMotionFiles) << endl;
#endif

    cout<< "ENV dmax = "<<ENV.getDouble(Env::dmax)<<endl;
    cout<< "p3d_env_dmax"<< p3d_get_env_dmax()<<endl<<endl;

    cout << "------------ AGENTS INFO ------------"<<endl<<endl;
    for (unsigned int i=0;i<_agents.size();++i){
        cout<<getAgentStruct(i)->robot->getName()<<endl;
        cout<<"\tis human: "<<getAgentStruct(i)->is_human<<endl;
        cout<<"\treach: "<<getAgentStruct(i)->reach()<<endl;
        cout<<"\tshoulder z"<<getAgentStruct(i)->shoulderHeight()<<endl;
        cout<<"\tsafe dist:"<<getAgentStruct(i)->safe_dist<<endl;
    }
    cout<<endl;


    bool dumpAll=false;
    if(dumpAll){
        cout << "============ MHO variable end========" <<endl<<endl;

        // ------------------------------------------------------------------
        // ENV
        // ------------------------------------------------------------------

        cout << "---------------------------" << endl;
        cout << "--------- Env      --------"<<endl;
        cout << "---------------------------" << endl<<endl;
        cout << "---------- bool -----------" << endl;
        for( int i=0;i<=Env::isOneToOne;i++)
        {
            cout << "Key : " << i;
            cout << " , Value : " << ENV.getBool(Env::boolParameter(i)) << endl;
        }
        cout << "---------- int  ------------" << endl;
        for( int i=0;i<=Env::mod_traj_selected;i++)
        {
            cout << "Key : " << i;
            cout << " , Value : " << ENV.getInt(Env::intParameter(i)) << endl;
        }
        cout << "--------    double   --------" << endl;
        for( int i=0;i<=Env::selected_endConf;i++)
        {
            cout << "Key : " << i;
            cout << " , Value : " << ENV.getDouble(Env::doubleParameter(i)) << endl;
        }

        cout << "--------   string    ----------" << endl;
        for( int i=0;i<=Env::ActiveGrid;i++)
        {
            cout << "Key : " << i;
            cout << " , Value : " << ENV.getString(Env::stringParameter(i)).toStdString() << endl;
        }
        // ------------------------------------------------------------------
        // PlanParam
        // ------------------------------------------------------------------

        cout<<endl<<endl;
        cout << "---------------------------" << endl;
        cout << "--------- PlanParam--------"<<endl;
        cout << "---------------------------" << endl<<endl;

        cout << "---------- bool -----------" << endl;
        for( int i=0;i<=PlanParam::env_trajRos;i++)
        {
            cout << "Key : " << i;
            cout << " , Value : " << PlanEnv->getBool(PlanParam::boolParameter(i)) << endl;
        }
        cout << "---------- int  ------------" << endl;
        for( int i=0;i<PlanParam::env_anglePow;i++)
        {
            cout << "Key : " << i;
            cout << " , Value : " << PlanEnv->getInt(PlanParam::intParameter(i)) << endl;
        }
        cout << "--------    double   --------" << endl;
        for( int i=0;i<=PlanParam::cellsOpacity;i++)
        {
            cout << "Key : " << i;
            cout << " , Value : " << PlanEnv->getDouble(PlanParam::doubleParameter(i)) << endl;
        }
        cout << "--------   string    ----------" << endl;
        for( int i=0;i<=PlanParam::titi;i++)
        {
            cout << "Key : " << i;
            cout << " , Value : " << PlanEnv->getString(PlanParam::stringParameter(i)).toStdString() << endl;
        }
    }
}

void MultiHandOver::setDefaultEnvVar(){
    ENV.setBool(Env::HRIPlannerCS,true);
    ENV.setBool(Env::enableHri,true);
    ENV.setBool(Env::isCostSpace,true);

    ENV.setBool(Env::useBallDist,false);
    ENV.setBool(Env::useBoxDist,true);

    //cout<<"coeff joint :"<<ENV.getDouble(Env::coeffJoint)<<" -- index OK:"<<(Env::coeffJoint ==22)<<endl;
    //cout<<"coeff confo :"<<ENV.getDouble(Env::coeffConfo)<<" -- index OK:"<<(Env::coeffConfo ==24)<<endl;
    //cout<<"coeff ArmPr :"<<ENV.getDouble(Env::coeffArmPr)<<" -- index OK:"<<(Env::coeffArmPr ==25)<<endl;
    ENV.setDouble(Env::coeffJoint,0.5);
    p3d_set_env_dmax(0.1);

    API_activeGrid = this->getPlanGrid();
    ENV.setBool(Env::drawGrid,true);
}

bool MultiHandOver::initFromConfXML(istream &is){
    ConfigFileParser px(this);
    return px.configure_from_xml(is);
}

bool MultiHandOver::initFromConfXML(string path){
    ifstream fs;
    fs.open(path.c_str());
    bool ok=(fs);
    return ok && this->initFromConfXML(fs);
}

std::string listFrontierTesters(FrontierTester* frontier_tester){
    std::stringstream ss;
    vector<pair<AbstractHandOverTest*,bool> > testers=frontier_tester->listTesters();
    for(unsigned int i=0;i<testers.size();i++){
        if(testers[i].second){
            ss<<"*";
        }else{
            ss<<" ";
        }
        ss<<testers[i].first->getName()<< " [" << testers[i].first->getRequiredCheckStatus()<<"->"<< testers[i].first->getCheckStatusLevel() << "]"<<endl;
    }
    return ss.str();

}

void MultiHandOver::initToolSet(){
    AbstracHandOverObjectCheck *ho_check;
    FrontierTester *frontier_tester;
    frontier_tester=new FrontierTester();
    if(_tool_set){
        delete _tool_set;
        _tool_set=0;
    }
    if(_use_rrt_for_frontier_check){
        ho_check=new MHO::HandOverWindowCheckerRRT(this->_object);
    }else{
        ho_check=new MHO::HandOverDirectChecker(this->_object);
    }
    ho_check->timeLimit(1);
    frontier_tester->addTester(new HandOverTestObject(ho_check,&this->_mho_agents),true);
    frontier_tester->addTester(new HandOverTestDistance(&this->_mho_agents),true);
    frontier_tester->addTester(new HandOverTestGrabs(&this->_mho_agents,&this->_object),true);
    string file_dir = getenv("HOME_MOVE3D");
    file_dir+="/statFiles/MhoConfs";
    frontier_tester->addTester(new DummyHandOverTest(FrontierCells::FCS_GRAB_CONFS,FrontierCells::FCS_HANDOVER),false);
    frontier_tester->addTester(new HandOverTestPreDefined(&this->_mho_agents,file_dir,this->getGridCellSize()),true);
    frontier_tester->addTester(new HandOverTestTrajectory(&this->_mho_agents),true);
    frontier_tester->addTester(new HandOverTestVisible(&this->_mho_agents,this->_object),true);
    for(unsigned int i=0;i<_handover_testers_defaults_set.size();i++){
        if(frontier_tester->getTester(_handover_testers_defaults_set[i]) != 0){
            frontier_tester->addDefault(_handover_testers_defaults_set[i]);
        }else{
            M3D_WARN("MHO WARNING: specified an unknown default handover tester : "<<_handover_testers_defaults_set[i]);
        }
    }
    M3D_INFO("MHO : list frontier testers\n"<<listFrontierTesters(frontier_tester));
    _tool_set=new ToolSet(frontier_tester,new MHO::SolutionTools(this),this,this->getFrontierMaxCheckStatus());
}

MHO::MultiHandOverHeuristic* MultiHandOver::createAstarHeuristic(){
    AStarHeuristic_t h = this->getAstarHeuristic();
    MHO::MultiHandOverHeuristic* heuristic;
    assert(this->getToolSet() && this->getToolSet()->solutionTools);
    switch(h){
    case DIST_BEST_AGENT:
        heuristic = new MHO::MultiHandOverHeuristic(this,this->getToolSet()->solutionTools);
        break;
    case BEST_HANDOVER_SERIES:
        M3D_WARN("MHO WARNING: using unstable heurisic MultiHandOverHeuristicClosestFrontier (BEST_HANDOVER_SERIES)");
        heuristic = new MHO::MultiHandOverHeuristicClosestFrontier(this, this->getToolSet()->solutionTools);
        break;
    default:
        M3D_WARN("heuristic not or badly set. Falling back to default (DIST_BEST_AGENT) (found: "<<this->getAstarHeuristic()<<')');
        heuristic = new MHO::MultiHandOverHeuristic(this,this->getToolSet()->solutionTools);
    }
    return heuristic;
}

//#define DEBUG_TEST_MHO_MULTI_AGENT_PATH
#ifdef DEBUG_TEST_MHO_MULTI_AGENT_PATH
unsigned int MultiHandOver::run(){
        this->setDrawAgentAccessibilityGrid(false);
        Chronometer chrono;
        vector<double> times(4,0);
        if(_first_run){
            this->computeGrid();
        }
        times[0]=chrono.getMilliSec();
        if(_first_run){
            this->computeFrontiers();
            _first_run=false;
        }
        times[1]=chrono.getMilliSec();
        initToolSet();
        if(_check_frontier_before){
            this->checkAllMetaFrontiersWithObject();
        }
        times[2]=chrono.getMilliSec();
        this->computeAgentsGraph();
        MultiPathPlanneLWADisplayInterface disp(true,this->_draw_lwa,_grid);

        vector<unsigned int> agents;
        agents.push_back(this->_first_agent_node->getId());
        agents.push_back(this->_target_agent_node->getId());
        vector<MultiAgentCell*> start_pos;
        start_pos.push_back(this->getPlanGrid()->getAgentStructAt(agents[0])->getOrigin());
        start_pos.push_back(this->getPlanGrid()->getAgentStructAt(agents[1])->getOrigin());
        vector<MultiAgentCell*> goal_pos;//goal positions are swapped postions of agents
        goal_pos.push_back(this->getPlanGrid()->getAgentStructAt(agents[1])->getOrigin());
        //goal_pos.push_back(this->getPlanGrid()->getAgentStructAt(agents[0])->getOrigin());
        goal_pos.push_back(0);
        //Eigen::Vector2i v_goal1,v_goal2;
        //v_goal1=start_pos[0]->getCoord()+Eigen::Vector2i(1,0);
        //v_goal2=start_pos[1]->getCoord()+Eigen::Vector2i(-1,0);
        //goal_pos.push_back(dynamic_cast<MultiAgentCell*>(getPlanGrid()->getCell(v_goal1)));
        //goal_pos.push_back(dynamic_cast<MultiAgentCell*>(getPlanGrid()->getCell(v_goal2)));
        //MultiPathLWANode *start=new MultiPathLWANode(agents,start_pos,ToolSet::distanceCostStatic,_tool_set);
        //MultiPathLWANode *goal=new MultiPathLWANode(agents,goal_pos,ToolSet::distanceCostStatic,_tool_set);
        //mho::LazyWeightedAstar lwa(start,goal,this->getEpsilonForLWAstar(),&disp,-1);
        MultiPathPlanner mpp;
        mpp.setAgents(agents);
        mpp.setDisplay(true,_draw_lwa,_grid);
        mpp.setStart(start_pos);
        mpp.setGoal(goal_pos);
        mpp.setSolutionTools(this->_tool_set->solutionTools);
        for(unsigned int i=0;i<3;i++){
            std::vector<MultiAgentCell*> &pos=(i!=1 ? start_pos : goal_pos);
            for(unsigned int a=0;a<pos.size();a++){
                if(pos[a]){
                    unsigned int a_i=agents[a];
                    Robot* r=getAgentStruct(a_i)->robot;
                    Configuration q=*r->getCurrentPos();
                    for(unsigned int q_i=0;q_i<2;q_i++){
                        q[6+q_i]=pos[a]->getCenter()[q_i];
                    }
                    r->setAndUpdate(q);
                    g3d_draw_allwin_active();
                }
            }
            usleep(200*1000);
        }
        //if(lwa.compute()){
        if(mpp.compute()){
            M3D_INFO( "solution found" );
            std::deque<mho::LWANodeInterface*> path=mpp.getPath();
            for(unsigned int i=0;i<path.size();i++){
                MultiPathLWANode *node=dynamic_cast<MultiPathLWANode*>(path[i]);
                for(unsigned int a=0;node->agent_index(a)>=0;a++){
                    unsigned int a_i=node->agent_index(a);
                    Robot* r=getAgentStruct(a_i)->robot;
                    Configuration q=*r->getCurrentPos();
                    for(unsigned int q_i=0;q_i<2;q_i++){
                        q[6+q_i]=node->cell(a)->getCenter()[q_i];
                    }
                    r->setAndUpdate(q);
                    g3d_draw_allwin_active();
                }
                usleep(200*1000);
            }
        }else{
            M3D_INFO( "not found");
        }
        return 0;
}

#else

unsigned int MultiHandOver::runScriptFromEnvVar(){
    M3D_INFO( "MHO: running script $MOVE3D_SCRIPT");
    char * script=getenv("MOVE3D_SCRIPT");
    if(script){
        return this->runScript(string(script));
    }else{
        M3D_ERROR("MHO : variable $MOVE3D_SCRIPT not set");
        return 1;
    }
}

unsigned int MultiHandOver::runScript(string filename){
    ifstream file;
    M3D_INFO( "MHO: running script "<<filename);

    string file_path=filename.substr(0,filename.find_last_of("/")+1);
    M3D_INFO("script: "<<filename<< " - path: "<<file_path);
    assert(file_path.at(file_path.size()-1)=='/');
    assert(filename.compare(0,file_path.size()-1,file_path,0,file_path.size()-1) == 0);
    file.open(filename.c_str());
    bool dry_run=true;
    for(unsigned int loopcnt=0;loopcnt<2;loopcnt++){
        bool error=false,initialised=false,do_exit=false;
        int n_line=0;
        while(!!file && !error && !do_exit){
            unsigned int max_buff=256;
            char buff[max_buff];
            file.getline(buff,max_buff);
            string line(buff);
            n_line++;
            if(line.size()==0 || line.compare(0,1,"#")==0){
            }else if(line.compare(0,3,"dl=")==0){
                //if(line.indexOf(QRegExp(QString("\".*\"")))!=-1){

                //}
            }else if(line.compare(0,5,"conf=")==0){
                if(!initialised){
                    M3D_ERROR("The scripts asks for configuring before initializing, you have to correct this manually");
                }
                //read conf file path
                //int i=line.indexOf(QRegExp(QString("\".*\"")));
                unsigned int i=line.find("\"",4);
                if(i!=string::npos){
                    string conf_file_name=line.substr(i);
                    if(conf_file_name.size()<=2){
                        error=true;
                        M3D_INFO("string too short");
                    }
                    M3D_INFO("file name to parse: " << conf_file_name);
                    if(!error){
                        conf_file_name=conf_file_name.substr(1,conf_file_name.size()-2);
                        if(conf_file_name.compare(0,1,"/")!=0){
                            conf_file_name=file_path+conf_file_name;
                        }
                    }
                    if(!error && dry_run){
                        ifstream conf_file(conf_file_name.c_str());
                        if(!conf_file){
                            M3D_ERROR("Can't open file "<<conf_file_name);
                            error=true;
                        }
                        conf_file.close();
                    }else{
                        this->initFromConfXML(conf_file_name);
                    }
                }
            }else if(line.compare(0,4,"init")==0){
                initialised=true;
                if(!dry_run){
                    this->setDefaultEnvVar();
                }
            }else if(line.compare(0,3,"run")==0){
                if(!initialised){
                    M3D_ERROR("The scripts asks for running before initializing, you have to correct this manually");
                }
                if(line.compare(4,7,"epsilon")==0){
                    //run the same problem with different epsilon values
                    std::vector<double> eps;
                    std::vector<string> v_str;
                    size_t pos=4+7;
                    int n_runs=1;
                    while(pos != line.size()){
                        if(line[pos]==' '){
                            v_str.push_back(string(""));
                        }else if(v_str.size()){
                            v_str.back().push_back(line[pos]);
                        }
                        pos++;
                    }
                    unsigned int token=0;
                    while(token<v_str.size() && v_str[token][0]!='x'){
                        bool conv_ok;
                        errno=0;
                        double v=strtod(v_str[token].c_str(),0);
                        conv_ok=(errno==0);
                        if(!conv_ok){
                            M3D_ERROR( "MHO: error while reading conf file at line " << n_line <<" : " <<line<< " --- token = "<<v_str[token] );
                            M3D_ERROR( "expected a float value");
                            error=true;
                        }else{
                            eps.push_back(v);
                        }
                        token++;
                    }
                    if(token<v_str.size() && v_str[token][0]=='x'){
                        bool conv_ok;
                        errno=0;
                        double v=strtol(v_str[token].c_str()+1,0,10);
                        conv_ok=(errno==0);
                        if(!conv_ok){
                            M3D_ERROR( "MHO: error while reading conf file at line " << n_line <<" : " <<line<< " --- token = "<<v_str[token] );
                            M3D_ERROR( "expected an integer value after 'x'");
                            error=true;
                        }else{
                            n_runs=v;
                        }
                    }
                    if(!error && !dry_run && n_runs >0 ){
                        bool save_random_f_t=this->_random_first_target_agents;
                        for(unsigned int r=0; r<static_cast<unsigned int>(n_runs); r++){
                            for(unsigned int i=0;i<eps.size();i++){
                                this->setEpsilonForLWAstar(eps[i]);
                                this->resetAgentsConf();
                                this->resetManipulablesConf();
                                this->run();
                                //only first run is made with current setting for randomize first and target agents
                                this->setRandomizeFirstTarget(false);
                            }
                            this->setRandomizeFirstTarget(save_random_f_t);
                        }
                    }else if(!error){
                        std::stringstream ss;
                        for(unsigned int i=0;i<eps.size();i++){
                            ss<<eps[i]<<", ";
                        }
                        M3D_INFO("will run with epsilon = "<<ss.str()<<" - "<<n_runs<<" times");
                    }
                }else{
                    size_t i =line.find_first_of("0123456789");
                    int n_runs=1;
                    if(i!=string::npos){
                        string n_line=line.substr(i);
                        bool conv_ok;
                        errno=0;
                        n_runs=strtol(n_line.c_str(),0,10);
                        conv_ok=(errno==0);
                        if(!conv_ok){
                            M3D_ERROR("MHO: error while reading conf file at line " << n_line <<" : " <<line);
                            error=true;
                        }else{
                            M3D_INFO( "MHO: will run " << n_runs <<" times");
                        }
                    }

                    if(!error && !dry_run){
                        for ( int i=0;i<n_runs;++i){
                            this->resetAgentsConf();
                            this->resetManipulablesConf();
                            this->run();
                        }
                    }
                }
            }else if(line.compare(0,4,"quit")==0){
                if(!dry_run){
                    M3D_INFO( "MHO: Cleanly quiting after reading quit command from script file (.conf)");
                    exit(0);
                }else{
                    do_exit=true;
                }
            }else{
                M3D_ERROR("Error: The script contains invalid data on line "<<n_line<<" : " << line );
                error=true;
            }
        }
        if(dry_run && !error){
            dry_run=false;
            file.clear();
            file.seekg(0);
        }else{
            break;
        }
    }

    //if dry run is still set, there was an error
    return (dry_run ? 1 : 0);
}

boost::mt19937 rng_random_agents;

unsigned int MultiHandOver::run(){
    if(_compute_traj_only){
        this->computeHandOverTrajectory(_solution_to_compute_traj,_step_to_compute_traj);
        _compute_traj_only=false;
    }else if(_play_solution_task){
        this->playSolutionTask(_solution_index_to_draw);
        _play_solution_task=false;
    }else if(_play_solution_task_traj){
        if(_play_solution_sync_step_by_step){
            bool ok=this->playSolutionTaskSyncStep(_solution_index_to_draw);
            _solution_index_to_draw = -1;
            if(!ok){
                _play_solution_sync_step_by_step=false;
                _play_solution_task_traj=false;
            }
        }else{
            unsigned int i=this->_solution_index_to_draw;
            if(i < _solutions_as_tasks.size() ){
                Task * tk = _solutions_as_tasks[i];
                unsigned int j=_solution_step_index_to_draw;
                if( j < tk->getSubTasks().size()){
                    SubTask *stk = tk->getSubTask(j);
                    this->setSubTaskTrajToRobot(stk);
                    //go to the next navigation or manip task:
                    do{
                        _solution_step_index_to_draw++;
                    }while(_solution_step_index_to_draw < tk->getSubTasks().size()
                           && !tk->getSubTask(_solution_step_index_to_draw)->isNav()
                           && !tk->getSubTask(_solution_step_index_to_draw)->isManip());
                }else{
                    _play_solution_task_traj=false; //end
                }
            }
        }
    }else{
        //this->dumpVar();
        //this->createGrid();
        //this->setAgentIndexToDrawGrid(this->_agents.size());
        //this->setDrawAgentAccessibilityGrid(true);
        //g3d_draw_allwin_active();
        if(_random_first_target_agents){
            boost::uniform_int<> unif(0,_agents.size()-1);
            boost::variate_generator<boost::mt19937&, boost::uniform_int<> > die(rng_random_agents, unif);
            int a(0),b(0);
            //avoid having target == first
            while(a==b){
                a=die();
                b=die();
            }
            setFirstAgent(_agents[a]);
            setTargetAgent(_agents[b]);
        }
        initToolSet();
        this->setDrawAgentAccessibilityGrid(false);
        Chronometer chrono;
        vector<double> times(4,0);
        if(_first_run){
            this->computeGrid();
        }
        this->testForReinit();
        times[0]=chrono.getMilliSec();
        if(_first_run){
            ifstream ifile("frontiers.mho");
            bool ok=false;
            if(ifile){
                M3D_INFO("reading frontiers from file");
                ok=this->_grid->loadFrontiers(ifile);
                if(ok){
                    this->_progression_state = max(_progression_state, FRONTIERS_DISTANCE);
                }
                ifile.close();
            }
            if(!ok){
                M3D_INFO( "computing frontiers");
                this->computeFrontiers();
                ofstream file("frontiers.mho");
                this->_grid->saveFrontiers(file);
                file.close();
            }
            _first_run=false;
            M3D_DEBUG( "...done");
        }
        times[1]=chrono.getMilliSec();
        if(_check_frontier_before){
            M3D_INFO( "checking frontiers" );
            this->checkAllMetaFrontiersWithObject();
            M3D_DEBUG( " ...done");
        }
        times[2]=chrono.getMilliSec();
        this->computeAgentsGraph();
        //return 0;
        switch (_search_algo){
        case RANDOM_SEARCH:
            this->searchSolutions();
            break;
        case WA_SEARCH:
            M3D_INFO( "MHO will use Weighted A*");
            M3D_WARN("WARNING: not using Lazy variant, the plan will be slow!");
            API_activeGrid=this->getPlanGrid();
            this->searchSolutionLWAstar(true);
            break;
        case LWA_SEARCH:
            //setDrawLWA(true);
            API_activeGrid=this->getPlanGrid();
            this->searchSolutionLWAstar(false);
            break;
        default:
            break;
        }
        //this->setAgentIndexToDrawGrid(this->_agents.size());
        //this->setDrawAgentAccessibilityGrid(true);
        //g3d_draw_allwin_active();
        //this->setDrawAgentAccessibilityGrid(false);

        times[3]=chrono.getMilliSec();
        //TODO paramerters for stat/info files creation
        //this->printMainTimes(times);
        //this->printSolutionsForest();
        //this->computeHandOverTrajectory(_best_solution);
    }
    return 0;
}
#endif


bool MultiHandOver::getSolutionDataForMHP(std::vector<MHO::MultiHandOverMhpHO> &hand_overs,
                                          std::vector<Eigen::Vector3d, Eigen::aligned_allocator<Eigen::Vector3d> > &path,
                                          double &cost, double &duration,
                                          std::string &main_agent_name){
    hand_overs.clear();
    MultiHandOverSolution *sol;
    if(this->_best_solution){
        sol=_best_solution;
    }else if(this->_solutions.size()>0){
        sol=_solutions.back();//the last computed
        //TODO maybe should not be in state where there is no best_solution state (for now its when computing LWA*)
    }else{
        return false; // no solutions
    }
    for(unsigned int i=0;i<sol->actions.size();++i){
        hand_overs.push_back(MultiHandOverMhpHO(sol,i));
        //cout<<"handover :"<<endl;
        //cout << "\tgiver : " << hand_overs.back().g_pose<<endl;
        //cout << "\treceiver : " << hand_overs.back().r_pose<<endl;
    }
    Robot *rob=global_Project->getActiveScene()->getActiveRobot();
    path.clear();
    main_agent_name=rob->getName();
    if(rob->isAgent() && rob->getHriAgent() && !rob->getHriAgent()->is_human){
        for(unsigned int i=0;i<sol->actions.size();i++){
            MultiHandOverSolution::Action &act=sol->actions[i];
            for(unsigned int ii=0;ii<2;++ii){
                if(act.robot(ii)==rob){
                    MultiHandOver3dPlanner mp(rob,false);
                    std::vector<Eigen::Vector2d,Eigen::aligned_allocator<Eigen::Vector2d> > path2d;
                    for(unsigned int ip=0;ip<act.path(ii).size();ip++){
                        path2d.push_back(act.path(ii)[ip]->getCenter());
                    }
                    mp.computePath2dToPath3dShortCut(path2d,path);
                    break;
                    //TODO the other trajectories for this robot??
                }
            }
        }
    }else{
        M3D_WARN("cannot plan for the active robot :"<<rob->getName());
    }

    cost=sol->cost;
    duration=sol->time;

    return true;

}

//bool MultiHandOver::getTasksForMHP(){
//    return true;
//}

void MultiHandOver::setToComputeHOTraj(bool activate, int solution, int step){
    _compute_traj_only=activate;
    if(activate){
        if(solution>=0 && static_cast<unsigned int>(solution)<_solutions.size()){
            _solution_to_compute_traj=_solutions[solution];
        }else if (solution==-1){
            _solution_to_compute_traj=_best_solution;
        }
        _step_to_compute_traj=step;
    }
}

unsigned int MultiHandOver::findAgents(){
    M3D_DEBUG( "MultiHandOver::findAgents()");
    //reset all data
    this->_agents.clear();
    this->_mho_agents.clear();
    this->_agents_init_confs.clear();
    this->_agents_target_confs.clear();
    this->_first_agent=0;
    this->_target_agent=0;
    this->_target_agent_node=0;
    this->_first_agent_node=0;
    this->_manipulable.clear();

    for (unsigned int i = 0; i < global_Project->getActiveScene()->getNumberOfRobots(); i++)
    {
        Robot *r=global_Project->getActiveScene()->getRobot(i);
        if (r->isAgent()){
            M3D_DEBUG( "found agent "<< r->getName());

            this->_agents.push_back(r);
            //this->_agents_use_costs.push_back(1);
            //this->_agents_speeds.push_back(1);
            this->_mho_agents.push_back(MHO::MultiHandOverAgent(r,1,1,this->_agents.size()-1));
            _mho_agents.back().have_final_position=true;//TODO have_final_position should be a setting
            //this->_agents_init_confs.push_back(*r->getCurrentPos());
            //TODO target might be different
            //this->_agents_target_confs.push_back(*r->getCurrentPos());
        }else{
            this->_manipulable.push_back(r);//TODO find manipulable object with findManipulableObjects(), not here
            this->_manipulables_init_confs.push_back(*r->getCurrentPos());
        }
    }
    this->fetchAgentsConfsInitGoTo();
    return this->_agents.size();
}

unsigned int MultiHandOver::findManipulableObjects(){
    //TODO find manipulable objects
    return _manipulable.size();
}

void MultiHandOver::fetchAgentsConfsInitGoTo(){
    _agents_target_confs.clear();
    _agents_init_confs.clear();
    _manipulables_init_confs.clear();
    for(unsigned int i=0; i <_agents.size();i++){
        this->_agents_init_confs.push_back(*_agents[i]->getCurrentPos());
        this->_agents_target_confs.push_back(*_agents[i]->getGoTo());
    }
    for(unsigned int i=0;i< _manipulable.size();i++){
        _manipulables_init_confs.push_back(*_manipulable[i]->getCurrentPos());
    }
}

/// \brief reset the agents to their initial configuration
void MultiHandOver::resetAgentsConf(){
    for (unsigned int i=0;i<_agents.size();++i){
        _agents[i]->setAndUpdate(_agents_init_confs[i]);
    }
}
void MultiHandOver::resetManipulablesConf(){
    for (unsigned int i=0;i<_manipulable.size();++i){
        _manipulable[i]->setAndUpdate(_manipulables_init_confs[i]);
    }
}

Configuration MultiHandOver::getAgentInitConf(unsigned int a){
    if(a>=_agents_init_confs.size()){
        throw exception();
    }else{
        return _agents_init_confs[a];
    }
}

Configuration MultiHandOver::getAgentTargetConf(unsigned int a){
    if(a>=_agents_target_confs.size()){
        throw exception();
    }else{
        return _agents_target_confs[a];
    }
}
MultiAgentCell* MultiHandOver::getAgentTargetCell(unsigned int a){
    if(! getAgentStruct(a)->have_final_position){
        return 0;
    }
    Configuration q=getAgentTargetConf(a);
    Eigen::Vector2d v(q.at(6+0),q.at(6+1));
    return dynamic_cast<MultiAgentCell*>(_grid->getCell(v));
}

MultiAgentCell* MultiHandOver::getAgentInitCell(unsigned int a){
    Configuration q=getAgentInitConf(a);
    Eigen::Vector2d v(q.at(6+0),q.at(6+1));
    return dynamic_cast<MultiAgentCell*>(_grid->getCell(v));
}

void MultiHandOver::setTargetAgent(std::string name){
    for (unsigned int i=0;i<this->_agents.size();++i){
        if(_agents[i]->getName() == name){
            this->setTargetAgent(_agents[i]);
            break;
        }
    }
}
void MultiHandOver::setFirstAgent(std::string name){
    for (unsigned int i=0;i<this->_agents.size();++i){
        if(_agents[i]->getName() == name){
            this->setFirstAgent(_agents[i]);
            break;
        }
    }
}
void MultiHandOver::setTargetAgent(p3d_rob *rob){
    for (unsigned int i=0;i<this->_agents.size();++i){
        if(_agents[i]->getRobotStruct() == rob){
            this->setTargetAgent(_agents[i]);
            break;
        }
    }
}
void MultiHandOver::setFirstAgent(p3d_rob *rob){
    for (unsigned int i=0;i<this->_agents.size();++i){
        if(_agents[i]->getRobotStruct() == rob){
            this->setFirstAgent(_agents[i]);
            break;
        }
    }
}

void MultiHandOver::setObject(p3d_rob *obj){
    for (unsigned int i=0;i<this->_manipulable.size();++i){
        if(_manipulable[i]->getRobotStruct() == obj){
            this->setObject(_manipulable[i]);
            break;
        }
    }
}
void MultiHandOver::setFirstAgentNode(){
    for (vector<mho::Node*>::iterator it=_agents_nodes.begin();it!=_agents_nodes.end();++it){
        if(_agents[(*it)->getId()]==this->_first_agent){
            _first_agent_node=*it;
            return;
        }
    }
}

void MultiHandOver::setTargetAgentNode(){
    for (vector<mho::Node*>::iterator it=_agents_nodes.begin();it!=_agents_nodes.end();++it){
        if(_agents[(*it)->getId()]==this->_target_agent){
            _target_agent_node=*it;
            return;
        }
    }
}

double MultiHandOver::getAgentUseCost(Robot *r){
    vector<Robot*>::iterator f=find(_agents.begin(),_agents.end(),r);
    double c=numeric_limits<double>::infinity();
    if(f!=_agents.end()){
        unsigned int i= f-_agents.begin();
        c = this->getAgentUseCost(i);
    }
    return c;
}

double MultiHandOver::getTimeHO(unsigned int a1, unsigned int a2){
    MultiHandOverAgent *m1,*m2;
    m1=&_mho_agents[a1];
    m2=&_mho_agents[a2];
    if(m1->is_human && m2->is_human){
        return _h_h_ho_time;
    }else if(!m1->is_human && !m2->is_human){
        return _r_r_ho_time;
    }else{
        return _r_h_ho_time;
    }
}

void MultiHandOver::computeLongestSolutionDistance(){
    double dist(0),time(0),use_cost(0);
    //TODO we have to take into account if the agents have to go back or not
    //TODO we can make a more precise estimation, because we are sure that the first
    //traj is at most dmax and the last if it goes back to initial is dmax (for each
    //agent)
    int max_paths_per_agent=2;
    //if go back to init, max_paths_per_agent=3
    for (unsigned int i=0;i<this->getAgents().size();++i){
        if(this->getAgentUseCost(i) < numeric_limits<double>::infinity()){
            //2*dmax is an upper bound for the real Dmax from any point
            double tmax,cmax;
            double dmax=max_paths_per_agent*2*this->_grid->getAgentStructAt(i)->getMaxDist();
            tmax=dmax/this->getAgentSpeed(i);
            cmax=tmax*this->getAgentUseCost(i);
            dist+=dmax;
            time+=tmax;
            use_cost+=cmax;
        }
    }
    this->_max_dist=dist;
    this->_max_time=time;
    this->_max_use_cost=use_cost;
    M3D_INFO("maximum measures for cost normalisation :"
           "\n\t* dist = "<<_max_dist
        << "\n\t* time = "<<_max_time
        << "\n\t* use = "<<_max_use_cost);
}

unsigned int MultiHandOver::computeAgentsGraph(void){
    //TO DO remove this awfull display:
    //ofstream of_graph;
    //of_graph.open("graph.dot");
    M3D_DEBUG( "MultiHandOver::computeAgentsGraph()");
    if(_first_agent==0 || _target_agent==0){
        M3D_ERROR( "MultiHandOver::computeAgentsGraph() : first and target agents not set");
        return 0;
    }
    this->_agents_frontiers_dist_graph.clear();
    this->_agents_nodes.clear();
    for(unsigned int i=0;i<this->_agents.size();++i){
        mho::Node n=mho::Node(i);
        _agents_nodes.push_back(dynamic_cast<mho::Node*>(_agents_frontiers_dist_graph.addNode(&n)));
        if(_agents[i]==_first_agent){
            _first_agent_node=_agents_nodes.back();
        }
        if (_agents[i]==_target_agent){
            _target_agent_node=_agents_nodes.back();
        }
    }

    //TO DO remove this awfull display:
    //of_graph<<"Graph AgentsGraph{"<<endl;
    for(unsigned int i=0;i<this->_agents_nodes.size();++i){
        //of_graph<<"   "<<_agents[_agents_nodes[i]->getId()]->getName()<<" [ pos = \""<< -_grid->getAgentStructAt(i)->reachable()[0]->getCenter().x()/3 <<","<< -_grid->getAgentStructAt(i)->reachable()[0]->getCenter().y()/4 <<"!\"]" <<endl;
        for(unsigned int j=0;j<i;++j){
            if(this->handOverPossible(_agents_nodes[i]->getId(),_agents_nodes[j]->getId())){
                _agents_frontiers_dist_graph.connectNodes(_agents_nodes[i],_agents_nodes[j],1.);
                //TO DO remove this awfull display:
                //of_graph<<"   "<<_agents[_agents_nodes[i]->getId()]->getName()<<"--"<<_agents[_agents_nodes[j]->getId()]->getName()<<";"<<endl;
            }
        }

    }
    //TO DO remove this awfull display:
    //of_graph<<"}"<<endl;
    //of_graph.close();
    //system("sed -i 's/HERAKLES_HUMAN/H/g;s/PR2_ROBOT/R/g' graph.dot;dot -Kfdp -Tpng graph.dot > graph.png;eog graph.png &");

    M3D_DEBUG( "MultiHandOver::computeAgentsGraph() found "<<this->_agents_frontiers_dist_graph.size()<<" agents.");

    return this->_agents_frontiers_dist_graph.size();
}

unsigned int MultiHandOver::computeRawSolutions(float cost_majoration){
    M3D_DEBUG( "MultiHandOver::computeRawSolutions()");
    M3D_DEBUG( "first agent is "<< _first_agent->getName()<< ", target is  "<< _target_agent->getName());
    _raw_solution.clear();
    pair<double,deque<mho::NodeInterface*> > dijk_ret = mho::dijkstra(&_agents_frontiers_dist_graph,_first_agent_node,_target_agent_node);
    M3D_DEBUG("dijkstra OK");
    M3D_DEBUG("computing all paths...");
    if(dijk_ret.second.size()!=0){
        _raw_solution = mho::enumerate_paths(&_agents_frontiers_dist_graph,_first_agent_node,_target_agent_node,dijk_ret.first*cost_majoration);
    }

    M3D_DEBUG( "MultiHandOver::computeRawSolutions() found "<<_raw_solution.size() << " solutions.");

    return _raw_solution.size();
}

bool MultiHandOver::handOverPossible(unsigned int a1, unsigned int a2){
    bool possible(false);
    if(getAgentUseCost(a1)+getAgentUseCost(a2) < numeric_limits<double>::infinity()){
        switch (_progression_state){
        case FRONTIERS_DISTANCE:
            //if(_grid->getFrontier(a1,a2).size()>0){
            if(_grid->getFrontierCells(a1,a2).size()>0 || !_grid->getAgentStructAt(a1)->frontierFullyChecked(_grid->getAgentStructAt(a2))){
                possible=true;
            }
            break;
        default:
            M3D_DEBUG("handOverPossible: no information");
            possible=false;
        }
    }
    return possible;
}

void MultiHandOver::initGrid(){
    std::vector<double> envsize(4);
    // initCostSpace
    envsize[0] = XYZ_ENV->box.x1; envsize[1] = XYZ_ENV->box.x2;
    envsize[2] = XYZ_ENV->box.y1; envsize[3] = XYZ_ENV->box.y2;
    if(!_grid || this->getGridCellSize() != _grid->getCellSize().x() || this->getGridCellSize() != _grid->getCellSize().y()){
        this->_grid=new MultiAgentGrid(this->_cellSize,envsize,this->_agents);
    }else{
        _grid->setAgents(_agents);
        _grid->resetData();
    }
    this->_grid->init();
}

void MultiHandOver::computeGrid(void){
    if(this->getGridCellSize() != _grid->getCellSize().x() || this->getGridCellSize()!=_grid->getCellSize().y()){
        M3D_DEBUG( "(re-)computing grid because of different cell size set in MHO");
        this->initGrid();
    }
    //this->getPlanGrid()->computeAccessibility();
    //this->getPlanGrid()->computeDistances();
    this->getPlanGrid()->computeAccessAndDistCombined();
    //TODO other stuff
}

double MultiHandOver::getShoulderHeight(unsigned int agent){
    //Robot *r=_agents[agent];
    //HRI_MANIP* manip=r->getHriAgent()->manip;
    //return (*r->getJoint(manip->shoulder_joint_ids[0])->getAbsPos())[2][3];
    return getAgentStruct(agent)->shoulderHeight();
}

void MultiHandOver::checkAllMetaFrontiersWithObject(){
    map<MultiAgentGrid::MetaAgentInGrid::s_ptr,unsigned int> meta_agents;
    for (unsigned int a1=0;a1<_agents.size();++a1){
        meta_agents[_grid->getAgentStructAt(a1)->getMetaAgent()]=a1;
    }
    M3D_DEBUG("checkAllMetaFrontiersWithObject : "<<meta_agents.size() << " meta agents");
    for(map<MultiAgentGrid::MetaAgentInGrid::s_ptr,unsigned int>::iterator a1=meta_agents.begin();a1!=meta_agents.end();++a1){
        //find frontiers with all other meta agents and itself
        for(map<MultiAgentGrid::MetaAgentInGrid::s_ptr,unsigned int>::iterator a2=a1;a2!=meta_agents.end();++a2){
            this->checkMetaFrontiersWithObject(a1->second,a2->second);
        }
    }
}

bool MultiHandOver::checkMetaFrontiersWithObject(unsigned int a1,unsigned int a2){
    bool ret=false;
    MultiAgentGrid::MetaAgentInGrid::s_ptr m1=_grid->getAgentStructAt(a1)->getMetaAgent(),
            m2=_grid->getAgentStructAt(a2)->getMetaAgent();

    vector<MultiAgentCell*> & fcells=m1->frontiers(m2);
    for(unsigned int i_cell=0;i_cell<fcells.size();++i_cell){
        MultiAgentCell *c=fcells[i_cell];
        vector<MetaFrontierPtr> & mfrts=c->getMetaFrontier(a1,m2);
        for(unsigned int i_frts=0;i_frts<mfrts.size();++i_frts){
            if(checkMetaFrontierWithObject(a1,a2,mfrts[i_frts])){
                //the frontier is ok
                m1->frontiers_ok(m2).push_back(static_cast<MetaFrontierPtr>(mfrts[i_frts])->cellOf(m1));
                m2->frontiers_ok(m1).push_back(static_cast<MetaFrontierPtr>(mfrts[i_frts])->cellOf(m2));
                ret=true;
            }
        }
    }
    return ret;
}

bool MultiHandOver::checkMetaFrontierWithObject(unsigned int a1,unsigned int a2,MetaFrontierPtr mf){
    bool st=false;
    //FrontierPtr f=mf->getCellA()->createRealFrontierFor(a1,a2,mf);
    FrontierPtr f=mf->cellOf(_grid->getAgentStructAt(a1)->getMetaAgent())->createRealFrontierFor(a1,a2,mf);
    st=_tool_set->frontierTester->test(f,FrontierCells::FCS_OBJECT);
    //st=checkFrontierWithObject(*f);
    if(st){
        mf->setCheckedOk(true);
        mf->setObjectPosition(f->object_position);
    }
    return st;
}

bool MultiHandOver::randomCheckFrontiersWithObject(shared_ptr<FrontierCells> & frontier_found, unsigned int a1, unsigned int a2,unsigned int max_it){
    bool found=false;
    unsigned int k=0;
    //HandOverFinder hofinder(this,_grid,a1,a2,_grid->getAgentStructAt(a1)->getOrigin(),_grid->getAgentStructAt(a2)->getOrigin(),0,0 );
    //hofinder.setCostFunction(this->navigationRelatedCost );
    //hofinder.findFrom2Points();
    while(!found && k++<max_it){
        if(_check_frontier_before){
            vector<MultiAgentCell*> &fc=_grid->getFrontierCheckedOk(a1,a2);
            if(fc.size()){
                MultiAgentCell* cell=fc[rand()%fc.size()];
                vector<MetaFrontierPtr> meta_frontiers =cell->getMetaFrontier(a1,a2);
                if(meta_frontiers.size()){
                    frontier_found=cell->createRealFrontierFor(a1,a2,meta_frontiers[rand()%meta_frontiers.size()]);
                    found=_tool_set->frontierTester->test(frontier_found,FrontierCells::FCS_OBJECT);
                    //found=checkFrontierWithObject(*frontier_found);
                    assert(frontier_found->check_status>=FrontierCells::FCS_OBJECT);
                }
            }
        }else{
            frontier_found = _grid->getRandomFrontier(a1,a2);
            if(frontier_found){
                //found=checkFrontierWithObject(*frontier_found);
                found=_tool_set->frontierTester->test(frontier_found,FrontierCells::FCS_OBJECT);
            }
        }
    }
    return found;
}

Eigen::Vector3d point_to_Vector3d(p3d_point &p){
    Eigen::Vector3d v;
    v[0]=p.x;
    v[1]=p.y;
    v[2]=p.z;
    return v;
}


Configuration MultiHandOver::confForBBAxisAlign(Robot * object, Eigen::Vector3d const & main_dir, Eigen::Vector3d const & second_dir){
    using namespace Eigen;
    Vector3d dir1=main_dir,dir2=second_dir,dir3=-main_dir.cross(second_dir);
    if(this->_object_bb_vects.size()==0){
        p3d_oriented_BB * bb= p3d_get_oriented_BB_obj(*_object->getRobotStruct()->o); // the bounding box of the object
        //the 3 main vectors describing the bounding box
        _object_bb_vects.resize(3);
        _object_bb_vects[0] = point_to_Vector3d(bb->BottomFrontRight) - point_to_Vector3d(bb->BottomBackRight);//front x
        _object_bb_vects[1] = point_to_Vector3d(bb->BottomFrontLeft) - point_to_Vector3d(bb->BottomFrontRight);//left y
        _object_bb_vects[2] = point_to_Vector3d(bb->UpFrontRight) - point_to_Vector3d(bb->BottomFrontRight);//up z
        vector<double> areas(3);
        areas[0]=_object_bb_vects[2].cross(_object_bb_vects[1]).norm();
        areas[1]=_object_bb_vects[2].cross(_object_bb_vects[0]).norm();
        areas[2]=_object_bb_vects[1].cross(_object_bb_vects[0]).norm();
        _object_directions_sorted_for_ho.clear();
        _object_directions_sorted_for_ho.reserve(3);
        for(unsigned int i=0;i<3;i++){
            for(unsigned int j=0;j<3;j++){
                if(i!=j && areas[i]<=areas[j]){
                    for(unsigned int k=0;k<3;k++){
                        if(j!=k && areas[j]<=areas[k]){
                            _object_directions_sorted_for_ho.push_back(i);
                            _object_directions_sorted_for_ho.push_back(j);
                            _object_directions_sorted_for_ho.push_back(k);
                            break;
                        }
                    }
                    if(_object_directions_sorted_for_ho.size()!=0){break;}
                }
            }
            if(_object_directions_sorted_for_ho.size()!=0){break;}
        }
    }

    p3d_matrix4 xform={{1, 0, 0, 0}, {0, 1, 0, 0}, {0, 0, 1, 0}, {0, 0, 0, 1}};
    dir1.normalize();
    dir2.normalize();
    dir3.normalize();
    xform[0][_object_directions_sorted_for_ho[0]]=dir1.x();
    xform[1][_object_directions_sorted_for_ho[0]]=dir1.y();
    xform[2][_object_directions_sorted_for_ho[0]]=dir1.z();

    xform[0][_object_directions_sorted_for_ho[1]]=dir2.x();
    xform[1][_object_directions_sorted_for_ho[1]]=dir2.y();
    xform[2][_object_directions_sorted_for_ho[1]]=dir2.z();

    xform[0][_object_directions_sorted_for_ho[2]]=dir3.x();
    xform[1][_object_directions_sorted_for_ho[2]]=dir3.y();
    xform[2][_object_directions_sorted_for_ho[2]]=dir3.z();

    Configuration conf = *object->getCurrentPos();
    double foo[3];
    p3d_mat4ExtractPosReverseOrder(xform,&foo[0],&foo[1],&foo[2],&conf[9],&conf[10],&conf[11]);
    return conf;
}

void MultiHandOver::setOrientationObjectForHO(Eigen::Vector3d & dir1){
    using namespace Eigen;
    vector<Vector3d> dirs(3);
    dirs[0]=Vector3d(1,0,0);
    dirs[1]=Vector3d(0,1,0);
    dirs[2]=Vector3d(0,0,1);

    //Configuration q=this->confForBBAxisAlign(_object,dirs[i],dirs[j]);
    Vector3d dir2;
    dir2=dirs[2]/*z*/.cross(dir1);
    dir2=dir2.cross(dir1);
    if(dir2.dot(dirs[2])<0){
        dir2=-dir2;
    }
    //<=> dir2 = z x d1 x d1
    Configuration q=this->confForBBAxisAlign(_object,dir1,dir2);

    _object->setAndUpdate(q);
}

bool MultiHandOver::checkFrontierWithObject(FrontierCells & frontier){
    assert(!"OBSOLETE FUNCTION");
    using namespace Eigen;
    bool ret=false;
    if(!_object){ cerr<< __FUNCTION__ << " : no object defined"<<endl; return false;}
    if(!frontier.cellA()->isReachable(frontier.agentA_ind) || !frontier.cellB()->isReachable(frontier.agentB_ind)){
        M3D_ERROR( ""<<__FUNCTION__ << " : invalid frontier (Reachability check failed)");
        return false;
    }
    assert(frontier.cellA()->isReachable(frontier.agentA_ind) && frontier.cellB()->isReachable(frontier.agentB_ind));
    if(frontier.check_status<FrontierCells::FCS_OBJECT){
        MHO::AbstracHandOverObjectCheck* ho_check=new MHO::HandOverDirectChecker(_object,frontier.agentA->robot(),frontier.agentB->robot());
        unsigned int a1=frontier.agentA_ind;
        unsigned int a2=frontier.agentB_ind;
        Eigen::Vector3d start_pos,end_pos;
        start_pos[0]=frontier.cellA()->getCenter()[0];
        start_pos[1]=frontier.cellA()->getCenter()[1];
        end_pos[0]=frontier.cellB()->getCenter()[0];
        end_pos[1]=frontier.cellB()->getCenter()[1];
        start_pos[2]=getShoulderHeight(a1);
        end_pos[2]=getShoulderHeight(a2);
        ho_check->reach(this->_mho_agents[a1].reach(),this->_mho_agents[a2].reach());
        ho_check->shoulders(start_pos,end_pos);
        ho_check->timeLimit(1);
        // ##############################################
        //TODO redondant
        // ##############################################
        // ##############################################
        bool ret_ho_check=ho_check->check();

        p3d_col_deactivate_robot_robot(frontier.agentA->robot()->getRobotStruct(),_object->getRobotStruct());
        p3d_col_deactivate_robot_robot(frontier.agentB->robot()->getRobotStruct(),_object->getRobotStruct());

        Configuration q_save = *_object->getCurrentPos();

        Eigen::Vector3d direction = end_pos - start_pos;

        this->setOrientationObjectForHO(direction);

        shared_ptr<Configuration> q_temp = _object->getCurrentPos();
        shared_ptr<Configuration> q_middle(new Configuration(_object));

        double length=direction.norm();
        double step_size = _object_bb_vects[_object_directions_sorted_for_ho[0]].norm();
        double dist_thresh=0.03;//TODO distance tolerance for object moves to check frontiers
        double middle_reach=(length-_grid->getArmReach(a2)+_grid->getArmReach(a2))/2;
        unsigned int iteration(0);
        bool collision=false;
        direction.normalize();

        q_temp->at(0+6)=start_pos[0];
        q_temp->at(1+6)=start_pos[1];
        q_temp->at(2+6)=start_pos[2];

        while(!collision && iteration*step_size<length+step_size){
            q_temp->at(0+6)+=direction[0]*step_size;
            q_temp->at(1+6)+=direction[1]*step_size;
            q_temp->at(2+6)+=direction[2]*step_size;
            _object->setAndUpdate(*q_temp);
            if(_draw_object_frontiers_checks){
                g3d_draw_allwin_active();
            }
            p3d_vector3 v1,v2;//only for arguments, unused
            double dist=p3d_col_robot_environment_distance(_object->getRobotStruct(),v1,v2);
            if(dist < dist_thresh || _object->isInCollisionWithOthersAndEnv()){
                collision=true;
            }
            /*unsigned int i_rob(0);
            while(!collision && i_rob<XYZ_ENV->nr){
                //check the distances with other robots/objects
                if(XYZ_ENV->robot[i_rob] != frontier.agentA->robot()->getRobotStruct() && XYZ_ENV->robot[i_rob] != frontier.agentB->robot()->getRobotStruct()){
                    dist=p3d_col_robot_robot_distance(_object->getRobotStruct(),XYZ_ENV->robot[i_rob],v1,v2);
                    if(dist<dist_thresh){
                        collision=true;
                    }
                }
                ++i_rob;
            }*/
            if(iteration*step_size<=middle_reach && (iteration+1)*step_size>middle_reach){
                *q_middle=*q_temp;
            }
            iteration++;
        }
        //cout<<endl;
        frontier.check_status=FrontierCells::FCS_OBJECT;
        //frontier.object_ok=!collision;
        frontier.status_ok[FrontierCells::FCS_OBJECT]=!collision;
        assert(!collision == ret_ho_check);
        if(collision){
        }else{
            Eigen::Vector3d mid(q_middle->at(6+0),q_middle->at(6+1),q_middle->at(6+2));
            frontier.object_position=q_middle;
            frontier.grabA=mid-direction*step_size/2;
            frontier.grabB=mid+direction*step_size/2;
            //cout <<frontier.object_position->at(0)<<frontier.object_position->at(1)<<frontier.object_position->at(2)<<endl;
        }
        _object->setAndUpdate(q_save);
        if(collision && this->_use_rrt_for_frontier_check){
            MHO::HandOverWindowCheckerRRT hochecker(_object,frontier.agentA->robot(),frontier.agentB->robot(),true,start_pos,end_pos,_grid->getArmReach(a1),_grid->getArmReach(a2));
            hochecker.timeLimit(0.1);
            bool is_ok=hochecker.check();
            if(is_ok){
                frontier.object_position = hochecker.getQMiddle();

                Eigen::Vector3d mid(q_middle->at(6+0),q_middle->at(6+1),q_middle->at(6+2));
                frontier.grabA=mid;
                frontier.grabB=mid;
            }
            ret= is_ok;
        }else{
            ret= !collision;
        }
        p3d_col_activate_robot_robot(frontier.agentA->robot()->getRobotStruct(),_object->getRobotStruct());
        p3d_col_activate_robot_robot(frontier.agentB->robot()->getRobotStruct(),_object->getRobotStruct());
    }else{
        ret= frontier.status_ok[FrontierCells::FCS_OBJECT];//=object_ok;
    }
    return ret;
}

bool MultiHandOver::handOverFindConfs(FrontierPtr frontier){
    assert("OBSOLETE FUNCTION");
    bool ok=true;
    Eigen::Vector3d point;
    configPt q=0;
    if(frontier->check_status>=FrontierCells::FCS_GRAB_CONFS){
        if(!frontier->status_ok[FrontierCells::FCS_GRAB_CONFS]){
            ok=false;
        }else{
            //TODO costs
            ok=true;
        }
    }else{
        //M3D_INFO("check conf HO begin");
        //TODO check this out: (remove in the if/else at the begining of the for)
        //p3d_col_deactivate_robot_robot(frontier->agentA->robot()->getRobotStruct(),frontier->agentB->robot()->getRobotStruct());
        p3d_col_deactivate_robot(this->_object->getRobotStruct());
        for(unsigned int i_ag=0;i_ag<2;i_ag++){
            MultiAgentCell::AgentInCell *agent;
            MultiAgentCell *cell;
            confPtr_t frontier_conf;
            if(i_ag==0){
                agent=frontier->agentA;
                p3d_col_deactivate_robot_robot(agent->robot()->getRobotStruct(),frontier->agentB->robot()->getRobotStruct());
                cell=frontier->cellA();
                point=frontier->grabA;
                frontier->confA=confPtr_t(new Configuration(agent->robot()));
                frontier_conf=frontier->confA;
            }else{
                agent=frontier->agentB;
                //reactivate colisions to avoid confs where agents are colliding
                p3d_col_activate_robot_robot(agent->robot()->getRobotStruct(),frontier->agentA->robot()->getRobotStruct());
                cell=frontier->cellB();
                point=frontier->grabB;
                frontier->confB=confPtr_t(new Configuration(agent->robot()));
                frontier_conf=frontier->confB;
            }
            Configuration q_current_pos=*agent->robot()->getCurrentPos();
            MHO::MultiHandOver3dPlanner mhopl(agent->robot(),agent->agentInGrid->isHuman());
            //Configuration q_current_pos(*agent->robot()->getCurrentPos());
            //if(!agent->agentInGrid->isHuman && mhopl.grabConfCheck(q,cell->getCenter(),point) ){
            Eigen::Vector3d shoulder(cell->getCenter().x(),cell->getCenter().y(),_mho_agents[agent->index].shoulderHeight());
            if((shoulder-point).norm()>_mho_agents[agent->index].reach()){
                ok=false;
                //M3D_INFO( "object is too far to be grasp");
            }else if(mhopl.grabConfCheck(q,cell->getCenter(),point)){
                Configuration q_grab(agent->robot(),p3d_copy_config(agent->robot()->getRobotStruct(),q));
                agent->robot()->setAndUpdate(q_grab);
                *frontier_conf=q_grab;
                p3d_destroy_config(agent->robot()->getRobotStruct(),q);
                q=0;
                //TODO add conf cost, handover cost...
                //handover_cost=10;
                if(_draw_agents_holding_conf_during_search){
                    g3d_draw_allwin_active();
                }
            }else{
                //handover_cost=numeric_limits<double>::infinity();
                ok=false;
            }
            //TODO give the possibility not to put the agent at its previous conf when checking grab confs
            agent->robot()->setAndUpdate(q_current_pos);// (or not)
            if(_draw_agents_holding_conf_during_search){
                g3d_draw_allwin_active();
            }
            if(!ok){
                //M3D_INFO( "HO find conf fail at "<<i_ag<<" -- "<<agent->robot()->getName());
                break;//avoid computing too much
            }
        }
        frontier->check_status=FrontierCells::FCS_GRAB_CONFS;
    }
    p3d_col_activate_robot_robot(frontier->agentA->robot()->getRobotStruct(),frontier->agentB->robot()->getRobotStruct());
    p3d_col_activate_robot(this->_object->getRobotStruct());
    assert(q==0 && "q (double[]) have not been free'd");
    return ok;
}

bool MultiHandOver::handOverFindConfs(MHO::MultiHandOverSolution * solution, int frontier_index){
    assert("OBSOLETE FUNCTION");
    unsigned int start(0),end(solution->frontiers.size());
    bool ok=true;
    if(frontier_index>=0 && static_cast<unsigned int>(frontier_index)<solution->frontiers.size()){
        start=frontier_index; end=frontier_index+1;
    }
    //these loops contains break instructions /!\ --
    for (unsigned int i=start;i<end;++i){
        shared_ptr<FrontierCells> frontier = solution->frontiers[i];
        Eigen::Vector3d point;
        configPt q=0;
        double handover_cost=0,handover_time=0;
        if(frontier->check_status>=FrontierCells::FCS_GRAB_CONFS){
            if(!frontier->status_ok[FrontierCells::FCS_GRAB_CONFS]){
                ok=false;
            }else{
                //TODO costs
                ok=true;
            }
        }else{
            for(unsigned int i_ag=0;i_ag<2;i_ag++){
                MultiAgentCell::AgentInCell *agent;
                MultiAgentCell *cell;
                confPtr_t save_conf;
                if(i_ag==0){
                    agent=frontier->agentA;
                    p3d_col_deactivate_robot_robot(agent->robot()->getRobotStruct(),frontier->agentB->robot()->getRobotStruct());
                    cell=frontier->cellA();
                    point=frontier->grabA;
                    frontier->confA=confPtr_t(new Configuration(agent->robot()));
                    save_conf=frontier->confA;
                }else{
                    agent=frontier->agentB;
                    p3d_col_activate_robot_robot(agent->robot()->getRobotStruct(),frontier->agentA->robot()->getRobotStruct());
                    cell=frontier->cellB();
                    point=frontier->grabB;
                    frontier->confB=confPtr_t(new Configuration(agent->robot()));
                    save_conf=frontier->confB;
                }
                MHO::MultiHandOver3dPlanner mhopl(agent->robot(),agent->agentInGrid->isHuman());
                //Configuration q_current_pos(*agent->robot()->getCurrentPos());
                //if(!agent->agentInGrid->isHuman && mhopl.grabConfCheck(q,cell->getCenter(),point) ){
                if(mhopl.grabConfCheck(q,cell->getCenter(),point) ){
                    Configuration q_grab(agent->robot(),p3d_copy_config(agent->robot()->getRobotStruct(),q));
                    agent->robot()->setAndUpdate(q_grab);
                    *save_conf=q_grab;
                    p3d_destroy_config(agent->robot()->getRobotStruct(),q);
                    q=0;
                    //TODO add conf cost, handover cost...
                    //handover_cost=10;
                    if(_draw_agents_holding_conf_during_search){
                        g3d_draw_allwin_active();
                    }
                }else{
                    //handover_cost=numeric_limits<double>::infinity();
                    ok=false;
                }
                //TODO put back : agent->robot()->setAndUpdate(q_current_pos);// (or not)
                if(_draw_agents_holding_conf_during_search){
                    g3d_draw_allwin_active();
                }
                if(!ok){
                    p3d_col_activate_robot_robot(frontier->agentA->robot()->getRobotStruct(),frontier->agentB->robot()->getRobotStruct());
                    break;//avoid computing too much
                }
            }
            frontier->check_status=FrontierCells::FCS_GRAB_CONFS;
        }
        //TODO remove this check, change above code with this function
        //frontier->check_status=FrontierCells::FCS_OBJECT;
        //bool other_ok=this->handOverFindConfs(frontier);
        // //assert(other_ok==ok);
        //if(other_ok!=ok){
        //    M3D_INFO("bad handoverfindconfs");
        //}
        //ok=other_ok;
        // -------
        if(ok){
            MHO::MultiHandOverHO ho(&solution->actions[i],this);
            handover_cost=ho.computeCost();
            handover_time=ho.getTime();
            //whether it's a new solution or a change in the solution, we add or modify the cost
            if(frontier_index<0){
                solution->handover_times.push_back(handover_time);
                solution->handover_hri_costs.push_back(handover_cost);
            }else{
                solution->handover_times[i]=(handover_time);
                solution->handover_hri_costs[i]=(handover_cost);
            }
            assert(frontier->confA->at(1) + frontier->confB->at(1)+1);//only check if confs are set
            frontier->status_ok[FrontierCells::FCS_GRAB_CONFS]=true;
            assert(q==0 && "q (double[]) have not been free'd");
        }else{
            break;//avoid computing too much
        }
    }
    return ok;
}

//may keep a solution if the overall cost is higher or =, but some local cost is improved
MHO::MultiHandOverSolution* MultiHandOver::changeHandOverSolution(MHO::MultiHandOverSolution * solution){
    MultiHandOverSolution * new_solution = solution->cloneAsSon();
    assert(new_solution->parent==solution);
    unsigned int i=rand()%new_solution->actions.size();
    MultiHandOverSolution::Action & act=new_solution->actions[i];

    //select a neighboor cell for each agent
    FrontierPtr frontier = act.frontier()->neighbourFrontier(_grid->getHandOverReach(act.frontier()->agentA_ind,act.frontier()->agentB_ind));
    //if(frontier && checkFrontierWithObject(*frontier)){
    if(frontier && _tool_set->frontierTester->test(frontier,FrontierCells::FCS_OBJECT)){
        new_solution->frontiers[i]=frontier;//change to the new frontier
        //reset costs
        new_solution->cost=0;
        new_solution->handover_times[i]=0;
        for(unsigned int i_ho=0;i_ho<new_solution->handover_times.size();++i_ho){
            new_solution->cost+=new_solution->handover_times[i_ho];
        }
        if(this->handOverFindConfs(new_solution,i)){//check if there is a grab conf ok for this frontier and add its cost
            new_solution->navigation_trajs.clear();
            new_solution->nav_dists.clear();
            //new_solution->actions.clear();
            this->computeNavigationTrajectories(new_solution);//TODO avoid computing ALL trajectories, only 3 need to be changed
            this->computeSolutionCost(new_solution);//TODO check for LOCAL cost improvment
            this->insert_solution_sorted(new_solution);

            if(new_solution->cost<this->_best_solution->cost){this->_best_solution=new_solution;}
            //return new_solution->cost < solution->cost;
            return new_solution;
        }
    }
    return 0;//no new solution OK
}
MHO::MultiHandOverSolution* MultiHandOver::optimizeSolution(MHO::MultiHandOverSolution* solution){
    MultiHandOverSolution* current=solution->cloneAsSon();
    bool save_use_rrt=_use_rrt_for_frontier_check;
    this->setUseRrtInFrontierCheck(false);
    //if (solution->actions.size()>2){return current;}
#ifdef MHO_OLD_OPTIMIZE_SOLUTION
    for(unsigned int i=0;i<solution->actions.size();++i){
        MultiHandOverSolution::Action & act=solution->actions[i];
        //if the action is a frontier in open space (the 2 agents are in same reachable space)
        //we try to optimize the frontier
        if( act.frontier()->cellA()->isReachable(act.frontier()->agentB_ind) || act.frontier()->cellB()->isReachable(act.frontier()->agentA_ind)){
            current=this->optimizeFrontier(current,i,act.agent(0)->reach()+act.agent(1)->reach(),0);
        }
    }
#else
    double cost_improvement(0),prev_cost(solution->cost);
    int n=0;//used for stop condition when no improvement
    do{
        unsigned int i=rand()%current->actions.size();
        MultiHandOverSolution::Action & act=current->actions[i];
        //if the action is a frontier in open space (the 2 agents are in same reachable space)
        //we try to optimize the frontier
        if( act.frontier()->cellA()->isReachable(act.frontier()->agentB_ind) || act.frontier()->cellB()->isReachable(act.frontier()->agentA_ind)){
            current=this->optimizeFrontier(current,i,act.agent(0)->reach()+act.agent(1)->reach(),0);
        }else{
            current=this->optimizeFrontier(current,i,act.agent(0)->reach()+act.agent(1)->reach(),0);
        }
        cost_improvement=prev_cost-current->cost;
        prev_cost=current->cost;
        if(cost_improvement < solution->cost * 0.01){
            n++;
        }else{
            n=std::max(0.0,n-static_cast<double>(current->actions.size())/3*2);//TODO this constant factor should be settable
        }
    }while(current->actions.size()>1 && static_cast<unsigned int>(n)<current->actions.size());
    //if only one action, its useless to run several times
#endif
    this->insert_solution_sorted(current);
    if(current->cost<this->_best_solution->cost){this->_best_solution=current;}
    this->setUseRrtInFrontierCheck(save_use_rrt);

    return current;
}

MHO::MultiHandOverSolution* MultiHandOver::optimizeFrontier(MHO::MultiHandOverSolution* solution,unsigned int action_index,double keep_distance, double min_distance){
    //MHO::AbstractSolutionTools * solution_tools = new MHO::SolutionTools(this);
    //MHO::AbstracHandOverObjectCheck * ho_check = new MHO::HandOverWindowCheckerRRT(_object);
    //MHO::AbstracHandOverObjectCheck * ho_check = new MHO::HandOverDirectChecker(_object,solution->actions[action_index].robot(0),solution->actions[action_index].robot(1));
    //ho_check->timeLimit(1.);//ms
    MHO::AbstractFrontierOptimizer *frontier_optimizer=new MHO::FrontierOptimizerClimber(_grid,_tool_set);//MHO::FrontierOptimizerClimber(_grid,solution_tools,ho_check);
    MHO::AlgoLoggerInterface *log=new AlgoLoggerInterfaceFull(this,_grid,_draw_optimize_frontier);
    frontier_optimizer->setLogger(log);
    MultiHandOverSolution* ret=frontier_optimizer->optimize(solution,action_index,keep_distance,min_distance);
    delete log;
    //delete solution_tools;
    //delete ho_check;
    delete frontier_optimizer;
    return ret;
}

MultiAgentCell* MultiHandOver::findOptimalCellNullDistanceFrontier(MultiHandOverSolution *solution, unsigned int action_index){
    MultiHandOverSolution *aux_solution=solution->clone();
    MultiAgentCell* current;
    unsigned int a0,a1;
    MultiAgentCell* from0,*from1,*to1;//correspond to alternative origins for altDistances
    MultiHandOverSolution::Action & act=aux_solution->actions[action_index];
    bool found=false;
    double best_cost=numeric_limits<double>::infinity();
    MultiAgentCell *best_candidate;
    //TODO add for go back costs
    a0=act.agent(0)->index;
    a1=act.agent(1)->index;
    from0=act.from(0);
    from1=act.from(1);
    if(action_index+1<solution->actions.size()){
        to1=solution->actions[action_index+1].to(0);
    }else{
        to1=0;
    }
    current=act.to(0);
    if (!current->isReachable(a0) || !current->isReachable(a1)){
        M3D_WARN( "MultiHandOver::optimalCellNullDistanceFrontier() : this method is not applicable here as reachable spaces does not overlaps");
        return 0;//this method is not applicable here
    }
    aux_solution->frontiers[action_index]=FrontierPtr();//erase the frontier to avoid confusion
    act.path(0).clear();
    act.path(1).clear();
    //aux_solution->navigation_trajs[action_index*2].clear();
    //aux_solution->navigation_trajs[action_index*2+1].clear();
    act.setDist(0,0);
    act.setDist(1,0);
    //aux_solution->nav_dists[action_index*2]=0;
    //aux_solution->nav_dists[action_index*2+1]=0;
    aux_solution->handover_hri_costs[action_index]=0;
    aux_solution->handover_times[action_index]=0;
    best_candidate=current;
    while(!found){
        vector<MultiAgentCell*> neighb=current->getNeighbors(a0);
        for(unsigned int candidate=0;candidate<neighb.size();candidate++){
            MultiAgentCell* c=neighb[candidate];
            act.setDist(0,c->getAltDistance(a0,from0));
            act.setDist(1,c->getAltDistance(a1,from1));
            aux_solution->nav_dists[action_index*2+2]=c->getAltDistance(a1,to1);
            //double d0=c->getAltDistance(a0,from0);
            //double d1=c->getAltDistance(a1,from1);
            //double d1_1=c->getAltDistance(a1,to1);
            //double cost= _f_dist*(d0+d1+d1_1)
            //        + _f_time * (d0/getAgentSpeed(a0)+d1_1/getAgentSpeed(a1))
            //        + _f_agent_cost * ( d0/getAgentSpeed(a0)*getAgentUseCost(a0) + (d1+d1_1)/getAgentSpeed(a1)*getAgentUseCost(a1));
            double cost=computeSolutionCost(aux_solution);
            if(cost<best_cost){
                best_candidate=c;
                best_cost=cost;
            }
        }
        if(best_candidate==current){
            found=true;
        }else{
            current=best_candidate;//for next loop
        }
#ifdef DRAW_OPTIMIZE_FRONTIER
        if(_draw_optimize_frontier){
            resetAgentsConf();
            //_grid->setDrawOptimizeFrontier(solution->frontiers[action_index]->cellA(),solution->frontiers[action_index]->cellB(),false);
            _grid->setDrawOptimizeFrontier(best_candidate,best_candidate,found);
            g3d_draw_allwin_active();
        }
#endif
    }
    delete aux_solution;
    return best_candidate;
}

double MultiHandOver::computeNavigationTrajectories(MultiHandOverSolution *solution, int action_index){
    //note: we DO NOT USE the Action::to() and Action::from() here because they are
    //based on trajectories! we prefer getting them manually form reliable sources
    //(frontiers)
    bool compute_one_only=action_index>-1 && static_cast<unsigned int>(action_index)<solution->actions.size();
    if(!compute_one_only){
        //compute for all
        solution->nav_dists.assign(solution->frontiers.size()*2,-1);
        solution->navigation_trajs.assign(solution->frontiers.size()*2,vector<MultiAgentCell*>(0));
        solution->nav_start_dates.assign(solution->frontiers.size()*2,0);
        solution->navigation_trajs_to_target.assign(solution->agents.size(),vector<MultiAgentCell*>(0));
        solution->nav_dists_to_target.assign(solution->agents.size(),-1);
    }

    MultiAgentCell *from1=0,*to1=0,*from2=0,*to2=0;
    unsigned int a1,a2;
    unsigned int i_start=0 , i_end=solution->frontiers.size();
    if(compute_one_only){
        i_start=static_cast<unsigned int>(action_index);
        i_end=i_start+1;
    }
    for (unsigned int i=i_start;i<i_end;++i){
        shared_ptr<FrontierCells> frontier = solution->frontiers[i];
        from1=to2;//from1 is where a2 was previously (exception for i=0 below)
        //set a1 as the giver agent and a2 as the receiver
        //frontiers[i] correspond to the HO between agents[i] and agents[i+1],
        //agents[i] is the first of both in possession of the object
        if(frontier->agentA->robot()==solution->agents[i]->robot){
            a1=frontier->agentA_ind;
            a2=frontier->agentB_ind;
            to1=frontier->cellA();
            to2=frontier->cellB();
        }else{
            a2=frontier->agentA_ind;
            a1=frontier->agentB_ind;
            to2=frontier->cellA();
            to1=frontier->cellB();
        }
        if(i==0){
            from1=_grid->getAgentStructAt(a1)->getOrigin();//initial position
        }else if(compute_one_only){
            //from1=solution->actions[i-1].to(1); // as from1=to2 == NULL in that case
            FrontierPtr f2=solution->frontiers[i-1];
            from1=f2->cellOf(a1);
        }
        //from2 was never moved yet
        from2=_grid->getAgentStructAt(a2)->getOrigin();//initial position

        //compute traj for a1 and add cost
        assert(from1->isReachable(a1) && to1->isReachable(a1));
        if(from1->getDistanceAny(a1,to1)<numeric_limits<double>::infinity()){
            solution->navigation_trajs[i*2]=to1->getTrajAny(a1,from1);
            solution->nav_dists[i*2]=to1->getDistanceAny(a1,from1);
            //assert the traj starts and ends at the good cell
            assert(solution->navigation_trajs[i*2].front()==from1);
            assert(solution->navigation_trajs[i*2].back()==to1);
        }else{
            solution->navigation_trajs[i*2]=(_grid->computeOtherTrajectory(from1,to1,a1));
            solution->nav_dists[i*2]=(to1->getTempDist());
            //assert the traj starts and ends at the good cell
            assert(solution->navigation_trajs[i*2].front()==from1);
            assert(solution->navigation_trajs[i*2].back()==to1);
        }
        //the path to go to the end destination
        if(_mho_agents[a1].have_final_position){
            solution->navigation_trajs_to_target[i]=this->getAgentTargetCell(a1)->getTrajAny(a1,to1);
            solution->nav_dists_to_target[i]=this->getAgentTargetCell(a1)->getDistanceAny(a1,to1);
        }

        assert(solution->navigation_trajs[i*2].size());
        //idem for a2
        assert(from2->isReachable(a2) && to2->isReachable(a2));
        if(from2->getDistanceAny(a2,to2)<numeric_limits<double>::infinity()){
            solution->navigation_trajs[i*2+1]=to2->getTrajAny(a2,from2);
            solution->nav_dists[i*2+1]=to2->getDistanceAny(a2,from2);
            //assert the traj starts and ends at the good cell
            assert(solution->navigation_trajs[i*2+1].front()==from2);
            assert(solution->navigation_trajs[i*2+1].back()==to2);
        }else{
            solution->navigation_trajs[i*2+1]=(_grid->computeOtherTrajectory(from2,to2,a2));
            solution->nav_dists[i*2+1]=(to2->getTempDist());
            //assert the traj starts and ends at the good cell
            assert(solution->navigation_trajs[i*2+1].front()==from2);
            assert(solution->navigation_trajs[i*2+1].back()==to2);
        }
        assert(solution->navigation_trajs[i*2+1].size());
    }
    if(compute_one_only && static_cast<unsigned int>(action_index)+1 < solution->actions.size()){
        //the trajectory of the agent holding the object in next action
        //is not correct anymore, we change it
        unsigned int i=static_cast<unsigned int>(action_index)+1;
        FrontierPtr f2=solution->frontiers[action_index+1];
        from1=to2;
        a1=a2;
        to1=f2->cellOf(a1);
        assert(from1->isReachable(a1) && to1->isReachable(a1));
        if(from1->getDistanceAny(a1,to1) < numeric_limits<double>::infinity()){
            solution->navigation_trajs[i*2]=to1->getTrajAny(a1,from1);
            solution->nav_dists[i*2]=to1->getDistanceAny(a1,from1);
            //assert the traj starts and ends at the good cell
            assert(solution->navigation_trajs[i*2].front()==from1);
            assert(solution->navigation_trajs[i*2].back()==to1);
        }else{
            solution->navigation_trajs[i*2]=(_grid->computeOtherTrajectory(from1,to1,a1));
            solution->nav_dists[i*2]=(to1->getTempDist());
            //assert the traj starts and ends at the good cell
            assert(solution->navigation_trajs[i*2].front()==from1);
            assert(solution->navigation_trajs[i*2].back()==to1);
        }
        assert(solution->navigation_trajs[i*2].size());
    }else{
        //in all the other cases, we stopped at the last action, so we have to
        //make the last agent go to its target position
        if(_mho_agents[a2].have_final_position){
            solution->navigation_trajs_to_target.back()=this->getAgentTargetCell(a2)->getTrajAny(a2,to2);
            solution->nav_dists_to_target.back()=this->getAgentTargetCell(a2)->getDistanceAny(a2,to2);
        }
    }
    //TODO go back costs
    return solution->cost;
}

void MultiHandOver::computeHandOverTrajectory(MultiHandOverSolution *solution,int step)
{
    if(!solution){return ;}
    Robot *rob=global_Project->getActiveScene()->getActiveRobot();
    MultiHandOver3dPlanner mp(rob,rob->getHriAgent()->is_human);
    //mp.clearTraj();
    confPtr_t q_start,q_target,q_nav;
    vector<Eigen::Vector2d,Eigen::aligned_allocator<Eigen::Vector2d> > path;
    //find the agent index in the solution:
    if(step>=0 && static_cast<unsigned int >(step)<solution->actions.size()){
        MultiHandOverSolution::Action a=solution->actions[step];
        if(a.agent(0)->robot == rob){
            q_nav=confPtr_t(new Configuration(_agents_init_confs[a.agent(0)->index]));
            if(step>0){
                q_start=confPtr_t (new Configuration(*solution->actions[step-1].frontier()->confOf(rob)));
            }else{
                q_start=confPtr_t (new Configuration (*rob->getInitialPosition()));
            }
            q_target=confPtr_t(new Configuration(*a.frontier()->confOf(rob)));
            for (unsigned int t=0;t< a.path(0).size();++t){
                path.push_back(a.path(0)[t]->getCenter());
            }
            path.push_back(a.path(0).back()->getCenter());
        }else if ( a.agent(1)->robot== rob){
            q_nav=confPtr_t(new Configuration(_agents_init_confs[a.agent(1)->index]));
            q_start=confPtr_t(new Configuration(*rob->getInitialPosition()));
            q_target=confPtr_t(new Configuration(*a.frontier()->confOf(rob)));
            for (unsigned int t=0;t< a.path(1).size();++t){
                path.push_back(a.path(1)[t]->getCenter());
            }
            path.push_back(a.path(1).back()->getCenter());
        }else{
            M3D_WARN("MultiHandOver::computeHandOverTrajectory: "
                                 "the active robot is not in the solution at given step, not planning");
            return;
        }
        M3D_INFO("compute traj for solution step "<<step<<"/"<<solution->actions.size()<<" for robot "<<rob->getName());
        mp.computeTrajectory(q_start,q_nav,path,q_target);
    }
}

void MultiHandOver::computeSubTaskTrajectory(SubTask* sub_task){
    if(!sub_task){return;}

    global_Project->getActiveScene()->setActiveRobot(sub_task->agent()->robot->getName());
    Robot *rob=global_Project->getActiveScene()->getActiveRobot();
    assert(rob==sub_task->agent()->robot);
    MultiHandOver3dPlanner mp(rob,rob->getHriAgent()->is_human);
    //mp.clearTraj();
    confPtr_t q_start,q_target,q_nav;
    vector<Eigen::Vector2d,Eigen::aligned_allocator<Eigen::Vector2d> > path;
    //find the agent index in the solution:
    if(sub_task->isNav()){
        if(sub_task->waypoints().size() >=2){
            q_nav=confPtr_t(new Configuration(*sub_task->waypoints().front()));
            q_start=confPtr_t(new Configuration(*sub_task->waypoints().front()));
            q_target=confPtr_t(new Configuration(*sub_task->waypoints().back()));
        }else{
            q_nav=confPtr_t(new Configuration(_agents_init_confs[sub_task->agent()->index]));
            //q_start=confPtr_t(new Configuration(_agents_init_confs[sub_task->agent()->index]));
            q_start=confPtr_t(new Configuration(*rob->getCurrentPos()));
            //q_target=confPtr_t(new Configuration(_agents_init_confs[sub_task->agent()->index]));
            //if there is a next task with a position, set it at end position of this
            for(unsigned int n=0;n<sub_task->getNext().size();n++){
                if(sub_task->getNext()[n]->agent() == sub_task->agent()){
                    if(sub_task->getNext()[n]->isManip()){
                        q_target=confPtr_t(new Configuration(*sub_task->getNext()[n]->waypoints().front()));
                    }
                }
            }
        }
        for(unsigned int i=0;i<sub_task->waypoints2d().size();i++){
            path.push_back(sub_task->waypoints2d()[i]);
        }
        if(path.size()==1){//make a path of 2 step at least
            path.push_back(path.back());
        }
    }else if(sub_task->isManip()){
        q_nav=confPtr_t(new Configuration(*sub_task->waypoints().front()));
        q_start=confPtr_t(new Configuration(*sub_task->waypoints().front()));
        q_target=confPtr_t(new Configuration(*sub_task->waypoints().back()));
        path.push_back(Eigen::Vector2d(q_start->at(6+0),q_start->at(6+1)));
        path.push_back(Eigen::Vector2d(q_start->at(6+0),q_start->at(6+1)));
    }else{
        //nothing to compute
        return;
    }
    if(!sub_task->agent()->is_human){
        //move virtual object somewhere
        for(unsigned int qi=37;qi<49;qi++){
            q_start->getConfigStruct()[qi]=q_nav->getConfigStruct()[qi];
            if(q_target){
                q_target->getConfigStruct()[qi]=q_nav->getConfigStruct()[qi];
            }
        }
    }
    M3D_DEBUG( "compute traj for subtask " << sub_task->getId() <<" : " << SubTask::typeString(sub_task->type()) );
    if(mp.computeTrajectory(q_start,q_nav,path,q_target)){
        sub_task->traj(rob->getCurrentTraj());
        sub_task->setMarker(SubTask::HAS_FULL_TRAJ,true);
    }else{
        M3D_ERROR("MultiHandOver::computeHandOverTrajectory(): couldn't compute trajectory!");
        rob->getRobotStruct()->tcur=0;
    }

}

double MultiHandOver::navigationRelatedCost(double dist,unsigned int agent){
    double t;
    return this->navigationRelatedCostTime(dist,agent,t);
}
double MultiHandOver::navigationRelatedCostTime(double dist,unsigned int agent,double &time){
    double t=dist/getAgentSpeed(agent);
    time=t;
    return t*getAgentUseCost(agent);
}

double MultiHandOver::computeSolutionCost(MultiHandOverSolution *solution){
    double time_nav_cost(0),time_ho_cost(0),distance(0),time_nav(0),time_ho(0),obj_time(0), hri_cost(0),
            cost(0);
    //solution->avail_delay.assign(solution->agents.size(),0);
    for (unsigned int i=0;i<solution->actions.size();++i){
        assert (solution->actions[i].getSolution() == solution);
        MultiHandOverSolution::Action & act=solution->actions[i];
        //nav start date of the 1st agent: when it starts to move, its immediately
        //after the end of the previous HO [or 0 if first action]
        act.setStartDate(0,obj_time);
        act.setNavTime(0,act.dist(0)/act.agent(0)->speed);
        act.setNavTime(1,act.dist(1)/act.agent(1)->speed);
        obj_time+=act.dist(0)/act.agent(0)->speed;//tps parcourru par le holder
        double waiting = act.dist(1)/act.agent(1)->speed - obj_time; //time the holder waits for the other agent
        if (waiting >0){ // the holder waits
            obj_time+=waiting; // the object is then delayed of that waiting time
            if(i==0){
                act.setStartDate(0,act.startDate(0)+waiting); //delay the start of the first agent
                waiting = 0;
                //the first agent can wait at its starting position
            }
        }else {
            //nav start time of the second agent: when it should start moving
            //to be at the HO point on time
            act.setStartDate(1,-waiting);
            waiting=0;
        }
        assert(abs(act.startDate(1) + act.navTime(1) - obj_time ) < 0.00001);
        assert(abs(act.startDate(0) + act.navTime(0) - obj_time + waiting) < 0.00001);
        // start date of the HO is at the end of both nav
        act.setStartDateHO(obj_time);
        obj_time+=solution->handover_times[i];
        act.setEndDateHO(obj_time);
        for (unsigned int a=0;a<2;++a){
            distance+=act.dist(a);
            time_nav+=act.dist(a)/act.agent(a)->speed;
            time_nav_cost+=act.dist(a)/act.agent(a)->speed*act.agent(a)->use_cost;
            //TODO handover cost per agent is not as bellow:
            time_ho_cost+=solution->handover_times[i]*act.agent(a)->use_cost;
        }
        //TODO handover cost is not as bellow:
        time_ho+=solution->handover_times[i];
        //TODO hri cost computed here with max() instead of sum() (+)
        hri_cost=std::max(hri_cost,solution->handover_hri_costs[i]);
    }
    //for the paths to target positions
    for(unsigned int i=0;i<solution->nav_dists_to_target.size();++i){
        double d=solution->nav_dists_to_target[i];
        //if(d>0){
            double time=d/solution->agents[i]->speed;
            double start = solution->actions[min((int)i,(int)solution->actions.size()-1)].endDateHO();
            // [min()] : the frist and last agents go to their targets after their 1st (and unique) HO, unlike others
            solution->nav_start_end_times_to_target[i]=make_pair(start,start+time);
            distance+=d;
            time_nav+=time;
            time_nav_cost+=d/solution->agents[i]->speed*solution->agents[i]->use_cost;
        //}
    }
    //and the last agent is holding the object when going to its target so add the corresponding time to the solution duration
    if(solution->nav_dists_to_target.size() && solution->nav_dists_to_target.back()>0){
        obj_time+=solution->nav_dists_to_target.back()/solution->agents.back()->speed;
    }
    solution->time=obj_time;
    solution->cumul_time_nav=time_nav;
    solution->hri_cost=hri_cost;
    double norm_dist,norm_obj_time,norm_use_cost;
    norm_dist=distance/_max_dist;
    norm_obj_time=obj_time/_max_time; //TODO max object time is not really that, i guess...
    norm_use_cost=(time_ho_cost+time_nav_cost)/_max_use_cost;
    cost=_f_dist*norm_dist + _f_time * norm_obj_time + _f_agent_cost * norm_use_cost+
            _f_hri*hri_cost;
    if(isnan(cost)){
        cost=numeric_limits<double>::infinity();
    }
    solution->cost=cost;
    return cost;
}

boost::mt19937 rand_gen;
boost::normal_distribution<> u_dist(0,1);
boost::variate_generator<boost::mt19937, boost::normal_distribution<> > generator(rand_gen,u_dist);
MultiHandOverSolution * MultiHandOver::getLowCostRandomSolution(double sigma,int* index){
    float r=generator()*sigma;

    unsigned int i=std::floor(std::fabs(r));
    if(i<_solutions.size()){
        if(index!=0){*index=i;}
        return _solutions[i];
    }
    return this->getLowCostRandomSolution(sigma,index);
}

void MultiHandOver::clearSolutions(){
    for (unsigned int i=0;i<_solutions.size();++i){
        delete _solutions[i];
    }
    _solutions.clear();
    _best_solution=0;
}

void MultiHandOver::insert_solution_sorted(MultiHandOverSolution *sol){
    std::vector<MultiHandOverSolution*>::iterator it=std::lower_bound(_solutions.begin(),_solutions.end(),sol,MultiHandOverSolution::less);
    _solutions.insert(it,sol);
}

struct SearchIterationDataLog{
    int iteration;
    double duration;
    double duration_cumul;
    string action;
    double cost_found;
    double best_cost;
    double temperature;
    bool graph_cleared;
    int solutions_count;
    int cumul_search_new;
    int cumul_improve_solution;
    float improved_solution_rank;
    float sigma;
    bool shortest_path_search;
    SearchIterationDataLog(int iter,double duration_cumul,int sol_cnt,bool shortest_path):
        iteration(iter),duration(0),duration_cumul(duration_cumul),
        action(""),cost_found(-1),best_cost(-1),graph_cleared(false),
        solutions_count(sol_cnt),cumul_search_new(0),cumul_improve_solution(0),
        improved_solution_rank(-1),sigma(0),shortest_path_search(shortest_path)
    {}
    string gnuplot(){
        stringstream out("");
        char T='\t',G='\"';
        out << iteration<< T << temperature<< T << cost_found<< T
            << best_cost << T << duration<< T<<duration_cumul<< T
            << solutions_count<< T << cumul_search_new << T
            << cumul_improve_solution << T << improved_solution_rank << T
            << sigma << T
            << G<<action<<G<< T << shortest_path_search;
        return out.str();
    }
    static string gnuplot_legend(void){
        stringstream out("");
        char T='\t';
        out << "#";
        out << "iter"       <<T << "temp"       <<T << "cost"   <<T
            << "best"       <<T << "time"       <<T <<"cumul"   <<T
            << "solCnt"     <<T << "cumNew"     <<T
            << "cumImprov"  <<T << "improvRank" <<T <<"stdevNorm"   <<T
            << "action"     <<T << "shortest"   <<endl;
        return out.str();
    }
    SearchIterationDataLog generateNext(int iteration_nb,bool shortest_path){
        SearchIterationDataLog next(iteration_nb,this->duration_cumul,this->solutions_count,shortest_path);
        next.cumul_improve_solution=this->cumul_improve_solution;
        next.cumul_search_new=this->cumul_search_new;
        return next;
    }
};

void MultiHandOver::searchSolutions(){
    SearchIterationDataLog search_log_data(0,0,0,false);
    Chronometer chrono;//for iteration duration logging
    Chronometer main_chrono;//for stop condition
    ofstream ofplot;
    ofstream ofSols;
    char* home_path = getenv("HOME_MOVE3D");
    time_t time_start=time(0);
    struct tm *tod=localtime(&time_start);
    stringstream ss;
    ss<<home_path << '/' << tod->tm_year+1900<<'-'<<tod->tm_mon<<'-'<<tod->tm_mday<<'-'<<tod->tm_hour<<':'<<tod->tm_min<<':'<<tod->tm_sec;
    string path_sol_found(ss.str());
    string path_sol_plot(ss.str());
    path_sol_found.append("-search_sol_solutions_found.txt");
    ofSols.open(path_sol_found.c_str());
    path_sol_plot.append("-search_sol_plot.dat");
    ofplot.open(path_sol_plot.c_str());
    M3D_INFO("MultiHandOver::searchSolutions() : begin");
    this->computeLongestSolutionDistance();
    using namespace mho;
    this->clearSolutions();
    MultiHandOverSolution::resetID();//reset ID counter of solutions
    SimpleGraph *graph=_agents_frontiers_dist_graph.clone();
    mho::NodeInterface* first_node(0),*target_node(0);
    for (unsigned int i=0;i<graph->size() && (first_node ==0 || target_node==0);++i){
        if(_agents[graph->at(i)->getId()]==this->_first_agent){
            first_node=graph->at(i);
        }else if (_agents[graph->at(i)->getId()]==this->_target_agent){
            target_node=graph->at(i);
        }
    }
    double temperature=30,temp_step=0.1, overheat(_overheat_temperature);
    bool linear=false;
    if(linear){temperature=overheat-temp_step;}
    assert(overheat>0 && "overheat temperature must be > 0");
    int iterations = 0, iterations_max(_max_iterations);
    int time_limit(_time_limit);//ms
    bool find_shortest_path=true;
    ofplot<<"# file containing a log of the multihandover searching solution method"<<endl;
    ofplot<<SearchIterationDataLog::gnuplot_legend();
    main_chrono.reset();
    bool optimize=true;//TODO param
    //_solutions.reserve(iterations_max/5);//reserve some space for a large number of solutions ptrs
    while( (overheat<0 || temperature < overheat )
          && (iterations_max < 0 || iterations < iterations_max )
          && (time_limit < 0 || main_chrono.getMilliSec() < time_limit ) ){
        chrono.reset();
        search_log_data=search_log_data.generateNext(iterations,find_shortest_path);
        search_log_data.temperature=temperature;
        iterations++;

        this->resetAgentsConf();
        if(_draw_agents_holding_conf_during_search){
            g3d_draw_allwin_active();
        }
        if(static_cast<double>(rand())/RAND_MAX < std::exp(-5/temperature)){
            search_log_data.action="new";
            search_log_data.cumul_search_new++;
            //find new serie of agents to use (when T is high)
            //TODO reset costs to 1 in graph when temperature augments/is too high...?
            pair<cost_t,vector< mho::NodeInterface* > > sequence; //sequence of agents
            if(find_shortest_path){
                sequence = mho::dijkstra(graph,first_node,target_node,true);
                find_shortest_path=false;
            }else{
                sequence = mho::random_path(graph,first_node,target_node,numeric_limits<double>::infinity());
            }
            if(sequence.second.size()==0 || sequence.first == numeric_limits<double>::infinity()){
                //reset all costs to 1
                search_log_data.graph_cleared=true;
                for (unsigned int i=0;i<graph->edgesNumber();++i){
                    graph->edgeAt(i)->setCost(1);
                }
                if(!linear){
                    temperature+=temp_step;
                }
            }else if (sequence.second.size()==1){ // solution found with 1 agent, no handover needed!!! (why are we here so?)
                M3D_INFO("no handover needed!");
                break;
                //TODO check no handover needed sooner (set agents?)
            }else{
                unsigned int i=0;
                bool checkfrontier_ok=true;
                MultiHandOverSolution * solution = new MultiHandOverSolution;
                solution->cost=0;
                MultiHandOverAgent * agent_tmp=&(this->_mho_agents.at(sequence.second[i]->getId()));
                solution->agents.push_back(agent_tmp);
                do{
                    solution->agents.push_back(&this->_mho_agents[sequence.second.at(i+1)->getId()]);
                    //solution->frontiers.push_back(shared_ptr<FrontierCells>());
                    FrontierPtr frontier;
                    checkfrontier_ok=randomCheckFrontiersWithObject(frontier,sequence.second[i]->getId(),sequence.second.at(i+1)->getId(),50);
                    if(checkfrontier_ok){solution->addHandOverAction(frontier);}
                }while(checkfrontier_ok && ++i<sequence.second.size()-1);
                if(!checkfrontier_ok){
                    //set cost to inf for invalid edge (not definitive)
                    graph->updateCost(sequence.second[i],sequence.second.at(i+1),numeric_limits<cost_t>::infinity());
                    find_shortest_path=(_solutions.size()==0); // modification in the graph, restart with a shortest path search
                    //temperature+=temp_step;
                }else{
                    if(this->handOverFindConfs(solution)){
                        //compute navigation traj and motion traj, add costs
                        this->computeNavigationTrajectories(solution);
                        this->computeSolutionCost(solution);
                        this->insert_solution_sorted(solution);
                        search_log_data.cost_found=solution->cost;
                        ++search_log_data.solutions_count;
                        if(!_best_solution || solution->cost<this->_best_solution->cost){
                            this->_best_solution=solution;
                            assert(_best_solution==_solutions[0]);
                            //just found the best solution
                            M3D_DEBUG("better solution found : "<<*_best_solution);
                        }
                        if(optimize){
                            MultiHandOverSolution *optimized_sol=this->optimizeSolution(solution);
                            if(optimized_sol->cost<solution->cost){
                                M3D_DEBUG("solution optimized out");
                                search_log_data.cost_found=optimized_sol->cost;
                                ++search_log_data.solutions_count;
                            }
                        }
                        if(!linear){
                            temperature-=temp_step;
                        }
                    }else{
                        if(!linear){
                            temperature+=temp_step;
                        }
                    }
                }
                if(_best_solution){search_log_data.best_cost=_best_solution->cost;}
                //for(unsigned int i=0;i<sequence.second.size();++i){
                    //ofplot<<sequence.second[i]->getId() << " - ";
                //}
                ofSols<<solution->print_long()<<endl;
                if(solution->sons.size()){
                    ofSols<<solution->sons.front()->print_long()<<endl;
                }
                if(solution->cost>=numeric_limits<double>::infinity()){
                    delete solution;solution=0;
                }
            }
        }else{
            //improve existing solution (when T is low)
            //the improvement consist in a change in one handover that reduces
            //the cost, at least locally
            search_log_data.action="improve";
            search_log_data.cumul_improve_solution++;
            if(_solutions.size()>0){
                //T->max => sigma->1/2 * size    -    T-> 0 => sigma -> 0
                int index=-1;
                double sigma=temperature/overheat/2 * _solutions.size()+_solutions.size()/100;
                search_log_data.sigma = sigma/_solutions.size();
                MultiHandOverSolution *new_solution, * solution =
                        getLowCostRandomSolution(sigma,&index);
                search_log_data.improved_solution_rank=static_cast<float>(index)/search_log_data.solutions_count;
                if((new_solution=this->changeHandOverSolution(solution))!=0){
                    if(!linear){
                        temperature-=temp_step;
                    }
                    ++search_log_data.solutions_count;
                    search_log_data.cost_found=new_solution->cost;
                    ofSols<<new_solution->print_long()<<endl;
                }else{
                    if(!linear){
                        temperature+=temp_step;
                    }
                }
            }
            if(_best_solution){search_log_data.best_cost=_best_solution->cost;}
        }
        search_log_data.duration=chrono.getMilliSec();
        search_log_data.duration_cumul+=search_log_data.duration/1000.;
        ofplot<<search_log_data.gnuplot()<<endl;
        ofplot.flush();
        ofSols.flush();
        M3D_INFO("iter #"<<iterations<<" - T="<<temperature<<" - " << _solutions.size()<< " solutions found");
        if(linear){temperature-=temp_step;}
        this->resetAgentsConf();
        this->resetManipulablesConf();
    }
    ofplot.close();
    M3D_INFO("log written in "<<home_path);
    if(_best_solution){
        M3D_INFO( "best solution found : " << *_best_solution);
    }else{
        M3D_INFO("No solution found");
    }
    delete graph;
    M3D_INFO( "in " << iterations <<" iterations, "<<static_cast<double>(main_chrono.getMilliSec())/1000.<<" sec");
    M3D_DEBUG("MultiHandOver::searchSolutions() : end");
}


void MultiHandOver::searchSolutionLWAstar(bool no_lazy){
    // initialize stats
    _lwa_stats.push_back(vector<double>(SF_SIZE,-1));
    vector<double> & stats=_lwa_stats.back();
    stats[SF_EPSILON] = this->_epsilon;
    stats[SF_HRI_FACT] = this->_f_hri;
    stats[SF_AGENT_FACT] = this->_f_agent_cost;
    stats[SF_TIME_FACT] = this->_f_time;
    stats[SF_DIST_FACT] = this->_f_dist;
    Chronometer chrono;
    //this->clearSolutions();
    DisplayLWAstarInGrid display(_grid,_showText,_draw_lwa);
    this->_display_lwa_in_grid_obj=&display;
    mho::LazyWeightedAstar lwAstar(this->getEpsilonForLWAstar(),&display);//arg1:the epsilon for weighted A* (multiplies the heuristic)
    Robot* first=getFirstAgent(), *target=getTargetAgent();
    LazyWeightedAstarNode *start(0), *goal(0);
    MultiHandOverHeuristic* heuristic = this->createAstarHeuristic();

    this->computeLongestSolutionDistance();
    for(unsigned int i=0;i<_agents.size();++i){
        if(first==_agents[i]){
            start=new LazyWeightedAstarNode(_grid->getAgentStructAt(i)->getOrigin(),i,heuristic,_tool_set);
        }
        if(target==_agents[i]){
            goal=new LazyWeightedAstarNode(_grid->getAgentStructAt(i)->getTarget(),i,heuristic,_tool_set);
        }
        if(this->getAgentStruct(i)->use_cost<=1 && this->getAgentStruct(i)->have_final_position){
            _grid->computeAltDistancesForAgent(this->getAgentTargetCell(i),i);
        }
    }
    bool ok=true;
    if( !goal ){
        ok=false;
        M3D_ERROR("target agent not found in environment");
    }else if( !goal->cell() ){
        ok=false;
        M3D_ERROR("target agent is not in grid boundaries");
    }
    if( !start ){
        ok=false;
        M3D_ERROR("initial (holding) agent not found in environment");
    }else if( !start->cell() ){
        ok=false;
        M3D_ERROR("initial (holding) agent is not in grid boundaries");
    }
    if(!ok){
        //clear allocated nodes
        if(goal){delete goal;}
        if(start){delete start;}
    }
    if (ok){
        ok=lwAstar.compute(start,goal,-1,no_lazy);
        //get and save stats
        vector<pair<unsigned int,double> > times=display.getTimes();
        stats[SF_OK]=(ok ? 1 : 0);
        stats[SF_TIME_LWA] =  chrono.getMilliSec();
        stats[SF_TRUE_COST_NB] = times[5].first;
        stats[SF_TRUE_COST_MEAN]= times[5].second / times[5].first;
        //stats[SF_TRUE_COST_STDEV] = 1./0.;// TODO std dev of true cost comp time

        stats[SF_HEURISTIC_NB] = times[3].first;
        stats[SF_HEURISTIC_MEAN] = times[3].second / times[3].first;
        //stats[SF_HEURISTIC_STDEV] = 1./0.; // TODO std dev heuristic comp time

        stats[SF_TMP_COST_NB   ] = times[2].first;
        stats[SF_TMP_COST_MEAN ] = times[2].second / times[2].first;
        //stats[SF_TMP_COST_STDEV] = 1./0.; // TODO std dev tmp cost comp time

    }else {
        M3D_ERROR("MHO searchSolutionLWAstar error initializing the start and goal nodes\n"
                     "\t\tcheck agents exists and are inside grid boundaries");
    }
    if(ok){
        M3D_INFO("MultiHandOver: LWA* found a solution");
        M3D_INFO("heuristic times:");
        for(unsigned int i=0;i<3;i++){
            M3D_INFO('\t'<<heuristic->getTimes()[i]);
        }
        MultiHandOverSolution *sol =new MultiHandOverSolution();
        sol->object = this->_object;
        deque<mho::LWANodeInterface*> path=lwAstar.getPath();
        for(deque<mho::LWANodeInterface*>::iterator it=path.begin();it!=path.end();++it){
            LazyWeightedAstarNode *n=LazyWeightedAstarNode::cast(*it);
            if(sol->agents.size()==0 || sol->agents.back()->index!=n->agent()){
                //a new agent has to be added: add elements corresponding in vectors
                sol->agents.push_back(&_mho_agents[n->agent()]);
                sol->navigation_trajs.push_back(vector<MultiAgentCell*>(0));
                sol->nav_dists.push_back(0);
                if(sol->agents.size()>1){
                    //adding the trajectories for the second agent to come to the HO cell from its initial position
                    sol->navigation_trajs.back()=n->cell()->getTraj(n->agent());
                    sol->nav_dists.back()=n->cell()->getDistance(n->agent());


                    sol->navigation_trajs.push_back(vector<MultiAgentCell*>(0));
                    sol->nav_dists.push_back(0);
                    LazyWeightedAstarNode *prev=LazyWeightedAstarNode::cast(*(it-1));
                    vector<FrontierPtr> frontiers = n->cell()->getAgentAt(n->agent())->frontiers(prev->cell()->getAgentAt(prev->agent()));
                    sol->handover_hri_costs.push_back(0);
                    sol->handover_times.push_back(0);
                    sol->ho_start_end_dates.push_back(make_pair(-1,-1));
                    for(vector<FrontierPtr>::iterator f_it=frontiers.begin();f_it!=frontiers.end();++f_it){
                        if((*f_it)->cellOf(prev->agent()) == prev->cell()){
                            sol->addHandOverAction(*f_it);
                            break;
                        }
                    }

                    //compute HO stats
                    MHO::MultiHandOverHO ho(&sol->actions.back(),this);
                    sol->handover_hri_costs.back()=ho.computeCost();
                    sol->handover_times.back()=ho.getTime();// must be after compute cost, wich computes time also
                }
            }
            if(sol->navigation_trajs.back().size()){
                sol->nav_dists.back()+=sol->navigation_trajs.back().back()->computeDist(n->cell());
            }
            sol->navigation_trajs.back().push_back(n->cell());
        }
        sol->navigation_trajs_to_target.assign(sol->agents.size(),std::vector<MultiAgentCell*>(0));
        sol->nav_start_end_times_to_target.assign(sol->agents.size(),make_pair(-1,-1));
        sol->nav_dists_to_target.assign(sol->agents.size(),0);
        for(unsigned int i=0;i<sol->agents.size();++i){
            unsigned int a=sol->agents[i]->index;
            if(getAgentStruct(a)->have_final_position){
                MultiAgentCell *last;
                if(i!=sol->agents.size()-1){
                    //not the last agent
                    last=sol->navigation_trajs[i*2].back();
                    sol->navigation_trajs_to_target[i]=this->getAgentTargetCell(a)->getTrajAny(a,last);
                    sol->nav_dists_to_target[i]=this->getAgentTargetCell(a)->getDistanceAny(a,last);
                }else{
                    //last agent = the go back is already computed by the LWA* algo
                    last=sol->navigation_trajs.back().back();
                    sol->navigation_trajs_to_target.back()=sol->navigation_trajs.back();
                    sol->nav_dists_to_target.back()=sol->nav_dists.back();
                    //remove the go back from the navs
                    sol->navigation_trajs.pop_back();
                    sol->nav_dists.pop_back();
                }
            }
            sol->initial_confs.push_back(getAgentInitConf(sol->agents[i]->index));
            sol->target_confs.push_back(getAgentTargetConf(sol->agents[i]->index));
        }
        assert(sol->agents.size());
        assert(sol->navigation_trajs.size());

        sol->nav_start_dates.assign(sol->navigation_trajs.size(), 0);
        sol->nav_times.assign(sol->navigation_trajs.size(),0);


        this->computeSolutionCost(sol);
        this->_solutions.push_back(sol);
        M3D_INFO("solution found : ");
        for(unsigned int i=0;i<sol->actions.size();++i){
           M3D_INFO(""<<sol->actions[i].robot(0)->getName()<< "-"<<sol->actions[i].robot(1)->getName()<<" ; ");
        } 
        M3D_INFO("cost(sol) : "<<sol->cost);
        M3D_DEBUG("cost(solN):"<<sol->time*_f_time + sol->getTotalDist()*_f_dist+sol->hri_cost*_f_hri + sol->getTotalUseCost()*_f_agent_cost);
        M3D_DEBUG("cost(lwa*): "<< lwAstar.getPath().back()->g());
        _lwa_stats.back()[SF_COST_LWA] = sol->cost;
        _lwa_stats.back()[SF_AGENT_NB] = sol->agents.size();
        stats[SF_HO_NB]=sol->actions.size();
        stats[SF_EXEC_TIME]=sol->time;

        unsigned int nb_hu=0;
        for(unsigned int i=1;i<sol->agents.size()-1;i++){
            if(sol->agents[i]->is_human){
                nb_hu++;
            }
        }
        M3D_INFO("short stats:\n"<<"eps\tt\tc\th/a\n"
                 <<this->getEpsilonForLWAstar()<<'\t'<<sol->time<<'\t'<<sol->cost<<'\t'<<nb_hu<<'/'<<sol->agents.size()-2
                 <<"\nh/a: #human/#agents total in the solution, both excluded of first and target");

        if(this->usePostProcess()){
            this->postProcessSolution(sol);
        }

        //this->insert_solution_sorted(sol);
        //if(!this->_best_solution || sol->cost<this->_best_solution->cost){
            //this->_best_solution=sol;
        //}
    }else{
        M3D_INFO("MultiHandOver: LWA* failed no solution found");
    }
    _lwa_stats.back()[SF_TIME_TOT] = chrono.getMilliSec();
    _lwa_stats.back()[SF_TIME_POSTPROC] = _lwa_stats.back()[SF_TIME_TOT] - _lwa_stats.back()[SF_TIME_LWA];
    _lwa_stats.back()[SF_AGENT_FROM] = this->_first_agent_node->getId();
    _lwa_stats.back()[SF_AGENT_TO] = this->_target_agent_node->getId();
    unsigned int sf_ho_cost_nb;
    _tool_set->getStatsCheckHoFull(stats[SF_HO_COST_MEAN],stats[SF_HO_COST_STDEV],sf_ho_cost_nb);
    stats[SF_HO_COST_NB] = sf_ho_cost_nb;
    printLWAStats();

    M3D_INFO("LWA* search ended in "<<chrono.getMilliSec()<<" ms");
    M3D_INFO("Counts of checks:");
    for(unsigned int i=0;i<FrontierCells::FCS_ENUM_SIZE;i++){
        M3D_INFO("\t*"<<i<<"\t"<<
                     this->_tool_set->frontierTester->getCheckCount(static_cast<FrontierCells::FrontierCheckStatus>(i))<<
                     "\t"<<this->_tool_set->frontierTester->getCheckTime(static_cast<FrontierCells::FrontierCheckStatus>(i)));
    }
    this->_display_lwa_in_grid_obj=0;
    delete heuristic;
    return;
}


void MultiHandOver::postProcessSolution(MultiHandOverSolution* sol)
{
    //convert to Task
    Task* task=new Task(sol);
    M3D_INFO("task created with "<<task->getSubTasks().size()<<" sub tasks");
    map<MultiHandOverAgent*,MultiAgentCell*> state_begin;
    for(vector<MultiHandOverAgent>::iterator it=this->_mho_agents.begin();it!=_mho_agents.end();++it){
        state_begin[&(*it)]=_grid->getAgentStructAt(it->index)->getOrigin();
    }
    //add the idle sub tasks to fill the task
    task->fillWithIdle(state_begin);
    //display
    //task->plotGanttChart();
    M3D_INFO("gantt chart generated");
    MultiPathCheck mpcheck(_grid,task,state_begin);
    bool ok=false;
    while(!ok){
        //check for collisions
        M3D_TRACE( "MHO: checking for collisions btw tasks");
        ok=mpcheck.check();
        M3D_TRACE( "MHO: ...check done");
        //create groups of collision to compute solution
        vector< set<SubTask*> > collision_groups;
        for(unsigned int i=0;i<mpcheck.getTasksInCollision().size();i++){
            pair<SubTask*,SubTask*> stks=mpcheck.getTasksInCollision()[i];
            //display collisions
            M3D_INFO("collision btw "<< (stks.first ? stks.first->agent()->shortname : "?")
                          << " and " << (stks.second ? stks.second->agent()->shortname : "?") );

            //add to collision groups
            int group_found=-1;
            for(unsigned int g=0;g<collision_groups.size();++g){
                pair<bool,bool> found=make_pair(false,false);
                if(stks.first->type()!=SubTask::IDLE && collision_groups[g].find(stks.first) != collision_groups[g].end()){ found.first=true; }
                if(stks.second->type()!=SubTask::IDLE && collision_groups[g].find(stks.second) != collision_groups[g].end()){ found.second=true; }
                if(found.second || found.first ){
                    if(group_found!=-1){
                        //group already found, merge groups
                        collision_groups[group_found].insert(collision_groups[g].begin(),collision_groups[g].end());
                        collision_groups[g].clear();
                    }else{
                        if (!found.first && stks.first->type() != SubTask::IDLE){
                            collision_groups[g].insert(stks.first);
                        }
                        if(!found.second && stks.second->type() != SubTask::IDLE){
                            collision_groups[g].insert(stks.second);
                        }
                        group_found=g;
                    }
                }
            }
            if(group_found==-1){
                //create new group
                collision_groups.push_back(set<SubTask*>());
                collision_groups.back().insert(stks.first);
                collision_groups.back().insert(stks.second);
            }
        }
        for(unsigned int i=0;i<collision_groups.size();i++){
            if(collision_groups[i].size()==0){
                swap(collision_groups[i],collision_groups.back());
                collision_groups.pop_back();
                i--;
            }
        }
        //for(unsigned int i=0;i<collision_groups.size();i++){
        if(collision_groups.size()){
            unsigned int i=0;
            MultiPathPlanner mpp;
            //for final results
            vector<unsigned int> agents;
            vector<MultiAgentCell*> start,goal;
            //for intermediary sorting
            //a vector of 3 vectors containing agents of a same priority and their data
            //each of the 3 vector correspond to a priority from high (0) to low (2)
            vector< vector< pair< unsigned int , pair < MultiAgentCell* , MultiAgentCell* > > > > prio_vector(3);
            double begin=0;//numeric_limits<double>::infinity();
            for(set<SubTask*>::iterator it=collision_groups[i].begin();
                it!=collision_groups[i].end();it++)
            {

                MultiAgentCell* goal_cell;
                unsigned int prio=0;
                if((*it)->type()==SubTask::NAV_WITH_OBJ){
                    //highest prio
                    prio=0;
                }else if((*it)->type()==SubTask::IDLE){
                    prio=2;
                }else{prio=1;}
                if((*it)->type() != SubTask::IDLE){
                    goal_cell=dynamic_cast<MultiAgentCell*>(_grid->getCell((*it)->waypoints2d().back()));
                }else{
                    goal_cell=0;
                }
                prio_vector[prio].push_back(
                            make_pair((*it)->agent()->index,
                                      make_pair(dynamic_cast<MultiAgentCell*>(_grid->getCell((*it)->waypoints2d().front())),
                                                goal_cell)));
                //max, because otherwise an agent would have to do several tasks at the same time
                begin=max((*it)->beginDate(),begin);
            }
            //fill the final vectors to pass to mpp
            for(unsigned int prio=0;prio<prio_vector.size();prio++){
                for(unsigned int pa=0;pa<prio_vector[prio].size();pa++){
                    agents.push_back(prio_vector[prio][pa].first);
                    start.push_back(prio_vector[prio][pa].second.first);
                    goal.push_back(prio_vector[prio][pa].second.second);
                }
            }
            mpp.setAgents(agents);
            mpp.setStart(start);
            mpp.setGoal(goal);
            mpp.setSolutionTools(this->_tool_set->solutionTools);
            mpp.setDisplay(_showText,false,_grid);
            mpp.setHierarchy(true);
            if(mpp.compute(1000)){
                M3D_INFO("Post Process OK, multi agent path found");
            }else{
                M3D_ERROR("Post Process FAIL, no multi agent path found");
                break;
            }
            map<MultiHandOverAgent*,vector<SubTask*> > backup_prev_s_task, backup_next_s_task; // saves the previous and following sub tasks to update them in the new sub tasks
            map<MultiHandOverAgent*, double> backup_prev_end,backup_next_begin; // if a prev/next as already been set, we need to know about date to decide to update or not
            for(set<SubTask*>::iterator colt_it=collision_groups[i].begin();colt_it!=collision_groups[i].end();colt_it++){
                SubTask* to_del=*colt_it;
                M3D_DEBUG( "deleting task "<<to_del->agent()->shortname<<"-"<<SubTask::typeString(to_del->type())<<" id:"<<to_del->getId());
                if(to_del->type()!=SubTask::IDLE){
                    //save tasks except idle
                    if( backup_prev_s_task.find(to_del->agent()) == backup_prev_s_task.end()
                            || backup_prev_end[to_del->agent()] > to_del->beginDate()){
                        //not set yet OR set from a following sub task
                        backup_prev_end[to_del->agent()]=to_del->beginDate();
                        backup_prev_s_task[to_del->agent()]=to_del->getPrev();
                    }
                    if(backup_next_s_task.find(to_del->agent()) == backup_next_s_task.end() || backup_next_begin[to_del->agent()] < to_del->endDate()){
                        //not set yet OR set from a previous sub task
                        backup_next_begin[to_del->agent()]=to_del->endDate();
                        backup_next_s_task[to_del->agent()]=to_del->getNext();
                    }
                }
                task->eraseSubTask(*colt_it);
            }
            deque<mho::LWANodeInterface*> path=mpp.getPath();
            vector<vector<SubTask*> > new_sub_tasks(agents.size(),vector<SubTask*>(0));
            MultiPathLWANode* prev_n(0);
            double t=begin;
            for(unsigned int p=0;p<path.size();p++){
                //TODO add prev next to each
                MultiPathLWANode* n=MultiPathLWANode::cast(path[p]);
                //get the duration of the transition to n
                double d=n->getDurationMax();
                if(prev_n){
                    //for each agent in the solution
                    for(unsigned int ai=0;n->agent_index(ai)>=0;ai++){
                        // or : for(unsigned int ai=0;ai<agents.size();ai++){
                        SubTask *current_tk=(new_sub_tasks[ai].size() ? new_sub_tasks[ai].back() : 0);
                        SubTask* prev_tk=current_tk;
                        if(n->cell(ai)==prev_n->cell(ai)){
                            //not moving
                            if(current_tk && current_tk->type()!=SubTask::IDLE){
                                current_tk->extendToEndDate(t);
                                current_tk=0;
                            }
                            if(!current_tk){
                                //create a new subtask
                                current_tk=new SubTask(&_mho_agents[n->agent_index(ai)],SubTask::IDLE,t,t+d);
                                current_tk->addWaypoint(n->cell(ai)->getCenter());
                                if(prev_tk){
                                    current_tk->addPrevious(prev_tk);
                                    prev_tk->addNext(current_tk);
                                }else{
                                    //add the previous from the backup'ed ones
                                    vector<SubTask*> &previous_stks=backup_prev_s_task[current_tk->agent()];
                                    for(vector<SubTask*>::iterator nstk=previous_stks.begin();nstk!=previous_stks.end();nstk++){
                                        current_tk->addPrevious(*nstk);
                                        (*nstk)->addNext(current_tk);
                                        (*nstk)->replan(current_tk->beginDate() - (*nstk)->duration());
                                    }
                                }
                                new_sub_tasks[ai].push_back(current_tk);
                            }
                            assert(current_tk && current_tk->type() == SubTask::IDLE);
                            current_tk->extendToEndDate(t+d);
                        }else{
                            //it s a move
                            if(current_tk && !current_tk->isNav()){
                                //extend the currently created task
                                current_tk->extendToEndDate(t);
                                current_tk=0;
                            }
                            if(!current_tk){
                                //create a new subtask
                                //TODO type have to be more precise (With obj, to HO...)
                                current_tk=new SubTask(&_mho_agents[n->agent_index(ai)],SubTask::NAV_SYNC,t,t+d);
                                if(prev_tk){
                                    current_tk->addWaypoint(prev_n->cell(ai)->getCenter(),t);
                                    current_tk->addPrevious(prev_tk);
                                    prev_tk->addNext(current_tk);
                                }else{
                                    //add the previous from the backup'ed ones
                                    vector<SubTask*> &previous_stks=backup_prev_s_task[current_tk->agent()];
                                    for(vector<SubTask*>::iterator nstk=previous_stks.begin();nstk!=previous_stks.end();nstk++){
                                        current_tk->addPrevious(*nstk);
                                        (*nstk)->addNext(current_tk);
                                        (*nstk)->replan(current_tk->beginDate() - (*nstk)->duration());
                                        if((*nstk)->agent()==current_tk->agent()){
                                            //add the end position of the previous task of this agent as
                                            //begining of this
                                            current_tk->addWaypoint((*nstk)->waypoints2d().back(),t);
                                        }
                                    }
                                }

                                new_sub_tasks[ai].push_back(current_tk);
                            }
                            assert(current_tk && current_tk->isNav());
                            if(p==1){
                                //add the first and previous cell as waypoint
                                current_tk->addWaypoint(prev_n->cell(ai)->getCenter(),t);
                            }
                            current_tk->addWaypoint(n->cell(ai)->getCenter(),t);
                            current_tk->extendToEndDate(t+d);

                        }
                    }
                    //a test
                    for(unsigned int ai=0;n->agent_index(ai)>=0;ai++){
                        SubTask *cur_a=(new_sub_tasks[ai].size() ? new_sub_tasks[ai].back() : 0);
                        for(unsigned int ai_b=0;ai_b < ai ;ai_b++){
                            SubTask *cur_b=(new_sub_tasks[ai_b].size() ? new_sub_tasks[ai_b].back() : 0);
                            //assert there are no collisions in the computed tasks
                            assert(dynamic_cast<MultiAgentCell*>(_grid->getCell(cur_a->waypoints2d().back()))->
                                   checkIsInCollision(n->agent_index(ai),
                                                      dynamic_cast<MultiAgentCell*>(_grid->getCell(cur_b->waypoints2d().back())),
                                                      n->agent_index(ai_b))==false);
                        }

                    }

                }
                prev_n=n;
                t+=d;
            }
            M3D_INFO("new MPP parsed with duration of "<<t);
            //add the sub tasks to the main task
            for(unsigned int at=0;at<new_sub_tasks.size();at++){
                //add the next sub tasks form the backup'ed ones
                //they go to the next stk of each agent
                vector<SubTask*> &next_stks=backup_next_s_task[new_sub_tasks[at].back()->agent()];
                for(vector<SubTask*>::iterator nstk=next_stks.begin();nstk!=next_stks.end();nstk++){
                    new_sub_tasks[at].back()->addNext(*nstk);
                    (*nstk)->addPrevious(new_sub_tasks[at].back());
                    double save=new_sub_tasks[at].back()->endDate();
                    (*nstk)->replan(new_sub_tasks[at].back()->endDate());
                    if((*nstk)->beginDate() != save){
                        M3D_DEBUG("new sub-task has begin date ("<<(*nstk)->beginDate() << ") != from its previous sub-task former end date ("<<save<<")");
                    }
                }
                for(unsigned int tat=0;tat<new_sub_tasks[at].size();tat++){
                    task->addSubTask(new_sub_tasks[at][tat]);
                }
            }
            //reset all idle tasks
            //delete them
            //task->removeIdles();
            //fill the task with IDLE sub tasks
            task->fillWithIdle(state_begin);
            //this->playTaskStepByStep(task);
        }
        //task->plotGanttChart();
    }

    _solutions_as_tasks.push_back(task);
    //task->plotGanttChart();
    task->displayGraph();
}


// //////////////////////////////////////////////////////////
// //////////////   DRAWING, DISPLAY, LOG   /////////////////
// //////////////////////////////////////////////////////////

void MultiHandOver::playTaskStepByStep(Task* task){
    M3D_INFO("Playing task step by step");
    double time=task->begin().date();
    double speed_factor=ENV.getDouble(Env::showTrajFPS);
    for(Task::state_iterator it=task->begin();!it.isEnd();it.advance()){
        assert(it.date()>=time);
        usleep((it.date()-time)*1000000/speed_factor);
        for(map<MHO::MultiHandOverAgent*,Eigen::Vector2d>::iterator st_it=it.state().begin();
            st_it!= it.state().end();st_it++){

            Configuration q=*(st_it->first)->robot->getInitialPosition();
            q[6+0]=st_it->second[0];
            q[6+1]=st_it->second[1];
            st_it->first->robot->setAndUpdate(q);
        }
        g3d_draw_allwin_active();
        time=it.date();
    }
}

void MultiHandOver::computeTaskTrajs(Task* task){
    for(unsigned int i=0;task->getSubTask(i);i++){
        SubTask *subtask=task->getSubTask(i);
        this->computeSubTaskTrajectory(subtask);
    }
}

void MultiHandOver::setSubTaskTrajToRobot(SubTask* subtask){
    if(!subtask->getMarker(SubTask::HAS_FULL_TRAJ)){
        this->computeSubTaskTrajectory(subtask);
    }else{
        subtask->traj().replaceP3dTraj();
    }
    global_Project->getActiveScene()->setActiveRobot(subtask->agent()->robot->getName());
}

void MultiHandOver::playSolutionTrajSync(Task *task){
    if(!_mho_display){
        _mho_display = new MHO::MultiHandOverDisplay();
    }
    computeTaskTrajs(task);

    _mho_display->playSolutionTrajSyncRT(task);

}

bool  MultiHandOver::playSolutionTaskSyncStep(int solution){
    bool ok=false;
    if(!_mho_display){
        _mho_display = new MHO::MultiHandOverDisplay();
    }
    if (solution>=0 && static_cast<unsigned int>(solution) <_solutions_as_tasks.size()){
        Task* task=_solutions_as_tasks[solution];
        this->computeTaskTrajs(task);
        ok= _mho_display->stepForwardSolutionTrajSync(task);
    } else{
        ok= _mho_display->stepForwardSolutionTrajSync();
    }
    return ok;
}

void MultiHandOver::playSolutionTask(int solution){
    if (solution>=0 && static_cast<unsigned int>(solution) <_solutions_as_tasks.size()){
        //this->playTaskStepByStep(_solutions_as_tasks[solution]);
        this->playSolutionTrajSync(_solutions_as_tasks[solution]);
    }
}
void MultiHandOver::setPlaySolutionTask(int solution){
    if (solution>=0 && static_cast<unsigned int>(solution) <_solutions_as_tasks.size()){
        _play_solution_task=true;
        _solution_index_to_draw=solution;
    }
}

void MultiHandOver::setPlaySolutionTaskTraj(int solution){
    if (solution>=0 && static_cast<unsigned int>(solution) <_solutions_as_tasks.size()){
        _solution_step_index_to_draw=0;
        _play_solution_task_traj=true;
        _solution_index_to_draw=solution;
    }
}

void MultiHandOver::setPlaySolutionSyncStepByStep(int solution){
    _solution_index_to_draw=solution;
    _play_solution_sync_step_by_step = true;
    _play_solution_task_traj=true;
}

bool MultiHandOver::isEndPlaySolutionTaskTraj(){
    return !_play_solution_task_traj;
}

MultiHandOverSolution::CostData MultiHandOver::getSolutionDataStudio(int solution){
    MultiHandOverSolution::CostData cost_data;
    MultiHandOverSolution * sol;
    if(solution>=0 && static_cast<unsigned int>(solution)<_solutions.size()){
        sol=_solutions[solution];
    }else if(solution==-1 && _best_solution){
        sol=_best_solution;
    }else if (_solutions.size()){
        sol=_solutions.back();
    }
    if(sol){
        cost_data = MultiHandOverSolution::CostData(sol);
    }
    return cost_data;
}

double MultiHandOver::setSolutionStepRobotsConfs(int solution, int step,MultiHandOverSolution::CostData &cost_data){
    MultiHandOverSolution * sol;
    if(solution>=0 && static_cast<unsigned int>(solution)<_solutions.size()){
        sol=_solutions[solution];
    }else if(solution==-1 && _best_solution){
        sol=_best_solution;
    }else if(solution==-1 && _solutions.size()){
        sol=_solutions.back();
    }else{
        sol=0;
    }

    if(sol && step>=0 && static_cast<unsigned int>(step)<sol->frontiers.size()){
        cost_data = MultiHandOverSolution::CostData(sol);
        _solution_index_to_draw=solution;
        _solution_step_index_to_draw=step;
        this->resetAgentsConf();
        for(unsigned int i=0;i<=static_cast<unsigned int>(step);++i){
            //std::tr1::shared_ptr<FrontierCells> frt=sol->frontiers[i];
            MultiHandOverSolution::Action *act=&sol->actions[i];
            for(unsigned int a_i=0;a_i<2;a_i++){
                if(act->agentConf(a_i))
                    act->robot(a_i)->setAndUpdate(*act->agentConf(a_i));
            }
            if(act->objectPos()){
                this->_object->setAndUpdate(*act->objectPos());
            }

            //if(frt->confA)
            //    frt->agentA->robot()->setAndUpdate(*frt->confA);
            //if(frt->confB)
            //    frt->agentB->robot()->setAndUpdate(*frt->confB);
            //if(frt->object_position)
            //    this->_object->setAndUpdate(*frt->object_position);


            // if(frt->agentA->robot()->isInCollisionWithOthersAndEnv()){
            //     M3D_INFO( frt->agentA->robot()->getName() << "is in collision w o and env");
            // }
            // if(frt->agentA->robot()->isInCollision()){
            //     M3D_INFO( frt->agentA->robot()->getName() << "is in collision");
            // }
            // if(frt->agentB->robot()->isInCollisionWithOthersAndEnv()){
            //     M3D_INFO( frt->agentB->robot()->getName() << "is in collision w o and env");
            // }
            // if(frt->agentB->robot()->isInCollision()){
            //     M3D_INFO( frt->agentB->robot()->getName() << "is in collision");
            // }
            //if(i==static_cast<unsigned int>(step) && frt->object_position){
            //    this->_object->setAndUpdate(*frt->object_position);
            //}
        }
    }else{
        sol=0;
    }
    return (sol ? sol->cost : -1);
}

void MultiHandOver::getConfHandOver(int solution, int step,string robot_name, confPtr_t & conf){
    MultiHandOverSolution * sol;
    if(solution>=0 && static_cast<unsigned int>(solution)<_solutions.size()){
        sol=_solutions[solution];
    }else if(solution==-1 && _best_solution){
        sol=_best_solution;
    }else if(solution==-1 && _solutions.size()){
        sol=_solutions.back();
    }

    if(sol && step>=0 && static_cast<unsigned int>(step)<sol->frontiers.size()){
        std::tr1::shared_ptr<FrontierCells> frt=sol->frontiers[step];
        //makes a copy
        if(frt->agentA->robot()->getName() == robot_name){
            conf=confPtr_t(new Configuration(*frt->confA));
        }else if(frt->agentB->robot()->getName() == robot_name){
            conf=confPtr_t(new Configuration(*frt->confB));
        }else{
            conf=confPtr_t((Configuration*)0);
        }
    }
}

void MultiHandOver::draw2dPath(){
    MultiHandOverSolution * sol=0;
    if(this->_draw_2d_path){
        unsigned int s(_solution_index_to_draw),t(_solution_step_index_to_draw);
        if (s==static_cast<unsigned int>(-1)){sol=_best_solution;}
        else if(s<_solutions.size()){ sol=_solutions[s];}
        if(sol && t*2 < sol->navigation_trajs.size()){
            this->draw2dPath(sol->navigation_trajs[t*2],sol->actions[t].agent(0)->index);
        }
        if(sol && this->_draw_all_2d_pathes){
            for(unsigned int i=0;i<sol->actions.size();++i){
                this->draw2dPath(sol->actions[i].path(0),sol->actions[i].agent(0)->index);
                this->draw2dPath(sol->actions[i].path(1),sol->actions[i].agent(1)->index);
            }
            for(unsigned int i=0;i<sol->navigation_trajs_to_target.size();i++){
                this->draw2dPath(sol->navigation_trajs_to_target[i],sol->agents[i]->index,true);
            }
        }else if(sol && t*2+1 < sol->navigation_trajs.size()){
            this->draw2dPath(sol->navigation_trajs[t*2+1],sol->actions[t].agent(1)->index);
        }
    }
}
void MultiHandOver::draw2dPath(const std::vector<MultiAgentCell *> &path,int color_id,bool alt){
    if(!path.size()){
        return;
    }

    color_id=color_id%(tOrange-tBlue)+tBlue;

    for(unsigned int i=0;i<path.size()-1;i++)
    {
        glLineWidth(3.);
        g3d_drawOneLine(path[i]->getCenter()[0],      path[i]->getCenter()[1],      0.4,
                path[i+1]->getCenter()[0],    path[i+1]->getCenter()[1],    0.4,
                color_id, NULL);
        if(alt){
            glLineWidth(0.5);
        }else{
            glLineWidth(1.);
        }

        double colorvector[4];
        //colorvector[0] = colors[color_id][0];       //red
        //colorvector[1] = colors[color_id][1];       //green
        //colorvector[2] = colors[color_id][2];       //blue
        //colorvector[3] = 0.5;       //transparency
        g3d_get_color_vect(color_id,colorvector);
        colorvector[3]=0.25;

        //        glEnable(GL_BLEND);
        g3d_set_color(Any,colorvector);

        double radius=0.05;
        if(alt){radius=0.07;}
        g3d_draw_solid_sphere(path[i+1]->getCenter()[0],
                path[i+1]->getCenter()[1],
                0.4,
                radius, 20);
        //            M3D_INFO( "\n\n cell: \n" << m_2DPath[i] );
    }
}

void MultiHandOver::printSolutionsForest(){
    ofstream output;
    char *home_path=getenv("HOME_MOVE3D");
    string path(home_path);
    path.append("/sol_forest.dot");
    output.open(path.c_str());
    MultiHandOverSolution::printFamiliesTrees(_solutions,output);
    output.close();
}
void MultiHandOver::printMainTimes(std::vector<double> times){
    assert(times.size()==4);
    ofstream output;
    char *home_path=getenv("HOME_MOVE3D");
    time_t time_start=time(0);
    struct tm *tod=localtime(&time_start);
    stringstream ss;
    ss<<home_path << '/' << tod->tm_year+1900<<'-'<<setw(2)<<setfill('0')<<tod->tm_mon+1<<'-'<<tod->tm_mday<<'-'<<tod->tm_hour<<':'<<tod->tm_min<<':'<<tod->tm_sec;
    string path(ss.str());
    path.append("-main_times.csv");
    //append a letter if the file already exists
    //test iteratively for each letter until file doesn't exits
    char c='A';
    while (FILE *file = fopen(path.c_str(), "r")){
        fclose(file);// on succes, close the file and create new path name
        path.clear();
        path.append(ss.str());
        char s[3];
        s[0]='-';
        s[1]=c;
        s[3]=0;
        path.append(s);
        //path.append(c);
        path.append("-main_times.csv");
        c++;
    }
    output.open(path.c_str(),ios_base::app | ios_base::out);
    output<<"step\tduration(ms)\tcumul (s)"<<endl;
    output<<"grid\t"<<      times[0]            <<"\t"<<times[0]/1000<<endl;
    output<<"frontier\t"<<  times[1]-times[0]   <<"\t"<<times[1]/1000<<endl;
    output<<"check\t"<<     times[2]-times[1]   <<"\t"<<times[2]/1000<<endl;
    output<<"search\t"<<    times[3]-times[2]   <<"\t"<<times[3]/1000<<endl;
    output.close();
}

void MultiHandOver::printLWAStats(){
    ofstream output;
    char *home_path=getenv("HOME_MOVE3D");
    time_t time_start=time(0);
    struct tm *tod=localtime(&time_start);
    stringstream ss,prefix,indic,suffix;
    prefix<<home_path << '/' << tod->tm_year+1900<<'-' << setw(2) << setfill('0')<< tod->tm_mon+1<<'-'<< setw(2) << tod->tm_mday<<'-'<< setw(2) << tod->tm_hour<<':'<< setw(2) << tod->tm_min<<':'<< setw(2) << tod->tm_sec;
    int c=0;
    indic << '-' << setfill('0') << setw(2) << c ;
    suffix<<"-lwa-stats.csv";
    //append a number if the file already exists
    //test iteratively for each number until file doesn't exits
    ss << prefix.str() << indic.str() << suffix.str();
    string path(ss.str());
    while (FILE *file = fopen(path.c_str(), "r")){
        fclose(file);// on succes, close the file and create new path name
        path.clear();
        indic.str("");
        ss.str("");
        c++;
        indic << '-' << setfill('0') << setw(2) << c ;
        ss << prefix.str() << indic.str() << suffix.str();
        path=ss.str();
    }
    output.open(path.c_str());
    global_Project->getActiveScene()->getName();
    char *p3d_name=0 , *sce_name=0;
    p3d_name=realpath(XYZ_ENV->p3d_file_path,p3d_name);
    sce_name=realpath(XYZ_ENV->sce_file_path,sce_name);
    //output<<"# p3d: " << cwd << XYZ_ENV->p3d_file_path<<endl;
    //output<<"# sce: " << cwd << XYZ_ENV->sce_file_path<<endl;
    output<<"# p3d: " << p3d_name<<endl;
    output<<"# sce: " << sce_name<<endl;
    free(p3d_name);
    free(sce_name);
    for(unsigned int i=0;i<SF_SIZE;i++){
        output<<_lwa_stats.back()[i] << " ;\t";
    }
    output<<endl;
    output.close();
}

} // namespace HRICS
