#ifndef HRICS_DISTANCE2D_H
#define HRICS_DISTANCE2D_H

#include "API/planningAPI.hpp"

namespace HRICS {

class Distance2D
{
public:
    Distance2D(Robot* rob, std::vector<Robot*> humans, std::vector<Joint *> costJoints);
    Distance2D(bool guess_settings=false);
    ~Distance2D();

    double getCostMean();
    double getCostMax();
    double getCost();

    /**
     * @brief setMaxToMeanRatio
     * @param a
     * @throw invalid_argument when a not in [0,1]
     */
    void setMaxToMeanRatio(double a);
    double getMaxToMeanRatio(){return _maxToMeanRatio;}

    void setMaxCost(double max){_maxCost = max;}
    void setMinCost(double min){_minCost = min;}
    double getMaxCost(){return _maxCost;}
    double getMinCost(){return _minCost;}

    void setHumans(std::vector<Robot*> humans);
    void setRobot(Robot* rob){_rob=rob;}
    void setCostJoints(std::vector<Joint*> joits);

    void initHumanPos();
    void clearHumanPos();
private:

    Robot *_rob;
    std::vector<Joint*> _cost_joints;
    std::vector<Robot*> _humans;
    std::vector<Eigen::Vector2d> _human_pos;

    double _ignore_threshold;
    double _minCost; /// the cost to apply where the robot is at more than _ignore_threshold from humans
    double _maxCost; /// the cost when the robot is on the human
    double _maxToMeanRatio; ///< the ratio "a" when using getCost, computes the cost as c = a*cmax + (1-a)*cmean. needs to be in [0,1]

};

} // namespace HRICS

#endif // HRICS_DISTANCE2D_H
