#include "Logger.h"
#ifndef MOVE3D_LOG_OFF
#include <log4cxx/consoleappender.h>
#include <log4cxx/patternlayout.h>
#include <log4cxx/basicconfigurator.h>
#include <log4cxx/helpers/exception.h>
#include <log4cxx/defaultconfigurator.h>
#include <log4cxx/propertyconfigurator.h>
#include <log4cxx/config/propertysetter.h>
#include <log4cxx/ndc.h>
#include <log4cxx/mdc.h>
#endif
#include <stdio.h>
#include <iostream>
#include <cstdlib>

namespace logm3d {
INIT_MOVE3D_LOG("planners.logger");

#ifndef MOVE3D_LOG_OFF
bool autoConfigureFromFile(){
    bool ok=false;
    std::string m3dhome, def_file("logm3d.properties"),def_cxx_file("log4cxx.properties");
    std::string filename;

    char *g=getenv("HOME_MOVE3D");
    if(g){
        m3dhome=g;
        m3dhome+="/";
    }
    FILE *f;
    enum STATE {DEF_VAR_M=0,DEF_VAR_C,DEF_M,DEF_C,H_DEF_M,H_DEF_C,DEF_SIZE} ;
    STATE i=STATE(0);
    while(!ok && i<DEF_SIZE){

        f=NULL;
        filename.clear();
        switch (i){
        case DEF_VAR_M:
            g= getenv("LOGM3D_CONFIGURATION");
            if(g)
                filename =g;
            break;
        case DEF_VAR_C:
            g= getenv("LOG4CXX_CONFIGURATION");
            if(g)
                filename=g;
            break;
        case DEF_M:
            filename = def_file;
            break;
        case DEF_C:
            filename = def_cxx_file;
            break;
        case H_DEF_M:
            filename = m3dhome;
            if(filename.size())
                filename.append(def_file);
            break;
        case H_DEF_C:
            filename = m3dhome;
            if(filename.size())
                filename.append(def_cxx_file);
            break;
        default:
            filename.clear();
        }

        if(filename.size())
            std::cout << filename<<std::endl;
            f=fopen(filename.c_str(),"r");
        if(f){
            fclose(f);
            ok=true;
            break;
        }
        i=(STATE)((int)i +1);
    }
    if(ok){
        log4cxx::PropertyConfigurator::configure(filename);
        M3D_INFO("using log4cxx configuration file "<<filename);

        using namespace std;
        for(unsigned int i=0;i<log4cxx::Logger::getRootLogger()->getAllAppenders().size();i++){
            log4cxx::AppenderPtr ap = log4cxx::Logger::getRootLogger()->getAllAppenders()[i];
            ap->setOption("application","move3d");
            cout << "logger :"<<endl;
            cout << ap->getName() << endl;
            cout << "class: "<<ap->getClass().getName();
        }
    }


    return ok;
}
#endif

void initializePlannerLogger(){
#ifdef MOVE3D_LOG_OFF
    return;
#else
    bool ok=false;
    log4cxx::NDC::push("planners");
    if(log4cxx::Logger::getRootLogger()->getAllAppenders().size() == 0){
        ok=autoConfigureFromFile();
    }
    if(!ok){
        log4cxx::ConsoleAppender * console = new log4cxx::ConsoleAppender(log4cxx::LayoutPtr(new log4cxx::PatternLayout("%d{HH:mm:ss,SSS} \33[33m%-5p\33[0m %c - %m%n") ) );
        console->setOption("application","move3d");
        log4cxx::BasicConfigurator::configure(log4cxx::AppenderPtr(console));
        log4cxx::Logger::getRootLogger()->setLevel(log4cxx::Level::getDebug());
        M3D_INFO("using progammatically defined default logger settings for log4cxx");
    }
#endif
}


LoggerPtr getLogger(std::string name){
#ifdef MOVE3D_LOG_OFF
    return 0;
#else
    return log4cxx::Logger::getLogger(name);
#endif
}

}

