/*
 * EST.cpp
 *
 *  Created on: Sep 16, 2009
 *      Author: jmainpri
 */

#include "EST.hpp"

#include "Planner-pkg.h" 

#include "planEnvironment.hpp"

#include "API/Roadmap/node.hpp"
#include "API/Roadmap/graph.hpp"

using namespace std;
using namespace tr1;


/**
 * ****************************************************
 *      EST Expansion
 * ****************************************************
 */

/**
 * EST
 */
Node* ESTExpansion::getExpansionNode(vector<Node*>& nodes) {
	// Get id biased to high values
	double x = (double) (pow(p3d_random(0, 1), 3));
	unsigned int size = nodes.size();
	unsigned int id = (unsigned int) (x * (double) (size - 1));
	//	cout << "id = " <<  id << endl;
	//	printAllNodes(nodes);

	int MaxFail = 5;

	for (;;) {
		unsigned int i = 0;
		vector<Node*>::iterator it = nodes.begin();
		// Get member of the map
		for (; it != nodes.end(); it++, i++) {
			//		if (i == id)
			//		{
			Node* thisNode(*it);

			if (thisNode->getNbExpandFailed() < MaxFail) {
				//					cout << "id = " << i << endl;
				return thisNode;
			}

			//			if( thisNode->getNumberOfNeighbors() > 1 )
			//			{
			//				if( nodes.size() > 10)
			//				{
			//					nodes.erase(it);
			//				}
			//				return getExpansionNode(nodes);

			//				nodes.erase(it);
			//				return thisNode;
			//			}
			//			cout << "Size = " << nodes.size() << endl;

			//			printAllNodes(nodes);
			//		}
		}
		MaxFail++;
	}

	cout << "Problem not finding node" << endl;
	//	Node* expandNode;
	return nodes[id];

}

/**
 * EST
 */
shared_ptr<Configuration> ESTExpansion::getExpansionDirection(Node* expansionNode, Node* toComp) {
	shared_ptr<Configuration> q;

	q = m_Graph->getRobot()->shoot(true);
	return q;

}

/**
 * EST
 */
Node* ESTExpansion::expandProcessEST(Node* expansionNode,
		std::tr1::shared_ptr<Configuration> directionConfig, unsigned& nbCreatedNodes) {
	shared_ptr<Configuration> fromConfig = expansionNode->getConfiguration();
	shared_ptr<Configuration> toConfig;

	LocalPath directionLocalpath(fromConfig, directionConfig);

	double pathDelta = directionLocalpath.getParamMax() == 0. ? 1.
			: MIN(1., step() / directionLocalpath.getParamMax() );

	// Expansion control
	// Discards potential nodes that are to close to the graph
	if (ENV.getBool(Env::refinementControl) && !refinementControl(directionLocalpath, pathDelta,
			*expansionNode)) {
		nbCreatedNodes = 0;
		return 0;
	}

	double param(0);
	double paramMax = directionLocalpath.getParamMax();

	// Perform extension toward directionConfig
	param = step();

	if (param > paramMax) {
		toConfig = directionConfig;
	} else {
		toConfig = directionLocalpath.configAtParam(param);
	}

	LocalPath extensionLocalpath(fromConfig, toConfig);

	Node* ptrToNode(NULL);

	Node* neigh = m_Graph->nearestWeightNeighbour(expansionNode, toConfig, false,
			ENV.getInt(Env::DistConfigChoice));

	if (neigh->getConfiguration()->dist(*toConfig) < 2 * step() / 3) {
		return ptrToNode;
	}

	//	if (transitionTest(*expansionNode, extensionLocalpath))
	//	{
	if (extensionLocalpath.isValid()) {
		//		cout << "Adding node" << endl;

		Node* newNode = addNode(expansionNode, extensionLocalpath, pathDelta, ptrToNode, nbCreatedNodes);

		//			double sum = newNode->sumCost();
		//
		//			cout << "Sum = " << newNode->sumCost() << endl;
		//			cout << "Cost = " << newNode->getCost() << endl;
		newNode->setSelectCost(newNode->sumCost());

		//		cout << "nbCreatedNodes = " << nbCreatedNodes << endl;
		return newNode;
	}
	//	}
	else {
		return ptrToNode;
	}
}


/**
 * ****************************************************
 *             EST
 * ****************************************************
 */


/**
 * Initialize EST.
 * @return the number of nodes added to the graph
 */
unsigned EST::init()
{
	expansion = new ESTExpansion(_Graph);

	sortedNodes = _Start->getSortedNodes();
	this->addNodeToSet(_Start);
	_Start->setSortedNodes(sortedNodes);

	//	if(ENV.getBool(Env::biDir))
	//	{
	sortedNodes = _Goal->getSortedNodes();
	this->addNodeToSet(_Goal);
	_Goal->setSortedNodes(sortedNodes);
	//	}

	return TreePlanner::init();
}


/**
 * Perform a single expansion step of EST, growing the connected component to which the given node belongs.
 * @return the number of created nodes
 */
unsigned EST::expandOneStep(Node * fromComp)
{
	sortedNodes = fromComp->getSortedNodes();

	// get node for expansion
	Node * expansionNode = expansion->getExpansionNode(sortedNodes);

	// get direction
	confPtr_t directionConf = this->sampleExpansionDirection(expansionNode);

	// expansion
	unsigned nbAddedNodes(0);
	Node * extensionNode = expansion->expandProcessEST(expansionNode, directionConf, nbAddedNodes);

	if (nbAddedNodes > 0)
	{
		this->addNodeToSet(extensionNode);
		fromComp->setSortedNodes(sortedNodes);
	}
	else
	{
		expansionNode->setExpandFailed();
		m_last_node = expansion->getLasAddedNode();
	}

	return nbAddedNodes;
}


/**
 * Check the stopping conditions of EST.
 * @return TRUE if one of the conditions is satisfied, and FALSE otherwise
 */
bool EST::checkStopConditions()
{
	if (TreePlanner::checkStopConditions())
		return true;

	if (ENV.getBool(Env::expandToGoal))
		return (_Start->inSameComponent(_Goal));

	return false;
}


/**
 * Check the preconditions.
 * @return TRUE if all conditions are satisfied, and FALSE otherwise
 */
bool EST::preConditions()
{
	if (TreePlanner::preConditions())
	{
		if (ENV.getBool(Env::expandToGoal))
		{
			LocalPath direct(_Start->getConfiguration(), _Goal->getConfiguration());
			if (direct.isValid())
			{
				this->connectNodeToComp(this->getInit(), this->getGoal());
				cout << "Direct connection" << endl;
				return true;
			}
		}

		if (ENV.getBool(Env::expandToGoal))
		{
			if (this->connectNodeToComp(this->getInit(), this->getGoal()))
				return false;
		}
		return true;
	}
	else
		return false;
}


/**
 * Sample a direction for the expansion process (taking into account the goal bias, if relevant).
 * @param directionNode: contains the goal node when relevant
 * @return the configuration toward which the tree will be expanded
 */
confPtr_t EST::sampleExpansionDirection(Node * directionNode)
{
	// bias toward the goal configuration
	if (ENV.getBool(Env::isGoalBiased) && p3d_random(0., 1.) <= ENV.getDouble(Env::Bias))
	{
		directionNode = _Goal;
		return _q_goal;
	}
	// or random sampling in the entire C-Space
	else
		return _Graph->getRobot()->shoot();
}


/**
 * Try to connect the given node to the given component taking into account that the space is a cost space.
 * @return TRUE if the connection is successful, and FALSE otherwise
 */
bool EST::connectNodeToComp(Node * node, Node * compNode)
{
	bool savedIsMaxDis = PlanEnv->getBool(PlanParam::isMaxDisNeigh);
	PlanEnv->setBool(PlanParam::isMaxDisNeigh, false);
	Node * node2 = _Graph->nearestWeightNeighbour(compNode, node->getConfiguration(), false,
			ENV.getInt(Env::DistConfigChoice));
	p3d_SetIsMaxDistNeighbor(savedIsMaxDis);

	LocalPath path(node->getConfiguration(), node2->getConfiguration());

	if (!ENV.getBool(Env::costBeforeColl))
	{
		if (path.isValid())
		{
			if (path.getParamMax() <= expansion->step())
			{
				unsigned nbCreatedNodes(0);
				expansion->addNode(node, path, 1., node2, nbCreatedNodes);
				return true;
			}
		}
		return false;
	}
	else
	{
		if (path.getParamMax() <= expansion->step())
		{
			if (path.isValid())
			{
				unsigned nbCreatedNodes(0);
				expansion->addNode(node, path, 1., node2, nbCreatedNodes);
				return true;
			}
		}
		return false;
	}
}


class myNodeComparator
{

public:
	EST * ptrNode;

	bool operator()(Node * node1, Node * node2)
	{
		return (node1->getSelectCost() < node2->getSelectCost());
	}

} nodeCompObject;


/**
 * Add the given node to the sorted set.
 */
void EST::addNodeToSet(Node * node)
{
	sortedNodes.push_back(node);

	sort(sortedNodes.begin(), sortedNodes.end(), nodeCompObject);
}

