/*
 * planner_functions.cpp
 *
 *  Created on: Jul 26, 2009
 *      Author: jmainpri
 */

#include "../p3d/env.hpp"

#include "planEnvironment.hpp"
#include "plannerFunctions.hpp"
#include "ReplanningSimulator.hpp"
#include "../p3d/env.hpp"

#include "planner/planner.hpp"

#include "planner/PRM/PRM.hpp"
#include "planner/PRM/DirectedPRM.hpp"
#include "planner/PRM/Visibility.hpp"
#include "planner/PRM/ACR.hpp"
#include "planner/PRM/PerturbationRoadmap.hpp"

#include "planner/Diffusion/RRT.hpp"
#include "planner/Diffusion/EST.hpp"
#include "planner/Diffusion/Variants/Transition-RRT.hpp"
#include "planner/Diffusion/Variants/TransitionEdge-RRT.hpp"
#include "planner/Diffusion/Variants/ManhattanLike-RRT.hpp"
#include "planner/Diffusion/Variants/Multi-RRT.hpp"
#include "planner/Diffusion/Variants/Multi-T-RRT.hpp"
#include "planner/Diffusion/Variants/Threshold-RRT.hpp"
#include "planner/Diffusion/Variants/Multi-Threshold-RRT.hpp"
#include "planner/Diffusion/Variants/RRT-Star.hpp"
#include "planner/Diffusion/Variants/DirectedRRT.hpp"

#include "planner/TrajectoryOptim/trajectoryOptim.hpp"
#include "planner/TrajectoryOptim/Classic/costOptimization.hpp"
#include "planner/cost_space.hpp"

#ifdef HRI_COSTSPACE
#include "HRI_costspace/HRICS_costspace.hpp"
#include "HRI_costspace/RRT/HRICS_rrt.hpp"
#include "HRI_costspace/RRT/HRICS_rrtPlan.hpp"
#endif

#include "Util-pkg.h"
#include "P3d-pkg.h"
#include "move3d-headless.h"
#include "Planner-pkg.h"
#ifdef LIGHT_PLANNER
#include "LightPlanner-pkg.h"
#endif


using namespace std;
using namespace tr1;

// ---------------------------------------------------------------------------------
// Exctracted trajectory Id (used for UI)
// ---------------------------------------------------------------------------------
unsigned int trajId = 0;

// ---------------------------------------------------------------------------------
// Last trajectory
// ---------------------------------------------------------------------------------
API::Trajectory last_traj;

API::Trajectory p3d_get_last_trajectory()
{
  return last_traj;
}

// ---------------------------------------------------------------------------------
// Run Id (used for multi-run)
// ---------------------------------------------------------------------------------
unsigned int runId = 0;

void p3d_planner_functions_set_run_id( unsigned int idRun )
{
	runId = idRun;
}

unsigned int p3d_planner_functions_get_run_id()
{
  return runId;
}

// ---------------------------------------------------------------------------------
// RRT statistcis
// ---------------------------------------------------------------------------------
RRTStatistics rrt_statistics;

void p3d_get_rrt_statistics( RRTStatistics& stat )
{
  stat = rrt_statistics;
}

// ---------------------------------------------------------------------------------
// Trajectory statistcis
// ---------------------------------------------------------------------------------
TrajectoryStatistics traj_statistics;

void p3d_get_traj_statistics( TrajectoryStatistics& stat )
{
  stat = traj_statistics;
}

// ---------------------------------------------------------------------------------
// Set back costspace after RRT
// ---------------------------------------------------------------------------------
static bool set_costspace=false;

// ---------------------------------------------------------------------------------
// Extract Traj
// ---------------------------------------------------------------------------------
p3d_traj* p3d_extract_traj(bool is_traj_found, int nb_added_nodes, Graph* graph, confPtr_t q_source, confPtr_t q_target) 
{
	API::Trajectory * traj = NULL;
	p3d_rob * rob = graph->getRobot()->getRobotStruct();

	// if trajectory has been found, extract it from the graph
	if (is_traj_found)
	{
		// case of direct connection
		if (nb_added_nodes == 2)
		{
			vector<confPtr_t> configs;
			configs.push_back(q_source);
			configs.push_back(q_target);
			traj = new API::Trajectory(configs);
		}
		else
		{
			if (ENV.getBool(Env::isMultiRRT) && ENV.getBool(Env::isOneToOne))
				traj = dynamic_cast<MultiRRT*>(global_Move3DPlanner)->extractTrajectory(q_source, q_target);
			else
				traj = global_Move3DPlanner->extractTrajectory();
		}
	}

	trajId++;

	if (traj)
	{
		p3d_traj * result = NULL;
		result = traj->replaceP3dTraj(NULL);
		rob->tcur = result;
		char trajName[] = "Specific";
		g3d_add_traj(trajName, trajId, rob, rob->tcur);

		traj->costDeltaAlongTraj();

		/**  FlyCrane @ CATEC **/
//		ofstream catec("trajectory.txt");
//		vector<LocalPath *> localPaths = traj->getCourbe();
//		confPtr_t conf = localPaths[0]->getBegin();
//		Eigen::Vector3d * CM = conf->getQuadCMs();
//		catec << showpoint << fixed << setprecision(1);
//		catec << '1' << setw(8) << CM[0][0] << setw(8) << CM[0][1] << setw(8) << CM[0][2];
//		catec << "     2" << setw(8) << CM[1][0] << setw(8) << CM[1][1] << setw(8) << CM[1][2];
//		catec << "     3" << setw(8) << CM[2][0] << setw(8) << CM[2][1] << setw(8) << CM[2][2] << endl;
//		for (unsigned i(0); i < localPaths.size(); ++i) {
//			conf = localPaths[i]->getEnd();
//			CM = conf->getQuadCMs();
//			catec << '1' << setw(8) << CM[0][0] << setw(8) << CM[0][1] << setw(8) << CM[0][2];
//			catec << "     2" << setw(8) << CM[1][0] << setw(8) << CM[1][1] << setw(8) << CM[1][2];
//			catec << "     3" << setw(8) << CM[2][0] << setw(8) << CM[2][1] << setw(8) << CM[2][2] << endl;
//		}
//		catec.close();

		// Prevent segfault if p3d graph deleted outside
		delete traj;
		return result;
	}
	else
	{
        if (PlanEnv->getBool(PlanParam::showGTPDebug))
            cout << __FILE__ << " , " << __func__ << " : No traj found" << endl;
		return NULL;
	}
}

// ---------------------------------------------------------------------------------
// Allocates an RRT depending on env variables
// ---------------------------------------------------------------------------------
RRT* p3d_allocate_rrt(Robot* rob, Graph* graph) {
	RRT* rrt;

	if (ENV.getBool(Env::isManhattan))
		rrt = new ManhattanLikeRRT(rob, graph);

	else if (ENV.getBool(Env::isMultiRRT) && ENV.getBool(Env::isCostSpace)
			&& ENV.getBool(Env::costThresholdRRT))
		rrt = new MultiThresholdRRT(rob, graph);

	else if (ENV.getBool(Env::isMultiRRT) && ENV.getBool(Env::isCostSpace))
		rrt = new MultiTRRT(rob, graph);

	else if (ENV.getBool(Env::isMultiRRT))
		rrt = new MultiRRT(rob, graph);

	else if (ENV.getBool(Env::costStarRRT))
		rrt = new StarRRT(rob, graph);

#ifdef HRI_COSTSPACE
	else if (ENV.getBool(Env::HRIPlannerWS) && ENV.getBool(Env::HRIPlannerTRRT))
		rrt = new HRICS_RRT(rob, graph);

	else if (ENV.getBool(Env::HRIPlannerCS) && ENV.getBool(Env::HRIPlannerTRRT))
		rrt = new HRICS_RRTPlan(rob, graph);
#endif

	else if (ENV.getBool(Env::isCostSpace) && ENV.getBool(Env::costThresholdRRT))
		rrt = new ThresholdRRT(rob, graph);

	else if (ENV.getBool(Env::isCostSpace) && ENV.getBool(Env::useTRRT))
		rrt = new TransitionRRT(rob, graph);
    else if (ENV.getBool(Env::isCostSpace) && ENV.getBool(Env::useTERRT))
        rrt = new TransitionEdgeRRT(rob, graph);

    else if (PlanEnv->getBool(PlanParam::orientedGraph))
        rrt = new DirectedRRT(rob, graph);

	else
	{
		if (ENV.getBool(Env::isCostSpace) && (!ENV.getBool(Env::useTRRT)))
		{
			ENV.setBool(Env::isCostSpace, false);
			set_costspace = true;
		}

		rrt = new RRT(rob, graph);
	}

	return rrt;
}

// ---------------------------------------------------------------------------------
// Planner function (RRT) for connection with Manipulation planner
// ---------------------------------------------------------------------------------
p3d_traj* p3d_planner_function(p3d_rob* robotPt, configPt qs, configPt qg)
{
    char *filename=ENV.getLogFile();

    if(filename)
        ChronoOn();

    ENV.setBool(Env::armOnly,p3d_has_a_quad_joint(robotPt) && ENV.getBool(Env::armOnly));
    ENV.setBool(Env::quadOnly,p3d_has_a_quad_joint(robotPt) && ENV.getBool(Env::quadOnly));

    if(ENV.getBool(Env::quadOnly) && robotPt->lpl_type==QUADSPLINE)
        ENV.setBool(Env::isHybrid,false);

    if(ENV.getBool(Env::isHybrid) || ENV.getBool(Env::armOnly) || robotPt->lpl_type==LINEAR){
        p3d_set_robot_steering_method("Linear");
        PlanEnv->setBool(PlanParam::orientedGraph,false);
    }

    if(robotPt->lpl_type==QUADSPLINE)
        PlanEnv->setBool(PlanParam::orientedGraph,true);

	// Gets the robot pointer and the 2 configurations
	Robot* rob = global_Project->getActiveScene()->getRobotByName(robotPt->name);
	confPtr_t q_source(new Configuration(rob, qs));
	confPtr_t q_target(new Configuration(rob, qg));

	// Allocate the p3d_graph if does't exist
	if (!API_activeGraph || ENV.getInt(Env::nbMultiRun) > 1) {
		delete API_activeGraph;
		API_activeGraph = new Graph(rob);
	}
	Graph * graph = API_activeGraph;

	// Delete last global planner
	if (global_Move3DPlanner)
		delete global_Move3DPlanner;

	// Main Run functions of all RRTs
	// All RRTs are initilized with init and run here
	int nb_added_nodes = 0;
	RRT* rrt = p3d_allocate_rrt(rob, graph);
	global_Move3DPlanner = rrt;
	nb_added_nodes += rrt->setInit(q_source);
	nb_added_nodes += rrt->setGoal(q_target);
    //nb_added_nodes += rrt->init();
    rrt->init();
    for(int i=0;i<nb_added_nodes;i++){
        rrt->addConnectibleNode();
    }
	rrt->setInitialized(true);
	rrt->setRunId(runId);

	cout << "* PLANNING ***************" << endl;
    ChronoTimeOfDayOn();

    nb_added_nodes += rrt->run();

	ChronoTimeOfDayTimes(&(graph->getGraphStruct()->rrtTime));
	ChronoTimeOfDayOff();

	graph->getGraphStruct()->totTime = graph->getGraphStruct()->rrtTime;

	if (ENV.getBool(Env::drawGraph) && globalRePlanningEnv) {
		globalRePlanningEnv->storeGraphToDraw(*graph);
	}

	// Extract the trajectory if one exists, else return NULL
	p3d_traj* traj = p3d_extract_traj(rrt->trajFound(), nb_added_nodes, graph, q_source, q_target);

    if(filename){
        double CPU_time,ts;
        ChronoMicroTimes(&CPU_time,&ts);
        ChronoOff();
        cout<<"CPU_time="<<CPU_time<<endl;
        int success=(traj!=NULL);
        int nb_expansions=rrt->getNumberOfExpansion();
        int nb_nodes=graph->getNumberOfNodes();
        int nb_edges=graph->getNumberOfEdges();
        double flying_time=robotPt->tcur->range_param;
        double connectible_rate=100*(double)rrt->getNbConnectibleNodes()/(double)nb_added_nodes;
        FILE *f=fopen(filename,"a");
        if(f){
            fprintf(f,"%d %f %d %d %d %d %f %f %d\n",
                    success, //1
                    CPU_time, //2
                    nb_expansions, //3
                    nb_added_nodes, //4
                    nb_nodes, //5
                    nb_edges, //6
                    flying_time, //7
                    100-connectible_rate, //8
                    0); //9
            fclose(f);
        }
    }

    if (PlanEnv->getBool(PlanParam::showGTPDebug))
        cout << "** ** --------------------------" << endl;

	if (traj) {
		cout << "Success" << endl;
	} else {
		cout << "Fail" << endl;
	}

    if (PlanEnv->getBool(PlanParam::showGTPDebug))
    {
        cout << "TIME =" << graph->getGraphStruct()->rrtTime << " sec." << endl;
        cout << "NB NODES " << graph->getNumberOfNodes() << endl;
        cout << "NB EXPANSION " << rrt->getNumberOfExpansion() << endl;

        cout << "** ** --------------------------" << endl;
    }
    cout << "* PLANNING END ***************" << endl;

	rrt_statistics.runId = rrt->getRunId();
	rrt_statistics.succeeded = (traj != NULL);
	rrt_statistics.time = graph->getGraphStruct()->rrtTime;
	//rrt_statistics.cost = 0.0;
	rrt_statistics.nbNodes = graph->getNumberOfNodes();
	rrt_statistics.nbExpansions = rrt->getNumberOfExpansion();

	if (set_costspace)
		ENV.setBool(Env::isCostSpace, true);

	return traj;
}

p3d_traj* pathPt=NULL;

// ---------------------------------------------------------------------------------
// Smoothing function (Shortcut) for connection with Manipulation planner
// ---------------------------------------------------------------------------------
void p3d_smoothing_function( p3d_rob* robotPt, p3d_traj* traj, int nbSteps, double maxTime )
{
  cout << "** SMOOTHING ***************" << endl;

  if( robotPt == NULL )
  {
    cout << "robot not defined in " << __func__ << endl;
    return;
  }
  
  if( maxTime != -1 )
    PlanEnv->setDouble( PlanParam::timeLimitSmoothing, maxTime );

  Robot* rob = global_Project->getActiveScene()->getRobotByName( robotPt->name );  

  API::Trajectory t;
  
  if(PlanEnv->getBool(PlanParam::withDeformation) || PlanEnv->getBool(PlanParam::withShortCut) )
  {
    double optTime = 0.0;
    
    if( traj == NULL ) {
      cout << "trajectory NULL in " << __func__ << endl;
      return;
    }
//    cout << "traj->nlp = " << traj->nlp << endl;
//    cout << "traj->range_param = " << traj->range_param << endl;
    
    API::CostOptimization optimTrj( rob, traj );
    
    optimTrj.setRunId( runId );
#ifdef QT_LIBRARY
    optimTrj.setContextName( ENV.getString(Env::nameOfFile).toStdString() );
#endif
    
    if(PlanEnv->getBool(PlanParam::withDeformation))
    {
      optimTrj.resetCostComputed();
      optimTrj.runDeformation( nbSteps , runId );
      optTime += optimTrj.getTime();
    }
    
    confPtr_t startBackup = optimTrj.getBegin();
    confPtr_t endBackup = optimTrj.getEnd();

    if(PlanEnv->getBool(PlanParam::withShortCut))
    {
      optimTrj.resetCostComputed();
      optimTrj.runShortCut( nbSteps, runId );
      optTime += optimTrj.getTime();
    }
    
    confPtr_t start = optimTrj.getBegin();
    confPtr_t end = optimTrj.getEnd();
	if (!startBackup->equal(*start)) {
		cout << "start conformation has changed: " << endl;
		startBackup->print();
		start->print();
	}
	if (!endBackup->equal(*end)) {
		cout << "end conformation has changed: " << endl;
		endBackup->print();
		end->print();
	}

    optimTrj.replaceP3dTraj();
    
//    optimTrj.resetCostComputed();
//    cout << "optimTrj.getRangeMax() : " << optimTrj.getRangeMax()  << endl;
//
//    if( PlanEnv->getBool(PlanParam::trajComputeCostAfterPlannif) )
//    {
//      optimTrj.costStatistics( traj_statistics );
//
//      cout << "--- stats on traj ---" << endl;
//      cout << " length = " << traj_statistics.length << endl;
//      cout << " max = " << traj_statistics.max << endl;
//      cout << " average = " << traj_statistics.average << endl;
//      cout << " integral = " << traj_statistics.integral << endl;
//      cout << " mecha_work = " << traj_statistics.mecha_work << endl;
//      cout << "---------------------" << endl;
//    }
    
    last_traj = API::Trajectory(optimTrj);
  }
  
  if( PlanEnv->getBool( PlanParam::withStomp ) )
  {
    traj_optim_runStompNoReset( runId );
    
    t = API::Trajectory( rob, rob->getRobotStruct()->tcur );
    t.resetCostComputed();
    
    if( PlanEnv->getBool(PlanParam::trajComputeCostAfterPlannif) )
    {
      t.costStatistics( traj_statistics );
      
      cout << "--- stats on traj ---" << endl;
      cout << " length = " << traj_statistics.length << endl;
      cout << " max = " << traj_statistics.max << endl;
      cout << " average = " << traj_statistics.average << endl;
      cout << " integral = " << traj_statistics.integral << endl;
      cout << " mecha_work = " << traj_statistics.mecha_work << endl;
      cout << "---------------------" << endl;
    }
  }

  pathPt = robotPt->tcur;
}


void p3d_smoothing_function_dummy( p3d_rob* robotPt, p3d_traj* traj, int nbSteps, double maxTime )
{

}

// ---------------------------------------------------------------------------------
// set IK function
// ---------------------------------------------------------------------------------
configPt (*ext_generate_goal_configuration)();

void p3d_set_goal_solution_function( configPt (*fct)() )
{
  ext_generate_goal_configuration = fct;
}

// ---------------------------------------------------------------------------------
// Generate IK function
// ---------------------------------------------------------------------------------
bool p3d_generate_goal_soution( Configuration& q )
{
  if( ext_generate_goal_configuration != NULL )
  {
    configPt q_new = ext_generate_goal_configuration();
    
    q.Clear();
    
    if( q_new != NULL )
    {
      q.setConfiguration( q_new );
      return true;
    }
  }
  return false;
}

// ---------------------------------------------------------------------------------
// Tree Planners
// ---------------------------------------------------------------------------------
int p3d_run_rrt(p3d_rob* robotPt){

    int has_a_quad_joint=p3d_has_a_quad_joint(robotPt);

    p3d_traj* path = p3d_planner_function(robotPt,robotPt->ROBOT_POS,robotPt->ROBOT_GOTO);

    if( path && path->courbePt && !PlanEnv->getBool(PlanParam::stopPlanner) && PlanEnv->getBool(PlanParam::withSmoothing))
    {
        double max_iteration = PlanEnv->getInt(PlanParam::smoothMaxIterations);
        double max_time = PlanEnv->getDouble( PlanParam::timeLimitSmoothing );

        API::Smoothing *optimTrj=new API::Smoothing(new Robot(robotPt),path);
        double old_cost=optimTrj->cost();
        cout<<"Previous cost: "<<old_cost<<endl;

        if(ENV.getBool(Env::isHybrid) && !ENV.getBool(Env::armOnly)){
            cout<<"Switch to quad"<<endl;
            p3d_switch_to_quad(robotPt,path);
            delete optimTrj;
            optimTrj=new API::Smoothing(new Robot(robotPt),path);
            old_cost=optimTrj->cost();
            cout<<"New cost: "<<old_cost;
            if(ENV.getBool(Env::isCostSpace)) cout<<" (range_param="<<optimTrj->getRangeMax()<<")";
            cout<<endl;
            if((!ENV.getBool(Env::drawDisabled)) && ENV.getBool(Env::drawTraj)) g3d_draw_allwin_active();
        }

        if(has_a_quad_joint || (ENV.getBool(Env::isCostSpace) && global_costSpace->getSelectedCostName().compare("costFlyCrane")==0)){
            cout<<"Remove useless"<<endl;
            if(optimTrj->removeUselessNodes()>0){
                optimTrj->replaceP3dTraj();
                path=robotPt->tcur;
            }
            cout<<"New cost: "<<optimTrj->cost();
            if(ENV.getBool(Env::isCostSpace)) cout<<" (range_param="<<optimTrj->getRangeMax()<<")";
            if((!ENV.getBool(Env::drawDisabled)) && ENV.getBool(Env::drawTraj)) g3d_draw_allwin_active();
        }

        delete optimTrj;
        cout<<"Random shortcut"<<endl;
        p3d_smoothing_function(robotPt, path, max_iteration, max_time);
        optimTrj=new API::Smoothing(new Robot(robotPt),path);
        double new_cost=optimTrj->cost();
        cout<<"New cost: "<<new_cost<<endl;

        if((!ENV.getBool(Env::drawDisabled)) && ENV.getBool(Env::drawTraj)) g3d_draw_allwin_active();

        if(has_a_quad_joint || (ENV.getBool(Env::isCostSpace) && global_costSpace->getSelectedCostName().compare("costFlyCrane")==0)){
            cout<<"Remove useless"<<endl;
            if(optimTrj->removeUselessNodes()>0){
                optimTrj->replaceP3dTraj();
                path=robotPt->tcur;
            }
            new_cost=optimTrj->cost();
            cout<<"New cost: "<<new_cost<<endl;
            if((!ENV.getBool(Env::drawDisabled)) && ENV.getBool(Env::drawTraj)) g3d_draw_allwin_active();
        }

        cout<<"Smoothing gain: "<<100*(old_cost-new_cost)/old_cost<<" %"<<endl;

        if(ENV.getBool(Env::armOnly)){
            cout<<"Switch to quad"<<endl;
            p3d_switch_to_quad(robotPt,path);
            delete optimTrj;
            optimTrj=new API::Smoothing(new Robot(robotPt),path);
            cout<<"New cost: "<<optimTrj->cost();
            delete optimTrj;
        }
    }

    return (path != NULL);
}


int p3d_plan(p3d_rob* robotPt, configPt from, configPt to, int robotPart, bool cartesian) {//, Graph* fromGraph){

	// Gets the robot pointer and the 2 configurations
	Robot* rob = global_Project->getActiveScene()->getRobotByName(robotPt->name);
	configPt qstart, qgoal;

	qstart = from;
	qgoal = to;

    //if((strcmp(robotPt->name,"PR2_ROBOT") == 0)
    //	||(strcmp(robotPt->name,"BOB_ROBOT") == 0)
    //	||(strcmp(robotPt->name,"MAX_ROBOT") == 0))
    if( rob->getHriAgent()
            && rob->getHriAgent()->type == HRI_PR2 )
	{
		int armId = -1;
		if(robotPart == PLAN_RIGHT_ARM && cartesian)
		{
			armId = 0;
		}

		if(robotPart == PLAN_LEFT_ARM && cartesian)
		{
			armId = 1;
		}

        if(p3d_equal_config(robotPt,qstart,qgoal,false))
		{
			cout<<"Goal is the same as start"<<endl;
			return PLAN_STATUS_START_EQUALS_GOAL;
		}


		p3d_cntrt* ct;

  		// over all constraints
		for(int i=0; i<robotPt->cntrt_manager->ncntrts; i++) 
		{
			string name = robotPt->cntrt_manager->cntrts[i]->namecntrt;

			if( name == "p3d_fix_jnts_relpos" ){
				continue;
			}

			cout << "deactivate : " << name << endl;

    		// get constraint from the cntrts manager
			ct = robotPt->cntrt_manager->cntrts[i];
			p3d_desactivateCntrt( robotPt, ct );
		}

		if(robotPart == PLAN_BASE)
		{
			p3d_multiLocalPath_disable_all_groupToPlan( robotPt , false );
			p3d_multiLocalPath_set_groupToPlan( robotPt, 1, 1, false);
			fixAllJointsExceptBase( robotPt );
		}
		else if(robotPart == PLAN_FULL)
		{
			p3d_multiLocalPath_enable_all_groupToPlan( robotPt );
			unFixAllJoints(robotPt);
		}
		else if(robotPart == PLAN_RIGHT_ARM && !cartesian)
		{
			p3d_multiLocalPath_disable_all_groupToPlan( robotPt , false );
			p3d_multiLocalPath_set_groupToPlan( robotPt, 2, 1, false);
			fixAllJointsWithoutArm( robotPt , 0 );
		}
		else if(robotPart == PLAN_LEFT_ARM  && !cartesian)
		{
			p3d_multiLocalPath_disable_all_groupToPlan( robotPt , false );
			p3d_multiLocalPath_set_groupToPlan( robotPt, 2, 1, false);
			fixAllJointsWithoutArm( robotPt , 1 );
		}

		if(cartesian)
		{
			ArmManipulationData& mData = (*robotPt->armManipulationData)[armId];
			ct = mData.getCcCntrt();
			for(int i = 0; i < ct->npasjnts; i++){
				p3d_jnt * joint = ct->pasjnts[i];
				if(joint->type != P3D_FIXED){
					unFixJoint(robotPt, joint);
				}
			}
			if((!strcmp(ct->namecntrt, "p3d_kuka_arm_ik")) ||
				(!strcmp(ct->namecntrt, "p3d_lwr_arm_ik")) ||
				(!strcmp(ct->namecntrt, "p3d_pr2_arm_ik"))){
				p3d_jnt * joint = robotPt->joints[ct->argu_i[0]];
				unFixJoint(robotPt, joint);
			}
			mData.setCartesian( true );

			ArmManipulationData& mDataOther = (*robotPt->armManipulationData)[(armId+1)%2];
			mDataOther.setCartesian( false );
		}
		else
		{
			ArmManipulationData& mData = (*robotPt->armManipulationData)[0];
			mData.setCartesian( false );
			ArmManipulationData& mData1 = (*robotPt->armManipulationData)[1];
			mData1.setCartesian( false );
		}

			for (int i=0; i < int((*robotPt->armManipulationData).size()); i++)
			{
				ArmManipulationData& armData  = (*robotPt->armManipulationData)[i];

				desactivateTwoJointsFixCntrt(robotPt,armData.getManipulationJnt(), armData.getCcCntrt()->pasjnts[ armData.getCcCntrt()->npasjnts-1 ]);


				if (armData.getCartesian())
				{
					cout << "Arm Data " << i <<  " is set cartesian" << endl;
					activateCcCntrts(robotPt, i, false);
					ManipulationUtils::unfixManipulationJoints(robotPt, i);
					shootTheObjectArroundTheBase(robotPt, robotPt->baseJnt, armData.getManipulationJnt(), 2.0);
					armData.getManipulationJnt()->dist  = 0.10;

				} else {
					deactivateCcCntrts(robotPt, i);
					p3d_set_and_update_this_robot_conf(robotPt, qstart);
					setAndActivateTwoJointsFixCntrt(robotPt,armData.getManipulationJnt(), armData.getCcCntrt()->pasjnts[ armData.getCcCntrt()->npasjnts-1 ]);
					ManipulationUtils::fixManipulationJoints(robotPt, i, qstart, NULL);
					shootTheObjectInTheWorld(robotPt, armData.getManipulationJnt());
				}
			}
		

		if(p3d_isOutOfBounds(robotPt, qstart)){
			cout<<"Start is out of bounds"<<endl;
			return PLAN_STATUS_START_INVALID;
		}

		if(p3d_isOutOfBounds(robotPt, qgoal)){
			cout<<"Goal is out of bounds"<<endl;
			return PLAN_STATUS_GOAL_INVALID;
		}

		p3d_copy_config_into(robotPt, qstart, &robotPt->ROBOT_POS);

		p3d_copy_config_into(robotPt, qgoal, &robotPt->ROBOT_GOTO);

		p3d_set_and_update_this_robot_conf(robotPt, qstart);
	}

	confPtr_t q_source(new Configuration(rob, qstart));
	confPtr_t q_target(new Configuration(rob, qgoal));

  // Allocate the p3d_graph if does't exist
	if (!API_activeGraph || ENV.getInt(Env::nbMultiRun) > 1) {
		delete API_activeGraph;

		/*if(fromGraph != NULL)
		{
			API_activeGraph = fromGraph;
		}
		else
		{*/
			API_activeGraph = new Graph(rob);		
		//}
	}
	Graph * graph = API_activeGraph;

	// Delete last global planner
	if (global_Move3DPlanner)
		delete global_Move3DPlanner;

	// Main Run functions of all RRTs
	// All RRTs are initilized with init and run here
	int nb_added_nodes = 0;
	RRT* rrt = p3d_allocate_rrt(rob, graph);
	global_Move3DPlanner = rrt;
	nb_added_nodes += rrt->setInit(q_source);
	nb_added_nodes += rrt->setGoal(q_target);
	nb_added_nodes += rrt->init();
	rrt->setInitialized(true);
	rrt->setRunId(runId);

	cout << "* PLANNING ***************" << endl;
	ChronoTimeOfDayOn();

	nb_added_nodes += rrt->run();

	ChronoTimeOfDayTimes(&(graph->getGraphStruct()->rrtTime));
	ChronoTimeOfDayOff();

	graph->getGraphStruct()->totTime = graph->getGraphStruct()->rrtTime;

	if (ENV.getBool(Env::drawGraph) && globalRePlanningEnv) {
		globalRePlanningEnv->storeGraphToDraw(*graph);
	}

	// Extract the trajectory if one exists, else return NULL
	p3d_traj* traj = p3d_extract_traj(rrt->trajFound(), nb_added_nodes, graph, q_source, q_target);

    if (PlanEnv->getBool(PlanParam::showGTPDebug))
        cout << "** ** --------------------------" << endl;

	if (traj) {
		cout << "Success" << endl;
	} else {
		cout << "Fail" << endl;
	}

    if (PlanEnv->getBool(PlanParam::showGTPDebug))
    {
        cout << "TIME =" << graph->getGraphStruct()->rrtTime << " sec." << endl;
        cout << "NB NODES " << graph->getNumberOfNodes() << endl;
        cout << "NB EXPANSION " << rrt->getNumberOfExpansion() << endl;

        cout << "** ** --------------------------" << endl;
    }
    cout << "* PLANNING END ***************" << endl;

	rrt_statistics.runId = rrt->getRunId();
	rrt_statistics.succeeded = (traj != NULL);
	rrt_statistics.time = graph->getGraphStruct()->rrtTime;
	//rrt_statistics.cost = 0.0;
	rrt_statistics.nbNodes = graph->getNumberOfNodes();
	rrt_statistics.nbExpansions = rrt->getNumberOfExpansion();

	if (set_costspace)
		ENV.setBool(Env::isCostSpace, true);

	if(rrt_statistics.succeeded)
		return PLAN_STATUS_FOUND;
	
	return PLAN_STATUS_NOT_FOUND;
}

bool p3d_run_est(p3d_rob* robotPt)
{	
	// Gets the robot pointer and the 2 configurations
  Robot* rob = global_Project->getActiveScene()->getRobotByName(robotPt->name);
  confPtr_t q_source = rob->getInitialPosition();
  confPtr_t q_target = rob->getGoTo();
  
  // Allocate the p3d_graph if does't exist
  // Removes graph if it exists, creates a new graph , Allocate RRT
  Graph* graph = API_activeGraph =  new Graph(rob);
	
	printf("------------- Running EST --------------\n");
#ifdef LIST_OF_PLANNERS
	RRT* rrt = (RRT*)plannerlist[0];
#else
	EST* est = new EST(rob,graph);
#endif
    est->setInit(q_source);
    est->setGoal(q_target);
	int nb_added_nodes = est->init();
	
	graph = est->getActivGraph();
	
	cout <<"nb nodes " << graph->getNumberOfNodes() << endl;
	
	nb_added_nodes += est->run();
	ENV.setBool(Env::isRunning,false);
	
	printf("nb added nodes %d\n", nb_added_nodes);
	printf("nb nodes %zu\n",graph->getNodes().size());
	bool res = est->trajFound();
	
#ifndef LIST_OF_PLANNERS
	delete est;
#endif
	
	return res;
}


// ---------------------------------------------------------------------------------
// PRMs
// ---------------------------------------------------------------------------------
int p3d_run_prm(p3d_rob* robotPt)
{	
  // Gets the robot pointer and the 2 configurations
  Robot* rob = global_Project->getActiveScene()->getRobotByName(robotPt->name);
  confPtr_t q_source = rob->getInitialPosition();
  confPtr_t q_target = rob->getGoTo();
  
  // Allocate the p3d_graph if does't exist
  Graph* graph;
  if(!API_activeGraph)
  {
    API_activeGraph =  new Graph(rob);
  }
  graph = API_activeGraph;

	cout << "Initializing PRM " << endl;
  int nb_added_nodes=0;
  PRM* prm = new PRM(rob, graph);
  global_Move3DPlanner=prm;
	nb_added_nodes = prm->init();
	nb_added_nodes += prm->run();
	
	cout << "nb added nodes " << nb_added_nodes << endl;
	cout << "nb nodes " << graph->getNumberOfNodes() << endl;
  
  p3d_extract_traj( prm->trajFound(), nb_added_nodes, graph, q_source, q_target);
	delete prm;
	return nb_added_nodes;
}


int p3d_run_directed_prm(p3d_rob* robotPt){

    char *filename=ENV.getLogFile();


    if(filename)
        ChronoOn();

    // Gets the robot pointer and the 2 configurations
    Robot* rob = global_Project->getActiveScene()->getRobotByName(robotPt->name);

    // Allocate the p3d_graph if does't exist
    Graph* graph;
    if(!API_activeGraph){
        API_activeGraph=new Graph(rob);
    }
    graph=API_activeGraph;

    DirectedPRM* directed_prm=new DirectedPRM(rob,graph);
    global_Move3DPlanner=directed_prm;

    if(!directed_prm->initialize()){
        printf("error: p3d_run_directed_prm: initialization failed\n");
        return 0;
    }

    int nb_added=(int)directed_prm->run();

    cout << "nb added nodes " << nb_added << endl;
    cout << "nb nodes " << graph->getNumberOfNodes() << endl;

    directed_prm->extractSmoothAndReplaceTraj();
    int success=(int)directed_prm->trajFound();
    int nb_expansions=(int)directed_prm->getNumberOfExpansions();
    double shooting_failure_rate=directed_prm->getShootingFailureRate();
    double connectible_rate=directed_prm->getConnectibleRate();
    delete directed_prm;

    if(filename){
        double CPU_time,ts;
        ChronoMicroTimes(&CPU_time,&ts);
        ChronoOff();
        cout<<"CPU_time="<<CPU_time<<endl;
        int nb_nodes=graph->getNumberOfNodes();
        int nb_edges=graph->getNumberOfEdges();
        double flying_time=robotPt->tcur->range_param;
        FILE *f=fopen(filename,"a");
        if(f){
            fprintf(f,"%d %f %d %d %d %d %f %f %f\n",
                    success, //1
                    CPU_time, //2
                    nb_expansions, //3
                    nb_added, //4
                    nb_nodes, //5
                    nb_edges, //6
                    flying_time, //7
                    100-connectible_rate, //8
                    shooting_failure_rate); //9
            fclose(f);
        }
    }

    return nb_added;
}

int p3d_run_perturb_prm(p3d_rob* robotPt)
{
  if( !PlanEnv->getBool(PlanParam::withCurrentTraj) )
  {
    ENV.setBool(Env::isCostSpace,false);
    ENV.setBool(Env::expandToGoal,true);
    ENV.setBool(Env::biDir,true);
    
    p3d_run_rrt( robotPt );
    
    ENV.setBool(Env::biDir,false);
    ENV.setBool(Env::isCostSpace,true);
    ENV.setBool(Env::expandToGoal,false);
  }
  
  // Gets the robot pointer and the 2 configurations
  Robot* rob = global_Project->getActiveScene()->getRobotByName(robotPt->name);
  confPtr_t q_source = rob->getInitialPosition();
  confPtr_t q_target = rob->getGoTo();
  
  // Allocate the p3d_graph if does't exist
  // Removes graph if it exists, creates a new graph , Allocate RRT
  Graph* graph = API_activeGraph =  new Graph(rob);
	
	cout << "Initializing PRM " << endl;
  int nb_added_nodes=0;
  PerturbationRoadmap* prm = new PerturbationRoadmap(rob, graph);
	nb_added_nodes = prm->init();
	nb_added_nodes += prm->run();
  
	cout << "nb added nodes " << nb_added_nodes << endl;
	cout << "nb nodes " << graph->getNumberOfNodes() << endl;
  
  p3d_extract_traj( prm->trajFound(), nb_added_nodes, graph, q_source, q_target);
	delete prm;
	return nb_added_nodes;
}

int p3d_run_vis_prm(p3d_rob* robotPt)
{	
  // Gets the robot pointer and the 2 configurations
  Robot* rob = global_Project->getActiveScene()->getRobotByName(robotPt->name);
  confPtr_t q_source = rob->getInitialPosition();
  confPtr_t q_target = rob->getGoTo();
  
  // Allocate the p3d_graph if does't exist
  // Removes graph if it exists, creates a new graph , Allocate RRT
  Graph* graph = API_activeGraph =  new Graph(rob);
	
  int nb_added_nodes=0;
	Vis_PRM* vprm = new Vis_PRM(rob,graph);
	nb_added_nodes = vprm->init();
	nb_added_nodes += vprm->run();
	
	cout << "nb added nodes " << nb_added_nodes << endl;
	cout << "nb nodes " << graph->getNumberOfNodes() << endl;
  
  p3d_extract_traj( vprm->trajFound(), nb_added_nodes, graph, q_source, q_target);
	delete vprm;
	return nb_added_nodes;
}

int p3d_run_acr(p3d_rob* robotPt)
{	
  // Gets the robot pointer and the 2 configurations
  Robot* rob = global_Project->getActiveScene()->getRobotByName(robotPt->name);
  confPtr_t q_source = rob->getInitialPosition();
  confPtr_t q_target = rob->getGoTo();
  
  // Allocate the p3d_graph if does't exist
  // Removes graph if it exists, creates a new graph 
  Graph* graph = API_activeGraph =  new Graph(rob);
	
  int nb_added_nodes=0;
	ACR* acr = new ACR(rob, graph);
	nb_added_nodes = acr->init();
	nb_added_nodes += acr->run();
	
	cout << "nb added nodes " << nb_added_nodes << endl;
	cout << "nb nodes " << graph->getNumberOfNodes() << endl;
  
  p3d_extract_traj( acr->trajFound(), nb_added_nodes, graph, q_source, q_target);
	delete acr;
	return nb_added_nodes;
}

/**
 * Function To replace p3d_Learn in Case of C++ API Use
 */
void p3d_learn_cxx(int NMAX,
									 int (*fct_stop)(void), void (*fct_draw)(void)) {
	p3d_graph *G;
	int inode, ADDED;
	double tu, ts;
	int nbInitGraphNodes, nbGraphNodes;
	
	ChronoOn();
	
	if (!XYZ_GRAPH) G = p3d_create_graph();
	else           G = XYZ_GRAPH;
	/*debut modif fpilarde*/
	inode = 0;
  
  p3d_rob* robotPt = XYZ_GRAPH->rob;
	
#ifdef P3D_PLANNER
	p3d_set_planning_type(P3D_GLOBAL);
#else
	printf("P3D_PLANNER not compiled in %s in %s",__func__,__FILE__);
#endif
	
	ENV.setBool(Env::expandToGoal,false);
	
	if (p3d_get_MOTION_PLANNER() != P3D_DIFFUSION) {
		//while (inode < NMAX) {
		/* Call basic PRM or Visibility method */
		switch (p3d_get_MOTION_PLANNER()) {
				
			case 1:
				cout << "CXX_PLANNER c++ API : p3d_run_prm" << endl;
				ADDED = p3d_run_prm(robotPt);
				break;
				
			case 2:
				cout << "CXX_PLANNER c++ API : p3d_run_vis_prm" << endl;
				ADDED = p3d_run_vis_prm(robotPt);
				break;
				
			case P3D_ALL_PRM:
				cout << "CXX_PLANNER c++ API : p3d_run_acr" << endl;
				ADDED = p3d_run_acr(robotPt);
				break;
				
			default:
				PrintInfo(("p3d_learn : ERREUR : pas de planificateur global...\n"));
				return;
		}
		inode += ADDED;
	}
	
	else {
		nbInitGraphNodes = G->nnode;
		ADDED = p3d_run_rrt(robotPt);
		nbGraphNodes = G->nnode;
		inode  = nbGraphNodes - nbInitGraphNodes;
	}
	
#ifdef P3D_PLANNER
	p3d_set_planning_type(P3D_NONE);
#else
	printf("P3D_PLANNER not compiled in %s in %s",__func__,__FILE__);
#endif
	
	PrintInfo(("Pour la creation de %d noeuds : ", inode));
	
	ChronoTimes(&tu, &ts);
	G->time = G->time + tu;
	
#ifdef P3D_PLANNER
	/* When retrieving statistics;
	 Commit Jim; date: 01/10/2008 */
	if(getStatStatus()){
		G->stat->preTime += tu;
	}
#endif
	
	ChronoPrint("");
	ChronoOff();
	
	MY_ALLOC_INFO("After p3d_learn");
	p3d_print_info_graph(G);
}

/**
 * Function To replace p3d_specific_learn in Case of C++ API Use
 */
int p3d_specific_learn_cxx(double *qs, double *qg, int *iksols, int *iksolg,
													 int (*fct_stop)(void), void (*fct_draw)(void)) {
	
	p3d_graph *G;
	
	int       inode = 0, fail = 1, ADDED;
	double    tu, ts;
	
	int nbInitGraphNodes, nbGraphNodes;
#ifdef ENERGY
	int n_coldeg, icoldeg;
	double *coldeg_qs;
#endif
	/* Avoid some stupid errors */
	if (qs == NULL) {
		PrintInfo(("p3d_specific_learn : ERREUR : pas de configuration initiale\n"));
		return(FALSE);
	}
	
	if ((qg == NULL) && (ENV.getBool(Env::expandToGoal) == true)) {
		PrintInfo(("p3d_specific_learn : ERREUR : pas de configuration finale\n"));
		return(FALSE);
	}
	
	ENV.setBool(Env::expandToGoal,true);
	
	ChronoOn();
	
	if (!XYZ_GRAPH)  G = p3d_create_graph();
	else            G = XYZ_GRAPH;
  
   p3d_rob* robotPt = XYZ_GRAPH->rob;
	
	if (p3d_GetIsWeightedChoice() == TRUE) {
		p3d_init_root_weight(G);
	}
	if (ENV.getInt(Env::ExpansionNodeMethod) == RANDOM_IN_SHELL_METH) {
		p3d_init_pb_variables(G);
		
#ifdef ENERGY
		if (p3d_get_MOTION_PLANNER() ==  BIO_COLDEG_RRT) {
			n_coldeg = bio_get_num_collective_degrees();
			// init coldeg_q in Ns
			coldeg_qs = bio_alloc_coldeg_config(n_coldeg);
			for (icoldeg = 0; icoldeg < n_coldeg; icoldeg++) {
				coldeg_qs[icoldeg] = 0.0;
			}
			bio_copy_coldeg_q_in_N(Ns, coldeg_qs, n_coldeg);
			bio_destroy_coldeg_config(coldeg_qs, n_coldeg);
			// WARNING : currently Ng is not considered !!!
		}
#endif
		
		
	}
	ADDED = TRUE;
	if (p3d_get_MOTION_PLANNER() != P3D_DIFFUSION) {
		/* While solution does not exists, insert new nodes with basic PRM or Visibility or RRT */
		//     while ((Ns->numcomp != Ng->numcomp) && !p3d_compco_linked_to_compco(Ns->comp, Ng->comp)) {  modif fpilarde
		switch (p3d_get_MOTION_PLANNER()) {
				
			case P3D_BASIC:
				cout << "CXX_PLANNER c++ API : p3d_run_prm" << endl;
				ADDED = p3d_run_prm(robotPt);
				break;
				
			case P3D_ISOLATE_LINKING:
				cout << "CXX_PLANNER c++ API : p3d_run_vis_prm" << endl;
				ADDED = p3d_run_vis_prm(robotPt);
				break;
				
			case P3D_ALL_PRM:
				cout << "CXX_PLANNER c++ API : p3d_run_acr" << endl;
				ADDED = p3d_run_acr(robotPt);
				break;
				
#ifdef ENERGY
			case BIO_COLDEG_RRT:
				ADDED = bio_expand_coldeg_rrt(G, fct_stop);
				break;
#endif
			default:
				PrintInfo(("p3d_specific_learn : ERREUR : pas de planificateur global...\n"));
				return(FALSE);
		}
		
		inode += ADDED;
		
	} else {
		nbInitGraphNodes = G->nnode;
		cout << "CXX_PLANNER c++ API : p3d_run_rrt" << endl;
		ADDED = p3d_run_rrt(robotPt);
		nbGraphNodes = G->nnode;
		inode  = nbGraphNodes - nbInitGraphNodes;
	}
	
#ifdef P3D_PLANNER
  p3d_set_planning_type(P3D_NONE);
#else
	printf("P3D_PLANNER not compiled in %s in %s",__func__,__FILE__);
#endif
  
	PrintInfo(("Pour la creation de %d noeuds : ", inode));
    if (ENV.getBool(Env::printMightabilities))
        ChronoPrint("");
	
	ChronoTimes(&tu, &ts);
	G->time = G->time + tu;
#ifdef P3D_PLANNER
	/* When retrieving statistics;
	 Commit Jim; date: 01/10/2008 */
	if(getStatStatus()){
		G->stat->preTime += tu;
	}
	ChronoOff();
#endif
	
	p3d_print_info_graph(G);
    if (ENV.getBool(Env::printMightabilities))
        MY_ALLOC_INFO("After p3d_specific_learn");
	if (p3d_get_saveInfoInFile()) {
		//    save_infos_in_file(G, ADDED);
	}
	
	PrintInfo(("\n"));
	return(ADDED || !fail);
}
