#include "Geometry.h"

namespace m3dGeometry
{


Eigen::Vector3d p3dToEigen(const p3d_point v){
    Eigen::Vector3d r;
    r.x() =v.x;
    r.y() =v.y;
    r.z() =v.z;
    return r;
}

p3d_point getTranslation(const p3d_matrix4 t){
    p3d_point p;
    p.x=t[0][3];
    p.y=t[1][3];
    p.z=t[2][3];
    return p;
}

Eigen::Vector2d getConfBase2DPos(const Configuration &q){
    Eigen::Vector2d v;
    v[0] = q[6];
    v[1] = q[7];
    return v;
}

void setBaseOrientation(Robot * rob,const Eigen::Matrix3d & rot_mat){
    Eigen::Vector3d angles = rot_mat.eulerAngles(0,1,2);
    setBaseOrientation(rob,angles);
}
void setBaseOrientation(Robot *rob, const Eigen::Vector3d & angles){
    Configuration q = *rob->getCurrentPos();
    for(unsigned int i=0;i<3;i++){
        q.at(6+3+i)=angles[i];
    }
    rob->setAndUpdate(q);
}

void setBasePosition(Robot *rob, const Eigen::Vector3d & pos){
    Configuration q=*rob->getCurrentPos();
    for(unsigned int i=0;i<3;i++){
        q.at(6+i)=pos[i];
    }
    rob->setAndUpdate(q);
}

void setBase6D(Robot *rob,const Eigen::Vector3d &pos, const Eigen::Vector3d &angles){
    setBasePosition(rob,pos);
    setBaseOrientation(rob,angles);
}


void setBase4D(Robot *rob, const Eigen::Vector3d &pos, double theta){
    Eigen::Vector3d angles;
    angles.x()=angles.y()=0;
    angles.z()=theta;
    setBasePosition(rob,pos);
    setBaseOrientation(rob,angles);
}

void setBaseTransform(Robot* rob, const Eigen::Affine3d & tf){
    setBaseOrientation(rob,tf.rotation());
    setBasePosition(rob,tf.translation());
}


void setBaseFromConf(Robot *rob, const Configuration & conf){
    Configuration q = *rob->getCurrentPos();
    for(unsigned int i=0;i<6;i++){
        q[6+i] = conf[6+i];
    }
    rob->setAndUpdate(q);
}

}

