#ifndef GEOMETRY_H
#define GEOMETRY_H

#include "P3d-pkg.h"
#include "API/Device/robot.hpp"
#include <Eigen/Eigen>

/**
 * this namespace provides functions to switch between p3d and Eigen data,
 * and tools for geometric operations with Configuration too, like rotation, etc...
 */
namespace m3dGeometry
{

    /** @name EigenToP3d
     */
    ///@{

    p3d_point eigenToP3d(const Eigen::Vector3d & v);

    ///@}


    /** @name p3dToEigen
     */
    ///@{

    Eigen::Vector3d p3dToEigen(const p3d_point v);

    ///@}

    /// @name position
    ///@{

    /// returns the translation part of a transform represented as 4x4 matrix
    p3d_point getTranslation(const p3d_matrix4 t);

    /// returns the base 2d position (1st DOF)
    Eigen::Vector2d getConfBase2DPos(const Configuration &q);

    /// set only base rotation based on a rotation matrix
    /// (wrt Rx=Ry=Rz=0, NOT rotation from actual position)
    void setBaseOrientation(Robot* rob,const Eigen::Matrix3d & rot_mat);
    /// set from angles Rx Ry Rz given in a vector
    void setBaseOrientation(Robot *rob, const Eigen::Vector3d & angles);
    /// set only base position in 3D (X,Y,Z translations from origin)
    void setBasePosition(Robot *rob, const Eigen::Vector3d & pos);
    /// as setBaseOrientation() and setBasePosition() with angles Rx,Ry,Rz specified
    /// in a vector
    void setBase6D(Robot *rob,const Eigen::Vector3d &pos, const Eigen::Vector3d &angles);
    /// as setBase6D() but set Rx and Ry to 0, Rz to theta
    void setBase4D(Robot *rob, const Eigen::Vector3d &pos, double theta);
    /// as setBase6D() but with a transform
    void setBaseTransform(Robot* rob, const Eigen::Affine3d & tf);

    /// change the current configuration of rob so that its base joint position
    /// matches the given conf one.
    void setBaseFromConf(Robot *rob, const Configuration & conf);
    ///@}

}

#endif // GEOMETRY_H
