#include "MultiHandOverDisplay.h"
#include "Chronometer.hpp"

namespace HRICS{
namespace MHO {

MultiHandOverDisplay::MultiHandOverDisplay():
    fps(25), _current_task(0), _current_time(0)
{

}

MultiHandOverDisplay::~MultiHandOverDisplay()
{

}

bool MultiHandOverDisplay::updateSolutionTrajSyncStep(Task* task, double time)
{
    bool no_more_sub=true;
    for(unsigned int i=0; i < task->getSubTasks().size() ; i++){
        if(task->getSubTask(i)->getMarker(SubTask::HAS_FULL_TRAJ)){
            API::Trajectory &traj=task->getSubTask(i)->traj();
            if(task->getSubTask(i)->beginDate() <= time
                    && task->getSubTask(i)->endDate() >= time){
                no_more_sub=false;
                double pmax=traj.getRangeMax();
                double time_to_param = pmax / (task->getSubTask(i)->duration());
                double p= ( time - task->getSubTask(i)->beginDate() ) * time_to_param;

                confPtr_t q=traj.configAtParam(p);
                if(q){
                    traj.getRobot()->setAndUpdate(*q);
                }
            }else if(task->getSubTask(i)->endDate() < time){
                //sub task ended already
                //cout << "done: "<<i<<": "<<task->getSubTask(i)->agent()->shortname <<" - "<<SubTask::typeString(task->getSubTask(i)->type())<<endl;
                confPtr_t q=traj.getEnd();
                if(q){
                    traj.getRobot()->setAndUpdate(*q);
                }
            }else if(task->getSubTask(i)->beginDate() > time){
                //cout << "coming: "<<i<<": "<<task->getSubTask(i)->agent()->shortname <<" - "<<SubTask::typeString(task->getSubTask(i)->type())<<endl;
                no_more_sub = false;
            }
            //}else{
            //cout << "omitting: "<<i<<": "<<task->getSubTask(i)->agent()->shortname <<" - "<<SubTask::typeString(task->getSubTask(i)->type())<<endl;
        }
    }
    return !no_more_sub;
}

bool MultiHandOverDisplay::stepForwardSolutionTrajSync(Task *task){
    bool ok;
    if(task){
        this->_current_task=task;
        this->_current_time=task->begin().date();
    }

    ok = updateSolutionTrajSyncStep(_current_task,_current_time);
    g3d_draw_allwin_active();
    _current_time += ENV.getDouble(Env::showTrajFPS) / this->fps;
    return ok;
}


void MultiHandOverDisplay::playSolutionTrajSyncRT(Task *task){
    cout<<"Playing task traj sync"<<endl;
    double time=task->begin().date();
    double speed_factor=ENV.getDouble(Env::showTrajFPS);
    bool finish=false;
    double cur_time=time;
    double step = speed_factor / this->fps;
    Chronometer chrono;
    unsigned int skip=0;

    while(!finish){
        chrono.reset();
        finish=true;
        if(skip==0){
            //cout << "t: "<<cur_time<<endl;
            finish=!this->updateSolutionTrajSyncStep(task,cur_time);
        }else{
            cout<<"skipping "<<skip<<endl;
            finish=false;
            skip--;
        }
        speed_factor=ENV.getDouble(Env::showTrajFPS);
        step = speed_factor / this->fps;
        cur_time += step;
        g3d_draw_allwin_active();
        long int remaining = 1e6 / this->fps - chrono.getMicroSec();
        //assert(remaining >0);
        //if too slow (remaining <0) skip some next computations
        if(remaining<0){
            skip = (-remaining ) / 1e6/this->fps +1;
        }else{
            usleep(remaining); //sleep the remaining time untill the next frame
        }
    }
    cout<<"Playing task traj sync ended"<<endl;

}

} // namespace MHO

}
