#include "LazyWAstar.h"
#include <limits>
#include <deque>
#include <algorithm>
#include <cmath>
#include <iostream>

#include <assert.h>
#include "Chronometer.hpp"

using namespace std;

namespace mho {

/*
LWAEdge::LWAEdge() :
    Edge(),
    _is_true_cost(0)
{ }

LWAEdge::LWAEdge(NodeInterface *n1, NodeInterface *n2, double c,bool true_cost):
    Edge(n1,n2,c),
    _is_true_cost(true_cost)
{ }

bool LWAEdge::isTrueCost(){
    return _is_true_cost;
}

void LWAEdge::setTrueCost(double cost){
    _is_true_cost=true;
    setCost(cost);
}
void LWAEdge::setTmpCost(double cost){
    _is_true_cost=false;
    setCost(cost);
}
*/

LWANodeBase::LWANodeBase():
    _parent(0),
    _oldest_clone(this),
    _clones(0)
{
    resetStatus();
    _clones.push_back(this);
    setTrueCost(false);
}

LWANodeBase::LWANodeBase(LWANodeBase *other):
    _is_true_cost(other->_is_true_cost),
    _g(other->_g), _h(other->_h),
    _open(other->_open),_closed(other->_closed),
    _parent(other->parent()),
    _oldest_clone(other->_oldest_clone),
    _clones(other->_clones)
{
}

LWANodeBase::~LWANodeBase(){
    removeClone(this);
}

LWANodeInterface *LWANodeBase::clone(){
    LWANodeBase *clone =new LWANodeBase(this);
    this->addClone(clone);
    return clone;
}

void LWANodeBase::addClone(LWANodeBase *clone){
    clone->_oldest_clone=this->_oldest_clone;
    clone->_oldest_clone->_clones.push_back(clone);
}

void LWANodeBase::updateOldest(LWANodeBase *new_oldest){
    if(new_oldest==this){
        _clones=_oldest_clone->_clones;
        _oldest_clone=this;
    }else{
        _oldest_clone=new_oldest;
        _clones.clear();
    }
}

void LWANodeBase::removeClone(LWANodeBase* removed){
    if(_oldest_clone==this){
        remove(_clones.begin(),_clones.end(),removed);
        if(removed==this && _clones.size()){
            //new oldest is first clone
            for(unsigned int i=0;i<_clones.size();++i){
                dynamic_cast<LWANodeBase*>(_clones[i])->updateOldest(dynamic_cast<LWANodeBase*>(_clones.front()));
            }
        }
    }else{
        _oldest_clone->removeClone(removed);
    }
}

bool LWANodeBase::haveLowerTrueCostClones(){
    bool found(0);
    for(vector<LWANodeInterface*>::iterator it=this->_oldest_clone->_clones.begin();it<this->_oldest_clone->_clones.end();++it){
        if((*it) != this && (*it)->isTrueCost() && (*it)->g() <= this->g()){
            found=true;
            break;
        }
    }
    return found;
}

LazyWeightedAstar::~LazyWeightedAstar(){
    this->reset();
}

struct IsSameAndClosed{
    bool operator()(LWANodeInterface *other){return other->isClosed() && ref->isSameConf(other);}
    LWANodeInterface* ref;
};

void LazyWeightedAstar::reset(){
    //reset data
    while(open.size()){
        delete open.top();
        open.pop();
    }
    for(vector<LWANodeInterface*>::iterator it=closed.begin();it!=closed.end();++it){
        delete *it;
    }
    closed.clear();
    _goal_found=0;
}

bool LazyWeightedAstar::compute(LWANodeInterface* start, LWANodeInterface *goal, double time_limit_ms, bool no_lazy){
    //reset data
    _start=start;
    _goal=goal;
    _no_lazy=no_lazy;
    setTimeLimit(time_limit_ms);
    return compute();
}

bool LazyWeightedAstar::compute(){
    Chronometer main_chrono,chrono2,chrono3;
    vector<pair<unsigned int,double> > times(10,make_pair(0,0.0));
    //0: find similar closed
    //1: node expansion
    //2: cost estimation (included in 1)
    //3: heuristic comp (included in 1)
    //4: true cost comp and treat
    //5: true cost comp (included in 4)

    open.push(_start);
    _start->h(_start->computeHeuristic(_goal));
    _start->g(0);
    _start->setTrueCost(true);
    _goal->g(numeric_limits<double>::infinity());
    _goal->h(0);
    _interface->onOpen(_start);
    //while (open.size() && !(_goal->isTrueCost() && _goal->f() <= open.top()->f())){
    main_chrono.reset();
    while (open.size() && !(open.top()->isSameConf(_goal) && open.top()->isTrueCost())
           && (getTimeLimit()<0 || main_chrono.getMilliSec()<getTimeLimit()) ){
        //while the cheapest is not the goal, with true cost
        //and the time limit is not over
        LWANodeInterface *s = open.top();
        open.pop();
        IsSameAndClosed pred; pred.ref=s;
        chrono2.reset();
        bool ok=(find_if(closed.begin(),closed.end(),pred) == closed.end());
        times[0].first++;
        times[0].second+=chrono2.getMilliSec();
        if(!ok){
            //if a similar (same conf) node have been closed, it means that it has a lower cost
            //so not interesting, just remove from open and ignore it
            continue;
        }else if( s->isTrueCost()){
            chrono2.reset();
            s->close();
            closed.push_back(s);
            _interface->onClose(s);
            vector<LWANodeInterface*> sons=s->sons();
            times[9].first++;
            times[9].second+=chrono2.getMilliSec();
            for (unsigned int i=0;i<sons.size();++i){
                chrono3.reset();
                LWANodeInterface * son;
                //if(sons[i]->isSameConf(_goal)){
                    //son=_goal->clone();
                //}else{
                    son=sons[i]->clone();
                //}
                times[6].first++;
                times[6].second+=chrono3.getMilliSec();
                if(!son->isClosed()){ //TODO always true, check this (issue with the sons() Node method implementation)
                    son->parent(s);
                    chrono3.reset();
                    if(this->_no_lazy){
                        son->g(s->g() + son->computeTrueCost(s));
                        son->setTrueCost(true);
                    }else{
                        son->g( s->g() + son->edgeCostEstimationTo(s));
                    }
                    times[2].first++;
                    times[2].second+=chrono3.getMilliSec();
                    chrono3.reset();
                    bool have_lower_clones=son->haveLowerTrueCostClones();
                    times[8].first++;
                    times[8].second+=chrono3.getMilliSec();
                    if(!have_lower_clones){
                        chrono3.reset();
                        son->h(this->_epsilon * son->computeHeuristic(_goal));
                        times[3].first++;
                        times[3].second+=chrono3.getMilliSec();
                        chrono3.reset();
                        open.push(son);
                        _interface->onOpen(son);
                        times[7].first++;
                        times[7].second+=chrono3.getMilliSec();
                    }else if(son!=_goal){
                        delete son;
                    }
                }
                delete sons[i];
            }
            times[1].first++;
            times[1].second+=chrono2.getMilliSec();
        }else{
            assert(!this->_no_lazy);//we should not be here if we don't use lazy variant
            chrono2.reset();
            double c=s->computeTrueCost(s->parent());
            times[5].first++;
            times[5].second+=chrono2.getMilliSec();
            if (c < numeric_limits<double>::infinity()){
                s->setTrueCost(true);
                s->g(s->parent()->g() + c);
                if(!s->haveLowerTrueCostClones()){
                    double h=s->h();
                    s->h(_epsilon * s->computeHeuristic(_goal));
                    if(h!=s->h()){
                        std::cout << " THERE ARE DIFFERENT EPSILON HERE!"<<std::endl; //TODO : if this text never shows, remove above assignment to h
                    }
                    open.push(s);
                    _interface->onOpen(s);
                }else{
                    delete s;
                }
            }else{
                s->g(numeric_limits<double>::infinity());
                _interface->onClose(s);
                delete s;
            }
            times[4].first++;
            times[4].second+=chrono2.getMilliSec();
        }
    }
    _goal_found=0;
    if(open.size() && open.top()->isSameConf(_goal) && open.top()->isTrueCost()){
        _goal_found=open.top();
        //_goal->parent(open.top()->parent());
        //_goal->g(open.top()->g());
        LWANodeInterface* cur=_goal_found;
        while(cur){
            if( !(cur->g()+cur->h()/_epsilon - _goal_found->g() < 0.000001)){
            //    cur->computeHeuristic(_goal);
                cout << "heuristic error: "<< cur->g() << " ; " << cur->h()/_epsilon << " ; " << cur->g() + cur->h() / _epsilon << " ; " << cur->f() << " ; " << _goal_found->g() << endl;
            }
            //assert(cur->g()+cur->h()/_epsilon - _goal_found->g() < 0.00001);
            assert(cur->isTrueCost());
            assert(cur->g()!=numeric_limits<double>::infinity());

            cur=cur->parent();
        }
        _interface->onFound(_goal_found);
    }
    _interface->onEnd(times);
    return _goal_found && _goal_found->parent()!=0;
}

deque<LWANodeInterface*> LazyWeightedAstar::getPath(){
    deque<LWANodeInterface*> path;
    if(_goal_found){
        path.push_front(_goal_found);
        while (path.front() && path.front() != _start){
            path.push_front(path.front()->parent());
        }
    }
    return path;
}

void LazyWeightedAstar::setInterface(DisplaySearchAstarInterface *interface){
    if(_interface && _owns_interface){
        delete _interface;
    }
    if(interface==0){
        _interface=new DisplaySearchAstarInterface();
        _owns_interface=true;
    }else{
        _interface=interface;
        _owns_interface=false;
    }
}
}
